/*************************************************************************
 *
 * Parallaxis Compiler Projects
 * ----------------------------
 *
 * Name	  :  SYSTEM_MPL.h
 * Date	  :  Mon Sep 4 02:23:53 1995
 * Desc	  :  System header file for Parallaxis programs. Is automatically
 *	     included in every generated MPL-Program.
 * Author :  keller
 *
 * Description:
 * ------------
 * System header file for Parallaxis-III programs compiled for MPL. This
 * file is automatically included by every generated MPL-program.
 * This file contains declarations for external functions and procedures as
 * they are internally used by the generated program. It also consists of
 * functions, type declarations and macros that build the system library and
 * the SYSTEM module.
 *
 * Convention:
 * Function and macro names ending with "V" are the vector versions of those
 * not ending with "V". Example:
 *   Write:  normal scalar version of Write
 *   WriteV: vector version of Write (plural argument)
 ************************************************************************/



/************************************************************************/
/*									*/
/*	If the library function alloca is available and the C programs	*/
/*	are compiled with the flag -DStackAlloc, then the memory space	*/
/*	for open array value parameters will be allocated in the stack	*/
/*	frame of the corresponding procedure. This temporary space	*/
/*	will be freed automatically when the procedure returns.		*/
/*	Otherwise, malloc and free will be used to allocate and		*/
/*	deallocate memory space for open array value parameters.	*/
/*									*/
/*	By default open array value parameters are copied using the	*/
/*	library function memcpy. This can be switched to bcopy by	*/
/*	compiling with the flag -DBCOPY on systems without memcpy.	*/
/*									*/
/************************************************************************/

# ifdef __STDC__
# define ARGS(parameters)	parameters
# else
# define ARGS(parameters)	()
# endif

#include <mpl.h>
#include <stdlib.h>
#include <stdio.h>
#include <mp_libc.h>
#include <string.h>  /* for memcpy in COPY_OPEN_ARRAY */

/*
 * Definition of standard constants
 */

#define FALSE		((BOOLEAN)0)
#define TRUE		((BOOLEAN)1)
#define NIL		(void *) 0L
#define EOL		'\n'
#define pi		3.1415926536

#ifndef NULL
#define NULL		(void *) 0L
#endif

/*
 * Definition of basic types
 */

typedef long		INTEGER;
typedef unsigned long	CARDINAL;
typedef unsigned char	BOOLEAN;
typedef unsigned char	CHAR;
typedef double		REAL;

/*
 * Definition of standard types
 */

typedef unsigned long	BITSET;
typedef void		(*PROC)();
typedef unsigned char	WORD;
typedef WORD		BYTE;
typedef char		*ADDRESS;

/*
 * Definition of parallaxis types
 */

typedef unsigned short USHORT;

typedef USHORT PPE;			  /* Physical PE, 0..65535 */
typedef USHORT PPE_LAYER;		  /* Id-number for a layer of PPEs */
typedef CARDINAL VPE;			  /* Virtual PE */

typedef struct {			  /* The position of a virtual PE */
     PPE_LAYER	layer;			  /* (VPE) is given by the layer and */
     PPE	ppe;			  /* the PPE number */
} VPE_POS;

typedef plural VPE_POS * plural VPE_POS_PTR;

#define CONTAINER_N_SIZE 5

typedef plural struct C_1 {
     VPE_POS pe;
     plural struct C_1 *next;
} CONTAINER_1;

typedef CONTAINER_1 *CONTAINER_1_PTR;

typedef plural struct C_N {
     VPE_POS pe[CONTAINER_N_SIZE];
     plural struct C_N *next;
} CONTAINER_N;

typedef CONTAINER_N *CONTAINER_N_PTR;

typedef struct {
     plural CONTAINER_1_PTR *act_single;
     plural USHORT *act_multiple_num;
     plural CONTAINER_N_PTR *act_multiple;
     plural USHORT *num_1;
     plural CONTAINER_1_PTR *single;
     plural CONTAINER_N_PTR *multiple;
} CONNECTION;

typedef struct P_C {
     INTEGER parameter;
     CONNECTION *connections;
     struct P_C *next;
} PARAMETER_CONNECTION;

plural struct {
     CARDINAL num_1;
     CONTAINER_1_PTR single;
} FREELIST;

typedef plural BOOLEAN ACTIVESET;	  /* active set data */

typedef struct {
     INTEGER  lower;			  /* lower bound of dimension */
     INTEGER  upper;			  /* upper bound of dimension */
     CARDINAL size;			  /* number of VPEs in this dimension */
} DIMENSION;

typedef struct {
     CARDINAL number;			  /* Id-Number of this group */
     CARDINAL VPEs;			  /* Number of virtual PEs */
     CARDINAL layers;			  /* Number of layers (depends on nproc) */
     ACTIVESET *activeset;		  /* Pointer to active set */
} CONFIGGROUP;	   

typedef struct {
     char *name;			  /* name of this configuration */
     CARDINAL rank;			  /* number of dimensions */
     DIMENSION *dim;			  /* dimensions itself (vector) */
     CONFIGGROUP *configgroup;		  /* pointer to the configgroup */
} CONFIGURATION;


/*----------------------------------------------------------------------*/

/*
 * Definition of standard functions without side-effects. They are executed
 * right within the PE-loop (no pre-compution via help vars).
 */

/*
 * ABS is overloaded in Parallaxis, therefore must be splitted into several
 * versions in MPL.
 *  ABS	  -> Macro    used when expression is C-constant (allows backend
 *						     compiler optimization)
 *  ABSC  -> Macro    used when expression is CARDINAL
 *  ABSI  -> Function used when expression is INTEGER
 *  ABSR  -> Function used when expression is REAL
 *  ABSCV -> Macro    used when expression is CARDINAL vector
 *  ABSIV -> Function used when expression is INTEGER vector
 *  ABSRV -> Function used when expression is REAL vector
 */

#define ABS(x)		((x) < 0 ? -(x) : (x))
#define ABSC(x)		((CARDINAL) (x))
extern  INTEGER ABSI 	ARGS((INTEGER x));
extern  REAL    ABSR 	ARGS((REAL x));
#define ABSCV(x)	((plural CARDINAL) (x))
extern  plural  INTEGER ABSIV 	ARGS((plural INTEGER x));
extern  plural  REAL    ABSRV 	ARGS((plural REAL x));


#define ADR(x)		((ADDRESS) &(x))
#define ADRV(x)		((ADDRESS) (x))
extern  CHAR CAP 	ARGS((CHAR ch));
extern  plural CHAR CAPV  ARGS((plural CHAR ch));
#define CHR(x)		((CHAR) (x))
#define CHRV(x)		((plural CHAR) (x))
#define EVEN(x)		(!(BOOLEAN)((x) & 01))
#define EVENV(x)	(!(plural BOOLEAN)((x) & 01))
#define FLOAT(x)	((REAL) (x))
#define FLOATV(x)	((plural REAL) (x))
#define ODD(x)		((BOOLEAN)((x) & 01))
#define ODDV(x)		((plural BOOLEAN)((x) & 01))
#define ORD(x)		((CARDINAL) (x))
#define ORDV(x)		((plural CARDINAL) (x))
#define TRUNC(x)	((INTEGER) (x))
#define TRUNCV(x)	((plural INTEGER) (x))
#define SQRT(x)		sqrt(x)
#define SQRTV(x)	p_sqrt(x)
#define EXP(x)		exp(x)
#define EXPV(x)		p_exp(x)
#define LN(x)		log(x)
#define LNV(x)		p_log(x)
#define SIN(x)		sin(x)
#define SINV(x)		p_sin(x)
#define ARCSIN(x)	asin(x)
#define ARCSINV(x)	p_asin(x)
#define COS(x)		cos(x)
#define COSV(x)		p_cos(x)
#define ARCCOS(x)	acos(x)
#define ARCCOSV(x)	p_acos(x)
#define TAN(x)		tan(x)
#define TANV(x)		p_tan(x)
#define ARCTAN(x)	atan(x)
#define ARCTANV(x)	p_atan(x)
#define CEILING(x)	((INTEGER) ceil(x))
#define CEILINGV(x)	((plural INTEGER) p_ceil(x))
#define FLOOR(x)	((INTEGER) floor(x))
#define FLOORV(x)	((plural INTEGER) p_floor(x))
#define ROUND(x)	((INTEGER) rint(x))
#define ROUNDV(x)	((plural INTEGER) p_rint(x))



#define ARCTAN2(x,y)    atan2(x, y)
#define ARCTAN2V(x,y)   p_atan2(x, y)
#define VAL(T,x)	((T) (x))

#define MIN_INTEGER	(-2147483648L)
#define MAX_INTEGER	2147483647L
#define MIN_CARDINAL	((CARDINAL)0L)
#define MAX_CARDINAL	((CARDINAL)4294967295L)

#define MIN_BOOLEAN	FALSE
#define MAX_BOOLEAN	TRUE

#define MIN_CHAR	((CHAR)'\0')
#define MAX_CHAR	((CHAR)'\377')

#define MIN_REAL	4.94065645841246544e-324
#define MAX_REAL	1.79769313486231470e+308

#define RandomCard()	((CARDINAL) random())
#define RandomCardV()	((plural CARDINAL) p_random())
#define RandomInt()	((INTEGER) (random() & MAX_INTEGER))
#define RandomIntV()	((plural INTEGER) (p_random() & MAX_INTEGER))
#define RandomReal()	((REAL) (random()/((REAL)MAX_CARDINAL+1.0)))
#define RandomRealV()	((plural REAL) (p_random()/((REAL)MAX_CARDINAL+1.0)))
#define RandomBool()	((BOOLEAN) (random() & MAX_BOOLEAN))
#define RandomBoolV()	((plural BOOLEAN) (p_random() & MAX_BOOLEAN))
#define RandomChar()	((CHAR) (random() & MAX_CHAR))
#define RandomCharV()	((plural CHAR) (p_random() & MAX_CHAR))


/* Remark: ID works OK this way if it's always precomputed via a help */
/* variable. The configuration of ID() must be the same as the one of the */
/* surrounding PElayer-loop. */
#define ID(conf)        (PElayer*nproc+iproc+1)
#define RANK(conf)      (conf.rank)
#define LEN(conf)       (conf.configgroup->VPEs)

extern CARDINAL DIM2ID ARGS((CONFIGURATION conf, INTEGER dims[]));
extern plural CARDINAL DIM2IDV ARGS((CONFIGURATION conf, plural INTEGER *dims[]));
extern plural INTEGER DIM ARGS((CONFIGURATION conf, CARDINAL dim));
extern CARDINAL LEN1 ARGS((CONFIGURATION conf, CARDINAL dim));
#define UPPER(conf,d)   (conf.dim[(d)-1].upper)
#define LOWER(conf,d)   (conf.dim[(d)-1].lower)

/* Predefined Reduce Functions */
/* FIRST & LAST use MIN & MAX on IDs */
#define SUM(x,y,z)     ((*x) = (*y) + (*z))
#define PRODUCT(x,y,z) ((*x) = *(y) * (*z))
#define MIN(x,y,z)     ((*x) = (*y) < (*z) ? (*y) : (*z))
#define MAX(x,y,z)     ((*x) = (*y) > (*z) ? (*y) : (*z))
#define AND(x,y,z)     ((*x) = (*y) && *(z))
#define OR(x,y,z)      ((*x) = (*y) || (*z))

/*----------------------------------------------------------------------*/

/*
 * Definition of parallaxis standard procedures
 */

extern void argv ARGS((CARDINAL index, CHAR arg[]));
extern void CloseInput ARGS((void));
extern void CloseOutput ARGS((void));
extern void OpenInput ARGS((CHAR *s));
extern void OpenOutput ARGS((CHAR *s));
extern void WriteLn(void);
extern void Write (CHAR c);
extern void WriteV (CONFIGURATION conf, plural CHAR c[]);
extern void WriteString(CHAR *s);
extern void WriteStringV(CONFIGURATION conf, plural CHAR * plural s[]);
extern void WriteInt(INTEGER i, CARDINAL len);
extern void WriteIntV(CONFIGURATION conf, plural INTEGER i[], CARDINAL len);
extern void WriteCard(INTEGER c, CARDINAL len);
extern void WriteCardV(CONFIGURATION conf, plural INTEGER c[], CARDINAL len);
extern void WriteBool(BOOLEAN b, CARDINAL len);
extern void WriteBoolV(CONFIGURATION conf, plural BOOLEAN b[], CARDINAL len);
extern void WriteReal(REAL r, CARDINAL len);
extern void WriteRealV(CONFIGURATION conf, plural REAL r[], CARDINAL len);
extern void WriteFixPt(REAL r, CARDINAL len, CARDINAL dec);
extern void WriteFixPtV(CONFIGURATION conf, plural REAL r[], CARDINAL len, 
			CARDINAL dec);


#define DEC(x)		(x)--
#define DEC2(x,n)	x -= n
#define DECV(conf,x)	(x)--
#define DEC2V(conf,x,n)	x -= n
#define INC(x)		(x)++
#define INC2(x,n)	x += n
#define INCV(conf,x)	(x)++
#define INC2V(conf,x,n)	x += n

#define EXCL(s,i)	s &= ~(0X1L << (i))
#define INCL(s,i)	s |= 0X1L << (i)

extern void ID2DIM ARGS((CONFIGURATION conf, CARDINAL id, INTEGER dims[]));

#define ROUTER TRUE
#define NO_ROUTER !(ROUTER)

/* Look in implementation on how to use 'WITH_ROUTER' */

extern void ID2DIMV ARGS((CONFIGURATION conf, plural CARDINAL id[], plural INTEGER *dims[], BOOLEAN WITH_ROUTER));

/*----------------------------------------------------------------------*/

/*
 *	Definition of (some) set operators
 */

#define SYSTEM_MaxSet		(sizeof (unsigned long) * 8 - 1)

#define SET_ELEM(el)		(0X1L << (el))
#define SET_cRNG(lo,hi) \
	((lo) <= (hi) ? ~0X0L >> (lo) << (lo) + SYSTEM_MaxSet - (hi) >> SYSTEM_MaxSet - (hi) : 0X0L)
#define SET_RANGE(lo,hi)	SET_RANGE1((CARDINAL)(lo), (CARDINAL)(hi))

extern unsigned long SET_RANGE1 ARGS((CARDINAL lo, CARDINAL hi));

#define SET_DIFF(s1,s2)		((s1) & ~(s2))
#define IN(x,s)			((BOOLEAN)((s) >> (x) & 0X1L))
#define SET_IS_SUBSET1(s1,s2)	((BOOLEAN)!((s1) & ~(s2)))
#define SET_IS_SUBSET2(s1,s2)	((BOOLEAN)!((s2) & ~(s1)))

/*
 *	Definition of compiler constants
 */

#define SYSTEM_ALIGN		8
#define SYSTEM_MASK		(~(SYSTEM_ALIGN - 1))

/*----------------------------------------------------------------------*/

/*
 *	Definition of compiler types
 */

typedef ADDRESS		OPAQUE;

/*
 *	Definition of compiler variables
 */

extern unsigned int SYSTEM_argc;  /* number of command line arguments */
extern char **SYSTEM_argv;	  /* command line arguments */
extern char **SYSTEM_envp;	  /* environment */

extern FILE *SYSTEM_stdin;	  /* current input file */
extern FILE *SYSTEM_stdout;	  /* current output file */
extern char termCH;
extern BOOLEAN Done;		  /* status variable for I/O */

extern CARDINAL PElayer;	  /* index for layers in vector */
				  /* expressions  */

extern CONFIGURATION DPUConfig;   /* for reduce functions */

/*----------------------------------------------------------------------*/

/*
 *	Definition of compiler macros and functions
 */

/* Initialization of the main module and exiting of the main module */

extern void INIT_MODULE ARGS((int argc, char *argv[], char *envp[]));
extern void EXIT_MODULE ARGS((void));


/* Initialization of a configuration: if not already done allocate memory */
/* for active_set and set number of PE-layers, set dimensions and compute */
/* number of PEs for configgroups. These values can only be determined at */
/* runtime as they require `nproc', or may depend on dynamic values. */

extern void SET_CONFIG_BOUNDS ARGS((CONFIGURATION *config, CARDINAL dim,
				    INTEGER lwb, INTEGER upb));
#if 0
#define INIT_CONFIGURATION(config) \
if (config.configgroup->activeset==NULL)\
{ \
     config.configgroup->layers =\
	  (config.configgroup->VPEs + nproc -1) / nproc;\
     config.configgroup->activeset = (ACTIVESET *) \
	  p_malloc(config.configgroup->layers * sizeof(ACTIVESET));\
     for (PElayer=0; PElayer<(config.configgroup->layers - 1); PElayer++)\
	  config.configgroup->activeset[PElayer] = TRUE;\
     config.configgroup->activeset[PElayer] = \
	  (iproc<(config.configgroup->VPEs - PElayer*nproc));\
}

#else
  extern void INIT_CONFIGURATION ARGS((CONFIGURATION config));
#endif

/* Initialize values of connection struct that depend on nproc */

extern void INIT_CONNECTION ARGS((CONNECTION *conn, CONFIGURATION));

/* Remember computed connections in container lists, has to be called */
/* inside a PElayer-loop and only on those PEs that establish a connection. */
/* The source of a connection is given by the ID of all active PEs. */

extern void REGISTER_LINK ARGS((CONNECTION *conn, plural CARDINAL *dest));

/* Read connection target from container lists, returns NULL if end of list */
/* is reached, then the next call will again return the first target. */
/* Call inside a PElayer-loop */

extern VPE_POS_PTR GET_LINK ARGS((CONNECTION *conn));

/* Same for parameterized connections */

extern void INIT_PARAMETER_CONNECTION ARGS((PARAMETER_CONNECTION *conn, INTEGER index, CONFIGURATION config));
extern void REGISTER_PARAMETER_LINK ARGS((PARAMETER_CONNECTION *conn, INTEGER index, plural CARDINAL dest[]));
extern VPE_POS_PTR GET_PARAMETER_LINK ARGS((PARAMETER_CONNECTION *conn, INTEGER index));

/* Access plural records or arrays on single PEs, act much like `proc'. */

extern void psend ARGS((char target[], CARDINAL id, plural char source[], CARDINAL size));
extern void pfetch ARGS ((plural char target[], CARDINAL id, char source[], CARDINAL size));

extern void reduce_DPU ARGS((ACTIVESET actives, plural char *values, void (*red_func)(), CARDINAL size));

/* Initialization of a global vector variable. Allocate memory for it. The */
/* number of PE-layers can only be determined at runtime as it requires */
/* `nproc'. */

#define INIT_VAR(var,config,type) \
       var = (type *) p_malloc(config.configgroup->layers * sizeof(type))


/* Header for loop over all PElayers of a config group */
#define FOR_ALL(config) \
     for (PElayer=0; PElayer<config.configgroup->layers; PElayer++)

/* Header for statement group working on the activeset of config group */ 
#define IF_ACTIVE(config) \
     if (config.configgroup->activeset[PElayer])

/* Build the loop header of a loop around a statement group working only */
/* on one configuration group. */

#define FOR_ALL_ACTIVE(config) \
     FOR_ALL(config)\
	  IF_ACTIVE(config)


/* Open an environment with a local active-set for IF, REPEAT, WHILE, */
/* LOOP, etc.  Copy the active-set to the local active-set and change the */
/* current active-set to the local one. OPEN_ENV must be called immediatly */
/* after an open brace as there are local variables declared. */

#define OPEN_ENV(config,n) \
     ACTIVESET activesetn[config.configgroup->layers], *oldactivesetn;\
     oldactivesetn = config.configgroup->activeset;\
     for (PElayer=0; PElayer<config.configgroup->layers; PElayer++)\
	  activesetn[PElayer] = oldactivesetn[PElayer];\
     config.configgroup->activeset = activesetn;

/* Open an environment with two local active-sets for IF with ELSIFs and */
/* CASE. Copy the active-set to both local active-sets and change the */
/* current active-set to a local one. OPEN_EXTENV must be called immediatly */
/* after an open brace as there are local variables declared. */

#define OPEN_EXTENV(config) \
     ACTIVESET activeset[config.configgroup->layers], *oldactiveset,\
               elseactiveset[config.configgroup->layers];\
     oldactiveset = config.configgroup->activeset;\
     for (PElayer=0; PElayer<config.configgroup->layers; PElayer++)\
	  elseactiveset[PElayer] = activeset[PElayer] = oldactiveset[PElayer];\
     config.configgroup->activeset = activeset;


/* Invert the active-set for the ELSE branch of an IF. This is also used by */
/* ELSIF branches. */

#define ELSE_ENV(config,n) \
     for (PElayer=0; PElayer<config.configgroup->layers; PElayer++)\
	  activesetn[PElayer] = (!activesetn[PElayer]) && oldactivesetn[PElayer];

/* Compute the active-set for an ELSIF branch. */

#define ELSIF_ENV(config) \
     for (PElayer=0; PElayer<config.configgroup->layers; PElayer++)\
	  elseactiveset[PElayer] = activeset[PElayer] = \
               (!activeset[PElayer]) && elseactiveset[PElayer];


/* Close the current environment and restore the active-set to the one */
/* before the environment. The memory for the active-set is automatically */
/* released by the closing brace that should follow CLOSE_ENV. */

#define CLOSE_ENV(config,n) \
     config.configgroup->activeset = oldactivesetn;



#ifdef StackAlloc

#define OPEN_ARRAY_LOCALS	char *FREE_POINTER;
#define ALLOC_OPEN_ARRAYS(size, arrays) \
	FREE_POINTER = alloca((int)((size) + (arrays) * (SYSTEM_ALIGN - 1)));
#define FREE_OPEN_ARRAYS

#else

#define OPEN_ARRAY_LOCALS	char *BLOCK_POINTER, *FREE_POINTER;
#define ALLOC_OPEN_ARRAYS(size, arrays) \
	BLOCK_POINTER = FREE_POINTER = \
		malloc((unsigned)((size) + (arrays) * (SYSTEM_ALIGN - 1)));
#define FREE_OPEN_ARRAYS	free(BLOCK_POINTER);

#endif

#ifdef BCOPY
#define COPY_OPEN_ARRAY(array, elems, type) \
	{ int ARRAY_SIZE = elems * sizeof(type); \
	  bcopy((char *)array, FREE_POINTER, ARRAY_SIZE); \
	  array = FREE_POINTER; \
	  FREE_POINTER += (ARRAY_SIZE + (SYSTEM_ALIGN - 1)) & SYSTEM_MASK; \
	}
#else
#define COPY_OPEN_ARRAY(array, elems, type) \
	{ int ARRAY_SIZE = elems * sizeof(type); \
	  array = (type *)memcpy(FREE_POINTER, (char *)array, ARRAY_SIZE); \
	  FREE_POINTER += (ARRAY_SIZE + (SYSTEM_ALIGN - 1)) & SYSTEM_MASK; \
	}
#endif

#define FOR_LIMIT_UP(last, step, min) \
	((last) < (min) + ((step) - 1) ? (min) : (last) - ((step) - 1))

#define FOR_LIMIT_DOWN(last, step, max) \
	((last) > (max) + ((step) + 1) ? (max) : (last) - ((step) + 1))


/* Given the ID of a virtual PE, compute the number of the PE layer and the */
/* number of the physical PE. These will change when using 2-dim. layout for */
/* PEs to use xnet. #### */

#define ID2layer(id)   ((id) / nproc)
#define ID2proc(id)    ((id) % nproc)

extern void CaseError ARGS((char file[], int line));
extern void ReturnError ARGS((char file[], int line));
extern void ReturnErrorV ARGS((CONFIGURATION config, char file[], int line));

/*
 *	Definition of library functions
 */

#ifdef StackAlloc
extern char *alloca();
#else
/*
extern char *malloc();
extern void free();
*/
#endif

#ifndef BCOPY
/*extern char *memcpy();*/
#endif
/*extern char *strncpy();*/
extern void exit();
