/***********************************************************************
 *  Projectname:	Parallaxis-III
 *  Filename:		p3config.h
 *  Last change:	Mon May 29 20:46:58 1995
 *  Author:  		Hartmut Keller
 ***********************************************************************
 *  Description:
 *  ------------
 *  This file contains the default values that may have to be modified
 *  when installing Parallaxis, e.g. some path definitions. Most of
 *  the macros can also be set by giving appropriate -D-switches to
 *  the comiler call, for example in the Makefile. On many systems,
 *  the given default values usually can be used without any change.
 ***********************************************************************/


/*----------- Common settings --------------------*/

/*
 * Set PATHSEPARATOR to the separator character for subdirectories in
 * paths. On MS-DOS-like file systems you may want to change this to
 * the character '\'.
 */
#ifndef PATHSEPARATOR
#define PATHSEPARATOR '/'
#endif

/*
 * Set VALUESEPARATORS to a string of characters that should be
 * allowed as separators for path lists. On MS-DOS-like file systems
 * you may want to leave out the colon `:' as this is usually the
 * drive separator in pathnames.
 */
#ifndef VALUESEPARATORS
#define VALUESEPARATORS ":;, "
#endif



/*----------- Settings for C-version (default version) -----------*/

/*
 * Set P3C to the name of the Parallaxis->C compiler.
 */
#ifndef P3C
#define P3C "p3c"
#endif

/*
 * Set P3CC to the name of the back-end C-compiler (usually "cc" or
 * "gcc"). 
 */
#ifndef P3CC
#define P3CC "cc"
#endif

/*
 * Set P3INC to the default path where the includes for C and
 * Parallaxis can be found. This can also be a list of paths,
 * seperated by one of the characters of VALUESEPARATORS.
 */
#ifndef P3INC
#define P3INC "/usr/local/include"
#endif

/*
 * Set P3LIB to the default path where the libraries for C and
 * Parallaxis can be found. This can also be a list of paths,
 * seperated by one of the characters of VALUESEPARATORS.
 */
#ifndef P3LIB
#define P3LIB "/usr/local/lib"
#endif

/*
 * Set STANDARDLIB to the default libraries that should be linked with
 * Parallaxis programs compiled with the C-version.
 */
#ifndef STANDARDLIB
#define STANDARDLIB "-lp3 -lm"
#endif



/*----------- Settings for PVM-version -----------*/

#ifdef PVM_VERSION
/*
 * Set P3C_PVM to the name of the Parallaxis->PVM compiler.
 */
#ifndef P3C_PVM
#define P3C_PVM "p3c_PVM"
#endif

/*
 * Set P3CC_PVM to the name of the back-end C-compiler (usually "cc"
 * or "gcc"). 
 */
#ifndef P3CC_PVM
#define P3CC_PVM "cc"
#endif

/*
 * Set P3INC_PVM to the default path where the includes for C, PVM and
 * Parallaxis can be found. This can also be a list of paths,
 * seperated by one of the characters of VALUESEPARATORS.
 */
#ifndef P3INC_PVM
#define P3INC_PVM "/usr/local/include,$PVM_ROOT/include"
#endif

/*
 * Set P3LIB_PVM to the default path where the libraries for C, PVM
 * and Parallaxis can be found. This can also be a list of paths,
 * seperated by one of the characters of VALUESEPARATORS.
 */
#ifndef P3LIB_PVM
#define P3LIB_PVM "/usr/local/lib,$PVM_ROOT/lib/$PVM_ARCH"
#endif

/*
 * Set STANDARDLIB_PVM to the default libraries that should be linked
 * with Parallaxis programs compiled with the PVM-version.
 */
#ifndef STANDARDLIB_PVM
#define STANDARDLIB_PVM "-lp3_PVM -lm -lpvm3"
#endif
#endif /* PVM_VERSION */



/*----------- Settings for MPL-version -----------*/

#ifdef MPL_VERSION
/*
 * Set P3C_MPL to the name of the Parallaxis->MPL compiler
 */
#ifndef P3C_MPL
#define P3C_MPL "p3c_MPL"
#endif

/*
 * Set P3CC_MPL to the name of the back-end MPL-compiler.
 */
#ifndef P3CC_MPL
#define P3CC_MPL "mpl_cc"
#endif

/*
 * Set P3INC_MPL to the default path where the includes for MPL and
 * Parallaxis can be found. This can also be a list of paths,
 * seperated by one of the characters of VALUESEPARATORS.
 */
#ifndef P3INC_MPL
#define P3INC_MPL "/usr/local/include"
#endif

/*
 * Set P3LIB_MPL to the default path where the libraries for MPL
 * and Parallaxis can be found. This can also be a list of paths,
 * seperated by one of the characters of VALUESEPARATORS.
 */
#ifndef P3LIB_MPL
#define P3LIB_MPL "/usr/local/lib"
#endif

/*
 * Set STANDARDLIB_MPL to the default libraries that should be linked
 * with Parallaxis programs compiled with the MPL-version.
 */
#ifndef STANDARDLIB_MPL
#define STANDARDLIB_MPL "-lp3_MPL -lm"
#endif
#endif /* MPL_VERSION */


/*----------- Settings for C-compiler switches -----------*/

#define sOPTnolink      "-c"
#define sOPTcverbose    "-v"
#define sOPTcdebug      "-g"
#define sOPTcnowarnings "-w"
#define sOPTinclude     "-I"
#define sOPTlibpath     "-L"
#define sOPTlibrary     "-l"
#define sOPToutput      "-o"

