/**********************************************************************/
/*                                                                    */
/* File   : shared_macros.h                                           */
/* Author : Juergen Wakunda                                           */
/*          (wakundju@hermes.informatik.uni-stuttgart.de              */
/* Date   : 18 January 1995                                           */
/*                                                                    */
/* In this little header-file are all types and macros defined,       */
/* that are needed by several modules. They are located all in this   */
/* file and if they change, they have to be changed only here.        */
/*                                                                    */
/* This file is included from SYSTEM_PVM.h, start.c and Graphics.h    */
/*                                                                    */
/**********************************************************************/

#ifndef _shared_macros_h
#define _shared_macros_h
/* ---------------------------------------------------------------------- */
/*                          begin shared_macros                           */
/* ---------------------------------------------------------------------- */

#include <pvm3.h>


/* ------------------------------------------------------------- */
/* definitions for I/O                                           */
/* ------------------------------------------------------------- */

#define INIT_TAG          1
#define TIME_TAG          2
#define IO_TAG            3
#define INTER_TASK_TAG    4         /* for inter-task communication */

#define STD_STRIDE 1      /* standard stride */

#define PVM_PKINTEGER(x,y)  (void) pvm_pklong(x, y, STD_STRIDE)
#define PVM_PKCARDINAL(x,y) (void) pvm_pklong((long*) x, y, STD_STRIDE)
#define PVM_PKBOOLEAN(x,y)  (void) pvm_pkbyte((char*) x, y, STD_STRIDE)
#define PVM_PKREAL(x,y)     (void) pvm_pkdouble(x, y, STD_STRIDE)
#define PVM_PKCHAR(x,y)     (void) pvm_pkbyte((char*) x, y, STD_STRIDE)
#define PVM_PKSTRING(x)     (void) pvm_pkstr((char*) x)
#define PVM_PKBITSET(x,y)   (void) pvm_pklong((long*) x, y, STD_STRIDE)

#define PVM_UPKINTEGER(x,y)  (void) pvm_upklong(x, y, STD_STRIDE)
#define PVM_UPKCARDINAL(x,y) (void) pvm_upklong((long*) x, y, STD_STRIDE)
#define PVM_UPKBOOLEAN(x,y)  (void) pvm_upkbyte((char*) x, y, STD_STRIDE)
#define PVM_UPKREAL(x,y)     (void) pvm_upkdouble(x, y, STD_STRIDE)
#define PVM_UPKCHAR(x,y)     (void) pvm_upkbyte((char*) x, y, STD_STRIDE)
#define PVM_UPKSTRING(x)     (void) pvm_upkstr((char*) x)
#define PVM_UPKBITSET(x,y)   (void) pvm_upklong((long*) x, y, STD_STRIDE)

#define PVM_MCAST(x,y,z)     (void) pvm_mcast(x, y, z)

#define PVM_PKSIG(SIG) { int sig = SIG; pvm_pkint(&sig, 1, 1); }

/* ------------------------------------------------------------ */
/* message-signals for I/O   */

/* file-I/O-signals          */
#define OpenInputSig          1
#define OpenOutputSig         2
#define CloseInputSig         3
#define CloseOutputSig        4
/* write-signals             */
#define WriteCardSig          5
#define WriteIntSig           6
#define WriteRealSig          7
#define WriteFixPtSig         8
#define WriteBoolSig          9
#define WriteCharSig         10
#define WriteStringSig       11
#define WriteCardVectSig     12
#define WriteIntVectSig      13
#define WriteRealVectSig     14
#define WriteFixPtVectSig    15
#define WriteBoolVectSig     16
#define WriteCharVectSig     17
#define WriteStringVectSig   18
#define WriteLnSig           19
/* two error-signals         */
#define CaseErrorSig         20
#define ReturnErrorSig       21
/* read-signals              */
#define ReadCardSig          22
#define ReadIntSig           23
#define ReadRealSig          24
#define ReadBoolSig          25
#define ReadCharSig          26
#define ReadStringSig        27
#define ReadCardVectSig      28
#define ReadIntVectSig       29
#define ReadRealVectSig      30
#define ReadBoolVectSig      31
#define ReadCharVectSig      32
#define ReadStringVectSig    33
/* two error-signals         */
#define IdRangeErrorSig      34
#define PeRangeErrorSig      35
/* graphic-signals           */
#define BeginGraphicsSig     36
#define OpenWindowSig        37
#define CloseWindowSig       38
#define SelectWindowSig      39
#define GetScreenSizeSig     40
#define GetWindowSizeSig     41
#define SetAreaSig           42
#define SetgAreaSig          43
#define SetbAreaSig          44
#define SetAreaXYZSig        45
#define SetgAreaXYZSig       46
#define SetbAreaXYZSig       47
#define SetLineSig           48
#define SetgLineSig          49
#define SetbLineSig          50
#define SetLineZSig          51
#define SetgLineZSig         52
#define SetbLineZSig         53
#define SetColumnSig         54
#define SetgColumnSig        55
#define SetbColumnSig        56
#define SetColumnZSig        57
#define SetgColumnZSig       58
#define SetbColumnZSig       59
#define SetColorSig          60
#define SetgColorSig         61
#define SetbColorSig         62
#define SetPixelSig          63
#define GetPixelSig          64
#define LineSig              65
#define DrawAtSig            66
#define DrawSig              67
#define DrawStringSig        68
#define DrawIntSig           69
#define DrawCardSig          70
#define DrawRealSig          71
#define DrawFixPtSig         72
#define DrawBoolSig          73
#define DrawLnSig            74

/* two termination-signals   */
#define TermErrorSig         98  /* termination with error */
#define TermSig              99  /* normal termination     */



/* ---------------------------------------------------------------------- */
/*                           end shared_macros                            */
/* ---------------------------------------------------------------------- */
#endif
