/*************************************************************************
 *
 * Parallaxis Compiler Projects
 * ----------------------------
 *
 * name  :  FileIO.h
 * desc  :  supply layer on top of C stream io
 * author:  lampke
 *
 ************************************************************************/

#ifndef FileIO_h
#define FileIO_h

#include <stdio.h>
#include "parallaxis.h"

#ifdef MSDOS
#define PATHDELIM       "\\"
#define PATHDELIMCHAR   '\\'
#define PATHLISTDELIM   ';'
#else
#define PATHDELIM       "/"
#define PATHDELIMCHAR   '/'
#define PATHLISTDELIM   ' '
#endif

#define DEFFILE_EXT     ".pd"
#define MODFILE_EXT     ".pm"

#define CHEADER_EXT     ".h"
#define CMODULE_EXT     ".c"
#define CROSSFILE_EXT   ".px"
#define MPLHEADER_EXT   ".h"
#define MPLMODULE_EXT   ".m"

#define FILESTOHANDLE   64

#define FileIO_ERRORHANDLE FILESTOHANDLE

/*************************************************************************
 * name :  OpenInput
 * in   :  file 'FileName' will be opened for input
 * out  :  FILE * handle to open file
 *         NULL, if error
 ************************************************************************/
extern FILE *FileIO_OpenInput(char *FileName);

/*************************************************************************
 * name :  OpenInputHandle
 * in   :  file 'FileName' will be opened for input
 * out  :  int handle to open file
 *         FileIO_ERRORHANDLE, if error
 ************************************************************************/
extern int FileIO_OpenInputHandle(char *FileName);

/*************************************************************************
 * name :  OpenOutput
 * in   :  file 'FileName' will be opened for output
 * out  :  FILE * handle to open file
 *         NULL, if error
 ************************************************************************/
extern FILE *FileIO_OpenOutput(char *FileName);

/*************************************************************************
 * name :  OpenOutputHandle
 * in   :  file 'FileName' will be opened for output
 * out  :  int handle to open file
 *         FileIO_ERRORHANDLE, if error
 ************************************************************************/
extern int FileIO_OpenOutputHandle(char *FileName);

/*************************************************************************
 * name :  Close
 * in   :  FILE * handle of file to be closed
 * out  :
 ************************************************************************/
extern void FileIO_Close(FILE *File);

/*************************************************************************
 * name :  CloseHandle
 * in   :  int handle of file to be closed
 * out  :
 ************************************************************************/
extern void FileIO_CloseHandle(int File);

/*************************************************************************
 * name :  CloseAll
 * in   :
 * out  :  number of files, which have been closed
 ************************************************************************/
extern SHORT FileIO_CloseAll(void);

extern void FileIO_Init(void);

#endif
