/*************************************************************************
 *
 * Parallaxis Compiler Projects
 * ----------------------------
 *
 * name  :  LoopStack.c
 * desc  :  stack for parallel statements nested in LOOP
 * author:  weiss
 *
 ************************************************************************/

#include <stdlib.h>
#include "DynArray.h"			  /* Reuse-Library */
#include "Idents.h"			  /* Reuse-Library */
#include "parallaxis.h"
#include "LoopStack.h"

static BOOLEAN isInitialised = FALSE;

/*
 * local
 */

#define InitialSize 16

static BOOLEAN isInLOOP = FALSE;

static long StackPtr, StackSize;

static tLoopStack *Stack;

/*
 * global function implementation
 */

/*---------------------------------------------------------------------------*/

void LoopStack_Push(tIdent Ident, tIdent parloop_config)
{
  if (StackPtr == StackSize) {
    ExtendArray((char **) &Stack, &StackSize, sizeof(tLoopStack));
  }
  Stack[StackPtr].parloop_config = parloop_config;
  Stack[StackPtr].label = Ident;
  Stack[StackPtr].config_num = 0;
  Stack[StackPtr].first = NULL;
  Stack[StackPtr].last = NULL;
  StackPtr++;
  isInLOOP = TRUE;
}

/*---------------------------------------------------------------------------*/

void LoopStack_OpenEnv(tIdent Ident)
{
  if (isInLOOP) {
    tLoopStack *top = LoopStack_Top();
    tConfig_List *conf = (tConfig_List *) malloc(sizeof(tConfig_List));

    conf->config = Ident;
    conf->next = top->first;
    if (conf->next != NULL) {
      conf->next->prev = conf;
    }
    conf->prev = NULL;
    top->config_num++;
    top->first = conf;
    if (top->last == NULL) {
      top->last = top->first;
    }
  }
}

/*---------------------------------------------------------------------------*/

void LoopStack_CloseEnv()
{
  if (isInLOOP) {
    tLoopStack *top = LoopStack_Top();
    tConfig_List *conf = top->first->next;

    free(top->first);
    top->config_num--;
    top->first = conf;
    if (top->first != NULL) {
      top->first->prev = NULL;
    } else {
      top->last = NULL;
    }
  }
}

/*---------------------------------------------------------------------------*/

tIdent LoopStack_GetTopLabel()
{
  return LoopStack_Top()->label;
}

/*---------------------------------------------------------------------------*/

ULONG LoopStack_GetTopConfigNum()
{
  if (isInLOOP) {
    return LoopStack_Top()->config_num;
  } else {
    return 0;
  }
}

/*---------------------------------------------------------------------------*/

tConfig_List *LoopStack_GetTopFirstConfig()
{
  return LoopStack_Top()->first;
}

/*---------------------------------------------------------------------------*/

tIdent LoopStack_GetTopActConfig(tConfig_List **list)
{
  tConfig_List *act = *list;

  *list = act->next;
  return act->config;
}

/*---------------------------------------------------------------------------*/

BOOLEAN LoopStack_IsTopLastConfig(tConfig_List **list)
{
  tIdent config = (*list)->config;
  tConfig_List *loop_list = (*list)->next;

  while (loop_list != NULL) {
    if (config == loop_list->config) {
      *list = (*list)->next;
      return FALSE;
    }
    loop_list = loop_list->next;
  }
  return TRUE;
}

/*---------------------------------------------------------------------------*/

tIdent LoopStack_GetTopLoopConfig()
{
  return LoopStack_Top()->parloop_config;
}

/*---------------------------------------------------------------------------*/

ULONG LoopStack_GetTopLastConfig(tConfig_List **list)
{
  tLoopStack *top = LoopStack_Top();
  tConfig_List *loop_list = top->last;
  ULONG num = 1;

  while (loop_list != NULL && loop_list->config != top->parloop_config) {
    num++;
    loop_list = loop_list->prev;
  }
  *list = loop_list;
  return num;
}  

/*---------------------------------------------------------------------------*/

ULONG LoopStack_GetTopNextLoopConfig(tConfig_List **list, ULONG num)
{
  tLoopStack *top = LoopStack_Top();
  tConfig_List *loop_list = (*list)->prev;

  num++;
  while (loop_list != NULL && loop_list->config != top->parloop_config) {
    num++;
    loop_list = loop_list->prev;
  }
  *list = loop_list;
  return num;
}  
  
/*---------------------------------------------------------------------------*/

void LoopStack_Pop()
{
  StackPtr--;
  if (StackPtr == 0) {
    isInLOOP = FALSE;
  }
}

/*---------------------------------------------------------------------------*/

tLoopStack *LoopStack_Top()
{
  if (StackPtr > 0) {
    return &Stack[StackPtr-1];
  } else {
    return NULL;
  }
}

/*---------------------------------------------------------------------------*/

void LoopStack_Free()
{
  ReleaseArray((char **) &Stack, &StackSize, sizeof(tLoopStack));
}

/*---------------------------------------------------------------------------*/

void LoopStack_Init()
{
  if (!isInitialised) {
    isInitialised = TRUE;

    StackSize = InitialSize;
    StackPtr = 0;
    MakeArray((char **) &Stack, &StackSize, sizeof(tLoopStack));
  }
}
