/*************************************************************************
 *
 * Parallaxis Compiler Projects
 * ----------------------------
 *
 * name  :  Source.c
 * desc  :  Source code is read in blocks when ever scanner
 *          need a new junk.
 *          Actual filename is stored in Scanner.Attribute.Position
 * author:  lampke
 *
 ************************************************************************/

#ifndef Source_h
#define Source_h

#if defined __STDC__ | defined __cplusplus
#define ARGS(parameters)       parameters
#else
#define ARGS(parameters)       ()
#endif

extern int  BeginSource  ARGS((char * FileName));

   /*
      BeginSource is called from the scanner to open files.
      If not called input is read form standard input.
   */

extern int  GetLine      ARGS((int File, char * Buffer, int Size));

   /*
      GetLine is called to fill a buffer starting at address 'Buffer'
      with a block of maximal 'Size' characters. Lines are terminated
      by newline characters (ASCII = 0xa). GetLine returns the number
      of characters transferred. Reasonable block sizes are between 128
      and 2048 or the length of a line. Smaller block sizes -
      especially block size 1 - will drastically slow down the scanner.
   */

extern void CloseSource  ARGS((int File));

   /*
      CloseSource is called from the scanner at end of file respectively
      at end of input. It can be used to close files.
   */

extern void Source_Init( void );

#endif
