/*************************************************************************
 *
 * Parallaxis Compiler Projects
 * ----------------------------
 *
 * name  :  GeneralIO.h
 * desc  :  supply basic functions for file handling
 * author:  lampke
 *
 ************************************************************************/

#ifndef GeneralIO_h
#define GeneralIO_h

#include <stdio.h>
#include "Idents.h"			  /* Reuse-Library */
#include "parallaxis.h"
#include "Positions.h"

/*************************************************************************
 * name :  AddSearchPath
 * in   :  'path' will be added to internal path table
 * out  :
 ************************************************************************/
extern void GeneralIO_AddSearchPath(char *path);

/*************************************************************************
 * name :  CheckOpen
 * in   :  Pointer to stream and filename
 * out  :  function exits, if File is not open
 ************************************************************************/
extern void GeneralIO_CheckOpen(FILE *File, char *FileName);

/*************************************************************************
 * name :  CheckModFile
 * in   :  Program file name with or without extension
 * out  :  TRUE and full filename, if program file is available
 *         FALSE otherwise
 ************************************************************************/
extern BOOLEAN GeneralIO_CheckModFile(char *ProgName, char *FileName);

/*************************************************************************
 * name :  CheckDefFile
 * in   :  ModuleName, for which the definition file should be opened
 * out  :  TRUE and full filename, if definition file is available
 *         FALSE otherwise
 ************************************************************************/
extern BOOLEAN GeneralIO_CheckDefFile(tIdent ModuleName, char *FileName);

/*************************************************************************
 * name :  RealName
 * in   :  whole name of the file
 * out  :  name of the file without path and extension
 ************************************************************************/
extern tIdent GeneralIO_RealName(tIdent Ident, tPosition Position);

/*************************************************************************
 * name :  OpenCHeader
 * in   :
 * out  :
 ************************************************************************/
extern FILE *GeneralIO_OpenCHeader(tIdent Ident);

/*************************************************************************
 * name :  OpenCModule
 * in   :
 * out  :
 ************************************************************************/
extern FILE *GeneralIO_OpenCModule(tIdent Ident);

/*************************************************************************
 * name :  OpenCrossreferenceFile
 * in   :
 * out  :
 ************************************************************************/
extern FILE *GeneralIO_OpenCrossreferenceFile(tIdent Ident);

/*************************************************************************
 * name :  OpenMPLHeader
 * in   :
 * out  :
 ************************************************************************/
extern FILE *GeneralIO_OpenMPLHeader(tIdent Ident);

/*************************************************************************
 * name :  OpenMPLModule
 * in   :
 * out  :
 ************************************************************************/
extern FILE *GeneralIO_OpenMPLModule(tIdent Ident);

extern void GeneralIO_Init(void);

#endif
