/*
**  Interface module between Parallaxis-III and X11R5 Graphics Library.
**
**  Written by Joerg Stippa in September 1994.
**  Contact address: stippa@hermes.informatik.uni-stuttgart.de
**		 or: stippa@isa.de
**
**  Modified file "Graphics_PVM" for use in PVM-Version of Parallaxis-III 
**  by Juergen Wakunda in February 1995.
**  Contact address: wakundju@hermes.informatik.uni-stuttgart.de
*/


#include "SYSTEM_PVM.h"
#include "Graphics.h"	/* Definitions for the graphics routines */
#include <malloc.h>

/* Global variable for error indication in Graphics-Module */
CARDINAL Graphics_GraphicsError;


/*
**  Create a new window with given size in width x height.
**  Name the window as 'title'.
**  Positioning is handled by the user via mouse.
*/
Graphics_window Graphics_OpenWindow
#ifdef __STDC__
(CHAR title[], CARDINAL O_1, CARDINAL width, CARDINAL height)
#else
(title, O_1, width, height)
CHAR title[];
CARDINAL O_1;
CARDINAL width, height;
#endif
{
  Graphics_window WindowId;

  if (me == 0) {
    (void) pvm_initsend(PvmDataDefault);
    PVM_PKSIG(OpenWindowSig);
    PVM_PKSTRING(title);
    PVM_PKCARDINAL(&O_1, 1);
    PVM_PKCARDINAL(&width, 1);
    PVM_PKCARDINAL(&height, 1);
    (void) pvm_send(io_tid, IO_TAG);
  }

  (void) pvm_recv(io_tid, IO_TAG);
  PVM_UPKCARDINAL(&WindowId, 1);
  PVM_UPKCARDINAL(&Graphics_GraphicsError, 1);
  return WindowId;
}


/*
**  Destroy a window and release all memory
*/
void Graphics_CloseWindow
#ifdef __STDC__
(CARDINAL window_num)
#else
(window_num)
CARDINAL window_num;
#endif
{
  if (me == 0) {
    (void) pvm_initsend(PvmDataDefault);
    PVM_PKSIG(CloseWindowSig);
    PVM_PKCARDINAL(&window_num, 1);
    (void) pvm_send(io_tid, IO_TAG);
  }

  (void) pvm_recv(io_tid, IO_TAG);
  PVM_UPKCARDINAL(&Graphics_GraphicsError, 1);
}



/*
**  Choose a window to be the active one.
*/

void Graphics_SelectWindow
#ifdef __STDC__
(CARDINAL window_num)
#else
(window_num)
CARDINAL window_num;
#endif
{
  if (me == 0) {
    (void) pvm_initsend(PvmDataDefault);
    PVM_PKSIG(SelectWindowSig);
    PVM_PKCARDINAL(&window_num, 1);
    (void) pvm_send(io_tid, IO_TAG);
  }

  (void) pvm_recv(io_tid, IO_TAG);
  PVM_UPKCARDINAL(&Graphics_GraphicsError, 1);
}



/*
**  Query the X window system for the current screen size
*/
void Graphics_GetScreenSize
#ifdef __STDC__
(CARDINAL *width, CARDINAL *height)
#else
(width, height)
CARDINAL *width, *height;
#endif
{
  if (me == 0) {
    (void) pvm_initsend(PvmDataDefault);
    PVM_PKSIG(GetScreenSizeSig);
    (void) pvm_send(io_tid, IO_TAG);
  }

  (void) pvm_recv(io_tid, IO_TAG);
  PVM_UPKCARDINAL(width, 1);
  PVM_UPKCARDINAL(height, 1);
  PVM_UPKCARDINAL(&Graphics_GraphicsError, 1);
}



/*
**  Query the system for the size of the active window.
*/

void Graphics_GetWindowSize
#ifdef __STDC__
(CARDINAL *width, CARDINAL *height)
#else
(width, height)
CARDINAL *width, *height;
#endif
{
  if (me == 0) {
    (void) pvm_initsend(PvmDataDefault);
    PVM_PKSIG(GetWindowSizeSig);
    (void) pvm_send(io_tid, IO_TAG);
  }

  (void) pvm_recv(io_tid, IO_TAG);
  PVM_UPKCARDINAL(width, 1);
  PVM_UPKCARDINAL(height, 1);
  PVM_UPKCARDINAL(&Graphics_GraphicsError, 1);
}


/* - - - - - - - - - - - - - - - - - - - - - - - */

/* here are some help-functions which are needed several times */

/* Packs some data of a CONFIGURATION in a message (Name, Rank, Lower- and */
/* Upperbounds). */
/* This procedure makes no pvm_initsend and pvm_send. It only packs the */
/* data. The I/O-process-equivalent is UnpackConfiguration. */

void PackConfiguration
#ifdef __STDC__
(CONFIGURATION Config)
#else
(Config)
CONFIGURATION Config;
#endif
{
  CARDINAL Rank, i;

  PVM_PKSTRING(Config.Name);
  Rank = RANK(Config);
  PVM_PKCARDINAL(&Rank, 1);

  for (i = 0; i < RANK(Config); i++) {
    INTEGER lower, upper;

    lower = LOWER(Config, RANK(Config) - i);
    upper = UPPER(Config, RANK(Config) - i);
    PVM_PKINTEGER(&lower, 1);
    PVM_PKINTEGER(&upper, 1);
  }
}


/* Composes and sends a message containing the activestatus of all PEs of a */
/* CONFIGURATION. */
/* The I/O-process-equivalent is ReceiveActiveStatus. */

void SendActiveStatus
#ifdef __STDC__
(CONFIGURATION Config)
#else
(Config)
CONFIGURATION Config;
#endif
{
  CARDINAL task_index, i;
  BOOLEAN BoolTrue  = TRUE;
  BOOLEAN BoolFalse = FALSE;
  BOOLEAN BoolValue;

  if (me == 0) {
    (void) pvm_initsend(PvmDataDefault);

    FOR_ALL_LOCAL_PEs(Config) {
      if (LOCAL_IS_ACTIVE(Config, PE)) {
	PVM_PKBOOLEAN(&BoolTrue, 1);
      } else {
	PVM_PKBOOLEAN(&BoolFalse, 1);
      }
    }

    for (task_index = 1; task_index < Config.u_proc; task_index++) {
      (void) pvm_recv(task_ids[task_index], INTER_TASK_TAG);
      for (i = 0; i < Config.ID_range[task_index].num_PEs; i++) {
	PVM_UPKBOOLEAN(&BoolValue, 1);
	PVM_PKBOOLEAN(&BoolValue, 1);
      }
    }

    (void) pvm_send(io_tid, IO_TAG);
  }


  if (me >= 1 && me < Config.u_proc) {
    (void) pvm_initsend(PvmDataDefault);
    FOR_ALL_LOCAL_PEs(Config) {
      if (LOCAL_IS_ACTIVE(Config, PE)) {
	PVM_PKBOOLEAN(&BoolTrue, 1);
      } else {
	PVM_PKBOOLEAN(&BoolFalse, 1);
      }
    }
    (void) pvm_send(parent_tid, INTER_TASK_TAG);
  }
}


/* Composes and sends a message, containing a RGB-colorvector. */
/* The I/O-process-equivalent is ReceiveColorVector. */

void SendColorVector
#ifdef __STDC__
(ImageIO_color *c, CONFIGURATION Config)
#else
(c, Config)
ImageIO_color *c;
CONFIGURATION Config;
#endif
{
  CARDINAL task_index, i, dummy;

  if (me == 0) {
    (void) pvm_initsend(PvmDataDefault);
    FOR_ALL_LOCAL_PEs(Config) {
      PVM_PKCARDINAL(&c[PE].red, 1);
      PVM_PKCARDINAL(&c[PE].green, 1);
      PVM_PKCARDINAL(&c[PE].blue, 1);
    }

    for (task_index = 1; task_index < Config.u_proc; task_index++) {
      (void) pvm_recv(task_ids[task_index], INTER_TASK_TAG);
      for (i = 0; i < Config.ID_range[task_index].num_PEs; i++) {
	PVM_UPKCARDINAL(&dummy, 1); PVM_PKCARDINAL(&dummy, 1);
	PVM_UPKCARDINAL(&dummy, 1); PVM_PKCARDINAL(&dummy, 1);
	PVM_UPKCARDINAL(&dummy, 1); PVM_PKCARDINAL(&dummy, 1);
      }
    }
    (void) pvm_send(io_tid, IO_TAG);
  }

  if (me >= 1 && me < Config.u_proc) {
    (void) pvm_initsend(PvmDataDefault);
    FOR_ALL_LOCAL_PEs(Config) {
      PVM_PKCARDINAL(&c[PE].red, 1);
      PVM_PKCARDINAL(&c[PE].green, 1);
      PVM_PKCARDINAL(&c[PE].blue, 1);
    }
    (void) pvm_send(parent_tid, INTER_TASK_TAG);
  }
}


/* Composes and sends a message, containing a grayvector. */
/* The I/O-process-equivalent is ReceiveGrayVector. */

void SendGrayVector
#ifdef __STDC__
(ImageIO_gray *c, CONFIGURATION Config)
#else
(c, Config)
ImageIO_gray *c;
CONFIGURATION Config;
#endif
{
  CARDINAL task_index, i, dummy;

  if (me == 0) {
    (void) pvm_initsend(PvmDataDefault);
    FOR_ALL_LOCAL_PEs(Config) {
      PVM_PKCARDINAL(&c[PE], 1);
    }

    for (task_index = 1; task_index < Config.u_proc; task_index++) {
      (void) pvm_recv(task_ids[task_index], INTER_TASK_TAG);
      for (i = 0; i < Config.ID_range[task_index].num_PEs; i++) {
	PVM_UPKCARDINAL(&dummy, 1); PVM_PKCARDINAL(&dummy, 1);
      }
    }
    (void) pvm_send(io_tid, IO_TAG);
  }

  if (me >= 1 && me < Config.u_proc) {
    (void) pvm_initsend(PvmDataDefault);
    FOR_ALL_LOCAL_PEs(Config) {
      PVM_PKCARDINAL(&c[PE], 1);
    }
    (void) pvm_send(parent_tid, INTER_TASK_TAG);
  }
}


/* Composes and sends a message, containing a binaryvector. */
/* The I/O-process-equivalent is ReceiveBinaryVector. */

void SendBinaryVector
#ifdef __STDC__
(ImageIO_binary *c, CONFIGURATION Config)
#else
(c, Config)
ImageIO_binary *c;
CONFIGURATION Config;
#endif
{
  CARDINAL task_index, i;
  BOOLEAN dummy;

  if (me == 0) {
    (void) pvm_initsend(PvmDataDefault);
    FOR_ALL_LOCAL_PEs(Config) {
      PVM_PKBOOLEAN(&c[PE], 1);
    }

    for (task_index = 1; task_index < Config.u_proc; task_index++) {
      (void) pvm_recv(task_ids[task_index], INTER_TASK_TAG);
      for (i = 0; i < Config.ID_range[task_index].num_PEs; i++) {
	PVM_UPKBOOLEAN(&dummy, 1); PVM_PKBOOLEAN(&dummy, 1);
      }
    }
    (void) pvm_send(io_tid, IO_TAG);
  }

  if (me >= 1 && me < Config.u_proc) {
    (void) pvm_initsend(PvmDataDefault);
    FOR_ALL_LOCAL_PEs(Config) {
      PVM_PKBOOLEAN(&c[PE], 1);
    }
    (void) pvm_send(parent_tid, INTER_TASK_TAG);
  }
}


/* - - - - - - - - - - - - - - - - - - - - - - - */


/*
**  Receive a single linear array of color information in c.
**  PE configuration is given in OPEN_VECTOR.
**  Copy color info and make it suitable for X. Finally call common code
**  for filling an area (max 2-dim.) appropriately.
**  Color's RGB range between 'zero' and 255 inclusive.
*/

void Graphics_SetArea
#ifdef __STDC__
(ImageIO_color c[], CONFIGURATION OPEN_VECTOR)
#else
(c, OPEN_VECTOR)
ImageIO_color c[];
CONFIGURATION OPEN_VECTOR;
#endif
{
  if (me == 0) {
    (void) pvm_initsend(PvmDataDefault);
    PVM_PKSIG(SetAreaSig);
    PackConfiguration(OPEN_VECTOR);
    (void) pvm_send(io_tid, IO_TAG);
  }

  SendActiveStatus(OPEN_VECTOR);
  SendColorVector(c, OPEN_VECTOR);

  (void) pvm_recv(io_tid, IO_TAG);
  PVM_UPKCARDINAL(&Graphics_GraphicsError, 1);
}



/*
**  The same as for Graphics_SetArea(), but handles grey scale.
**  Grey ranges between 'zero' and 255 inclusive.
*/

void Graphics_SetgArea
#ifdef __STDC__
(ImageIO_gray c[], CONFIGURATION OPEN_VECTOR)
#else
(c, OPEN_VECTOR)
ImageIO_gray c[];
CONFIGURATION OPEN_VECTOR;
#endif
{
  if (me == 0) {
    (void) pvm_initsend(PvmDataDefault);
    PVM_PKSIG(SetgAreaSig);
    PackConfiguration(OPEN_VECTOR);
    (void) pvm_send(io_tid, IO_TAG);
  }
   
  SendActiveStatus(OPEN_VECTOR);
  SendGrayVector(c, OPEN_VECTOR);

  (void) pvm_recv(io_tid, IO_TAG);
  PVM_UPKCARDINAL(&Graphics_GraphicsError, 1);
}



/*
** Even fewer colors are used, i.e. black and white presentation.
** Anything else is like Graphics_SetArea().
*/

void Graphics_SetbArea
#ifdef __STDC__
(ImageIO_binary c[], CONFIGURATION OPEN_VECTOR)
#else
(c, OPEN_VECTOR)
ImageIO_binary c[];
CONFIGURATION OPEN_VECTOR;
#endif
{
  if (me == 0) {
    (void) pvm_initsend(PvmDataDefault);
    PVM_PKSIG(SetbAreaSig);
    PackConfiguration(OPEN_VECTOR);
    (void) pvm_send(io_tid, IO_TAG);
  }

  SendActiveStatus(OPEN_VECTOR);
  SendBinaryVector(c, OPEN_VECTOR);

  (void) pvm_recv(io_tid, IO_TAG);
  PVM_UPKCARDINAL(&Graphics_GraphicsError, 1);
}



/*
**  Now, some pictures are too small or too big to enjoy.
**  Handle zooming/shrinking.
*/
void Graphics_SetAreaXYZ
#ifdef __STDC__
(ImageIO_color c[], INTEGER x, INTEGER y, INTEGER zoom, CONFIGURATION OPEN_VECTOR)
#else
(c, x, y, zoom, OPEN_VECTOR)
ImageIO_color c[];
INTEGER x;
INTEGER y;
INTEGER zoom;
CONFIGURATION OPEN_VECTOR;
#endif
{
  if (me == 0) {
    (void) pvm_initsend(PvmDataDefault);
    PVM_PKSIG(SetAreaXYZSig);
    PVM_PKINTEGER(&x, 1);
    PVM_PKINTEGER(&y, 1);
    PVM_PKINTEGER(&zoom, 1);
    PackConfiguration(OPEN_VECTOR);
    (void) pvm_send(io_tid, IO_TAG);
  }

  SendActiveStatus(OPEN_VECTOR);
  SendColorVector(c, OPEN_VECTOR);

  (void) pvm_recv(io_tid, IO_TAG);
  PVM_UPKCARDINAL(&Graphics_GraphicsError, 1);
}



/*
**  Have a look at Graphics_SetAreaXYZ.
**  Should be the same code, but with grey scale than with color.
*/

void Graphics_SetgAreaXYZ
#ifdef __STDC__
(ImageIO_gray c[], INTEGER x, INTEGER y, INTEGER zoom, CONFIGURATION OPEN_VECTOR)
#else
(c, x, y, zoom, OPEN_VECTOR)
ImageIO_gray c[];
INTEGER x;
INTEGER y;
INTEGER zoom;
CONFIGURATION OPEN_VECTOR;
#endif
{
  if (me == 0) {
    (void) pvm_initsend(PvmDataDefault);
    PVM_PKSIG(SetgAreaXYZSig);
    PVM_PKINTEGER(&x, 1);
    PVM_PKINTEGER(&y, 1);
    PVM_PKINTEGER(&zoom, 1);
    PackConfiguration(OPEN_VECTOR);
    (void) pvm_send(io_tid, IO_TAG);
  }

  SendActiveStatus(OPEN_VECTOR);
  SendGrayVector(c, OPEN_VECTOR);

  (void) pvm_recv(io_tid, IO_TAG);
  PVM_UPKCARDINAL(&Graphics_GraphicsError, 1);
}



/*
**  You should know what I want to say: ...
*/

void Graphics_SetbAreaXYZ
#ifdef __STDC__
(ImageIO_binary c[], INTEGER x, INTEGER y, INTEGER zoom, CONFIGURATION OPEN_VECTOR)
#else
(c, x, y, zoom, OPEN_VECTOR)
ImageIO_binary c[];
INTEGER x;
INTEGER y;
INTEGER zoom;
CONFIGURATION OPEN_VECTOR;
#endif
{
  if (me == 0) {
    (void) pvm_initsend(PvmDataDefault);
    PVM_PKSIG(SetbAreaXYZSig);
    PVM_PKINTEGER(&x, 1);
    PVM_PKINTEGER(&y, 1);
    PVM_PKINTEGER(&zoom, 1);
    PackConfiguration(OPEN_VECTOR);
    (void) pvm_send(io_tid, IO_TAG);
  }

  SendActiveStatus(OPEN_VECTOR);
  SendBinaryVector(c, OPEN_VECTOR);

  (void) pvm_recv(io_tid, IO_TAG);
  PVM_UPKCARDINAL(&Graphics_GraphicsError, 1);
}


/*
**  Hey, something new has to be looked at!
**  In this case no difference between CONFIGURATIONs is made, all are
**  treated as linear arrays. Start drawing pixels at the left of the
**  window and expand to the right. The result will be a horizontal
**  line with 'line' pixel distance from the top of the window.
*/

void Graphics_SetLine
#ifdef __STDC__
(ImageIO_color c[], CARDINAL line, CONFIGURATION OPEN_VECTOR)
#else
(c, line, OPEN_VECTOR)
ImageIO_color c[];
CARDINAL line;
CONFIGURATION OPEN_VECTOR;
#endif
{
  if (me == 0) {
    (void) pvm_initsend(PvmDataDefault);
    PVM_PKSIG(SetLineSig);
    PVM_PKCARDINAL(&line, 1);

    PackConfiguration(OPEN_VECTOR);
    (void) pvm_send(io_tid, IO_TAG);
  }

  SendActiveStatus(OPEN_VECTOR);
  SendColorVector(c, OPEN_VECTOR);

  (void) pvm_recv(io_tid, IO_TAG);
  PVM_UPKCARDINAL(&Graphics_GraphicsError, 1);
}


/*
**  Yes, you guess right. The same as above, but with gray scale
**  colors.
*/

void Graphics_SetgLine
#ifdef __STDC__
(ImageIO_gray c[], CARDINAL line, CONFIGURATION OPEN_VECTOR)
#else
(c, line, OPEN_VECTOR)
ImageIO_gray c[];
CARDINAL line;
CONFIGURATION OPEN_VECTOR;
#endif
{
  if (me == 0) {
    (void) pvm_initsend(PvmDataDefault);
    PVM_PKSIG(SetgLineSig);
    PVM_PKCARDINAL(&line, 1);

    PackConfiguration(OPEN_VECTOR);
    (void) pvm_send(io_tid, IO_TAG);
  }

  SendActiveStatus(OPEN_VECTOR);
  SendGrayVector(c, OPEN_VECTOR);

  (void) pvm_recv(io_tid, IO_TAG);
  PVM_UPKCARDINAL(&Graphics_GraphicsError, 1);
}



/*
**  "No, not again!" you might say, but it is true. The same, but for
**  black and white.
*/

void Graphics_SetbLine
#ifdef __STDC__
(ImageIO_binary c[], CARDINAL line, CONFIGURATION OPEN_VECTOR)
#else
(c, line, OPEN_VECTOR)
ImageIO_binary c[];
CARDINAL line;
CONFIGURATION OPEN_VECTOR;
#endif
{
  if (me == 0) {
    (void) pvm_initsend(PvmDataDefault);
    PVM_PKSIG(SetbLineSig);
    PVM_PKCARDINAL(&line, 1);

    PackConfiguration(OPEN_VECTOR);
    (void) pvm_send(io_tid, IO_TAG);
  }

  SendActiveStatus(OPEN_VECTOR);
  SendBinaryVector(c, OPEN_VECTOR);

  (void) pvm_recv(io_tid, IO_TAG);
  PVM_UPKCARDINAL(&Graphics_GraphicsError, 1);
}



/*
**  Like SetAreaXYZ there should be the same for SetLine (and SetColumn).
**  Just create a new configuration describing a single line (column)
**  and call SetAreaXYZ.
*/

void Graphics_SetLineZ
#ifdef __STDC__
(ImageIO_color c[], CARDINAL line, INTEGER offset, INTEGER zoom, CONFIGURATION OPEN_VECTOR)
#else
(c, line, offset, zoom, OPEN_VECTOR)
ImageIO_color c[];
CARDINAL line;
INTEGER offset;
INTEGER zoom;
CONFIGURATION OPEN_VECTOR;
#endif
{
  if (me == 0) {
    (void) pvm_initsend(PvmDataDefault);
    PVM_PKSIG(SetLineZSig);
    PVM_PKCARDINAL(&line, 1);
    PVM_PKINTEGER(&offset, 1);
    PVM_PKINTEGER(&zoom, 1);
    PackConfiguration(OPEN_VECTOR);
    (void) pvm_send(io_tid, IO_TAG);
  }

  SendActiveStatus(OPEN_VECTOR);
  SendColorVector(c, OPEN_VECTOR);

  (void) pvm_recv(io_tid, IO_TAG);
  PVM_UPKCARDINAL(&Graphics_GraphicsError, 1);
}



/*
**  Just like SetLineZ, but for gray values.
*/

void Graphics_SetgLineZ
#ifdef __STDC__
(ImageIO_gray c[], CARDINAL line, INTEGER offset, INTEGER zoom, CONFIGURATION OPEN_VECTOR)
#else
(c, line, offset, zoom, OPEN_VECTOR)
ImageIO_gray c[];
CARDINAL line;
INTEGER offset;
INTEGER zoom;
CONFIGURATION OPEN_VECTOR;
#endif
{
  if (me == 0) {
    (void) pvm_initsend(PvmDataDefault);
    PVM_PKSIG(SetgLineZSig);
    PVM_PKCARDINAL(&line, 1);
    PVM_PKINTEGER(&offset, 1);
    PVM_PKINTEGER(&zoom, 1);
    PackConfiguration(OPEN_VECTOR);
    (void) pvm_send(io_tid, IO_TAG);
  }

  SendActiveStatus(OPEN_VECTOR);
  SendGrayVector(c, OPEN_VECTOR);

  (void) pvm_recv(io_tid, IO_TAG);
  PVM_UPKCARDINAL(&Graphics_GraphicsError, 1);
}



/*
**  Don't feel worried. Here is the routine for black and white pixels.
*/

void Graphics_SetbLineZ
#ifdef __STDC__
(ImageIO_binary c[], CARDINAL line, INTEGER offset, INTEGER zoom, CONFIGURATION OPEN_VECTOR)
#else
(c, line, offset, zoom, OPEN_VECTOR)
ImageIO_binary c[];
CARDINAL line;
INTEGER offset;
INTEGER zoom;
CONFIGURATION OPEN_VECTOR;
#endif
{
  if (me == 0) {
    (void) pvm_initsend(PvmDataDefault);
    PVM_PKSIG(SetbLineZSig);
    PVM_PKCARDINAL(&line, 1);
    PVM_PKINTEGER(&offset, 1);
    PVM_PKINTEGER(&zoom, 1);
    PackConfiguration(OPEN_VECTOR);
    (void) pvm_send(io_tid, IO_TAG);
  }

  SendActiveStatus(OPEN_VECTOR);
  SendBinaryVector(c, OPEN_VECTOR);

  (void) pvm_recv(io_tid, IO_TAG);
  PVM_UPKCARDINAL(&Graphics_GraphicsError, 1);
}



/*
**  How amazing. The above routines not for lines but for columns.
**  Do I really need to explain what's going on?
**  You should have understood the above code, so I will not comment
**  the following routines for drawing columns instead of rows.
*/

void Graphics_SetColumn
#ifdef __STDC__
(ImageIO_color c[], CARDINAL col, CONFIGURATION OPEN_VECTOR)
#else
(c, col, OPEN_VECTOR)
ImageIO_color c[];
CARDINAL col;
CONFIGURATION OPEN_VECTOR;
#endif
{
  if (me == 0) {
    (void) pvm_initsend(PvmDataDefault);
    PVM_PKSIG(SetColumnSig);
    PVM_PKCARDINAL(&col, 1);

    PackConfiguration(OPEN_VECTOR);
    (void) pvm_send(io_tid, IO_TAG);
  }

  SendActiveStatus(OPEN_VECTOR);
  SendColorVector(c, OPEN_VECTOR);

  (void) pvm_recv(io_tid, IO_TAG);
  PVM_UPKCARDINAL(&Graphics_GraphicsError, 1);
}



void Graphics_SetgColumn
#ifdef __STDC__
(ImageIO_gray c[], CARDINAL col, CONFIGURATION OPEN_VECTOR)
#else
(c, col, OPEN_VECTOR)
ImageIO_gray c[];
CARDINAL col;
CONFIGURATION OPEN_VECTOR;
#endif
{
  if (me == 0) {
    (void) pvm_initsend(PvmDataDefault);
    PVM_PKSIG(SetgColumnSig);
    PVM_PKCARDINAL(&col, 1);

    PackConfiguration(OPEN_VECTOR);
    (void) pvm_send(io_tid, IO_TAG);
  }

  SendActiveStatus(OPEN_VECTOR);
  SendGrayVector(c, OPEN_VECTOR);

  (void) pvm_recv(io_tid, IO_TAG);
  PVM_UPKCARDINAL(&Graphics_GraphicsError, 1);
}



void Graphics_SetbColumn
#ifdef __STDC__
(ImageIO_binary c[], CARDINAL col, CONFIGURATION OPEN_VECTOR)
#else
(c, col, OPEN_VECTOR)
ImageIO_binary c[];
CARDINAL col;
CONFIGURATION OPEN_VECTOR;
#endif
{
  if (me == 0) {
    (void) pvm_initsend(PvmDataDefault);
    PVM_PKSIG(SetbColumnSig);
    PVM_PKCARDINAL(&col, 1);

    PackConfiguration(OPEN_VECTOR);
    (void) pvm_send(io_tid, IO_TAG);
  }

  SendActiveStatus(OPEN_VECTOR);
  SendBinaryVector(c, OPEN_VECTOR);

  (void) pvm_recv(io_tid, IO_TAG);
  PVM_UPKCARDINAL(&Graphics_GraphicsError, 1);
}



void Graphics_SetColumnZ
#ifdef __STDC__
(ImageIO_color c[], CARDINAL col, INTEGER offset, INTEGER zoom, CONFIGURATION OPEN_VECTOR)
#else
(c, col, offset, zoom, OPEN_VECTOR)
ImageIO_color c[];
CARDINAL col;
INTEGER offset;
INTEGER zoom;
CONFIGURATION OPEN_VECTOR;
#endif
{
  if (me == 0) {
    (void) pvm_initsend(PvmDataDefault);
    PVM_PKSIG(SetColumnZSig);
    PVM_PKCARDINAL(&col, 1);
    PVM_PKINTEGER(&offset, 1);
    PVM_PKINTEGER(&zoom, 1);
    PackConfiguration(OPEN_VECTOR);
    (void) pvm_send(io_tid, IO_TAG);
  }

  SendActiveStatus(OPEN_VECTOR);
  SendColorVector(c, OPEN_VECTOR);

  (void) pvm_recv(io_tid, IO_TAG);
  PVM_UPKCARDINAL(&Graphics_GraphicsError, 1);
}



void Graphics_SetgColumnZ
#ifdef __STDC__
(ImageIO_gray c[], CARDINAL col, INTEGER offset, INTEGER zoom, CONFIGURATION OPEN_VECTOR)
#else
(c, col, offset, zoom, OPEN_VECTOR)
ImageIO_gray c[];
CARDINAL col;
INTEGER offset;
INTEGER zoom;
CONFIGURATION OPEN_VECTOR;
#endif
{
  if (me == 0) {
    (void) pvm_initsend(PvmDataDefault);
    PVM_PKSIG(SetgColumnZSig);
    PVM_PKCARDINAL(&col, 1);
    PVM_PKINTEGER(&offset, 1);
    PVM_PKINTEGER(&zoom, 1);
    PackConfiguration(OPEN_VECTOR);
    (void) pvm_send(io_tid, IO_TAG);
  }

  SendActiveStatus(OPEN_VECTOR);
  SendGrayVector(c, OPEN_VECTOR);

  (void) pvm_recv(io_tid, IO_TAG);
  PVM_UPKCARDINAL(&Graphics_GraphicsError, 1);
}



void Graphics_SetbColumnZ
#ifdef __STDC__
(ImageIO_binary c[], CARDINAL col, INTEGER offset, INTEGER zoom, CONFIGURATION OPEN_VECTOR)
#else
(c, col, offset, zoom, OPEN_VECTOR)
ImageIO_binary c[];
CARDINAL col;
INTEGER offset;
INTEGER zoom;
CONFIGURATION OPEN_VECTOR;
#endif
{
  if (me == 0) {
    (void) pvm_initsend(PvmDataDefault);
    PVM_PKSIG(SetbColumnZSig);
    PVM_PKCARDINAL(&col, 1);
    PVM_PKINTEGER(&offset, 1);
    PVM_PKINTEGER(&zoom, 1);
    PackConfiguration(OPEN_VECTOR);
    (void) pvm_send(io_tid, IO_TAG);
  }

  SendActiveStatus(OPEN_VECTOR);
  SendBinaryVector(c, OPEN_VECTOR);

  (void) pvm_recv(io_tid, IO_TAG);
  PVM_UPKCARDINAL(&Graphics_GraphicsError, 1);
}


/*
**  Start a request for the color used in the subsequent drawing
**  commands. This is a scalar routine, requiring only one color!
*/

void Graphics_SetColor
#ifdef __STDC__
(ImageIO_color c)
#else
(c)
ImageIO_color c;
#endif
{
  if (me == 0) {
    (void) pvm_initsend(PvmDataDefault);
    PVM_PKSIG(SetColorSig);
    PVM_PKCARDINAL(&(c.red), 1);
    PVM_PKCARDINAL(&(c.green), 1);
    PVM_PKCARDINAL(&(c.blue), 1);
    (void) pvm_send(io_tid, IO_TAG);
  }

  (void) pvm_recv(io_tid, IO_TAG);
  PVM_UPKCARDINAL(&Graphics_GraphicsError, 1);
}



/*
**  Set a gray scale color for subsequent drawing commands.
*/

void Graphics_SetgColor
#ifdef __STDC__
(ImageIO_gray c)
#else
(c)
ImageIO_gray c;
#endif
{
  if (me == 0) {
    (void) pvm_initsend(PvmDataDefault);
    PVM_PKSIG(SetgColorSig);
    PVM_PKCARDINAL(&c, 1);
    (void) pvm_send(io_tid, IO_TAG);
  }

  (void) pvm_recv(io_tid, IO_TAG);
  PVM_UPKCARDINAL(&Graphics_GraphicsError, 1);
}



/*
**  Handle black and white request.
*/

void Graphics_SetbColor
#ifdef __STDC__
(ImageIO_binary c)
#else
(c)
ImageIO_binary c;
#endif
{
  if (me == 0) {
    (void) pvm_initsend(PvmDataDefault);
    PVM_PKSIG(SetPixelSig);
    PVM_PKBOOLEAN(&c, 1);
    (void) pvm_send(io_tid, IO_TAG);
  }

  (void) pvm_recv(io_tid, IO_TAG);
  PVM_UPKCARDINAL(&Graphics_GraphicsError, 1);
}


void Graphics_SetPixel
#ifdef __STDC__
(CARDINAL x, CARDINAL y)
#else
(x, y)
CARDINAL x;
CARDINAL y;
#endif
{
  if (me == 0) {
    (void) pvm_initsend(PvmDataDefault);
    PVM_PKSIG(SetPixelSig);
    PVM_PKCARDINAL(&x, 1);
    PVM_PKCARDINAL(&y, 1);
    (void) pvm_send(io_tid, IO_TAG);
  }

  (void) pvm_recv(io_tid, IO_TAG);
  PVM_UPKCARDINAL(&Graphics_GraphicsError, 1);
}


ImageIO_color Graphics_GetPixel
#ifdef __STDC__
(CARDINAL x, CARDINAL y)
#else
(x, y)
CARDINAL x;
CARDINAL y;
#endif
{
  ImageIO_color rgb_pixel;
    
  if (me == 0) {
    (void) pvm_initsend(PvmDataDefault);
    PVM_PKSIG(GetPixelSig);
    PVM_PKCARDINAL(&x, 1);
    PVM_PKCARDINAL(&y, 1);
    (void) pvm_send(io_tid, IO_TAG);
  }

  (void) pvm_recv(io_tid, IO_TAG);
  PVM_UPKCARDINAL(&Graphics_GraphicsError, 1);
  PVM_UPKCARDINAL(&rgb_pixel.red, 1);
  PVM_UPKCARDINAL(&rgb_pixel.green, 1);
  PVM_UPKCARDINAL(&rgb_pixel.blue, 1);

  return rgb_pixel;
}


void Graphics_Line
#ifdef __STDC__
(CARDINAL x1, CARDINAL y1, CARDINAL x2, CARDINAL y2)
#else
(x1, y1, x2, y2)
CARDINAL x1, y1, x2, y2;
#endif
{
  if (me == 0) {
    (void) pvm_initsend(PvmDataDefault);
    PVM_PKSIG(LineSig);
    PVM_PKCARDINAL(&x1, 1);
    PVM_PKCARDINAL(&y1, 1);
    PVM_PKCARDINAL(&x2, 1);
    PVM_PKCARDINAL(&y2, 1);
    (void) pvm_send(io_tid, IO_TAG);
  }

  (void) pvm_recv(io_tid, IO_TAG);
  PVM_UPKCARDINAL(&Graphics_GraphicsError, 1);
}


void Graphics_DrawAt
#ifdef __STDC__
(CARDINAL x, CARDINAL y)
#else
(x, y)
CARDINAL x, y;
#endif
{
  if (me == 0) {
    (void) pvm_initsend(PvmDataDefault);
    PVM_PKSIG(DrawAtSig);
    PVM_PKCARDINAL(&x, 1);
    PVM_PKCARDINAL(&y, 1);
    (void) pvm_send(io_tid, IO_TAG);
  }

  (void) pvm_recv(io_tid, IO_TAG);
  PVM_UPKCARDINAL(&Graphics_GraphicsError, 1);
}  


void Graphics_Draw
#ifdef __STDC__
(CHAR c)
#else
(c)
CHAR c;
#endif
{
  if (me == 0) {
    (void) pvm_initsend(PvmDataDefault);
    PVM_PKSIG(DrawSig);
    PVM_PKCHAR(&c, 1);
    (void) pvm_send(io_tid, IO_TAG);
  }

  (void) pvm_recv(io_tid, IO_TAG);
  PVM_UPKCARDINAL(&Graphics_GraphicsError, 1);
}


void Graphics_DrawString
#ifdef __STDC__
(CHAR s[], CARDINAL O_2)
#else
(s, O_2)
CHAR s[];
CARDINAL O_2;
#endif
{
  if (me == 0) {
    (void) pvm_initsend(PvmDataDefault);
    PVM_PKSIG(DrawStringSig);
    PVM_PKCARDINAL(&O_2, 1);
    PVM_PKSTRING(s);
    (void) pvm_send(io_tid, IO_TAG);
  }

  (void) pvm_recv(io_tid, IO_TAG);
  PVM_UPKCARDINAL(&Graphics_GraphicsError, 1);
}


void Graphics_DrawInt
#ifdef __STDC__
(INTEGER i, INTEGER l)
#else
(i, l)
INTEGER i, l;
#endif
{
  if (me == 0) {
    (void) pvm_initsend(PvmDataDefault);
    PVM_PKSIG(DrawIntSig);
    PVM_PKINTEGER(&i, 1);
    PVM_PKINTEGER(&l, 1);
    (void) pvm_send(io_tid, IO_TAG);
  }

  (void) pvm_recv(io_tid, IO_TAG);
  PVM_UPKCARDINAL(&Graphics_GraphicsError, 1);
}


void Graphics_DrawCard
#ifdef __STDC__
(CARDINAL c, CARDINAL l)
#else
(c, l)
CARDINAL c, l;
#endif
{
  if (me == 0) {
    (void) pvm_initsend(PvmDataDefault);
    PVM_PKSIG(DrawCardSig);
    PVM_PKCARDINAL(&c, 1);
    PVM_PKCARDINAL(&l, 1);
    (void) pvm_send(io_tid, IO_TAG);
  }

  (void) pvm_recv(io_tid, IO_TAG);
  PVM_UPKCARDINAL(&Graphics_GraphicsError, 1);
}


void Graphics_DrawReal
#ifdef __STDC__
(REAL r, CARDINAL l)
#else
(r, l)
REAL r;
CARDINAL l;
#endif
{
  if (me == 0) {
    (void) pvm_initsend(PvmDataDefault);
    PVM_PKSIG(DrawRealSig);
    PVM_PKREAL(&r, 1);
    PVM_PKCARDINAL(&l, 1);
    (void) pvm_send(io_tid, IO_TAG);
  }

  (void) pvm_recv(io_tid, IO_TAG);
  PVM_UPKCARDINAL(&Graphics_GraphicsError, 1);
}


void Graphics_DrawFixPt
#ifdef __STDC__
(REAL r, CARDINAL l, CARDINAL m)
#else
(r, l, m)
REAL r;
CARDINAL l, m;
#endif
{
  if (me == 0) {
    (void) pvm_initsend(PvmDataDefault);
    PVM_PKSIG(DrawFixPtSig);
    PVM_PKREAL(&r, 1);
    PVM_PKCARDINAL(&l, 1);
    PVM_PKCARDINAL(&m, 1);
    (void) pvm_send(io_tid, IO_TAG);
  }

  (void) pvm_recv(io_tid, IO_TAG);
  PVM_UPKCARDINAL(&Graphics_GraphicsError, 1);
}


void Graphics_DrawBool
#ifdef __STDC__
(BOOLEAN b, CARDINAL l)
#else
(b, l)
BOOLEAN b;
CARDINAL l;
#endif
{
  if (me == 0) {
    (void) pvm_initsend(PvmDataDefault);
    PVM_PKSIG(DrawBoolSig);
    PVM_PKBOOLEAN(&b, 1);
    PVM_PKCARDINAL(&l, 1);
    (void) pvm_send(io_tid, IO_TAG);
  }

  (void) pvm_recv(io_tid, IO_TAG);
  PVM_UPKCARDINAL(&Graphics_GraphicsError, 1);
}


void Graphics_DrawLn
#ifdef __STDC__
(void)
#else
()
#endif
{
  if (me == 0) {
    (void) pvm_initsend(PvmDataDefault);
    PVM_PKSIG(DrawLnSig);
    (void) pvm_send(io_tid, IO_TAG);
  }

  (void) pvm_recv(io_tid, IO_TAG);
  PVM_UPKCARDINAL(&Graphics_GraphicsError, 1);
}


void BEGIN_Graphics
#ifdef __STDC__
(void)
#else
()
#endif
{
  if (me == 0) {
    (void) pvm_initsend(PvmDataDefault);
    PVM_PKSIG(BeginGraphicsSig);
    (void) pvm_send(io_tid, IO_TAG);
  }
}



