# ifndef yySymbols
# define yySymbols

# if defined __STDC__ | defined __cplusplus
# define ARGS(parameters)	parameters
# else
# define ARGS(parameters)	()
# endif

/* line 16 "symbols.cg" */


#include  <stdio.h>
#include  <string.h>
#include  "Idents.h"			  /* Reuse-Library */
#include  "parallaxis.h"
#include  "Errors.h"
#include  "Options.h"
#include  "Values.h"



# ifndef bool
# define bool char
# endif
# define NoSymbols (tSymbols) 0L
# define kSymObject 1
# define kSymNoObject 2
# define kSymConst 3
# define kSymEnumLiteral 4
# define kSymModule 5
# define kSymField 6
# define kSymProcBody 7
# define kSymProcHead 8
# define kSymTypeDecl 9
# define kSymOpaque 10
# define kSymVar 11
# define kSymConfig 12
# define kSymConnect 13
# define kSymStdProc 14
# define kSymExpr 15
# define kSymType 16
# define kSymNoType 17
# define kSymError 18
# define kSymInteger 19
# define kSymCardinal 20
# define kSymReal 21
# define kSymBoolean 22
# define kSymChar 23
# define kSymBitset 24
# define kSymProc 25
# define kSymWord 26
# define kSymAddress 27
# define kSymIntCard 28
# define kSymNil 29
# define kSymString 30
# define kSymStringChar 31
# define kSymVoid 32
# define kSymStdProcType 33
# define kSymQualident 34
# define kSymConstructor 35
# define kSymOpaqueType 36
# define kSymConfiguration 37
# define kSymVector 38
# define kSymArray 39
# define kSymEnumeration 40
# define kSymPointer 41
# define kSymProcType 42
# define kSymRecord 43
# define kSymSet 44
# define kSymSubrange 45
# define kSymObjects 46
# define kSymElem 47
# define kSymUnion 48
# define kSymCObjects 49
# define kSymTypes 50
# define kSymEnv 51
# define kSymSelectors 52
# define kSymStringPar 53
# define kSymConfigs 54
# define kSymConnects 55
# define kSymLocalIds 56

typedef unsigned char Symbols_tKind;
typedef unsigned short Symbols_tMark;
typedef unsigned short Symbols_tLabel;
typedef union Symbols_Node * tSymbols;
typedef void (* Symbols_tProcTree) ();
/* line 30 "symbols.cg" */


typedef tSymbols tEnv       ;	/* type to represent environments	*/
typedef tSymbols tObject    ;	/* type for object descriptions		*/
typedef tSymbols tObjects   ;	/* type to repr. lists of object descr.	*/
typedef tSymbols tCObjects  ;	/* type to repr. lists of object descr.	*/
typedef tSymbols tType      ;	/* representation of data types		*/
typedef tSymbols tTypes     ;	/* type to repr. formal parameter lists	*/
typedef tSymbols tSelectors ;	/* type to repr. record field selectors	*/
typedef tSymbols tStrings   ;	/* type to repr. lists of string params.*/
typedef tSymbols tConfigs   ;   /* type to repr. lists of configurations*/
typedef tSymbols tConnects  ;   /* type to repr. list of connections    */
typedef tSymbols tLocalIds  ;   /* type to repr. list of local identif. */

typedef enum {
  cVoid
} tVoid;

#define NoEnv        NoSymbols	/* empty environment		*/
#define NoObjects    NoSymbols	/* empty list of object descr.	*/
#define NoCObjects   NoSymbols	/* empty list of object descr.	*/
#define NoTypes      NoSymbols	/* empty formal parameter list	*/
#define NoSelectors  NoSymbols	/* empty list of record field	*/
#define NoStrings    NoSymbols	/* empty list of string params.	*/
#define NoConfigs    NoSymbols  /* empty list of configurations */
#define NoConnects   NoSymbols  /* empty list of connections    */
#define NoLocalIds   NoSymbols  /* empty list of local identif. */

/*
 *   Standard functions
 */
typedef enum {
   eSystemAbs    ,
   eSystemCap    ,
   eSystemChr    ,
   eSystemDim    ,   /* Parallaxis */
   eSystemDim2Id ,   /* Parallaxis */
   eSystemEven   ,   /* Parallaxis */
   eSystemFloat  ,
   eSystemHigh   ,
   eSystemId     ,   /* Parallaxis */
   eSystemId2Dim ,   /* Parallaxis */
   eSystemLen    ,   /* Parallaxis */
   eSystemLower  ,   /* Parallaxis */
   eSystemMax    ,   /* also Parallaxis reduce */
   eSystemMin    ,   /* also Parallaxis reduce */
   eSystemOdd    ,
   eSystemOrd    ,
   eSystemRank   ,   /* Parallaxis */
   eSystemSize   ,
   eSystemTrunc  ,
   eSystemUpper  ,   /* Parallaxis */
   eSystemVal    ,
   eSystemDec    ,
   eSystemExcl   ,
   eSystemHalt   ,
   eSystemInc    ,
   eSystemIncl   ,
   eSystemNew    ,
   eSystemDispose,
   eSystemAdr    ,
   eSystemTSize,
/*
 * Reduce functions
 */
   eSystemSum,
   eSystemProduct,
   eSystemAnd,
   eSystemOr,
   eSystemFirst,
   eSystemLast,
/*
 * MathLib
 */
   eSystemSqrt,
   eSystemExponent,
   eSystemLn,
   eSystemSin,
   eSystemArcSin,
   eSystemCos,
   eSystemArcCos,
   eSystemTan,
   eSystemArcTan,
   eSystemArcTan2,
   eSystemCeiling,
   eSystemFloor,
   eSystemRound,
   eSystemCardRandom,
   eSystemIntRandom,
   eSystemRealRandom,
   eSystemBoolRandom,
   eSystemCharRandom,
 /*
 * SystemIO
 */
   eSystemOpenOutput,
   eSystemCloseOutput,
   eSystemOpenInput,
   eSystemCloseInput,
   eSystemDone,
   eSystemEOL,
   eSystemTermCH,
   eSystemWrite,
   eSystemWriteInt,
   eSystemWriteCard,
   eSystemWriteReal,
   eSystemWriteFixPt,
   eSystemWriteBoolean,
   eSystemWriteString,
   eSystemWriteLn,
   eSystemRead,
   eSystemReadInt,
   eSystemReadCard,
   eSystemReadReal,
   eSystemReadBoolean,
   eSystemReadString,
/*
 * Commandline Arguments
 */
   eSystemArgc,
   eSystemArgv
} tStdProc;


/*
 *  default Object and Type
 */
extern tObject    NoObject;	    /* default, error, or no object	  */
extern tType      NoType;	    /* default, error, or no type         */

extern tObject    OpenVectorConfig; /* stores name of open vector configs */

extern tObjects   Predefs;	    /* predefined objects		  */

extern tObject    ModuleSYSTEM;	    /* module SYSTEM			  */

extern tIdent
  IdentALLOC,			    /* identifier ALLOCATE		  */
  IdentDEALLOC,			    /* identifier DEALLOCATE		  */
  IdentSYSTEM;			    /* identifier of module SYSTEM	  */

/*
 * predefined types
 */
extern tType
  TypeINTEGER,
  TypeCARDINAL,
  TypeREAL,
  TypeBOOLEAN,
  TypeCHAR,
  TypeBITSET,
  TypePROC,
  TypeWORD,
  TypeADDRESS,
  TypeIntCard,
  TypeNil,
  TypeString,
  TypeStringChar,
  TypeVoid;

/*
 * procedure to link two object lists
 */
extern tObjects Symbols_Union(tObjects Objects1, tObjects Objects2);
extern tObjects Symbols_AddConfig(tConfigs Configs, tType Type);
/*
 * procedures to extend object and type descriptions
 */
extern tObject mSymConstSetValue    (tObject Object, tValue Value);
extern tObject mSymModuleSetObjects (tObject Object, tObjects Objects);
extern tObject mSymModuleSetLocals  (tObject Object, tObjects Locals);
extern tObject mSymProcSetLocals    (tObject Object, tObjects Locals, BOOLEAN IsExported);
extern tObject mSymTypeDeclSetType  (tObject Object, tType Type);
extern tObject mSymOpaqueSetType    (tObject Object, tType Type);
extern tObject mSymConfigGroup      (tObject Object, INDEX NewConfigGroup);
extern tObject mSymConfigTypeView   (tObject Object, tType Type, tObject View);
extern tObject mSymConfigTypes      (tObject Object, tTypes Types);
extern tObject mSymConfigBounds     (tObject Object, tEnv Env);
extern tObject mSymConfigConnections(tObject Object, tObjects Connections);
extern tObject mSymConnectConfigs   (tObject Object, tObject Source, tObject Dest);

extern tType mSymOpaqueTypeFullType (tType Type, tType FullType);
extern tType mSymVectorBaseType     (tType Type, BOOLEAN IsOpen, tType BaseType);
extern tType mSymVectorConfig       (tType Type, tObject Config);
extern tType mSymArrayIndexElem     (tType Type, BOOLEAN IsOpen, tType IndexType, tType ElemType);
extern tType mSymEnumerationLiteral (tType Type, tObjects EnumLiterals, INDEX MaxValue);
extern tType mSymPointerTargetType  (tType Type, tType TargetType);
extern tType mSymProcTypeFormalsResult (tType Type, tTypes FormalTypes, tType ResultType);
extern tType mSymRecordFields       (tType Type, tObjects Fields);
extern tType mSymSetBaseType        (tType Type, tType BaseType);
extern tType mSymSubrangeBaseType   (tType Type, tType BaseType);
extern tType mSymSubrangeBounds     (tType Type, tValue Lwb, tValue Upb);

extern tObjects mSymConnectElems    (ADDRESS Actuals);

extern tVoid mSymVoid1(tObject Object, tVoid Void);
extern tVoid mSymVoid2(tVoid Void1, tVoid Void2);

extern tIdent  Symbols_DefineCIdent(tObject Object, tIdent CIdent);
extern BOOLEAN Symbols_NestedUseVar(tObject Object, USHORT Level);

/*
 *      procedures for identification
 */
extern tType   Symbols_GroundType   (tType Type);
extern tObject Symbols_Identify     (tIdent Ident, tEnv Env);
extern tObject Symbols_Identify2    (tIdent Ident, tObjects Objects);
extern BOOLEAN Symbols_IsDeclared   (tIdent Ident, tObjects Objects);
extern tIdent  Symbols_LookUp       (tObject M2Object, tCObjects Objects);
extern BOOLEAN Symbols_IsRedeclared (tIdent Ident, tLocalIds LocalIds);
extern BOOLEAN Symbols_ConnectionIsDeclared (tObject Object, tConnects Connects, tIdent *CIdent);
extern INDEX   Symbols_GetConfigGroup (tObject Object, INDEX ConfigGroupIn, tObject View, tPosition Pos, tIdent *Ident);

/*
 *      procedures for configurations and connections
 */
extern tObjects Symbols_GetConnections (tIdent Ident , tEnv Env);
extern INDEX    Symbols_GetConfigRank  (tObject Object);
extern tType    Symbols_GetReduceType  (tType FktType, tType Type);
extern tType    Symbols_GetMoveReceiveType(tType ExprType, tObject ConnectObj);
extern tType    Symbols_GetProcType    (tType Type);
extern tType    Symbols_GetDimType     (tType Type, tTypes Types);
extern tType    Symbols_ExtendToVectorType(tType Vector, tType Scalar);
extern tObjects Symbols_MakeVectorList (tObjects Objects, tType Type);

/*
 *      procedures to collect object descriptions
 */
extern tObjects  Symbols_Filter     (tObjects tObjects);
extern tCObjects Symbols_Pointers   (BOOLEAN GlobalPtrs, tObjects Objects);
extern void      Symbols_OpenArrays (tObjects Objects, tCObjects *ValueOpens, tCObjects *VAROpens);

/*
 *      procedures for 'safe' access to object and type descriptions
 */
extern tSelectors Symbols_GetSelectors (tObject Object);
extern tObjects   Symbols_GetExport1   (tObject Object);
extern tObjects   Symbols_GetExport2   (tObject Object);
extern tObjects   Symbols_GetObjects   (tObject Object);
extern tType      Symbols_GetType      (tObject Object);

extern tIdent     Symbols_GetTypeIdent       (tType Type);
extern tType      Symbols_GetVectorBaseType  (tType Type);
extern tType      Symbols_GetIndexType       (tType Type);
extern tType      Symbols_GetElemType        (tType Type);
#if 0
extern tType      Symbols_GetArraysElemType  (tType VType, tType Type);
#else
extern tType      Symbols_GetArraysElemType  (tType Type);
#endif
extern tType      Symbols_GetTargetType      (tType Type);
extern tType      Symbols_GetSetBaseType     (tType Type);
extern tType      Symbols_GetSubrangeBaseType(tType Type);
extern tTypes     Symbols_GetFormals         (tType Type);
extern tType      Symbols_GetResultType      (tType Type);
extern tType      Symbols_GetResultType2     (tType Type, tTypes Actuals);
extern tType      Symbols_GetResultType3     (tType Type, tTypes Actuals);
extern tObjects   Symbols_GetFields          (tType Type);

extern BOOLEAN    Symbols_Head   (tTypes Types, tType *Type);
extern tTypes     Symbols_Tail   (tTypes Types);
extern INDEX      Symbols_Count  (tTypes Types);

extern tObject    Symbols_GetLiteral   (tObjects Objects, USHORT Index);

/*
 *      predicates on objects and types
 */
extern BOOLEAN Symbols_IsExported   (tObject Object);
extern BOOLEAN Symbols_IsOpenArray  (tObject Object);
extern BOOLEAN Symbols_IsVAR        (tObject Object);
extern BOOLEAN Symbols_IsProcedure  (tObject Object);
extern BOOLEAN Symbols_IsOfType     (tObject Object);

extern BOOLEAN Symbols_IsForward    (USHORT PosPointerTo, tObject TypeObj);

extern BOOLEAN Symbols_IsBoolType   (tType Type);
extern BOOLEAN Symbols_IsIntType    (tType Type);
extern BOOLEAN Symbols_IsInt        (tType Type);
extern BOOLEAN Symbols_IsRealType   (tType Type);
extern BOOLEAN Symbols_IsCharType   (tType Type);
extern BOOLEAN Symbols_IsSetType    (tType Type);
extern BOOLEAN Symbols_IsEnumType   (tType Type);
extern BOOLEAN Symbols_IsStringType (tType Type);
extern BOOLEAN Symbols_IsVectorType (tType Type);
extern BOOLEAN Symbols_IsArrayType  (tType Type);
extern BOOLEAN Symbols_IsArrayOfInt (tType Type);
extern BOOLEAN Symbols_IsINTEGER    (tType Type);
extern BOOLEAN Symbols_IsCARDINAL   (tType Type);
extern BOOLEAN Symbols_IsOpen       (tType Type);
extern BOOLEAN Symbols_IsFunctioncall(tType Type);
extern BOOLEAN Symbols_IsRandom(tType Type);
extern BOOLEAN Symbols_IsReduceFunction(tType Type);

/*
 *  procedures for output during developement
 */
extern void Symbols_WriteEnv     (FILE *f, tEnv Env);
extern void Symbols_WriteObjects (FILE *f, tObjects Objects);



# ifndef Symbols_NodeHead
# define Symbols_NodeHead
# endif
typedef struct { Symbols_tKind yyKind; Symbols_tMark yyMark; Symbols_NodeHead } Symbols_tNodeHead;
typedef struct { Symbols_tNodeHead yyHead; tIdent Ident; tIdent CIdent; } ySymObject;
typedef struct { Symbols_tNodeHead yyHead; tIdent Ident; tIdent CIdent; } ySymNoObject;
typedef struct { Symbols_tNodeHead yyHead; tIdent Ident; tIdent CIdent; tValue Value; } ySymConst;
typedef struct { Symbols_tNodeHead yyHead; tIdent Ident; tIdent CIdent; tSymbols Type; INDEX Index; } ySymEnumLiteral;
typedef struct { Symbols_tNodeHead yyHead; tIdent Ident; tIdent CIdent; tSymbols ExportList; tSymbols Objects; tSymbols Locals; } ySymModule;
typedef struct { Symbols_tNodeHead yyHead; tIdent Ident; tIdent CIdent; tSymbols Type; tSymbols Selectors; } ySymField;
typedef struct { Symbols_tNodeHead yyHead; tIdent Ident; tIdent CIdent; tSymbols Type; BOOLEAN IsExported; tSymbols Locals; } ySymProcBody;
typedef struct { Symbols_tNodeHead yyHead; tIdent Ident; tIdent CIdent; tSymbols Type; } ySymProcHead;
typedef struct { Symbols_tNodeHead yyHead; tIdent Ident; tIdent CIdent; INDEX TypePos; tSymbols Type; } ySymTypeDecl;
typedef struct { Symbols_tNodeHead yyHead; tIdent Ident; tIdent CIdent; INDEX TypePos; tSymbols Type; } ySymOpaque;
typedef struct { Symbols_tNodeHead yyHead; tIdent Ident; tIdent CIdent; tSymbols Type; BOOLEAN IsVAR; INDEX Level; BOOLEAN NestedUse; BOOLEAN IsDimIndex; ADDRESS TypeTree; } ySymVar;
typedef struct { Symbols_tNodeHead yyHead; tIdent Ident; tIdent CIdent; INDEX ConfigGroup; INDEX Level; BOOLEAN NestedUse; BOOLEAN IsGeneric; BOOLEAN IsRespecified; BOOLEAN Visited; tSymbols Type; tSymbols Indexes; tSymbols Connections; tSymbols View; ULONG Rank; LONG Len; } ySymConfig;
typedef struct { Symbols_tNodeHead yyHead; tIdent Ident; tIdent CIdent; INDEX Level; BOOLEAN NestedUse; BOOLEAN IndexUsed; tSymbols SourceConfig; tSymbols DestConfig; } ySymConnect;
typedef struct { Symbols_tNodeHead yyHead; tIdent Ident; tIdent CIdent; INDEX StdProc; tSymbols Type; } ySymStdProc;
typedef struct { Symbols_tNodeHead yyHead; tIdent Ident; tIdent CIdent; ADDRESS Expr; } ySymExpr;
typedef struct { Symbols_tNodeHead yyHead; } ySymType;
typedef struct { Symbols_tNodeHead yyHead; } ySymNoType;
typedef struct { Symbols_tNodeHead yyHead; INDEX ErrorType; LONG Info1; LONG Info2; } ySymError;
typedef struct { Symbols_tNodeHead yyHead; } ySymInteger;
typedef struct { Symbols_tNodeHead yyHead; } ySymCardinal;
typedef struct { Symbols_tNodeHead yyHead; } ySymReal;
typedef struct { Symbols_tNodeHead yyHead; } ySymBoolean;
typedef struct { Symbols_tNodeHead yyHead; } ySymChar;
typedef struct { Symbols_tNodeHead yyHead; } ySymBitset;
typedef struct { Symbols_tNodeHead yyHead; } ySymProc;
typedef struct { Symbols_tNodeHead yyHead; } ySymWord;
typedef struct { Symbols_tNodeHead yyHead; } ySymAddress;
typedef struct { Symbols_tNodeHead yyHead; } ySymIntCard;
typedef struct { Symbols_tNodeHead yyHead; } ySymNil;
typedef struct { Symbols_tNodeHead yyHead; } ySymString;
typedef struct { Symbols_tNodeHead yyHead; } ySymStringChar;
typedef struct { Symbols_tNodeHead yyHead; } ySymVoid;
typedef struct { Symbols_tNodeHead yyHead; INDEX StdProc; } ySymStdProcType;
typedef struct { Symbols_tNodeHead yyHead; tSymbols Object; } ySymQualident;
typedef struct { Symbols_tNodeHead yyHead; tSymbols TypeObj; } ySymConstructor;
typedef struct { Symbols_tNodeHead yyHead; tSymbols TypeObj; tSymbols Type; } ySymOpaqueType;
typedef struct { Symbols_tNodeHead yyHead; tSymbols TypeObj; tSymbols Config; } ySymConfiguration;
typedef struct { Symbols_tNodeHead yyHead; tSymbols TypeObj; tSymbols Config; tSymbols Type; BOOLEAN IsOpen; } ySymVector;
typedef struct { Symbols_tNodeHead yyHead; tSymbols TypeObj; tIdent StructId; tSymbols IndexType; tSymbols ElemType; BOOLEAN IsOpen; } ySymArray;
typedef struct { Symbols_tNodeHead yyHead; tSymbols TypeObj; INDEX MaxValue; tSymbols Objects; } ySymEnumeration;
typedef struct { Symbols_tNodeHead yyHead; tSymbols TypeObj; tSymbols Type; } ySymPointer;
typedef struct { Symbols_tNodeHead yyHead; tSymbols TypeObj; tSymbols Types; tSymbols Type; } ySymProcType;
typedef struct { Symbols_tNodeHead yyHead; tSymbols TypeObj; tIdent StructId; tSymbols Objects; } ySymRecord;
typedef struct { Symbols_tNodeHead yyHead; tSymbols TypeObj; tSymbols Type; } ySymSet;
typedef struct { Symbols_tNodeHead yyHead; tSymbols TypeObj; ADDRESS LwbExpr; ADDRESS UpbExpr; tSymbols Type; tValue Lwb; tValue Upb; } ySymSubrange;
typedef struct { Symbols_tNodeHead yyHead; } ySymObjects;
typedef struct { Symbols_tNodeHead yyHead; tIdent Ident; BOOLEAN IsPseudoObj; tSymbols Object; tSymbols Next; } ySymElem;
typedef struct { Symbols_tNodeHead yyHead; tSymbols Objects1; tSymbols Objects2; } ySymUnion;
typedef struct { Symbols_tNodeHead yyHead; tObject M2Object; tSymbols Next; tIdent CObject; } ySymCObjects;
typedef struct { Symbols_tNodeHead yyHead; BOOLEAN IsVAR; BOOLEAN IsLval; tSymbols Type; tSymbols Next; } ySymTypes;
typedef struct { Symbols_tNodeHead yyHead; tSymbols Objects; tSymbols HiddenEnv; } ySymEnv;
typedef struct { Symbols_tNodeHead yyHead; tIdent Selector; tSymbols Next; } ySymSelectors;
typedef struct { Symbols_tNodeHead yyHead; tIdent CString; tSymbols FormalType; ADDRESS M2String; tSymbols Next; } ySymStringPar;
typedef struct { Symbols_tNodeHead yyHead; tSymbols Object; tSymbols Next; tIdent ActiveIdent; tIdent ElseIdent; } ySymConfigs;
typedef struct { Symbols_tNodeHead yyHead; tSymbols Object; tSymbols Next; tIdent CIdent; } ySymConnects;
typedef struct { Symbols_tNodeHead yyHead; tIdent Ident; tSymbols Next; } ySymLocalIds;

union Symbols_Node {
 Symbols_tKind Kind;
 Symbols_tNodeHead yyHead;
 ySymObject SymObject;
 ySymNoObject SymNoObject;
 ySymConst SymConst;
 ySymEnumLiteral SymEnumLiteral;
 ySymModule SymModule;
 ySymField SymField;
 ySymProcBody SymProcBody;
 ySymProcHead SymProcHead;
 ySymTypeDecl SymTypeDecl;
 ySymOpaque SymOpaque;
 ySymVar SymVar;
 ySymConfig SymConfig;
 ySymConnect SymConnect;
 ySymStdProc SymStdProc;
 ySymExpr SymExpr;
 ySymType SymType;
 ySymNoType SymNoType;
 ySymError SymError;
 ySymInteger SymInteger;
 ySymCardinal SymCardinal;
 ySymReal SymReal;
 ySymBoolean SymBoolean;
 ySymChar SymChar;
 ySymBitset SymBitset;
 ySymProc SymProc;
 ySymWord SymWord;
 ySymAddress SymAddress;
 ySymIntCard SymIntCard;
 ySymNil SymNil;
 ySymString SymString;
 ySymStringChar SymStringChar;
 ySymVoid SymVoid;
 ySymStdProcType SymStdProcType;
 ySymQualident SymQualident;
 ySymConstructor SymConstructor;
 ySymOpaqueType SymOpaqueType;
 ySymConfiguration SymConfiguration;
 ySymVector SymVector;
 ySymArray SymArray;
 ySymEnumeration SymEnumeration;
 ySymPointer SymPointer;
 ySymProcType SymProcType;
 ySymRecord SymRecord;
 ySymSet SymSet;
 ySymSubrange SymSubrange;
 ySymObjects SymObjects;
 ySymElem SymElem;
 ySymUnion SymUnion;
 ySymCObjects SymCObjects;
 ySymTypes SymTypes;
 ySymEnv SymEnv;
 ySymSelectors SymSelectors;
 ySymStringPar SymStringPar;
 ySymConfigs SymConfigs;
 ySymConnects SymConnects;
 ySymLocalIds SymLocalIds;
};

extern tSymbols SymbolsRoot;
extern unsigned long Symbols_HeapUsed;
extern char * Symbols_PoolFreePtr, * Symbols_PoolMaxPtr;
extern unsigned short Symbols_NodeSize [56 + 1];
extern char * Symbols_NodeName [56 + 1];

extern void (* Symbols_Exit) ();
extern tSymbols Symbols_Alloc ();
extern tSymbols MakeSymbols ARGS((Symbols_tKind yyKind));
extern bool Symbols_IsType ARGS((register tSymbols yyt, register Symbols_tKind yyKind));

extern tSymbols mSymObject ARGS((tIdent pIdent));
extern tSymbols mSymNoObject ARGS((tIdent pIdent));
extern tSymbols mSymConst ARGS((tIdent pIdent));
extern tSymbols mSymEnumLiteral ARGS((tIdent pIdent, tSymbols pType, INDEX pIndex));
extern tSymbols mSymModule ARGS((tIdent pIdent, tSymbols pExportList));
extern tSymbols mSymField ARGS((tIdent pIdent, tSymbols pType, tSymbols pSelectors));
extern tSymbols mSymProcBody ARGS((tIdent pIdent, tSymbols pType));
extern tSymbols mSymProcHead ARGS((tIdent pIdent, tSymbols pType));
extern tSymbols mSymTypeDecl ARGS((tIdent pIdent, INDEX pTypePos));
extern tSymbols mSymOpaque ARGS((tIdent pIdent, INDEX pTypePos));
extern tSymbols mSymVar ARGS((tIdent pIdent, tSymbols pType, BOOLEAN pIsVAR, INDEX pLevel, BOOLEAN pNestedUse, BOOLEAN pIsDimIndex, ADDRESS pTypeTree));
extern tSymbols mSymConfig ARGS((tIdent pIdent, INDEX pConfigGroup, INDEX pLevel, BOOLEAN pNestedUse, BOOLEAN pIsGeneric, BOOLEAN pIsRespecified, BOOLEAN pVisited));
extern tSymbols mSymConnect ARGS((tIdent pIdent, INDEX pLevel, BOOLEAN pNestedUse, BOOLEAN pIndexUsed));
extern tSymbols mSymStdProc ARGS((tIdent pIdent, INDEX pStdProc, tSymbols pType));
extern tSymbols mSymExpr ARGS((tIdent pIdent, ADDRESS pExpr));
extern tSymbols mSymType ARGS(());
extern tSymbols mSymNoType ARGS(());
extern tSymbols mSymError ARGS((INDEX pErrorType, LONG pInfo1, LONG pInfo2));
extern tSymbols mSymInteger ARGS(());
extern tSymbols mSymCardinal ARGS(());
extern tSymbols mSymReal ARGS(());
extern tSymbols mSymBoolean ARGS(());
extern tSymbols mSymChar ARGS(());
extern tSymbols mSymBitset ARGS(());
extern tSymbols mSymProc ARGS(());
extern tSymbols mSymWord ARGS(());
extern tSymbols mSymAddress ARGS(());
extern tSymbols mSymIntCard ARGS(());
extern tSymbols mSymNil ARGS(());
extern tSymbols mSymString ARGS(());
extern tSymbols mSymStringChar ARGS(());
extern tSymbols mSymVoid ARGS(());
extern tSymbols mSymStdProcType ARGS((INDEX pStdProc));
extern tSymbols mSymQualident ARGS((tSymbols pObject));
extern tSymbols mSymConstructor ARGS((tSymbols pTypeObj));
extern tSymbols mSymOpaqueType ARGS((tSymbols pTypeObj));
extern tSymbols mSymConfiguration ARGS((tSymbols pTypeObj, tSymbols pConfig));
extern tSymbols mSymVector ARGS((tSymbols pTypeObj));
extern tSymbols mSymArray ARGS((tSymbols pTypeObj, tIdent pStructId));
extern tSymbols mSymEnumeration ARGS((tSymbols pTypeObj));
extern tSymbols mSymPointer ARGS((tSymbols pTypeObj));
extern tSymbols mSymProcType ARGS((tSymbols pTypeObj));
extern tSymbols mSymRecord ARGS((tSymbols pTypeObj, tIdent pStructId));
extern tSymbols mSymSet ARGS((tSymbols pTypeObj));
extern tSymbols mSymSubrange ARGS((tSymbols pTypeObj, ADDRESS pLwbExpr, ADDRESS pUpbExpr));
extern tSymbols mSymObjects ARGS(());
extern tSymbols mSymElem ARGS((tIdent pIdent, BOOLEAN pIsPseudoObj, tSymbols pObject, tSymbols pNext));
extern tSymbols mSymUnion ARGS((tSymbols pObjects1, tSymbols pObjects2));
extern tSymbols mSymCObjects ARGS((tObject pM2Object, tSymbols pNext));
extern tSymbols mSymTypes ARGS((BOOLEAN pIsVAR, BOOLEAN pIsLval, tSymbols pType, tSymbols pNext));
extern tSymbols mSymEnv ARGS((tSymbols pObjects, tSymbols pHiddenEnv));
extern tSymbols mSymSelectors ARGS((tIdent pSelector, tSymbols pNext));
extern tSymbols mSymStringPar ARGS((tIdent pCString, tSymbols pFormalType, ADDRESS pM2String, tSymbols pNext));
extern tSymbols mSymConfigs ARGS((tSymbols pObject, tSymbols pNext));
extern tSymbols mSymConnects ARGS((tSymbols pObject, tSymbols pNext, tIdent pCIdent));
extern tSymbols mSymLocalIds ARGS((tIdent pIdent, tSymbols pNext));

extern void WriteSymbolsNode ARGS((FILE * yyyf, tSymbols yyt));
extern void WriteSymbols ARGS((FILE * yyyf, tSymbols yyt));
extern void QuerySymbols ARGS((tSymbols yyt));
extern void BeginSymbols ();
extern void CloseSymbols ();

# endif
