#include <stdio.h>
#include <string.h>
#include <malloc.h>
#include <signal.h>    /* for Ctrl-C handling */

#include "SYSTEM_C.h"
/* communication-signals (WriteCardSig, ...) and PVM-macros */
#include "shared_macros.h"
/* #include "pvm3.h"     is alread done in shared_macros */
#include "Graphics.h"


/* uncomment this line to test the recognition of the commandline-options */
/* #define TestOptions */


/* global variables */

int n_nodes;      /* number of nodes to start program on */
int mytid;        /* tid of this start-program */
int *task_ids;    /* task-ids of tasks */
CARDINAL MaxDiff;
CARDINAL MinPEs;
CARDINAL ** format_field; /* these two global variables are needed by the */
CARDINAL * f_length;      /* vector-write-routines */


struct {
  long u_sec;
  long u_msec;
  long s_sec;
  long s_msec;
} *tasktime;


/* --------------------------------------------------------------------- */
/* ---------------- I/O-Routines --------------------------------------- */
/* --------------------------------------------------------------------- */

/* the following I/O-routines use the routines of the */
/* singleprocessor-version of the SYSTEM_C-module. */
/* (these are the routines without "IO_"-prefix) */

#define MCAST_DONE()                        \
  (void) pvm_initsend(PvmDataDefault);      \
  PVM_PKBOOLEAN(&Done, 1);                  \
  PVM_MCAST(task_ids, n_nodes, IO_TAG)


void IO_OpenInput
#ifdef __STDC__
(void)
#else
()
#endif
{
  char filename[1024];

  /* receive filename */
  PVM_UPKSTRING(filename);
  
  OpenInput(filename);

  /* send back Done */
  MCAST_DONE();
}


void IO_OpenOutput
#ifdef __STDC__
(void)
#else
()
#endif
{
  char filename[1024];

  /* receive filename */
  PVM_UPKSTRING(filename);

  OpenOutput(filename);

  /* send back Done */
  MCAST_DONE();
}


void IO_CloseInput()
{
  CloseInput();

  /* send back Done */
  MCAST_DONE();
}


void IO_CloseOutput()
{
  CloseOutput();

  /* send back Done */
  MCAST_DONE();
}


void IO_WriteLn
#ifdef __STDC__
(void)
#else
()
#endif
{
  WriteLn();

  /* send back Done */
  MCAST_DONE();
}


void IO_WriteCard
#ifdef __STDC__
(void)
#else
()
#endif
{
  CARDINAL card;
  CARDINAL length;

  PVM_UPKCARDINAL(&card, 1);
  PVM_UPKCARDINAL(&length, 1);

  WriteCard(card, length);

  /* send back Done */
  MCAST_DONE();
}


void IO_WriteInt
#ifdef __STDC__
(void)
#else
()
#endif
{
  INTEGER int_value;
  CARDINAL length;

  PVM_UPKINTEGER(&int_value, 1);
  PVM_UPKCARDINAL(&length, 1);

  WriteInt(int_value, length);

  /* send back Done */
  MCAST_DONE();
}


void IO_WriteReal
#ifdef __STDC__
(void)
#else
()
#endif
{
  REAL real_value;
  CARDINAL length;

  PVM_UPKREAL(&real_value, 1);
  PVM_UPKCARDINAL(&length, 1);

  WriteReal(real_value, length);

  /* send back Done */
  MCAST_DONE();
}


void IO_WriteFixPt
#ifdef __STDC__
(void)
#else
()
#endif
{
  REAL real_value;
  CARDINAL length;
  CARDINAL decimals;

  PVM_UPKREAL(&real_value, 1);
  PVM_UPKCARDINAL(&length, 1);
  PVM_UPKCARDINAL(&decimals, 1);

  WriteFixPt(real_value, length, decimals);

  /* send back Done */
  MCAST_DONE();
}


void IO_WriteBool
#ifdef __STDC__
(void)
#else
()
#endif
{
  BOOLEAN bool_value;
  CARDINAL length;

  PVM_UPKBOOLEAN(&bool_value, 1);
  PVM_UPKCARDINAL(&length, 1);

  WriteBool(bool_value, length);

  /* send back Done */
  MCAST_DONE();
}


void IO_WriteChar
#ifdef __STDC__
(void)
#else
()
#endif
{
  CHAR char_value;

  PVM_UPKCHAR(&char_value, 1);

  Write(char_value);

  /* send back Done */
  MCAST_DONE();
}


void IO_WriteString
#ifdef __STDC__
(void)
#else
()
#endif
{
  CHAR string_value[1024];

  PVM_UPKSTRING(string_value);

  WriteString(string_value);

  /* send back Done */
  MCAST_DONE();
}


void IO_WriteCardVect
#ifdef __STDC__
(void)
#else
()
#endif
{
  CARDINAL u_proc, card_value, i;
  CARDINAL d_length;
  CARDINAL t_ind;           /* task-index */
  CARDINAL count;           /* counter over printed items since last */
			    /* line-break */
  CARDINAL item_count;      /* counter over totally number of items  printed */
  CARDINAL next_break;      /* number of items to print until next break */
  CARDINAL f_ind, num_breaks;
  BOOLEAN  Done_Sum;         /* sums the Done-status after each scalar WriteCard */
  CARDINAL length;


  PVM_UPKCARDINAL(&length, 1);
  PVM_UPKCARDINAL(&u_proc, 1);
  for (t_ind = 0; t_ind < u_proc; t_ind++) {
    PVM_UPKCARDINAL(&f_length[t_ind], 1);
    format_field[t_ind] = (CARDINAL*) malloc(sizeof(CARDINAL) *
					     (f_length[t_ind] +1));
    for (i=0; i<=f_length[t_ind]; i++)
      PVM_UPKCARDINAL(&(format_field[t_ind])[i], 1);
  }

  Done_Sum = TRUE;

  for (t_ind = 0; t_ind < u_proc; t_ind++) {
    f_ind = 0;
    num_breaks = f_length[t_ind];
    next_break = (format_field[t_ind])[f_ind]; f_ind++;

    item_count = 0;     /* init loop-counters */
    count = 0;
    /* process leading line-breaks */
    while (count == next_break) { 
      WriteLn();
      Done_Sum = Done_Sum && Done; /* sum the Done-status */
      next_break = (format_field[t_ind])[f_ind]; f_ind++;
    }

    PVM_UPKCARDINAL(&d_length, 1);
    while (item_count < d_length) {
      PVM_UPKCARDINAL(&card_value, 1);
      WriteCard(card_value, length);
      Done_Sum = Done_Sum && Done; /* sum the Done-status */
      Write(' ');                  /* space between the values */
      Done_Sum = Done_Sum && Done; /* sum the Done-status */
      item_count++;
      count++;

      /* check for line-breaks */
      while (count == next_break) { 
	WriteLn();
	Done_Sum = Done_Sum && Done; /* sum the Done-status */
	next_break = (format_field[t_ind])[f_ind]; f_ind++;
	count = 0;
      } /* end while (count == next_break) */
    } /* end while (count < n_items) */
  } /* end for (t_ind = 0 ... */

  Done = Done_Sum;
  /* send back Done */
  MCAST_DONE();

  for (t_ind = 0; t_ind < u_proc; t_ind++) {
    free(format_field[t_ind]);
  }
}



void IO_WriteIntVect
#ifdef __STDC__
(void)
#else
()
#endif
{
  INTEGER  int_value;
  CARDINAL u_proc, i;
  CARDINAL d_length;
  CARDINAL t_ind;           /* task-index */
  CARDINAL count;           /* counter over printed items since last */
			    /* line-break */
  CARDINAL item_count;      /* counter over totally number of items  printed */
  CARDINAL next_break;      /* number of items to print until next break */
  CARDINAL f_ind, num_breaks;
  BOOLEAN  Done_Sum;         /* sums the Done-status after each scalar WriteCard */
  CARDINAL length;


  PVM_UPKCARDINAL(&length, 1);
  PVM_UPKCARDINAL(&u_proc, 1);
  for (t_ind = 0; t_ind < u_proc; t_ind++) {
    PVM_UPKCARDINAL(&f_length[t_ind], 1);
    format_field[t_ind] = (CARDINAL*) malloc(sizeof(CARDINAL) *
					     (f_length[t_ind] +1));
    for (i=0; i<=f_length[t_ind]; i++)
      PVM_UPKCARDINAL(&(format_field[t_ind])[i], 1);
  }

  Done_Sum = TRUE;

  for (t_ind = 0; t_ind < u_proc; t_ind++) {
    f_ind = 0;
    num_breaks = f_length[t_ind];
    next_break = (format_field[t_ind])[f_ind]; f_ind++;

    item_count = 0;     /* init loop-counters */
    count = 0;
    /* process leading line-breaks */
    while (count == next_break) { 
      WriteLn();
      Done_Sum = Done_Sum && Done; /* sum the Done-status */
      next_break = (format_field[t_ind])[f_ind]; f_ind++;
    }

    PVM_UPKCARDINAL(&d_length, 1);
    while (item_count < d_length) {
      PVM_UPKINTEGER(&int_value, 1);
      WriteInt(int_value, length);
      Done_Sum = Done_Sum && Done; /* sum the Done-status */
      Write(' ');                  /* space between the values */
      Done_Sum = Done_Sum && Done; /* sum the Done-status */
      item_count++;
      count++;

      /* check for line-breaks */
      while (count == next_break) { 
	WriteLn();
	Done_Sum = Done_Sum && Done; /* sum the Done-status */
	next_break = (format_field[t_ind])[f_ind]; f_ind++;
	count = 0;
      } /* end while (count == next_break) */
    } /* end while (count < n_items) */
  } /* end for (t_ind = 0 ... */

  Done = Done_Sum;
  /* send back Done */
  MCAST_DONE();

  for (t_ind = 0; t_ind < u_proc; t_ind++) {
    free(format_field[t_ind]);
  }
}



void IO_WriteRealVect
#ifdef __STDC__
(void)
#else
()
#endif
{
  REAL     real_value;
  CARDINAL u_proc, i;
  CARDINAL d_length;
  CARDINAL t_ind;           /* task-index */
  CARDINAL count;           /* counter over printed items since last */
			    /* line-break */
  CARDINAL item_count;      /* counter over totally number of items  printed */
  CARDINAL next_break;      /* number of items to print until next break */
  CARDINAL f_ind, num_breaks;
  BOOLEAN  Done_Sum;         /* sums the Done-status after each scalar WriteCard */
  CARDINAL length;


  PVM_UPKCARDINAL(&length, 1);
  PVM_UPKCARDINAL(&u_proc, 1);
  for (t_ind = 0; t_ind < u_proc; t_ind++) {
    PVM_UPKCARDINAL(&f_length[t_ind], 1);
    format_field[t_ind] = (CARDINAL*) malloc(sizeof(CARDINAL) *
					     (f_length[t_ind] +1));
    for (i=0; i<=f_length[t_ind]; i++)
      PVM_UPKCARDINAL(&(format_field[t_ind])[i], 1);
  }

  Done_Sum = TRUE;

  for (t_ind = 0; t_ind < u_proc; t_ind++) {
    f_ind = 0;
    num_breaks = f_length[t_ind];
    next_break = (format_field[t_ind])[f_ind]; f_ind++;

    item_count = 0;     /* init loop-counters */
    count = 0;
    /* process leading line-breaks */
    while (count == next_break) { 
      WriteLn();
      Done_Sum = Done_Sum && Done; /* sum the Done-status */
      next_break = (format_field[t_ind])[f_ind]; f_ind++;
    }

    PVM_UPKCARDINAL(&d_length, 1);
    while (item_count < d_length) {
      PVM_UPKREAL(&real_value, 1);
      WriteReal(real_value, length);
      Done_Sum = Done_Sum && Done; /* sum the Done-status */
      Write(' ');       /* space between the values */
      Done_Sum = Done_Sum && Done; /* sum the Done-status */
      item_count++;
      count++;

      /* check for line-breaks */
      while (count == next_break) { 
	WriteLn();
	Done_Sum = Done_Sum && Done; /* sum the Done-status */
	next_break = (format_field[t_ind])[f_ind]; f_ind++;
	count = 0;
      } /* end while (count == next_break) */
    } /* end while (count < n_items) */
  } /* end for (t_ind = 0 ... */

  Done = Done_Sum;
  /* send back Done */
  MCAST_DONE();

  for (t_ind = 0; t_ind < u_proc; t_ind++) {
    free(format_field[t_ind]);
  }
}


void IO_WriteFixPtVect
#ifdef __STDC__
(void)
#else
()
#endif
{
  REAL     real_value;
  CARDINAL u_proc, i;
  CARDINAL d_length;
  CARDINAL t_ind;           /* task-index */
  CARDINAL count;           /* counter over printed items since last */
			    /* line-break */
  CARDINAL item_count;      /* counter over totally number of items  printed */
  CARDINAL next_break;      /* number of items to print until next break */
  CARDINAL f_ind, num_breaks;
  BOOLEAN  Done_Sum;         /* sums the Done-status after each scalar WriteCard */
  CARDINAL length;
  CARDINAL decimals;


  PVM_UPKCARDINAL(&length, 1);
  PVM_UPKCARDINAL(&decimals, 1);
  PVM_UPKCARDINAL(&u_proc, 1);
  for (t_ind = 0; t_ind < u_proc; t_ind++) {
    PVM_UPKCARDINAL(&f_length[t_ind], 1);
    format_field[t_ind] = (CARDINAL*) malloc(sizeof(CARDINAL) *
					     (f_length[t_ind] +1));
    for (i=0; i<=f_length[t_ind]; i++)
      PVM_UPKCARDINAL(&(format_field[t_ind])[i], 1);
  }

  Done_Sum = TRUE;

  for (t_ind = 0; t_ind < u_proc; t_ind++) {
    f_ind = 0;
    num_breaks = f_length[t_ind];
    next_break = (format_field[t_ind])[f_ind]; f_ind++;

    item_count = 0;     /* init loop-counters */
    count = 0;
    /* process leading line-breaks */
    while (count == next_break) { 
      WriteLn();
      Done_Sum = Done_Sum && Done; /* sum the Done-status */
      next_break = (format_field[t_ind])[f_ind]; f_ind++;
    }

    PVM_UPKCARDINAL(&d_length, 1);
    while (item_count < d_length) {
      PVM_UPKREAL(&real_value, 1);
      WriteFixPt(real_value, length, decimals);
      Done_Sum = Done_Sum && Done; /* sum the Done-status */
      Write(' ');       /* space between the values */
      Done_Sum = Done_Sum && Done; /* sum the Done-status */
      item_count++;
      count++;

      /* check for line-breaks */
      while (count == next_break) { 
	WriteLn();
	Done_Sum = Done_Sum && Done; /* sum the Done-status */
	next_break = (format_field[t_ind])[f_ind]; f_ind++;
	count = 0;
      } /* end while (count == next_break) */
    } /* end while (count < n_items) */
  } /* end for (t_ind = 0 ... */

  Done = Done_Sum;
  /* send back Done */
  MCAST_DONE();

  for (t_ind = 0; t_ind < u_proc; t_ind++) {
    free(format_field[t_ind]);
  }
}


void IO_WriteBoolVect
#ifdef __STDC__
(void)
#else
()
#endif
{
  BOOLEAN  bool_value;
  CARDINAL u_proc, i;
  CARDINAL d_length;
  CARDINAL t_ind;           /* task-index */
  CARDINAL count;           /* counter over printed items since last */
			    /* line-break */
  CARDINAL item_count;      /* counter over totally number of items  printed */
  CARDINAL next_break;      /* number of items to print until next break */
  CARDINAL f_ind, num_breaks;
  BOOLEAN  Done_Sum;         /* sums the Done-status after each scalar WriteCard */
  CARDINAL length;           /* length of output-format */


  PVM_UPKCARDINAL(&length, 1);
  PVM_UPKCARDINAL(&u_proc, 1);
  for (t_ind = 0; t_ind < u_proc; t_ind++) {
    PVM_UPKCARDINAL(&f_length[t_ind], 1);
    format_field[t_ind] = (CARDINAL*) malloc(sizeof(CARDINAL) *
					     (f_length[t_ind] +1));
    for (i=0; i<=f_length[t_ind]; i++)
      PVM_UPKCARDINAL(&(format_field[t_ind])[i], 1);
  }

  Done_Sum = TRUE;

  for (t_ind = 0; t_ind < u_proc; t_ind++) {
    f_ind = 0;
    num_breaks = f_length[t_ind];
    next_break = (format_field[t_ind])[f_ind]; f_ind++;

    item_count = 0;     /* init loop-counters */
    count = 0;
    /* process leading line-breaks */
    while (count == next_break) { 
      WriteLn();
      Done_Sum = Done_Sum && Done; /* sum the Done-status */
      next_break = (format_field[t_ind])[f_ind]; f_ind++;
    }

    PVM_UPKCARDINAL(&d_length, 1);
    while (item_count < d_length) {
      PVM_UPKBOOLEAN(&bool_value, 1);
      WriteBool(bool_value, length);
      Done_Sum = Done_Sum && Done; /* sum the Done-status */
      Write(' ');       /* space between the values */
      Done_Sum = Done_Sum && Done; /* sum the Done-status */
      item_count++;
      count++;

      /* check for line-breaks */
      while (count == next_break) { 
	WriteLn();
	Done_Sum = Done_Sum && Done; /* sum the Done-status */
	next_break = (format_field[t_ind])[f_ind]; f_ind++;
	count = 0;
      } /* end while (count == next_break) */
    } /* end while (count < n_items) */
  } /* end for (t_ind = 0 ... */

  Done = Done_Sum;
  /* send back Done */
  MCAST_DONE();

  for (t_ind = 0; t_ind < u_proc; t_ind++) {
    free(format_field[t_ind]);
  }
}


void IO_WriteCharVect
#ifdef __STDC__
(void)
#else
()
#endif
{
  CHAR     char_value;
  CARDINAL d_length, i;
  CARDINAL u_proc;
  CARDINAL t_ind;           /* task-index */
  CARDINAL count;           /* counter over printed items since last */
			    /* line-break */
  CARDINAL item_count;      /* counter over totally number of items  printed */
  CARDINAL next_break;      /* number of items to print until next break */
  CARDINAL f_ind, num_breaks;
  BOOLEAN  Done_Sum;         /* sums the Done-status after each scalar WriteCard */


  PVM_UPKCARDINAL(&u_proc, 1);
  for (t_ind = 0; t_ind < u_proc; t_ind++) {
    PVM_UPKCARDINAL(&f_length[t_ind], 1);
    format_field[t_ind] = (CARDINAL*) malloc(sizeof(CARDINAL) *
					     (f_length[t_ind] +1));
    for (i=0; i<=f_length[t_ind]; i++)
      PVM_UPKCARDINAL(&(format_field[t_ind])[i], 1);
  }

  Done_Sum = TRUE;

  for (t_ind = 0; t_ind < u_proc; t_ind++) {
    f_ind = 0;
    num_breaks = f_length[t_ind];
    next_break = (format_field[t_ind])[f_ind]; f_ind++;

    item_count = 0;     /* init loop-counters */
    count = 0;
    /* process leading line-breaks */
    while (count == next_break) { 
      WriteLn();
      Done_Sum = Done_Sum && Done; /* sum the Done-status */
      next_break = (format_field[t_ind])[f_ind]; f_ind++;
    }

    PVM_UPKCARDINAL(&d_length, 1);
    while (item_count < d_length) {
      PVM_UPKCHAR(&char_value, 1);
      Write(char_value);
      Done_Sum = Done_Sum && Done; /* sum the Done-status */
      item_count++;
      count++;

      /* check for line-breaks */
      while (count == next_break) { 
	WriteLn();
	Done_Sum = Done_Sum && Done; /* sum the Done-status */
	next_break = (format_field[t_ind])[f_ind]; f_ind++;
	count = 0;
      } /* end while (count == next_break) */
    } /* end while (count < n_items) */
  } /* end for (t_ind = 0 ... */

  Done = Done_Sum;
  /* send back Done */
  MCAST_DONE();

  for (t_ind = 0; t_ind < u_proc; t_ind++) {
    free(format_field[t_ind]);
  }
}



void IO_WriteStringVect
#ifdef __STDC__
(void)
#else
()
#endif
{
  CHAR     string[1024];
  CARDINAL u_proc, i;
  CARDINAL d_length;
  CARDINAL t_ind;           /* task-index */
  CARDINAL count;           /* counter over printed items since last */
			    /* line-break */
  CARDINAL item_count;      /* counter over totally number of items  printed */
  CARDINAL next_break;      /* number of items to print until next break */
  CARDINAL f_ind, num_breaks;
  BOOLEAN  Done_Sum;         /* sums the Done-status after each scalar WriteCard */


  PVM_UPKCARDINAL(&u_proc, 1);
  for (t_ind = 0; t_ind < u_proc; t_ind++) {
    PVM_UPKCARDINAL(&f_length[t_ind], 1);
    format_field[t_ind] = (CARDINAL*) malloc(sizeof(CARDINAL) *
					     (f_length[t_ind] +1));
    for (i=0; i<=f_length[t_ind]; i++)
      PVM_UPKCARDINAL(&(format_field[t_ind])[i], 1);
  }

  Done_Sum = TRUE;

  for (t_ind = 0; t_ind < u_proc; t_ind++) {
    PVM_UPKCARDINAL(&d_length, 1);
    f_ind = 0;
    num_breaks = f_length[t_ind];
    next_break = (format_field[t_ind])[f_ind]; f_ind++;

    item_count = 0;     /* init loop-counters */
    count = 0;
    /* process leading line-breaks */
    while (count == next_break) { 
      WriteLn();
      Done_Sum = Done_Sum && Done; /* sum the Done-status */
      next_break = (format_field[t_ind])[f_ind]; f_ind++;
    }

    while (item_count < d_length) {
      PVM_UPKSTRING(string);
      WriteString(string);
      Done_Sum = Done_Sum && Done; /* sum the Done-status */
      Write(' ');                  /* space between the values */
      Done_Sum = Done_Sum && Done; /* sum the Done-status */
      item_count++;
      count++;

      /* check for line-breaks */
      while (count == next_break) { 
	WriteLn();
	Done_Sum = Done_Sum && Done; /* sum the Done-status */
	next_break = (format_field[t_ind])[f_ind]; f_ind++;
	count = 0;
      } /* end while (count == next_break) */
    } /* end while (count < n_items) */
  } /* end for (t_ind = 0 ... */

  Done = Done_Sum;
  /* send back Done */
  MCAST_DONE();

  for (t_ind = 0; t_ind < u_proc; t_ind++) {
    free(format_field[t_ind]);
  }
}


void IO_CaseError
#ifdef __STDC__
(void)
#else
()
#endif
{
  char file[256];
  int line;

  (void) pvm_upkstr(file);
  (void) pvm_upkint(&line, 1, 1);
  (void) fprintf (stderr, "\"%s\", line %1d: case expression out of range\n", file, line);
  (void) pvm_exit();
  exit(1);
}


void IO_ReturnError
#ifdef __STDC__
(void)
#else
()
#endif
{
  char file[256];
  int line;

  (void) pvm_upkstr(file);
  (void) pvm_upkint(&line, 1, 1);
  (void) fprintf (stderr, "\"%s\", line %1d: missing return from function\n", file, line);
  (void) pvm_exit();
  exit(1);
}


int IdRangeError
#ifdef __STDC__
(void)
#else
()
#endif
{
  CARDINAL id;
  CHAR ConfigName[256];

  PVM_UPKCARDINAL(&id, 1);
  PVM_UPKSTRING(ConfigName);

  (void) fprintf(stderr, "id-range-error: no id %lu in Configuration %s\n", id, ConfigName);

  return 2;
}


int PeRangeError
#ifdef __STDC__
(void)
#else
()
#endif
{
  char file[256];
  int line;
  CARDINAL PeNumber;

  (void) pvm_upkstr(file);
  (void) pvm_upkint(&line, 1, 1);
  PVM_UPKCARDINAL(&PeNumber, 1);
  (void) fprintf (stderr, "\"%s\", line %1d: processor number %1lu out of range\n", file, line, PeNumber);
  return 1;
}


/* routines that communicate with process 0 */


void IO_ReadCard
#ifdef __STDC__
(void)
#else
()
#endif
{
  CARDINAL card_value;

  ReadCard(&card_value);

  /* send back value, termCH and Done */
  (void) pvm_initsend(PvmDataDefault);
  PVM_PKCARDINAL(&card_value, 1);
  PVM_PKCHAR(&termCH, 1);
  PVM_PKBOOLEAN(&Done, 1);
  PVM_MCAST(task_ids, n_nodes, IO_TAG);
}


void IO_ReadInt
#ifdef __STDC__
(void)
#else
()
#endif
{
  INTEGER int_value;

  ReadInt(&int_value);

  /* send back value, termCH and Done */
  (void) pvm_initsend(PvmDataDefault);
  PVM_PKINTEGER(&int_value, 1);
  PVM_PKCHAR(&termCH, 1);
  PVM_PKBOOLEAN(&Done, 1);
  PVM_MCAST(task_ids, n_nodes, IO_TAG);
}


void IO_ReadReal
#ifdef __STDC__
(void)
#else
()
#endif
{
  REAL real_value;

  ReadReal(&real_value);

  /* send back value, termCH and Done */
  (void) pvm_initsend(PvmDataDefault);
  PVM_PKREAL(&real_value, 1);
  PVM_PKCHAR(&termCH, 1);
  PVM_PKBOOLEAN(&Done, 1);
  PVM_MCAST(task_ids, n_nodes, IO_TAG);
}


void IO_ReadBool
#ifdef __STDC__
(void)
#else
()
#endif
{
  BOOLEAN bool_value;

  ReadBool(&bool_value);

  /* send back value, termCH and Done */
  (void) pvm_initsend(PvmDataDefault);
  PVM_PKBOOLEAN(&bool_value, 1);
  PVM_PKCHAR(&termCH, 1);
  PVM_PKBOOLEAN(&Done, 1);
  PVM_MCAST(task_ids, n_nodes, IO_TAG);
}


void IO_ReadChar
#ifdef __STDC__
(void)
#else
()
#endif
{
  char char_value;

  Read(&char_value);

  /* send back value and Done */
  (void) pvm_initsend(PvmDataDefault);
  PVM_PKCHAR(&char_value, 1);
  PVM_PKBOOLEAN(&Done, 1);
  PVM_MCAST(task_ids, n_nodes, IO_TAG);
}


void IO_ReadString
#ifdef __STDC__
(void)
#else
()
#endif
{
  CHAR string_value[1024];

  ReadString(string_value);

  /* send back value, termCH and Done */
  (void) pvm_initsend(PvmDataDefault);
  PVM_PKSTRING(string_value);
  PVM_PKCHAR(&termCH, 1);
  PVM_PKBOOLEAN(&Done, 1);
  PVM_MCAST(task_ids, n_nodes, IO_TAG);
}


#define READ_BODY(TYPE, PKROUTINE, READROUTINE)			\
  TYPE value;                                          		\
  CARDINAL u_proc;   /* number of used processors */		\
		     /* for actual configuration */    		\
  CARDINAL d_length; /* number of data-items to be read */	\
		     /* for actual task */			\
  BOOLEAN Done_Sum;						\
  CARDINAL t_ind, i;						\
  								\
								\
  /* unpack number of active tasks */				\
  PVM_UPKCARDINAL(&u_proc, 1);					\
  (void) pvm_initsend(PvmDataDefault);				\
  for (t_ind=0; t_ind<u_proc; t_ind++) {			\
    PVM_UPKCARDINAL(&d_length, 1);				\
    PVM_PKCARDINAL(&d_length, 1);				\
    /* read items */						\
    for (i=0; i<d_length; i++) {				\
      READROUTINE(&value);        				\
      Done_Sum = Done_Sum && Done;				\
      PKROUTINE(&value, 1);					\
    }								\
  }								\
  (void) pvm_send(task_ids[0], IO_TAG); 			\
								\
  Done = Done_Sum;						\
  (void) pvm_initsend(PvmDataDefault);				\
  PVM_PKCHAR(&termCH, 1);                                       \
  PVM_PKBOOLEAN(&Done, 1);					\
  PVM_MCAST(task_ids, n_nodes, IO_TAG);


void IO_ReadCardVect
#ifdef __STDC__
(void)
#else
()
#endif
{
  READ_BODY(CARDINAL, PVM_PKCARDINAL, ReadCard)
}


void IO_ReadIntVect
#ifdef __STDC__
(void)
#else
()
#endif
{
  READ_BODY(INTEGER, PVM_PKINTEGER, ReadInt)
}


void IO_ReadRealVect
#ifdef __STDC__
(void)
#else
()
#endif
{
  READ_BODY(REAL, PVM_PKREAL, ReadReal)
}


void IO_ReadBoolVect
#ifdef __STDC__
(void)
#else
()
#endif
{
  READ_BODY(BOOLEAN, PVM_PKBOOLEAN, ReadBool)
}


void IO_ReadCharVect
#ifdef __STDC__
(void)
#else
()
#endif
{
  CHAR value, dummy;
  CARDINAL u_proc;   /* number of used processors for actual configuration */
  BOOLEAN Done_Sum;
  CARDINAL t_ind, f_ind, items, next_break, count, i;
  CARDINAL *num_breaks, *item_count, **format_field;
  

  /* unpack number of tasks in configuraton */
  PVM_UPKCARDINAL(&u_proc, 1);
  num_breaks = (CARDINAL*) malloc(sizeof(CARDINAL) * u_proc);
  item_count = (CARDINAL*) malloc(sizeof(CARDINAL) * u_proc);
  format_field = (CARDINAL**) malloc(sizeof(CARDINAL*) * u_proc);

  /* unpack format-information for all tasks */
  (void) pvm_initsend(PvmDataDefault);
  for (t_ind=0; t_ind<u_proc; t_ind++) {
    PVM_UPKCARDINAL(&num_breaks[t_ind], 1);
    format_field[t_ind] = (CARDINAL*) malloc(sizeof(CARDINAL) *
					     (num_breaks[t_ind] + 1));
    for (i=0; i<num_breaks[t_ind] + 1; i++) {
      PVM_UPKCARDINAL(&((format_field[t_ind])[i]), 1);
    }
    PVM_UPKCARDINAL(&item_count[t_ind], 1);
  }

  for (t_ind=0; t_ind<u_proc; t_ind++) {
    PVM_PKCARDINAL(&item_count[t_ind], 1);
    /* read chars and pack them into a message */
    f_ind = 0;
    next_break = (format_field[t_ind])[f_ind]; f_ind++;
    items = 0;
    count = 0;
    /* process leading line-breaks */
    while (count == next_break) { 
      Read(&dummy);
      Done_Sum = Done_Sum && Done; /* sum the Done-status */
      next_break = (format_field[t_ind])[f_ind]; f_ind++;
    }
    
    while (items < item_count[t_ind]) {
      Read(&value);
      Done_Sum = Done_Sum && Done;
      PVM_PKCHAR(&value, 1);
      items++;
      count++;
      
      /* check for line-breaks */
      while (count == next_break) { 
	Read(&dummy);
	Done_Sum = Done_Sum && Done; /* sum the Done-status */
	next_break = (format_field[t_ind])[f_ind]; f_ind++;
	count = 0;
      } /* end while (count == next_break) */
    } /* end while (count < items) */
  } /* end for (t_ind = 0 ... */

  (void) pvm_send(task_ids[0], IO_TAG);

  Done = Done_Sum;
  (void) pvm_initsend(PvmDataDefault);
  PVM_PKBOOLEAN(&Done, 1);
  PVM_MCAST(task_ids, n_nodes, IO_TAG);

  free(num_breaks);
  free(item_count);
  for (t_ind = 0; t_ind < u_proc; t_ind++) {
    free(format_field[t_ind]);
  }
  free(format_field);
}


void IO_ReadStringVect
#ifdef __STDC__
(void)
#else
()
#endif
{
  CHAR value[1024];
  CARDINAL u_proc;   /* number of used processors */
		     /* for actual configuration */
  CARDINAL d_length; /* number of data-items to be read */
		     /* for actual task */
  BOOLEAN Done_Sum;
  CARDINAL t_ind, i;


  /* unpack number of active tasks */
  PVM_UPKCARDINAL(&u_proc, 1);
  (void) pvm_initsend(PvmDataDefault);
  for (t_ind=0; t_ind<u_proc; t_ind++) {
    PVM_UPKCARDINAL(&d_length, 1);
    PVM_PKCARDINAL(&d_length, 1);
    /* read items */
    for (i=0; i<d_length; i++) {
      ReadString(value);
      Done_Sum = Done_Sum && Done;
      PVM_PKSTRING(value);
    }
  }
  (void) pvm_send(task_ids[0], IO_TAG);

  Done = Done_Sum;
  (void) pvm_initsend(PvmDataDefault);
  PVM_PKCHAR(&termCH, 1);
  PVM_PKBOOLEAN(&Done, 1);
  PVM_MCAST(task_ids, n_nodes, IO_TAG);
}


int ErrorTermination
/* terminates all tasks */
#ifdef __STDC__
(void)
#else
()
#endif
{
  int exit_code, i;

  pvm_upkint(&exit_code, 1, 1);

  /* kill all tasks */
  for (i = 0; i < n_nodes; i++)
    pvm_kill(task_ids[i]);

  return exit_code;
}


int NormalTermination
/* receives user- and system-times for all tasks */
#ifdef __STDC__
(void)
#else
()
#endif
{
  int exit_code, i;

  pvm_upkint(&exit_code, 1, 1);

  for (i = 0; i < n_nodes; i++) {
    pvm_upklong(&tasktime[i].u_sec,  1, 1);
    pvm_upklong(&tasktime[i].u_msec, 1, 1);
    pvm_upklong(&tasktime[i].s_sec,  1, 1);
    pvm_upklong(&tasktime[i].s_msec, 1, 1);
  }

  return exit_code;
}



void IO_BeginGraphics
#ifdef __STDC__
(void)
#else
()
#endif
{
  BEGIN_Graphics();
}


void IO_OpenWindow
#ifdef __STDC__
(void)
#else
()
#endif
{
  Graphics_window WindowId;
  CHAR title[256];
  CARDINAL O_1;
  CARDINAL width, height;

  /* receive parameters */
  PVM_UPKSTRING(title);
  PVM_UPKCARDINAL(&O_1, 1);
  PVM_UPKCARDINAL(&width, 1);
  PVM_UPKCARDINAL(&height, 1);
  
  /* call graphics routine */
  WindowId = Graphics_OpenWindow(title, O_1, width, height);

  /* send back result and errors */
  (void) pvm_initsend(PvmDataDefault);
  PVM_PKCARDINAL(&WindowId, 1);
  PVM_PKCARDINAL(&Graphics_GraphicsError, 1);
  PVM_MCAST(task_ids, n_nodes, IO_TAG);
}


void IO_CloseWindow
#ifdef __STDC__
(void)
#else
()
#endif
{
  CARDINAL window_num;

  /* receive parameters */
  PVM_UPKCARDINAL(&window_num, 1);
  
  /* call graphics routine */
  (void) Graphics_CloseWindow(window_num);

  /* send back result and errors */
  (void) pvm_initsend(PvmDataDefault);
  PVM_PKCARDINAL(&Graphics_GraphicsError, 1);
  PVM_MCAST(task_ids, n_nodes, IO_TAG);
}


void IO_SelectWindow
#ifdef __STDC__
(void)
#else
()
#endif
{
  CARDINAL window_num;

  /* receive parameters */
  PVM_UPKCARDINAL(&window_num, 1);
  
  /* call graphics routine */
  (void) Graphics_SelectWindow(window_num);

  /* send back result and errors */
  (void) pvm_initsend(PvmDataDefault);
  PVM_PKCARDINAL(&Graphics_GraphicsError, 1);
  PVM_MCAST(task_ids, n_nodes, IO_TAG);
}


void IO_GetScreenSize
#ifdef __STDC__
(void)
#else
()
#endif
{
  CARDINAL width, height;

  /* receive parameters */
  
  /* call graphics routine */
  (void) Graphics_GetScreenSize(&width, &height);

  /* send back result and errors */
  (void) pvm_initsend(PvmDataDefault);
  PVM_PKCARDINAL(&width, 1);
  PVM_PKCARDINAL(&height, 1);
  PVM_PKCARDINAL(&Graphics_GraphicsError, 1);
  PVM_MCAST(task_ids, n_nodes, IO_TAG);
}


void IO_GetWindowSize
#ifdef __STDC__
(void)
#else
()
#endif
{
  CARDINAL width, height;

  /* receive parameters */
  
  /* call graphics routine */
  (void) Graphics_GetWindowSize(&width, &height);

  /* send back result and errors */
  (void) pvm_initsend(PvmDataDefault);
  PVM_PKCARDINAL(&width, 1);
  PVM_PKCARDINAL(&height, 1);
  PVM_PKCARDINAL(&Graphics_GraphicsError, 1);
  PVM_MCAST(task_ids, n_nodes, IO_TAG);
}


/* - - - - - - - - - - - - - - - - - - - - - - - */

/* here are some help-functions which are needed several times */


/* Unpacks some data of a CONFIGURATION in a message (Name, Rank, Lower- and */
/* Upperbounds) and initializes the CONFIGURATION and the stacks. */
/* This procedure makes no pvm_recv. It only unpacks the data. */
/* The equivalent in the parallaxis-program is PackConfiguration. */

void UnpackConfiguration
#ifdef __STDC__
(CONFIGURATION *Config)
#else
(Config)
CONFIGURATION *Config;
#endif
{
  CHAR ConfigName[256];
  CARDINAL Rank, i;

  PVM_UPKSTRING(ConfigName);
  PVM_UPKCARDINAL(&Rank, 1);
  INIT_CONFIGURATION(ConfigName, Config, Rank);

  for (i = 0; i < RANK(*Config); i++) {
    INTEGER lower, upper;

    PVM_UPKINTEGER(&lower, 1);
    PVM_UPKINTEGER(&upper, 1);
    SET_BOUNDS(Config, i, lower, upper);
  }

  INIT_STACKS(Config);
}


/* Receives a message containing the activestatus of all PEs of a CONFIGURATION. */
/* The equivalent in the parallaxis-program is SendActiveStatus. */

void ReceiveActiveStatus
#ifdef __STDC__
(CONFIGURATION Config)
#else
(Config)
CONFIGURATION Config;
#endif
{
  CARDINAL i;
  BOOLEAN ActiveState;

  /* receive activestatus */
  (void) pvm_recv(task_ids[0], IO_TAG);
  for (i = 0; i < LEN(Config); i++) {
    PVM_UPKBOOLEAN(&ActiveState, 1);
    if (ActiveState == FALSE) SET_INACTIVE(Config, i);
  }
}


/* Receives a message, containing a RGB-colorvector. */
/* The equivalent in the parallaxis-program is SendColorVector. */

void ReceiveColorVector
#ifdef __STDC__
(ImageIO_color **c, CONFIGURATION Config)
#else
(c, Config)
ImageIO_color **c;
CONFIGURATION Config;
#endif
{
  /* receive colorvector */
  *c = (ImageIO_color*) malloc(sizeof(ImageIO_color) * LEN(Config));
  (void) pvm_recv(task_ids[0], IO_TAG);
  for (PE = 0; PE < LEN(Config); PE++) {
    PVM_UPKCARDINAL(&((*c)[PE].red), 1);
    PVM_UPKCARDINAL(&((*c)[PE].green), 1);
    PVM_UPKCARDINAL(&((*c)[PE].blue), 1);
  }
}


/* Receives a message, containing a grayvector. */
/* The equivalent in the parallaxis-program SendGrayVector. */

void ReceiveGrayVector
#ifdef __STDC__
(ImageIO_gray **c, CONFIGURATION Config)
#else
(c, Config)
ImageIO_gray **c;
CONFIGURATION Config;
#endif
{
  /* receive grayvector */
  *c = (ImageIO_gray*) malloc(sizeof(ImageIO_gray) * LEN(Config));
  (void) pvm_recv(task_ids[0], IO_TAG);
  for (PE = 0; PE < LEN(Config); PE++) {
    PVM_UPKCARDINAL(&((*c)[PE]), 1);
  }
}


/* Receives a message, containing a binaryvector. */
/* The equivalent in the parallaxis-program SendBinaryVector. */

void ReceiveBinaryVector
#ifdef __STDC__
(ImageIO_binary **c, CONFIGURATION Config)
#else
(c, Config)
ImageIO_binary **c;
CONFIGURATION Config;
#endif
{
  /* receive binaryvector */
  *c = (ImageIO_binary*) malloc(sizeof(ImageIO_binary) * LEN(Config));
  (void) pvm_recv(task_ids[0], IO_TAG);
  for (PE = 0; PE < LEN(Config); PE++) {
    PVM_UPKBOOLEAN(&((*c)[PE]), 1);
  }
}


/* - - - - - - - - - - - - - - - - - - - - - - - */


void IO_SetArea
#ifdef __STDC__
(void)
#else
()
#endif
{
  CONFIGURATION Config;
  ImageIO_color *c;
 
  UnpackConfiguration(&Config);
  ReceiveActiveStatus(Config);
  ReceiveColorVector(&c, Config);

  /* call Graphics-routine */
  Graphics_SetArea(c, Config);

  /* send back result and errors */
  (void) pvm_initsend(PvmDataDefault);
  PVM_PKCARDINAL(&Graphics_GraphicsError, 1);
  PVM_MCAST(task_ids, n_nodes, IO_TAG);

  FREE_CONFIGURATION(&Config, TRUE);
}  


void IO_SetgArea
#ifdef __STDC__
(void)
#else
()
#endif
{
  CONFIGURATION Config;
  ImageIO_gray *c;
 
  UnpackConfiguration(&Config);
  ReceiveActiveStatus(Config);
  ReceiveGrayVector(&c, Config);

  /* call Graphics-routine */
  Graphics_SetgArea(c, Config);

  /* send back result and errors */
  (void) pvm_initsend(PvmDataDefault);
  PVM_PKCARDINAL(&Graphics_GraphicsError, 1);
  PVM_MCAST(task_ids, n_nodes, IO_TAG);

  FREE_CONFIGURATION(&Config, TRUE);
}  


void IO_SetbArea
#ifdef __STDC__
(void)
#else
()
#endif
{
  CONFIGURATION Config;
  ImageIO_binary *c;
 
  UnpackConfiguration(&Config);
  ReceiveActiveStatus(Config);
  ReceiveBinaryVector(&c, Config);

  /* call Graphics-routine */
  Graphics_SetbArea(c, Config);

  /* send back result and errors */
  (void) pvm_initsend(PvmDataDefault);
  PVM_PKCARDINAL(&Graphics_GraphicsError, 1);
  PVM_MCAST(task_ids, n_nodes, IO_TAG);

  FREE_CONFIGURATION(&Config, TRUE);
}  


void IO_SetAreaXYZ
#ifdef __STDC__
(void)
#else
()
#endif
{
  CONFIGURATION Config;
  ImageIO_color *c;
  INTEGER x, y, zoom;
 
  PVM_UPKINTEGER(&x, 1);
  PVM_UPKINTEGER(&y, 1);
  PVM_UPKINTEGER(&zoom, 1);

  UnpackConfiguration(&Config);
  ReceiveActiveStatus(Config);
  ReceiveColorVector(&c, Config);

  /* call Graphics-routine */
  Graphics_SetAreaXYZ(c, x, y, zoom, Config);

  /* send back result and errors */
  (void) pvm_initsend(PvmDataDefault);
  PVM_PKCARDINAL(&Graphics_GraphicsError, 1);
  PVM_MCAST(task_ids, n_nodes, IO_TAG);

  FREE_CONFIGURATION(&Config, TRUE);
}  


void IO_SetgAreaXYZ
#ifdef __STDC__
(void)
#else
()
#endif
{
  CONFIGURATION Config;
  ImageIO_gray *c;
  INTEGER x, y, zoom;
 
  PVM_UPKINTEGER(&x, 1);
  PVM_UPKINTEGER(&y, 1);
  PVM_UPKINTEGER(&zoom, 1);

  UnpackConfiguration(&Config);
  ReceiveActiveStatus(Config);
  ReceiveGrayVector(&c, Config);

  /* call Graphics-routine */
  Graphics_SetgAreaXYZ(c, x, y, zoom, Config);

  /* send back result and errors */
  (void) pvm_initsend(PvmDataDefault);
  PVM_PKCARDINAL(&Graphics_GraphicsError, 1);
  PVM_MCAST(task_ids, n_nodes, IO_TAG);

  FREE_CONFIGURATION(&Config, TRUE);
}  


void IO_SetbAreaXYZ
#ifdef __STDC__
(void)
#else
()
#endif
{
  CONFIGURATION Config;
  ImageIO_binary *c;
  INTEGER x, y, zoom;
 
  PVM_UPKINTEGER(&x, 1);
  PVM_UPKINTEGER(&y, 1);
  PVM_UPKINTEGER(&zoom, 1);

  UnpackConfiguration(&Config);
  ReceiveActiveStatus(Config);
  ReceiveBinaryVector(&c, Config);

  /* call Graphics-routine */
  Graphics_SetbAreaXYZ(c, x, y, zoom, Config);

  /* send back result and errors */
  (void) pvm_initsend(PvmDataDefault);
  PVM_PKCARDINAL(&Graphics_GraphicsError, 1);
  PVM_MCAST(task_ids, n_nodes, IO_TAG);

  FREE_CONFIGURATION(&Config, TRUE);
}  


void IO_SetLine
#ifdef __STDC__
(void)
#else
()
#endif
{
  CONFIGURATION Config;
  ImageIO_color *c;
  CARDINAL line;
 
  PVM_UPKCARDINAL(&line, 1);
  UnpackConfiguration(&Config);
  ReceiveActiveStatus(Config);
  ReceiveColorVector(&c, Config);

  /* call Graphics-routine */
  Graphics_SetLine(c, line, Config);

  /* send back result and errors */
  (void) pvm_initsend(PvmDataDefault);
  PVM_PKCARDINAL(&Graphics_GraphicsError, 1);
  PVM_MCAST(task_ids, n_nodes, IO_TAG);

  FREE_CONFIGURATION(&Config, FALSE);
}  


void IO_SetgLine
#ifdef __STDC__
(void)
#else
()
#endif
{
  CONFIGURATION Config;
  ImageIO_gray *c;
  CARDINAL line;
 
  PVM_UPKCARDINAL(&line, 1);
  UnpackConfiguration(&Config);
  ReceiveActiveStatus(Config);
  ReceiveGrayVector(&c, Config);

  /* call Graphics-routine */
  Graphics_SetgLine(c, line, Config);

  /* send back result and errors */
  (void) pvm_initsend(PvmDataDefault);
  PVM_PKCARDINAL(&Graphics_GraphicsError, 1);
  PVM_MCAST(task_ids, n_nodes, IO_TAG);

  FREE_CONFIGURATION(&Config, FALSE);
}  


void IO_SetbLine
#ifdef __STDC__
(void)
#else
()
#endif
{
  CONFIGURATION Config;
  ImageIO_binary *c;
  CARDINAL line;
 
  PVM_UPKCARDINAL(&line, 1);
  UnpackConfiguration(&Config);
  ReceiveActiveStatus(Config);
  ReceiveBinaryVector(&c, Config);

  /* call Graphics-routine */
  Graphics_SetbLine(c, line, Config);

  /* send back result and errors */
  (void) pvm_initsend(PvmDataDefault);
  PVM_PKCARDINAL(&Graphics_GraphicsError, 1);
  PVM_MCAST(task_ids, n_nodes, IO_TAG);

  FREE_CONFIGURATION(&Config, FALSE);
}  


void IO_SetLineZ
#ifdef __STDC__
(void)
#else
()
#endif
{
  CONFIGURATION Config;
  ImageIO_color *c;
  CARDINAL line;
  INTEGER offset, zoom;
 
  PVM_UPKCARDINAL(&line, 1);
  PVM_UPKINTEGER(&offset, 1);
  PVM_UPKINTEGER(&zoom, 1);
  UnpackConfiguration(&Config);
  ReceiveActiveStatus(Config);
  ReceiveColorVector(&c, Config);

  /* call Graphics-routine */
  Graphics_SetLineZ(c, line, offset, zoom, Config);

  /* send back result and errors */
  (void) pvm_initsend(PvmDataDefault);
  PVM_PKCARDINAL(&Graphics_GraphicsError, 1);
  PVM_MCAST(task_ids, n_nodes, IO_TAG);

  FREE_CONFIGURATION(&Config, FALSE);
}  


void IO_SetgLineZ
#ifdef __STDC__
(void)
#else
()
#endif
{
  CONFIGURATION Config;
  ImageIO_gray *c;
  CARDINAL line;
  INTEGER offset, zoom;
 
  PVM_UPKCARDINAL(&line, 1);
  PVM_UPKINTEGER(&offset, 1);
  PVM_UPKINTEGER(&zoom, 1);
  UnpackConfiguration(&Config);
  ReceiveActiveStatus(Config);
  ReceiveGrayVector(&c, Config);

  /* call Graphics-routine */
  Graphics_SetgLineZ(c, line, offset, zoom, Config);

  /* send back result and errors */
  (void) pvm_initsend(PvmDataDefault);
  PVM_PKCARDINAL(&Graphics_GraphicsError, 1);
  PVM_MCAST(task_ids, n_nodes, IO_TAG);

  FREE_CONFIGURATION(&Config, FALSE);
}  


void IO_SetbLineZ
#ifdef __STDC__
(void)
#else
()
#endif
{
  CONFIGURATION Config;
  ImageIO_binary *c;
  CARDINAL line;
  INTEGER offset, zoom;
 
  PVM_UPKCARDINAL(&line, 1);
  PVM_UPKINTEGER(&offset, 1);
  PVM_UPKINTEGER(&zoom, 1);
  UnpackConfiguration(&Config);
  ReceiveActiveStatus(Config);
  ReceiveBinaryVector(&c, Config);

  /* call Graphics-routine */
  Graphics_SetbLineZ(c, line, offset, zoom, Config);

  /* send back result and errors */
  (void) pvm_initsend(PvmDataDefault);
  PVM_PKCARDINAL(&Graphics_GraphicsError, 1);
  PVM_MCAST(task_ids, n_nodes, IO_TAG);

  FREE_CONFIGURATION(&Config, FALSE);
}  


void IO_SetColumn
#ifdef __STDC__
(void)
#else
()
#endif
{
  CONFIGURATION Config;
  ImageIO_color *c;
  CARDINAL col;
 
  PVM_UPKCARDINAL(&col, 1);
  UnpackConfiguration(&Config);
  ReceiveActiveStatus(Config);
  ReceiveColorVector(&c, Config);

  /* call Graphics-routine */
  Graphics_SetColumn(c, col, Config);

  /* send back result and errors */
  (void) pvm_initsend(PvmDataDefault);
  PVM_PKCARDINAL(&Graphics_GraphicsError, 1);
  PVM_MCAST(task_ids, n_nodes, IO_TAG);

  FREE_CONFIGURATION(&Config, FALSE);
}  


void IO_SetgColumn
#ifdef __STDC__
(void)
#else
()
#endif
{
  CONFIGURATION Config;
  ImageIO_gray *c;
  CARDINAL col;
 
  PVM_UPKCARDINAL(&col, 1);
  UnpackConfiguration(&Config);
  ReceiveActiveStatus(Config);
  ReceiveGrayVector(&c, Config);

  /* call Graphics-routine */
  Graphics_SetgColumn(c, col, Config);

  /* send back result and errors */
  (void) pvm_initsend(PvmDataDefault);
  PVM_PKCARDINAL(&Graphics_GraphicsError, 1);
  PVM_MCAST(task_ids, n_nodes, IO_TAG);

  FREE_CONFIGURATION(&Config, FALSE);
}  


void IO_SetbColumn
#ifdef __STDC__
(void)
#else
()
#endif
{
  CONFIGURATION Config;
  ImageIO_binary *c;
  CARDINAL col;
 
  PVM_UPKCARDINAL(&col, 1);
  UnpackConfiguration(&Config);
  ReceiveActiveStatus(Config);
  ReceiveBinaryVector(&c, Config);

  /* call Graphics-routine */
  Graphics_SetbColumn(c, col, Config);

  /* send back result and errors */
  (void) pvm_initsend(PvmDataDefault);
  PVM_PKCARDINAL(&Graphics_GraphicsError, 1);
  PVM_MCAST(task_ids, n_nodes, IO_TAG);

  FREE_CONFIGURATION(&Config, FALSE);
}  


void IO_SetColumnZ
#ifdef __STDC__
(void)
#else
()
#endif
{
  CONFIGURATION Config;
  ImageIO_color *c;
  CARDINAL col;
  INTEGER offset, zoom;
 
  PVM_UPKCARDINAL(&col, 1);
  PVM_UPKINTEGER(&offset, 1);
  PVM_UPKINTEGER(&zoom, 1);
  UnpackConfiguration(&Config);
  ReceiveActiveStatus(Config);
  ReceiveColorVector(&c, Config);

  /* call Graphics-routine */
  Graphics_SetColumnZ(c, col, offset, zoom, Config);

  /* send back result and errors */
  (void) pvm_initsend(PvmDataDefault);
  PVM_PKCARDINAL(&Graphics_GraphicsError, 1);
  PVM_MCAST(task_ids, n_nodes, IO_TAG);

  FREE_CONFIGURATION(&Config, FALSE);
}  


void IO_SetgColumnZ
#ifdef __STDC__
(void)
#else
()
#endif
{
  CONFIGURATION Config;
  ImageIO_gray *c;
  CARDINAL col;
  INTEGER offset, zoom;
 
  PVM_UPKCARDINAL(&col, 1);
  PVM_UPKINTEGER(&offset, 1);
  PVM_UPKINTEGER(&zoom, 1);
  UnpackConfiguration(&Config);
  ReceiveActiveStatus(Config);
  ReceiveGrayVector(&c, Config);

  /* call Graphics-routine */
  Graphics_SetgColumnZ(c, col, offset, zoom, Config);

  /* send back result and errors */
  (void) pvm_initsend(PvmDataDefault);
  PVM_PKCARDINAL(&Graphics_GraphicsError, 1);
  PVM_MCAST(task_ids, n_nodes, IO_TAG);

  FREE_CONFIGURATION(&Config, FALSE);
}  


void IO_SetbColumnZ
#ifdef __STDC__
(void)
#else
()
#endif
{
  CONFIGURATION Config;
  ImageIO_binary *c;
  CARDINAL col;
  INTEGER offset, zoom;
 
  PVM_UPKCARDINAL(&col, 1);
  PVM_UPKINTEGER(&offset, 1);
  PVM_UPKINTEGER(&zoom, 1);
  UnpackConfiguration(&Config);
  ReceiveActiveStatus(Config);
  ReceiveBinaryVector(&c, Config);

  /* call Graphics-routine */
  Graphics_SetbColumnZ(c, col, offset, zoom, Config);

  /* send back result and errors */
  (void) pvm_initsend(PvmDataDefault);
  PVM_PKCARDINAL(&Graphics_GraphicsError, 1);
  PVM_MCAST(task_ids, n_nodes, IO_TAG);

  FREE_CONFIGURATION(&Config, FALSE);
}  


void IO_SetColor
#ifdef __STDC__
(void)
#else
()
#endif
{
  ImageIO_color c;

  PVM_UPKCARDINAL(&c.red, 1);
  PVM_UPKCARDINAL(&c.green, 1);
  PVM_UPKCARDINAL(&c.blue, 1);

  Graphics_SetColor(c);

  /* send back result and errors */
  (void) pvm_initsend(PvmDataDefault);
  PVM_PKCARDINAL(&Graphics_GraphicsError, 1);
  (void) pvm_mcast(task_ids, n_nodes, IO_TAG);
}  


void IO_SetgColor
#ifdef __STDC__
(void)
#else
()
#endif
{
  ImageIO_gray c;

  PVM_UPKCARDINAL(&c, 1);

  Graphics_SetgColor(c);

  /* send back result and errors */
  (void) pvm_initsend(PvmDataDefault);
  PVM_PKCARDINAL(&Graphics_GraphicsError, 1);
  (void) pvm_mcast(task_ids, n_nodes, IO_TAG);
}  


void IO_SetbColor
#ifdef __STDC__
(void)
#else
()
#endif
{
  ImageIO_binary c;

  PVM_UPKBOOLEAN(&c, 1);

  Graphics_SetbColor(c);

  /* send back result and errors */
  (void) pvm_initsend(PvmDataDefault);
  PVM_PKCARDINAL(&Graphics_GraphicsError, 1);
  (void) pvm_mcast(task_ids, n_nodes, IO_TAG);
}  


void IO_SetPixel
#ifdef __STDC__
(void)
#else
()
#endif
{
  CARDINAL x, y;

  PVM_UPKCARDINAL(&x, 1);
  PVM_UPKCARDINAL(&y, 1);

  Graphics_SetPixel(x, y);

  /* send back result and errors */
  (void) pvm_initsend(PvmDataDefault);
  PVM_PKCARDINAL(&Graphics_GraphicsError, 1);
  (void) pvm_mcast(task_ids, n_nodes, IO_TAG);
}  


void IO_GetPixel
#ifdef __STDC__
(void)
#else
()
#endif
{
  CARDINAL x, y;
  ImageIO_color rgb_pixel;

  PVM_UPKCARDINAL(&x, 1);
  PVM_UPKCARDINAL(&y, 1);

  rgb_pixel = Graphics_GetPixel(x, y);

  /* send back result and errors */
  (void) pvm_initsend(PvmDataDefault);
  PVM_PKCARDINAL(&Graphics_GraphicsError, 1);
  PVM_PKCARDINAL(&rgb_pixel.red, 1);
  PVM_PKCARDINAL(&rgb_pixel.green, 1);
  PVM_PKCARDINAL(&rgb_pixel.blue, 1);
  (void) pvm_mcast(task_ids, n_nodes, IO_TAG);
}  


void IO_Line
#ifdef __STDC__
(void)
#else
()
#endif
{
  CARDINAL x1, y1, x2, y2;

  PVM_UPKCARDINAL(&x1, 1);
  PVM_UPKCARDINAL(&y1, 1);
  PVM_UPKCARDINAL(&x2, 1);
  PVM_UPKCARDINAL(&y2, 1);

  Graphics_Line(x1, y1, x2, y2);

  /* send back result and errors */
  (void) pvm_initsend(PvmDataDefault);
  PVM_PKCARDINAL(&Graphics_GraphicsError, 1);
  (void) pvm_mcast(task_ids, n_nodes, IO_TAG);
}  



void IO_DrawAt
#ifdef __STDC__
(void)
#else
()
#endif
{
  CARDINAL x, y;

  PVM_UPKCARDINAL(&x, 1);
  PVM_UPKCARDINAL(&y, 1);

  Graphics_DrawAt(x, y);

  /* send back result and errors */
  (void) pvm_initsend(PvmDataDefault);
  PVM_PKCARDINAL(&Graphics_GraphicsError, 1);
  (void) pvm_mcast(task_ids, n_nodes, IO_TAG);
}  


void IO_Draw
#ifdef __STDC__
(void)
#else
()
#endif
{
  CHAR c;

  PVM_UPKCHAR(&c, 1);

  Graphics_Draw(c);

  /* send back result and errors */
  (void) pvm_initsend(PvmDataDefault);
  PVM_PKCARDINAL(&Graphics_GraphicsError, 1);
  (void) pvm_mcast(task_ids, n_nodes, IO_TAG);
}  


void IO_DrawString
#ifdef __STDC__
(void)
#else
()
#endif
{
  CHAR *s;
  CARDINAL O_2;

  PVM_UPKCARDINAL(&O_2, 1);
  s = (CHAR *) malloc(sizeof(CHAR) * O_2);
  PVM_UPKSTRING(s);

  Graphics_DrawString(s, O_2);

  /* send back result and errors */
  (void) pvm_initsend(PvmDataDefault);
  PVM_PKCARDINAL(&Graphics_GraphicsError, 1);
  (void) pvm_mcast(task_ids, n_nodes, IO_TAG);
}  


void IO_DrawInt
#ifdef __STDC__
(void)
#else
()
#endif
{
  INTEGER i, l;

  PVM_UPKINTEGER(&i, 1);
  PVM_UPKINTEGER(&l, 1);

  Graphics_DrawInt(i, l);

  /* send back result and errors */
  (void) pvm_initsend(PvmDataDefault);
  PVM_PKCARDINAL(&Graphics_GraphicsError, 1);
  (void) pvm_mcast(task_ids, n_nodes, IO_TAG);
}  


void IO_DrawCard
#ifdef __STDC__
(void)
#else
()
#endif
{
  CARDINAL c, l;

  PVM_UPKCARDINAL(&c, 1);
  PVM_UPKCARDINAL(&l, 1);

  Graphics_DrawCard(c, l);

  /* send back result and errors */
  (void) pvm_initsend(PvmDataDefault);
  PVM_PKCARDINAL(&Graphics_GraphicsError, 1);
  (void) pvm_mcast(task_ids, n_nodes, IO_TAG);
}  


void IO_DrawReal
#ifdef __STDC__
(void)
#else
()
#endif
{
  REAL     r;
  CARDINAL l;

  PVM_UPKREAL(&r, 1);
  PVM_UPKCARDINAL(&l, 1);

  Graphics_DrawReal(r, l);

  /* send back result and errors */
  (void) pvm_initsend(PvmDataDefault);
  PVM_PKCARDINAL(&Graphics_GraphicsError, 1);
  (void) pvm_mcast(task_ids, n_nodes, IO_TAG);
}  


void IO_DrawFixPt
#ifdef __STDC__
(void)
#else
()
#endif
{
  REAL     r;
  CARDINAL l, m;

  PVM_UPKREAL(&r, 1);
  PVM_UPKCARDINAL(&l, 1);
  PVM_UPKCARDINAL(&m, 1);

  Graphics_DrawFixPt(r, l, m);

  /* send back result and errors */
  (void) pvm_initsend(PvmDataDefault);
  PVM_PKCARDINAL(&Graphics_GraphicsError, 1);
  (void) pvm_mcast(task_ids, n_nodes, IO_TAG);
}  


void IO_DrawBool
#ifdef __STDC__
(void)
#else
()
#endif
{
  BOOLEAN  b;
  CARDINAL l;

  PVM_UPKBOOLEAN(&b, 1);
  PVM_UPKCARDINAL(&l, 1);

  Graphics_DrawBool(b, l);

  /* send back result and errors */
  (void) pvm_initsend(PvmDataDefault);
  PVM_PKCARDINAL(&Graphics_GraphicsError, 1);
  (void) pvm_mcast(task_ids, n_nodes, IO_TAG);
}  


void IO_DrawLn
#ifdef __STDC__
(void)
#else
()
#endif
{
  Graphics_DrawLn();

  /* send back result and errors */
  (void) pvm_initsend(PvmDataDefault);
  PVM_PKCARDINAL(&Graphics_GraphicsError, 1);
  (void) pvm_mcast(task_ids, n_nodes, IO_TAG);
}  



/* ---------------------------------------------------------------------- */
/* ---------------------------------------------------------------------- */


/* I/O-signals are defined in "shared_macros.h" */

int MANAGE_IO
#ifdef __STDC__
(void)
#else
()
#endif
/* makes whole I/O from and to tasks on compute-nodes */
{
  int exit_code;
  int JobSig;

  /* central receive loop */
  /* receive message/do input and output */
  do {
    (void) pvm_recv(-1, IO_TAG);
    (void) pvm_upkint(&JobSig, 1, 1);

    switch (JobSig) {

    /* file open/close-operations */
    case OpenInputSig       : IO_OpenInput();
                              break;

    case OpenOutputSig      : IO_OpenOutput();
                              break;

    case CloseInputSig      : IO_CloseInput();
                              break;

    case CloseOutputSig     : IO_CloseOutput();
                              break;

    /* write-operations */
    case WriteLnSig         : IO_WriteLn();
                              break;

    case WriteCardSig       : IO_WriteCard();
                              break;

    case WriteIntSig        : IO_WriteInt();
                              break;

    case WriteRealSig       : IO_WriteReal();
                              break;

    case WriteFixPtSig      : IO_WriteFixPt();
                              break;

    case WriteBoolSig       : IO_WriteBool();
                              break;

    case WriteCharSig       : IO_WriteChar();
                              break;

    case WriteStringSig     : IO_WriteString();
                              break;

    case WriteCardVectSig   : IO_WriteCardVect();
                              break;

    case WriteIntVectSig    : IO_WriteIntVect();
                              break;

    case WriteRealVectSig   : IO_WriteRealVect();
                              break;

    case WriteFixPtVectSig  : IO_WriteFixPtVect();
                              break;

    case WriteBoolVectSig   : IO_WriteBoolVect();
                              break;

    case WriteCharVectSig   : IO_WriteCharVect();
                              break;

    case WriteStringVectSig : IO_WriteStringVect();
                              break;

    /* two compiler-errors */
    case CaseErrorSig       : IO_CaseError();
                              break;

    case ReturnErrorSig     : IO_ReturnError();
                              break;

    /* read-operations */
    case ReadCardSig        : IO_ReadCard();
                              break;

    case ReadIntSig         : IO_ReadInt();
                              break;

    case ReadRealSig        : IO_ReadReal();
                              break;

    case ReadBoolSig        : IO_ReadBool();
                              break;

    case ReadCharSig        : IO_ReadChar();
                              break;

    case ReadStringSig      : IO_ReadString();
                              break;

    case ReadCardVectSig    : IO_ReadCardVect();
                              break;

    case ReadIntVectSig     : IO_ReadIntVect();
                              break;

    case ReadRealVectSig    : IO_ReadRealVect();
                              break;

    case ReadBoolVectSig    : IO_ReadBoolVect();
                              break;

    case ReadCharVectSig    : IO_ReadCharVect();
                              break;

    case ReadStringVectSig  : IO_ReadStringVect();
                              break;

    /* graphic-operations */
    case BeginGraphicsSig   : IO_BeginGraphics();
                              break;

    case OpenWindowSig      : IO_OpenWindow();
                              break;

    case CloseWindowSig     : IO_CloseWindow();
                              break;

    case SelectWindowSig    : IO_SelectWindow();
                              break;

    case GetScreenSizeSig   : IO_GetScreenSize();
                              break;

    case GetWindowSizeSig   : IO_GetWindowSize();
                              break;

    case SetAreaSig         : IO_SetArea();
                              break;

    case SetgAreaSig        : IO_SetgArea();
                              break;

    case SetbAreaSig        : IO_SetbArea();
                              break;

    case SetAreaXYZSig      : IO_SetAreaXYZ();
                              break;

    case SetgAreaXYZSig     : IO_SetgAreaXYZ();
                              break;

    case SetbAreaXYZSig     : IO_SetbAreaXYZ();
                              break;

    case SetLineSig         : IO_SetLine();
                              break;

    case SetgLineSig        : IO_SetgLine();
                              break;

    case SetbLineSig        : IO_SetbLine();
                              break;

    case SetLineZSig        : IO_SetLineZ();
                              break;

    case SetgLineZSig       : IO_SetgLineZ();
                              break;

    case SetbLineZSig       : IO_SetbLineZ();
                              break;

    case SetColumnSig       : IO_SetColumn();
                              break;

    case SetgColumnSig      : IO_SetgColumn();
                              break;

    case SetbColumnSig      : IO_SetbColumn();
                              break;

    case SetColumnZSig      : IO_SetColumnZ();
                              break;

    case SetgColumnZSig     : IO_SetgColumnZ();
                              break;

    case SetbColumnZSig     : IO_SetbColumnZ();
                              break;

    case SetColorSig        : IO_SetColor();
                              break;

    case SetgColorSig       : IO_SetgColor();
                              break;

    case SetbColorSig       : IO_SetbColor();
                              break;

    case SetPixelSig        : IO_SetPixel();
                              break;

    case GetPixelSig        : IO_GetPixel();
                              break;

    case LineSig            : IO_Line();
                              break;

    case DrawAtSig          : IO_DrawAt();
                              break;

    case DrawSig            : IO_Draw();
                              break;

    case DrawStringSig      : IO_DrawString();
                              break;

    case DrawIntSig         : IO_DrawInt();
                              break;

    case DrawCardSig        : IO_DrawCard();
                              break;

    case DrawRealSig        : IO_DrawReal();
                              break;

    case DrawFixPtSig       : IO_DrawFixPt();
                              break;

    case DrawBoolSig        : IO_DrawBool();
                              break;

    case DrawLnSig          : IO_DrawLn();
                              break;
    /* errors */
    case IdRangeErrorSig    : exit_code = IdRangeError();
                              break;

    case PeRangeErrorSig    : exit_code = PeRangeError();
                              break;

    /* termination-signals */
    case TermErrorSig       : exit_code = ErrorTermination();
                              break;

    case TermSig            : exit_code = NormalTermination();
                              break;

    default                 : /* print error-message */
                              fprintf(stderr, "\n\n!!! error in i/o-process: "
				     "received unknown Signal (%d)\n", JobSig);

    } /* end switch */    

  } while (JobSig != TermSig && 
	   JobSig != TermErrorSig && 
	   JobSig != IdRangeErrorSig && 
	   JobSig != PeRangeErrorSig);

  return exit_code;
}


/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */


void Usage( char **argv )
{
  printf("Usage:\n"
	 "You have to specify at least the number of nodes to run the program on\n"
	 "and the name of the program.\n"
	 "usage: %s [-v] [-d] [-maxdiff diff] [-minpes min] "
	 "n name [Parameters]\n"
	 "-v            : verbose-mode, print information\n"
	 "-d            : debug-mode, start up program 'name' under debugger\n"
	 "-maxdiff diff : set the maximum difference of PEs between two nodes "
	                 "to diff\n"
	 "-minpes min   : set the minimum number of PEs on one node to min\n"
	 "n             : number of nodes to start the program on\n"
	 "name          : name of the program to start\n"
	 "Parameters    : parameters of the program name\n",
	 argv[0]);
}


void CheckStartOptions(char **argv, int argc, int *ArgNr,
		       BOOLEAN *VerboseMode, BOOLEAN *DebugMode, 
		       CARDINAL *MaxDiff, CARDINAL *MinPEs, int *n_nodes,
		       char *ProgName)
/* check the commandline-options and read in arguments */
{
  BOOLEAN OptionRecognized;
  BOOLEAN MaxDiffOption = FALSE;
  BOOLEAN MinPEsOption  = FALSE;

  *ArgNr = 1;

  if (argc < 3) {
    printf("Error: Not enough parameters!\n");
    Usage(argv);
    exit(-1);
  }

  /* set defaultvalues */
  *MaxDiff = 500;
  *MinPEs  = 100;
  *DebugMode = FALSE;
  *VerboseMode = FALSE;
  
  do {
    OptionRecognized = FALSE;

    if (*ArgNr >= argc) {
      printf("Error: not enough parameters specified.\n");
      Usage(argv);
      exit(-1);
    }

    if (strcmp(argv[*ArgNr], "-v")==0) {
      *VerboseMode = TRUE;
      (*ArgNr)++;
      OptionRecognized = TRUE;
    }
    else if (strcmp(argv[*ArgNr], "-d")==0) {
      *DebugMode = TRUE;
      (*ArgNr)++;
      OptionRecognized = TRUE;
    }
    else if (strcmp(argv[*ArgNr], "-maxdiff")==0) {
      sscanf(argv[++(*ArgNr)], "%lu", MaxDiff);
      if (*MaxDiff <= 0) {
	printf("Error in Option -maxdiff: number >= 1 required\n");
	exit(-1);
      } else {
	MaxDiffOption = TRUE;
	OptionRecognized = TRUE;
	(*ArgNr)++;
      }
    }
    else if (strcmp(argv[*ArgNr], "-minpes")==0) {
      sscanf(argv[++(*ArgNr)], "%lu", MinPEs);
      if (*MaxDiff <= 0) {
	printf("Error in Option -minpes: number >= 1 required\n");
	exit(-1);
      } else {
	MinPEsOption = TRUE;
	OptionRecognized = TRUE;
	(*ArgNr)++;
      }
    }
  } while (OptionRecognized);

  /* next option has to be the number of nodes */
  sscanf(argv[*ArgNr], "%u", n_nodes);
  (*ArgNr)++;
  if (*n_nodes < 1) {
    printf("Error in parameters: you have to specify\n"
	   "the number of nodes (with a positive number).\n");
    exit(-1);
  }

  /* next option has to be the program name */
  if (*ArgNr < argc) {
    sscanf(argv[*ArgNr], "%s", ProgName);
    /* ArgNr is not increased because it has to point to the program name */
    /* after all options are checked */
  } else {
    printf("Error in parameters: you have to specify the program name.\n");
    exit(-1);
  }

  /* further checks */
  if (MaxDiffOption && MinPEsOption && *MaxDiff <= *MinPEs) {
    printf("Error in option -maxdiff: value must be greater than minpes.\n");
    printf("maxdiff = %lu, minpes = %lu\n", *MaxDiff, *MinPEs);
    exit(-1);
  } else if (!(MaxDiffOption && MinPEsOption) && *MaxDiff <= *MinPEs) {
    *MinPEs = *MaxDiff -1;
  }
}



void KillPvmTasks
#ifdef __STDC__
(void)
#else
()
#endif
{
  CARDINAL i;

  /* kill all pvm-tasks */
  for (i = 0; i < n_nodes; i++)
    pvm_kill(task_ids[i]);
  printf("\nInterruption. All tasks killed\n");
  exit(-1);
}




/* --------------------------------------------------------------------- */
/* ---------------- Main Program --------------------------------------- */
/* --------------------------------------------------------------------- */


int main
#ifdef __STDC__
(int argc, char *argv[], char *envp[])
#else
(argc, argv, envp)
int argc;
char *argv[], *envp[];
#endif
{
  int n_childs;     /* number of processes spawned by pvm */
  int exit_code;    /* exit-code of parallel program */
  int ArgNr;        /* index in argv-array */
  BOOLEAN VerboseMode, DebugMode;
  CARDINAL MaxDiff, MinPEs;
  char ProgName[512];


  /* process commandline-options */
  CheckStartOptions(argv, argc, &ArgNr, &VerboseMode, &DebugMode, 
		    &MaxDiff, &MinPEs, &n_nodes, ProgName);

  /* options are all o.k. do your work now */
#ifdef TestOptions
  printf("\nTestmodus: Erkennung der Kommandozeilenoptionen\n"
	 "(um Testmodus zu entfernen in start.c #define TestOptions wegmachen)\n");
  printf("Debugmode: ");
  if (DebugMode)
    printf("eingeschaltet.\n");
  else
    printf("ausgeschaltet.\n");
  printf("Verbosemode: ");
  if (VerboseMode)
    printf("eingeschaltet.\n");
  else
    printf("ausgeschaltet.\n");
  printf("MaxDiff = %lu\n", MaxDiff);
  printf("MinPEs  = %lu\n", MinPEs );
  printf("Anzahl Knoten = %u\n", n_nodes);
  printf("Programmname: %s    nochmal anders: %s\n", argv[ArgNr], ProgName);
  if (++ArgNr < argc) {
    printf("eigene Optionen: ");
    while (ArgNr < argc) {
      printf("%s, ", argv[ArgNr]);
      ArgNr++;
    }
    printf("\n");
  } else {
    printf("keine weiteren Optionen\n");
  }

  printf("\nProgrammende\n");
  exit(0);
#endif

  /* handle Ctrl-C */
  signal(SIGINT, &KillPvmTasks);

  /* enroll in pvm */
  mytid = pvm_mytid();

  /* start tasks */
  if (VerboseMode) {
    printf("starting %s on %d node", ProgName, n_nodes);
    if (n_nodes > 1) printf("s");
    printf("\n\n");
  }

  task_ids = (int*) malloc(sizeof(int) * n_nodes);
  if (!DebugMode) {
    n_childs = pvm_spawn(ProgName,        /* name of the user-program        */
			 &argv[ArgNr+1],  /* pointer to command-line-args    */
			 PvmTaskDefault,  /* start task on any node          */
			 "",              /* no node specification necessary */
			 n_nodes,         /* number of tasks to start        */
			 task_ids );      /* get array of tids               */
  } else {
    /* start under debugger */
    n_childs = pvm_spawn(ProgName,        /* name of the user-program        */
			 &argv[ArgNr+1],  /* pointer to command-line-args    */
			 PvmTaskDebug,    /* start task under debugger       */
			 "",              /* no node specification necessary */
			 n_nodes,         /* number of tasks to start        */
			 task_ids );      /* get array of tids               */
  }

  if (n_childs != n_nodes) {
    (void) printf("Error: %d processes of %d could be started\n", 
	   n_childs, n_nodes);
    (void) pvm_exit();
    exit(-1); /* EXIT_MODULE(-1); geht nicht */
  }


  /* now the own arguments are checked, so the system can do what it wants */
  /* with them */
  INIT_MODULE(argc - ArgNr, &argv[ArgNr], envp);

  /* send INIT-message to all tasks */
  (void) pvm_initsend(PvmDataDefault);
  pvm_pkint(&n_nodes, 1, 1);
  (void) pvm_pkint(task_ids, n_nodes, 1);

  PVM_PKCARDINAL(&MaxDiff, 1);
  PVM_PKCARDINAL(&MinPEs,  1);

  (void) pvm_pkint(&SYSTEM_argc, 1, 1);
  {
    int i, l;

    for (i=0; i<=SYSTEM_argc; i++) {
      l = strlen(SYSTEM_argv[i]);
      (void) pvm_pkint(&l, 1, 1);
      (void) pvm_pkstr(SYSTEM_argv[i]);
    }
  }

  (void) pvm_mcast(task_ids, n_nodes, INIT_TAG);


  tasktime = malloc( sizeof(*tasktime) * n_nodes );

  /* allocate two arrays which are used by the vector-write-routines */
  format_field = malloc( sizeof(CARDINAL*) * n_nodes);
  f_length = malloc( sizeof(CARDINAL) * n_nodes);

  /* --------------------------------------------------------------------- */
  /* make whole I/O from and to tasks on nodes until work is done and then */
  /* get exit-code                                                         */
  exit_code = MANAGE_IO();
  /* --------------------------------------------------------------------- */

  /* free the two arrays */
  free(format_field);
  free(f_length);


  if (VerboseMode) {
    printf("\nReceived TERM-message with exit-code %d\n\n", exit_code);
  }

  if (VerboseMode && exit_code == 0) {
    int i;
    
    printf("Elapsed time in seconds:\n\t\t      User\t\tSystem\n");
    
    for (i = 0; i < n_nodes; i++) {
      printf("task %6d:\t%7ld.%06ld | %7ld.%06ld\n", i,
	     tasktime[i].u_sec, tasktime[i].u_msec, 
	     tasktime[i].s_sec, tasktime[i].s_msec);
    }
  } else if (VerboseMode) {
    printf ("No measured times available.\n");
  }

  free( tasktime );
    
  pvm_exit();
  /* EXIT_MODULE(exit_code); would be better! */
  EXIT_MODULE();
  return exit_code; /* is not executed because EXIT_MODULE calls exit(0) */
                    /* but so the compiler doesn`t give a warning */
}

