/*************************************************************************
 *
 * Parallaxis Compiler Projects
 * ----------------------------
 *
 * name  :  FileIO.c
 * desc  :  supply layer on top of C stream io
 * author:  lampke
 *
 ************************************************************************/

#include <stdio.h>
#include <sys/types.h>
#include <sys/stat.h>
#include "FileIO.h"

static BOOLEAN isInitialised = FALSE;

/*
 * local types
 */

/*
 * local variables
 */
static FILE   *FileTab[FILESTOHANDLE];
static BOOLEAN IsLineBuffered[FILESTOHANDLE];

/*
 * local function prototypes
 */
static BOOLEAN IsCharDev( int File );

/*
 * local function implementation
 */
/*************************************************************************
 * name :  IsCharDev
 * in   :  int handle of file to test
 * out  :  TRUE, id file is character device
 *         FALSE, otherwise
 ************************************************************************/
static BOOLEAN IsCharDev( int File )
{
   struct stat buf;

   fstat( File, &buf );
   return( ( S_IFCHR & buf.st_mode ) == S_IFCHR );
}

/*
 * Global variables
 */

/*
 * global function implementation
 */
/*------- OpenInput --------------------------------------------------------*/

FILE *FileIO_OpenInput( char *FileName )
{
   FILE *f;
   SHORT slot;

   f = fopen( FileName, "r" );
   if ( f != NULL ) {
      slot = fileno( f );
      if ( slot < FILESTOHANDLE ) {
         FileTab[slot] = f;
         IsLineBuffered[slot] = IsCharDev( slot );
      } else {
         f = NULL;
      } /* endif */
   } /* endif */

   return f;
}

/*------- OpenInputHandle --------------------------------------------------*/

int FileIO_OpenInputHandle( char *FileName )
{
   FILE *f;
   SHORT slot;

   f = fopen( FileName, "r" );
   if ( f != NULL ) {
      slot = fileno( f );
      if ( slot < FILESTOHANDLE ) {
         FileTab[slot] = f;
         IsLineBuffered[slot] = IsCharDev( slot );
      } else {
         slot = FileIO_ERRORHANDLE;
      } /* endif */
   } else {
      slot = FileIO_ERRORHANDLE;
   } /* endif */

   return slot;
}

/*------- OpenOutput -------------------------------------------------------*/

FILE *FileIO_OpenOutput( char *FileName )
{
   FILE *f;
   SHORT slot;

   f = fopen( FileName, "w" );
   if ( f != NULL ) {
      slot = fileno( f );
      if ( slot < FILESTOHANDLE ) {
         FileTab[slot] = f;
         IsLineBuffered[slot] = FALSE;
      } else {
         f = NULL;
      } /* endif */
   } /* endif */

   return f;
}

/*------- OpenOutputHandle -------------------------------------------------*/

int FileIO_OpenOutputHandle( char *FileName )
{
   FILE *f;
   SHORT slot;

   f = fopen( FileName, "w" );
   if ( f != NULL ) {
      slot = fileno( f );
      if ( slot < FILESTOHANDLE ) {
         FileTab[slot] = f;
         IsLineBuffered[slot] = FALSE;
      } else {
         slot = FileIO_ERRORHANDLE;
      } /* endif */
   } else {
      slot = FileIO_ERRORHANDLE;
   } /* endif */

   return slot;
}

/*------- Close ------------------------------------------------------------*/

void FileIO_Close( FILE *File )
{
   SHORT slot;

   if ( File != NULL ) {
      slot = fileno( File );
      if ( slot < FILESTOHANDLE ) {
         FileTab[slot] = NULL;
         IsLineBuffered[slot] = FALSE;
      } /* endif */
      fclose( File );
   }
}

/*------- CloseHandle ------------------------------------------------------*/

void FileIO_CloseHandle( int File )
{
   if ( FileTab[File] != NULL ) {
      fclose( FileTab[File] );
      FileTab[File] = NULL;
      IsLineBuffered[File] = FALSE;
   } /* endif */
}

/*------- CloseAll ---------------------------------------------------------*/

SHORT FileIO_CloseAll( void )
{
   SHORT count = 0;
   SHORT i;

   for (i=3; i<FILESTOHANDLE; i++) {
      if ( FileTab[i] != NULL ) {
         fclose( FileTab[i] );
         FileTab[i] = NULL;
         IsLineBuffered[i] = FALSE;
         count++;
      } /* endif */
   } /* endfor */

   return count;
}

void FileIO_Init( void )
{
   SHORT i;

   if ( !isInitialised ) {
      isInitialised = TRUE;
      FileTab[0]        = stdin;
      IsLineBuffered[0] = TRUE;
      FileTab[1] = stdout;
      IsLineBuffered[1] = TRUE;
      FileTab[2] = stderr;
      IsLineBuffered[2] = TRUE;
      for (i=3; i<FILESTOHANDLE; i++) {
         FileTab[i]        = NULL;
         IsLineBuffered[i] = FALSE;
      } /* endfor */
   } /* endif */
}

