#include <stdio.h>
#include <stdlib.h>			  /* for rand48() */
#include <time.h>			  /* for time() */
#include <string.h>

#include "SYSTEM_MPL.h"

#if 1
void SET_CONFIG_BOUNDS(CONFIGURATION *config, CARDINAL dim,
		       INTEGER lwb, INTEGER upb)
{
  config->dim[dim-1].lower = lwb;
  config->dim[dim-1].upper = upb;
  config->dim[dim-1].size = config->configgroup->VPEs * (upb-lwb+1);
  config->configgroup->VPEs = config->dim[dim-1].size;
}

void INIT_CONFIGURATION(CONFIGURATION config)
{
  if (config.configgroup->activeset==NULL) { 
    config.configgroup->layers = (config.configgroup->VPEs + nproc - 1) /
				 nproc;
    config.configgroup->activeset = (ACTIVESET *) 
		      p_malloc(config.configgroup->layers * sizeof(ACTIVESET));
    for (PElayer=0; PElayer<(config.configgroup->layers - 1); PElayer++) {
      config.configgroup->activeset[PElayer] = TRUE;
      config.configgroup->activeset[PElayer] = (iproc <
				   (config.configgroup->VPEs - PElayer*nproc));
    }
  }
}
#endif

void INIT_CONNECTION(CONNECTION *conn, CONFIGURATION config)
{
  conn->act_single = (plural CONTAINER_1_PTR *) p_calloc(config.configgroup->layers, sizeof(CONTAINER_1_PTR));
  conn->act_multiple_num = (plural USHORT *) p_calloc(config.configgroup->layers, sizeof(USHORT));
  conn->act_multiple = (plural CONTAINER_N_PTR *) p_calloc(config.configgroup->layers, sizeof(CONTAINER_N_PTR));
  conn->num_1 = (plural USHORT *) p_calloc(config.configgroup->layers, sizeof(USHORT));
  conn->single = (plural CONTAINER_1_PTR *) p_calloc(config.configgroup->layers, sizeof(CONTAINER_1_PTR));
  conn->multiple = (plural CONTAINER_N_PTR *) p_calloc(config.configgroup->layers, sizeof(CONTAINER_N_PTR));
}    

void REGISTER_LINK (CONNECTION *conn, plural CARDINAL dest[])
{
  CONTAINER_1_PTR plural Cont_1;

  if (conn->num_1[PElayer] = CONTAINER_N_SIZE-1) {
    CONTAINER_N_PTR plural Cont_N = (CONTAINER_N_PTR) p_malloc(sizeof(CONTAINER_N));
    USHORT i;

    for (i = 0; i < CONTAINER_N_SIZE-2; i++) {
      Cont_1 = conn->single[PElayer];
      Cont_N->pe[i].layer = Cont_1->pe.layer;	/* copy */
      Cont_N->pe[i].ppe = Cont_1->pe.ppe;
      conn->single[PElayer] = Cont_1->next;	/* delete from single list */
      Cont_1->next = FREELIST.single;		/* insert in FREELIST */
      FREELIST.single = Cont_1;
    }
    FREELIST.num_1 += CONTAINER_N_SIZE-1;
    Cont_N->pe[CONTAINER_N_SIZE-1].layer = ID2layer(dest[PElayer]);
    Cont_N->pe[CONTAINER_N_SIZE-1].ppe = ID2proc(dest[PElayer]);
    Cont_N->next = conn->multiple[PElayer];	/* insert in multiple list */
    conn->multiple[PElayer] = Cont_N;
    conn->num_1[PElayer] = 0;
    conn->act_single[PElayer] = NULL;
    conn->act_multiple[PElayer] = Cont_N;
  } else {
    if (FREELIST.num_1 > 0) {
      Cont_1 = FREELIST.single;
      FREELIST.single = Cont_1->next;
      FREELIST.num_1--;
    } else {
      Cont_1 = (CONTAINER_1_PTR) p_malloc(sizeof(CONTAINER_1));
    }
    Cont_1->pe.layer = ID2layer(dest[PElayer]); /* insert in single list */
    Cont_1->pe.ppe = ID2proc(dest[PElayer]);
    Cont_1->next = conn->single[PElayer];
    conn->single[PElayer] = Cont_1;
    conn->num_1[PElayer]++;
    conn->act_single[PElayer] = Cont_1;
  }
}

VPE_POS_PTR GET_LINK(CONNECTION *conn)
{
  VPE_POS_PTR pos;

  if (conn->act_single[PElayer] != NULL) {
    pos = &(conn->act_single[PElayer]->pe);
    conn->act_single[PElayer] = conn->act_single[PElayer]->next;
  } else {
    if (conn->act_multiple[PElayer] != NULL) {
      pos = &(conn->act_multiple[PElayer]->pe[conn->act_multiple_num[PElayer]]);
      if (conn->act_multiple_num[PElayer] = CONTAINER_N_SIZE-1) {
        conn->act_multiple[PElayer] = conn->act_multiple[PElayer]->next;
        conn->act_multiple_num[PElayer] = 0;
      } else {
        conn->act_multiple_num[PElayer]++;
      }
    } else {
      conn->act_single[PElayer] = conn->single[PElayer];
      conn->act_multiple[PElayer] = conn->multiple[PElayer];
      pos = NULL;
    }
  }
  return pos;
}

void INIT_PARAMETER_CONNECTION(PARAMETER_CONNECTION *conn, INTEGER index, CONFIGURATION config)
{
  if (conn->next = NULL) {
    conn->parameter = index;
    INIT_CONNECTION(conn->connections, config);
  } else {
    PARAMETER_CONNECTION *New_Conn = (PARAMETER_CONNECTION *)
				     malloc(sizeof(PARAMETER_CONNECTION));
    New_Conn->parameter = index;
    INIT_CONNECTION(New_Conn->connections, config);
    New_Conn->next = conn;
    conn = New_Conn;
  }
}

void REGISTER_PARAMETER_LINK(PARAMETER_CONNECTION *conn, INTEGER index, plural CARDINAL dest[])
{
  PARAMETER_CONNECTION *loop_connection = conn;

  while (loop_connection != NULL && loop_connection->parameter != index) {
    loop_connection = loop_connection->next;
  }
  if (loop_connection == NULL) {
    fprintf(stderr, "unknown connection index: %d\n", index);
    exit(1);
  }
  REGISTER_LINK(loop_connection->connections, dest);
}

VPE_POS_PTR GET_PARAMETER_LINK(PARAMETER_CONNECTION *conn, INTEGER index)
{
  PARAMETER_CONNECTION *loop_connection = conn;

  while (loop_connection != NULL && loop_connection->parameter != index) {
    loop_connection = loop_connection->next;
  }
  if (loop_connection == NULL) {
    fprintf(stderr, "unknown connection index: %d\n", index);
    exit(1);
  }
  return GET_LINK(loop_connection->connections);
}

void psend(char target[], CARDINAL id, plural char source[], CARDINAL size)
{
  CARDINAL i;

  for (i = 0; i < size; i++) {
    target[i] = proc[ID2proc(id)].source[ID2layer(id)*size+i];
  }
}

void pfetch(plural char target[], CARDINAL id, char source[], CARDINAL size)
{
  CARDINAL i;

  for (i = 0; i < size; i++) {
    proc[ID2proc(id)].target[ID2layer(id)*size+i] = source[i];
  }
}

plural char *xfetchpS(CARDINAL dist, plural char *from, CARDINAL size)
{
  CARDINAL i;
  plural char *to = p_malloc(size);

  for (i = 0; i < size; i++) {
    to[i] = xnetpS[dist].from[i];
  }
  return to;
}

void reduce_DPU(ACTIVESET actives, plural char *values, void (*red_func)(), CARDINAL size)
{
  CARDINAL i;

  for(i = 1; i < nyproc; i *= 2) {
    if ((iyproc+1)%(i*2) == 1) {
      ACTIVESET aset = actives && xnetpS[i].actives;

      if (aset) {
	DPUConfig.configgroup->activeset = &(aset);
	(*red_func)(values, values, xfetchpS(i, values, size), DPUConfig);
      }
    }
  }
}

/*======================================================================*/

/*
 *	Implementation of standard parallaxis functions
 */

/*----------------------------------------------------------------------*/

/*
 * Definition of ABS. ABS is overloaded in Parallaxis, therefore must be
 * splitted into several versions in MPL. The macro versions are defined in
 * the header file.
 *  ABS   -> Macro used when expression is C-constant (allows backend
 *           compiler optimization)
 *  ABSC  -> Macro used when expression is CARDINAL
 *  ABSCV -> Macro used when expression is CARDINAL vector
 *  ABSI  -> Function used when expression is INTEGER
 *  ABSIV -> Function used when expression is INTEGER vector
 *  ABSR  -> Function used when expression is REAL
 *  ABSRV -> Function used when expression is REAL vector
 */

INTEGER ABSI				  /* INTEGER version */
#ifdef __STDC__
(register INTEGER x)
#else
(x) register INTEGER x;
#endif
{
  return (x < 0 ? -x : x);
}

plural INTEGER ABSIV			  /* INTEGER vector version */
#ifdef __STDC__
(register plural INTEGER x)
#else
(x) register plural INTEGER x;
#endif
{
  return (x < 0 ? -x : x);
}

REAL ABSR				  /* REAL version */
#ifdef __STDC__
(register REAL x)
#else
(x) register REAL x;
#endif
{
  return (x < 0 ? -x : x);
}

plural REAL ABSRV			  /* REAL vector version */
#ifdef __STDC__
(register plural REAL x)
#else
(x) register plural REAL x;
#endif
{
  return (x < 0 ? -x : x);
}

CHAR CAP
#ifdef __STDC__
(register CHAR ch)
#else
(ch) register CHAR ch;
#endif
{
   return ((ch >= 'a') && (ch <= 'z') ? ch - 'a' + 'A' : ch);
}

plural CHAR CAPV
#ifdef __STDC__
(register plural CHAR ch)
#else
(ch) register plural CHAR ch;
#endif
{
   return ((ch >= 'a') && (ch <= 'z') ? ch - 'a' + 'A' : ch);
}

CARDINAL DIM2ID(CONFIGURATION conf, INTEGER dims[])
/* Expects an index and returns the corresponding ID number in the range */
/* from 1 to LEN(config). 'dim' is in the range 1..RANK(conf). */
{
  register CARDINAL id = 0,
		    dimension;

  for (dimension = RANK(conf); dimension >= 1; dimension--) {
    id = id * LEN1(conf, dimension) +
	 (dims[dimension-1] - LOWER(conf, dimension));
  }
  return id+1;
}

plural CARDINAL DIM2IDV(CONFIGURATION conf, plural INTEGER *dims[])
{
  register CARDINAL dimension;
  register plural CARDINAL id = 0;

  for (dimension = RANK(conf); dimension >= 1; dimension--) {
    id = id * LEN1(conf, dimension) +
	 (dims[dimension-1][PElayer] - LOWER(conf, dimension));
  }
  return id+1;
}

plural INTEGER DIM(CONFIGURATION conf, CARDINAL dim)
/* Returns a vector with the indices of the dimension 'dim'. */
/* 'dim' shall be in the range 1..RANK(conf). */
{
  register CARDINAL dimension,
		    divisor = 1;

  for (dimension = 1; dimension < dim; dimension++) {
    divisor *= LEN1(conf, dimension);
  }
  return ((ID(conf)-1) / divisor) % LEN1(conf, dim) + LOWER(conf, dim);
}

CARDINAL LEN1(CONFIGURATION conf, CARDINAL dim)
{
  return conf.dim[dim-1].upper-conf.dim[dim-1].lower+1;
}


/*======================================================================*/

/*
 *	Implementation of standard procedures
 */

extern void argv
#ifdef __STDC__
(CARDINAL index, CHAR *arg)
#else
(index, arg)
CARDINAL index;
CHAR *arg;
CARDINAL arg_length;
#endif
{
  (void) strncpy( (char *) arg, SYSTEM_argv[index], strlen(SYSTEM_argv[index]) );
}

void CloseInput()
{
  Done = (fclose(SYSTEM_stdin) == 0);
  SYSTEM_stdin = stdin;
}

void CloseOutput()
{
  Done = (fclose(SYSTEM_stdout) == 0);
  SYSTEM_stdout = stdout;
}

void OpenInput
#ifdef __STDC__
(CHAR *s)
#else
(s) CHAR *s;
#endif
{
  char filename[1024];

  if (strlen((char *) s) == 0) {
    fprintf(stdout, "Enter filename to read from: ");
    fflush(stdout);
    fscanf(stdin, "%s", filename);
  } else {
    strcpy(filename, (char *) s);
  }
  Done = ((SYSTEM_stdin = fopen(filename, "r")) != NULL);
}

void OpenOutput
#ifdef __STDC__
(CHAR *s)
#else
(s) CHAR *s;
#endif
{
  char filename[1024];

  if (strlen((char *) s) == 0) {
    fprintf(stdout, "Enter filename to write to: ");
    fflush(stdout);
    fscanf(stdin, "%s", filename);
  } else {
    strcpy(filename, (char *) s);
  }
  Done = ((SYSTEM_stdout = fopen(filename, "w")) != NULL);
}

/*-------------------- Write Commands ------------------------------*/

/*
 * WriteLn
 */

void WriteLn(void)
{
     fprintf(SYSTEM_stdout, "\n");
}



/*
 *   Write
 */

void Write(CHAR c)
{
     Done = (fprintf(SYSTEM_stdout, "%c", c) != EOF);
}

void WriteV (CONFIGURATION conf, plural CHAR c[])
{
     CARDINAL index[RANK(conf)];
     CARDINAL dimension;
     CARDINAL id=0;
     BOOLEAN Done = TRUE;

     for (dimension=0; dimension<RANK(conf); dimension++)
	  index[dimension] = 0;

     do
     {
	  Done = Done && (fprintf(SYSTEM_stdout, "%c", 
				  proc[ID2proc(id)].c[ID2layer(id)]) != EOF);
	  id++;
	  for (dimension=0; dimension<RANK(conf); dimension++)
	  {
	       if (++index[dimension] == conf.dim[dimension].size)
	       {
		    WriteLn();
		    index[dimension] = 0;
	       }
	  }
     } while (id < LEN(conf));
}


/*
 *   WriteString
 */

void WriteString(CHAR *s)
{
     Done = (fprintf(SYSTEM_stdout, "%s", s) != EOF);
}

void WriteStringV(CONFIGURATION conf, plural CHAR  * plural s[])
{
     
     CARDINAL index[RANK(conf)];
     CARDINAL dimension;
     CARDINAL id=0;

     for (dimension=0; dimension<RANK(conf); dimension++)
	  index[dimension] = 0;

     do
     {
	  if (iproc == ID2proc(id))
	       /* p_fprintf returns no success value, so we assume that it */
	       /* is always successful */
	       p_fprintf(SYSTEM_stdout, "%s", s[ID2layer(id)]);
	  id++;
	  for (dimension=0; dimension<RANK(conf); dimension++)
	  {
	       if (++index[dimension] == conf.dim[dimension].size)
	       {
		    WriteLn();
		    index[dimension] = 0;
	       }
	  }
     } while (id < LEN(conf));
     Done = TRUE;			  /* see comment about p_fprintf above */
}

/*
 *   WriteInt
 */

void WriteInt(INTEGER i, CARDINAL len)
{
     char format[16];

     sprintf(format, "%%%luld", len);	  /* create printf-format-string */
     Done = (fprintf(SYSTEM_stdout, format, i) != EOF);
}

void WriteIntV(CONFIGURATION conf, plural INTEGER i[], CARDINAL len)
{
     
     CARDINAL index[RANK(conf)];
     CARDINAL dimension;
     CARDINAL id=0;
     char format[16];

     sprintf(format, "%%%luld", len);	  /* create printf-format-string */

     Done = TRUE;

     for (dimension=0; dimension<RANK(conf); dimension++)
	  index[dimension] = 0;

     do
     {
	  Done = Done && (fprintf(SYSTEM_stdout, format, 
				 proc[ID2proc(id)].i[ID2layer(id)]) != EOF);
	  id++;
	  for (dimension=0; dimension<RANK(conf); dimension++)
	  {
	       if (++index[dimension] == conf.dim[dimension].size)
	       {
		    WriteLn();
		    index[dimension] = 0;
	       }
	  }
     } while (id < LEN(conf));
}

/*
 *   WriteCard
 */

void WriteCard(INTEGER c, CARDINAL len)
{
     char format[16];

     sprintf(format, "%%%lulu", len);	  /* create printf-format-string */
     Done = (fprintf(SYSTEM_stdout, format, c) != EOF);
}

void WriteCardV(CONFIGURATION conf, plural INTEGER c[], CARDINAL len)
{
     
     CARDINAL index[RANK(conf)];
     CARDINAL dimension;
     CARDINAL id=0;
     char format[16];

     sprintf(format, "%%%lulu", len);	  /* create printf-format-string */

     Done = TRUE;

     for (dimension=0; dimension<RANK(conf); dimension++)
	  index[dimension] = 0;

     do
     {
	  Done = Done && (fprintf(SYSTEM_stdout, format, 
				  proc[ID2proc(id)].c[ID2layer(id)]) != EOF);
	  id++;
	  for (dimension=0; dimension<RANK(conf); dimension++)
	  {
	       if (++index[dimension] == conf.dim[dimension].size)
	       {
		    WriteLn();
		    index[dimension] = 0;
	       }
	  }
     } while (id < LEN(conf));
}


/*
 *   WriteBool
 */

void WriteBool(BOOLEAN b, CARDINAL len)
{
     int i;

     Done = (fprintf(SYSTEM_stdout, "%s", b?"TRUE":"FALSE") != EOF);
     for (i=b?4:5; i<len; i++)
	  Done = Done && (fprintf(SYSTEM_stdout, " ") != EOF);
}

void WriteBoolV(CONFIGURATION conf, plural BOOLEAN b[], CARDINAL len)
{
     CARDINAL index[RANK(conf)];
     CARDINAL dimension;
     CARDINAL id=0;
     BOOLEAN DoneV = TRUE;

     Done = TRUE;

     for (dimension=0; dimension<RANK(conf); dimension++)
	  index[dimension] = 0;

     do
     {
	  WriteBool(proc[ID2proc(id)].b[ID2layer(id)], len);
	  DoneV = DoneV && Done;
	  id++;
	  for (dimension=0; dimension<RANK(conf); dimension++)
	  {
	       if (++index[dimension] == conf.dim[dimension].size)
	       {
		    WriteLn();
		    index[dimension] = 0;
	       }
	  }
     } while (id < LEN(conf));
     Done = DoneV;
}


/*
 *   WriteReal
 */

void WriteReal(REAL r, CARDINAL len)
{
     char format[16];

     sprintf(format, "%%%lue", len);	  /* create printf-format-string */
     Done = (fprintf(SYSTEM_stdout, format, r) != EOF);
}

void WriteRealV(CONFIGURATION conf, plural REAL r[], CARDINAL len)
{
     
     CARDINAL index[RANK(conf)];
     CARDINAL dimension;
     CARDINAL id=0;
     char format[16];

     sprintf(format, "%%%lue", len);	  /* create printf-format-string */

     Done = TRUE;

     for (dimension=0; dimension<RANK(conf); dimension++)
	  index[dimension] = 0;

     do
     {
	  Done = Done && (fprintf(SYSTEM_stdout, format, 
				  proc[ID2proc(id)].r[ID2layer(id)]) != EOF);
	  id++;
	  for (dimension=0; dimension<RANK(conf); dimension++)
	  {
	       if (++index[dimension] == conf.dim[dimension].size)
	       {
		    WriteLn();
		    index[dimension] = 0;
	       }
	  }
     } while (id < LEN(conf));
}


/*
 *   WriteFixPt
 */

void WriteFixPt(REAL r, CARDINAL len, CARDINAL dec)
{
     char format[32];

     sprintf(format, "%%%lu.%luf", len, dec); /* create printf-format-string */
     Done = (fprintf(SYSTEM_stdout, format, r) != EOF);
}

void WriteFixPtV(CONFIGURATION conf, plural REAL r[], CARDINAL len, CARDINAL dec)
{
     
     CARDINAL index[RANK(conf)];
     CARDINAL dimension;
     CARDINAL id=0;
     char format[32];

     sprintf(format, "%%%lu.%luf", len, dec); /* create printf-format-string */

     Done = TRUE;

     for (dimension=0; dimension<RANK(conf); dimension++)
	  index[dimension] = 0;

     do
     {
	  Done = Done && (fprintf(SYSTEM_stdout, format, 
				  proc[ID2proc(id)].r[ID2layer(id)]) != EOF);
	  id++;
	  for (dimension=0; dimension<RANK(conf); dimension++)
	  {
	       if (++index[dimension] == conf.dim[dimension].size)
	       {
		    WriteLn();
		    index[dimension] = 0;
	       }
	  }
     } while (id < LEN(conf));
}


/*
 *   ID2DIM
 */

void ID2DIM(CONFIGURATION conf, CARDINAL id, INTEGER dims[])
/* Returns the DIM values of PE 'id' */
{
  register CARDINAL dimension;

  id--;
  for (dimension = 1 ; dimension <= RANK(conf); dimension++) {
    dims[dimension-1] = id % LEN1(conf, dimension) +
			  LOWER(conf, dimension);
    id /= LEN1(conf, dimension);
  }
}

/* In most cases the ID values in 'id' don't correspond to the PE they reside */
/* on. Then we need the router to get the right DIM values, otherwise we don't. */

void ID2DIMV(CONFIGURATION conf, plural CARDINAL id[], plural INTEGER *dims[], BOOLEAN WITH_ROUTER)
{
  register CARDINAL dimension;

  for (dimension = 0; dimension < RANK(conf); dimension++) {
    all {
      plural INTEGER dimvals[conf.configgroup->layers];

      FOR_ALL(conf) {
	/* compute dimvalue for every id */
	dimvals[PElayer] = DIM(conf, dimension+1);
	dims[dimension][PElayer] = WITH_ROUTER ?
	  router[ID2proc(id[PElayer])].dimvals[ID2layer(id[PElayer])] :
	  dimvals[PElayer];
      }
    }
  }
}

/*======================================================================*/

/*
 *	Implementation of set operators
 */

unsigned long SET_RANGE1
#ifdef __STDC__
(register CARDINAL lo, register CARDINAL hi)
#else
(lo, hi) register CARDINAL lo, hi;
#endif
{
    return (lo <= hi ? ~0X0L >> lo << lo + SYSTEM_MaxSet - hi >> SYSTEM_MaxSet - hi : 0X0L);
}

/*
 *	Implementation of compiler functions
 */

void CaseError
#ifdef __STDC__
(char file[], int line)
#else
(file, line) char file[]; int line;
#endif
{
  (void) fprintf (stderr, "\"%s\", line %1d: case expression out of range\n", file, line);
  exit(1);
}

void ReturnError
#ifdef __STDC__
(char file[], int line)
#else
(file, line) char file[]; int line;
#endif
{
  (void) fprintf (stderr, "\"%s\", line %1d: missing return from function\n", file, line);
  exit(1);
}

void ReturnErrorV
#ifdef __STDC__
(CONFIGURATION config, char file[], int line)
#else
(config, file, line) CONFIGURATION config; char file[]; int line;
#endif
{
  FOR_ALL_ACTIVE(config) {
    (void) fprintf (stderr, "\"%s\", line %1d: missing return from function\n", file, line);
    exit(1);
  }
}

void InternalError
#ifdef __STDC__
(char *text)
#else
(text) char *text;
#endif
{
     (void) fprintf (stderr, "Internal Error (corrupt translation)!\n");
     (void) fprintf (stderr, "Reason: %s\n", text);
     (void) fprintf (stderr, "Please try to localize the problem and send "
		     "a bug report together with an\n"
		     "appropriate code fragment to the authors.\n");
     exit(1);
}

/*
 *	Main program
 */

unsigned int SYSTEM_argc;
char **SYSTEM_argv;
char **SYSTEM_envp;

FILE *SYSTEM_stdin;
FILE *SYSTEM_stdout;
char termCH;
BOOLEAN Done;

CARDINAL PElayer;

DIMENSION DPUDim[1];
CONFIGGROUP DPUGroup[1];
CONFIGURATION DPUConfig;

void INIT_MODULE(int argc, char *argv[], char *envp[])
{
  FREELIST.num_1 = 0;
  FREELIST.single = NULL;

  SYSTEM_argc = argc - 1;
  SYSTEM_argv = argv;
  SYSTEM_envp = envp;

  SYSTEM_stdin = stdin;
  SYSTEM_stdout = stdout;

  DPUDim->lower = 0;
  DPUDim->upper = nproc-1;
  DPUDim->size  = nproc;

  DPUGroup->number = 0;
  DPUGroup->VPEs   = nproc;
  DPUGroup->layers = 1;
  DPUGroup->activeset = NULL; /* activeset is set in reduce */

  DPUConfig.name = "DPUConfig";
  DPUConfig.rank = 1;
  DPUConfig.dim  = DPUDim;
  DPUConfig.configgroup = DPUGroup;

  srand48(time(NULL)); /* init random number generator */
}


void EXIT_MODULE(void)
{
  exit(0);
}


