#include <memory.h>
#include <stdio.h>
#include <string.h>
#include <time.h>
#include <math.h>
#include <stdlib.h>        /* qsort */
#include <sys/time.h>      /* for time-measuring */
#include <sys/resource.h>  /* dito               */

#include "SYSTEM_PVM.h"


/* uncomment this line to print out the distribution of PEs */
/* over the processors                                      */
/* #define TestIDdistribution */


/*********** Prototypes of procedures which are not exported ***********/

extern void INIT_ACTIVE_STACK ARGS((CONFIGURATION *Config));
/* Allocates memory for the ActiveStack-Pointer of Config and initializes */
/* all PEs to active.                                                     */

extern void FREE_ACTIVE_STACK ARGS((CONFIGURATION *Config));
/* frees the memory of the active stack of Config */

extern CARDINAL num_CR ARGS((CONFIGURATION Config));
/* returns the number of Carrige-Returns to print while printing data */
/* from the ID-range start_ID to end_ID */ 

extern void INIT_ACTIVE_STACK ARGS((CONFIGURATION *Config));
/* Allocates memory for the ActiveStack-Pointer of Config and initializes */
/* all PEs to active.                                                     */

extern void INIT_ELSE_STACK ARGS((CONFIGURATION *Config));
/* Allocates memory for the ElseStack-Pointer of Config and initializes */
/* all PEs to active.                                                   */

extern void FREE_ELSE_STACK ARGS((CONFIGURATION *Config));
/* frees the memory of the else stack of Config */

extern void POP_ELSE ARGS((CONFIGURATION Config));
/* Pop the previous else status. */

extern void Id_Range_Error ARGS((CARDINAL id, CONFIGURATION Config));
/* stops the program and prints an error-message: */
/* "Id-Range Error: no id (nr) in Configuration Config." */
/* This error occurs when accessing an PE by PE-selection like vector<<id>> */
/* not yet implemented in the start-process!!! */

extern void EndParallaxisError ARGS((int ExitCode));
/* end parallaxis program with exitcode */
/* this procedure can be called asynchronously from any task */
/* all other tasks are killed */

extern void EndParallaxis ARGS((int ExitCode, struct rusage Zeit1, long u_sec, long u_msec, long s_sec, long s_msec));
/* end parallaxis program with exitcode */
/* this procedure must be called synchronously from every task */

#define IS_IN_LOCAL_DIM_RANGE(Config,dim) IS_IN_LOCAL_ID_RANGE(Config, DIM2ID(Config, dim))
/* checks if an given array of dimension-indexes is in the local range of */
/* id-numbers that this (real) processor holds */

extern BOOLEAN IS_IN_CONFIG ARGS((CONFIGURATION Config, INTEGER *IndexArray));

extern GENERIC_CONNECTIONS_PTR ALL_GENERIC_CONNECTIONS;
extern GENERIC_PARAMETER_CONNECTIONS_PTR ALL_GENERIC_PARAMETER_CONNECTIONS;
extern long u_sec;
extern long u_msec;
extern long s_sec;
extern long s_msec;      /* for elapsed time */
extern struct rusage StartTime;


/************ Implementation of standard functions *************/

CARDINAL argc()
{
  return SYSTEM_argc;
}


void argv
#ifdef __STDC__
(CARDINAL index, CHAR arg[])
#else
(index, arg)
CARDINAL index;
CHAR *arg;
#endif
{
  if (index <= argc()) {
    (void) strcpy(arg, SYSTEM_argv[index]);
  }
}


INTEGER ABSI
#ifdef __STDC__
(register INTEGER x)
#else
(x) register INTEGER x;
#endif
{
  return (x < 0 ? -x : x);
}


REAL ABSR
#ifdef __STDC__
(register REAL x)
#else
(x) register REAL x;
#endif
{
  return (x < 0 ? -x : x);
}


CHAR CAP
#ifdef __STDC__
(register CHAR ch)
#else
(ch) register CHAR ch;
#endif
{
   return (ch >= 'a' && ch <= 'z' ? ch - 'a' + 'A' : ch);
}

CARDINAL MOD1
#ifdef __STDC__
(INTEGER x, INTEGER y)
#else
(x, y)
INTEGER x, y;
#endif
{
  return ((x % y < 0) ? x % y + (y < 0 ? -y : y) : x % y );
}

/* Implementation of parallaxis standard functions **************************/

REAL ArcTan2
#ifdef __STDC__
(REAL x, REAL y)
#else
(x, y)
REAL x, y;
#endif
{
  return x == 0 && y == 0 ? 0 : atan2(x, y);
}

/*
 *	Implementation of set operators
 */

unsigned long SET_RANGE1
#ifdef __STDC__
(register CARDINAL lo, register CARDINAL hi)
#else
(lo, hi) register CARDINAL lo, hi;
#endif
{
  return (lo <= hi ? ~0X0L >> lo << (lo + SYSTEM_MaxSet - hi) >> (SYSTEM_MaxSet - hi) : 0X0L);
}



CARDINAL RandomCard
#ifdef __STDC__
(void)
#else
()
#endif
{
  CARDINAL rand_num;

  if (me == 0) {
    rand_num = RandomCard1;
    (void) pvm_initsend(PvmDataDefault);
    PVM_PKCARDINAL(&rand_num, 1);
    (void) pvm_mcast(&task_ids[1], n_tasks - 1, INTER_TASK_TAG);
  } else {
    (void) pvm_recv(task_ids[0], INTER_TASK_TAG);
    PVM_UPKCARDINAL(&rand_num, 1);
  }
  return rand_num;
}


INTEGER RandomInt
#ifdef __STDC__
(void)
#else
()
#endif
{
  INTEGER rand_num;

  if (me == 0) {
    rand_num = RandomInt1;
    (void) pvm_initsend(PvmDataDefault);
    PVM_PKINTEGER(&rand_num, 1);
    (void) pvm_mcast(&task_ids[1], n_tasks - 1, INTER_TASK_TAG);
  } else {
    (void) pvm_recv(task_ids[0], INTER_TASK_TAG);
    PVM_UPKINTEGER(&rand_num, 1);
  }
  return rand_num;
}


REAL RandomReal
#ifdef __STDC__
(void)
#else
()
#endif
{
  REAL rand_num;

  if (me == 0) {
    rand_num = RandomReal1;
    (void) pvm_initsend(PvmDataDefault);
    PVM_PKREAL(&rand_num, 1);
    (void) pvm_mcast(&task_ids[1], n_tasks - 1, INTER_TASK_TAG);
  } else {
    (void) pvm_recv(task_ids[0], INTER_TASK_TAG);
    PVM_UPKREAL(&rand_num, 1);
  }
  return rand_num;
}


BOOLEAN RandomBool
#ifdef __STDC__
(void)
#else
()
#endif
{
  BOOLEAN rand_num;

  if (me == 0) {
    rand_num = RandomBool1;
    (void) pvm_initsend(PvmDataDefault);
    PVM_PKBOOLEAN(&rand_num, 1);
    (void) pvm_mcast(&task_ids[1], n_tasks - 1, INTER_TASK_TAG);
  } else {
    (void) pvm_recv(task_ids[0], INTER_TASK_TAG);
    PVM_UPKBOOLEAN(&rand_num, 1);
  }
  return rand_num;
}


CHAR RandomChar
#ifdef __STDC__
(void)
#else
()
#endif
{
  CHAR rand_num;

  if (me == 0) {
    rand_num = RandomChar1;
    (void) pvm_initsend(PvmDataDefault);
    PVM_PKCHAR(&rand_num, 1);
    (void) pvm_mcast(&task_ids[1], n_tasks - 1, INTER_TASK_TAG);
  } else {
    (void) pvm_recv(task_ids[0], INTER_TASK_TAG);
    PVM_UPKCHAR(&rand_num, 1);
  }
  return rand_num;
}





/* Parallaxis specific procedures ----------------------------------------- */

void ID2DIM
#ifdef __STDC__
(CONFIGURATION Config, CARDINAL Id, INTEGER Dim[])
#else
(Config, Id, Dim) CONFIGURATION Config; CARDINAL Id; INTEGER Dim[];
#endif
{
  CARDINAL Dimension;

  if (1 <= Id && Id <= LEN(Config)) {
    Id -= 1;
    for (Dimension = 1 ; Dimension <= RANK(Config); Dimension++) {
      Dim[Dimension - 1] = Id % LEN1(Config, Dimension) +
        LOWER(Config, Dimension);
      Id /= LEN1(Config, Dimension);
    }
  }
}


/* Functions for Communication-Management  --------------------------------- */

/* I/O-tags are defined in "shared_macros.h" */

void INIT_PVM
/* Enrolls this process in PVM and starts (n_tasks) child-processes.    */
/* In mytid the tid of the calling process is given and in task_ids the */
/* tids of the started tasks are given.                                 */
/* Passes the first message to all tasks which contains all task_ids.   */
#ifdef __STDC__
(int *my_tid, int *parent_tid, int *io_tid, CARDINAL *n_tasks, int
 **task_ids, int *me, CARDINAL *MaxDiff, CARDINAL *MinPEs)
#else
(my_tid, parent_tid, io_tid, n_tasks, task_ids, me, MaxDiff, MinPEs) 
  int *my_tid;
  int *parent_tid; 
  int *io_tid;
  CARDINAL *n_tasks;
  int **task_ids;
  int *me;
  CARDINAL *MaxDiff;
  CARDINAL *MinPEs;
#endif
{
  /* enroll in pvm */
  *my_tid = pvm_mytid();
  *io_tid = pvm_parent();

  /* check if this program is not started directly */
  if (*io_tid == PvmNoParent) {
    fprintf(stderr, "This program can not be started directly\n"
	   "Start this program with the 'start'-command:\n"
	   "Usage: start n name [arguments ...]\n"
	   "where n is the number of nodes to start the program name on.\n"
	   "Example: start 5 %s\n", SYSTEM_argv[0]);
    pvm_exit();
    exit(-1);
  }

  /* receive INIT-message from start-process */
  (void) pvm_recv(*io_tid, INIT_TAG);
  PVM_UPKCARDINAL(n_tasks, 1);
  /* allocate memory for (n_tasks) task-ids */
  *task_ids = (int*) malloc(sizeof(int) * (*n_tasks));
  (void) pvm_upkint(*task_ids, *n_tasks, 1);

  /* compute own index (me) in tids-array (task_ids) */
  for(*me=0; *me<*n_tasks; (*me)++)    if(*my_tid == (*task_ids)[*me]) { break; }

  /* set parent_tid to tid of task 0 */
  *parent_tid = *task_ids[0];     /* DON'T CHANGE THIS LINE! */

  /* unpack MaxDiff and MinPEs */
  PVM_UPKCARDINAL(MaxDiff, 1);
  PVM_UPKCARDINAL(MinPEs,  1);

  (void) pvm_upkint(&SYSTEM_argc, 1, 1);
  SYSTEM_argv = (char **) malloc(sizeof(char*) * (SYSTEM_argc + 1));
  {
    int i, l;

    for (i=0; i<=SYSTEM_argc; i++) {
      (void) pvm_upkint(&l, 1, 1);
      SYSTEM_argv[i] = (char*) malloc(sizeof(char) * (l+1));
      (void) pvm_upkstr(SYSTEM_argv[i]);
    }
  }

} /* end INIT_PVM */


/* two functions for use in the construct that is written for the */
/* LOAD-statement */

CARDINAL NUM_ACTIVES_OF_PREDS
#ifdef __STDC__
(CONFIGURATION Config)
#else
(Config)
CONFIGURATION Config;
#endif
/* Returns the number of active PEs of all predesessor processors in Config */
/* and uses tag for communication. This call is part of the LOAD-statement */
/* and it is absolutely necessary that a corresponding call of */
/* NUM_ELEMS_LOADED follows, otherwise program execution will stop in a */
/* deadlock. */
{
  CARDINAL actives;

  if (me == 0) {
    /* no predesessor, return 0 */
    return 0;
  } else {
    if (me > 0 && me < Config.u_proc) {
      /* predesessor exists, and processor belongs to Config */
      /* receive number of active PEs of Configuration from call of */
      /* NUM_ELEMS_LOADED of predesessor */
      (void) pvm_recv(task_ids[me-1], INTER_TASK_TAG);
      PVM_UPKCARDINAL(&actives, 1);
      return actives;
    } else {
      /* processor does not belong to this Configuration */
      return 0;
    }
  }
}


extern CARDINAL NUM_ELEMS_LOADED 
#ifdef __STDC__
(CONFIGURATION Config, CARDINAL n_elems, BOOLEAN send_counter)
#else
(Config, n_elems, send_counter)
CONFIGURATION Config; CARDINAL n_elems; BOOLEAN send_counter;
#endif
/* Returns the number of elements loaded in the array (only if send_counter */
/* is TRUE, otherwise returns 0). n_elems is the number of elems loaded so */
/* far in all predesessors and this processor himself. */
/* This call is the other part of the LOAD-statement and it */
/* is absolutely necessary that a corresponding call of NUM_ACTIVES_OF_PREDS */
/* has been executed before, otherwise program execution will stop in a */
/* deadlock. */
{
  CARDINAL elems_loaded;

  if (me+1 < Config.u_proc) {
    /* i have a successor which belongs to this Config */
    /* send him number of elems */
    (void) pvm_initsend(PvmDataDefault);
    PVM_PKCARDINAL(&n_elems, 1);
    (void) pvm_send(task_ids[me + 1], INTER_TASK_TAG);
    if (send_counter) {
      (void) pvm_recv(task_ids[Config.u_proc - 1], INTER_TASK_TAG);
      PVM_UPKCARDINAL(&elems_loaded, 1);
      return elems_loaded;
    } else {
      /* no data exchange, return 0 (has no special meaning) */
      return 0;
    }

  } else {
    if (me+1 == Config.u_proc) {
      /* i am the last processor which belongs to Config */
      /* send n_elems to neigbours */
      if (send_counter) {
	if (n_tasks > 1) {
	  /* pack n_elems */
	  (void) pvm_initsend(PvmDataDefault);
	  PVM_PKCARDINAL(&n_elems, 1);
	}
	if (me > 0) {
	  /* send to left neighbours */
	  (void) pvm_mcast(task_ids, Config.u_proc - 1, INTER_TASK_TAG);
	}
	if (n_tasks > Config.u_proc) {
	  /* send to right neighbours */
	  (void) pvm_mcast(&task_ids[Config.u_proc], 
			   n_tasks - Config.u_proc, INTER_TASK_TAG);
	}
	/* i was the last processor of Config, so return n_elems */
	return n_elems;
      } else {
	/* no data exchange, return 0 (has no special meaning) */
	return 0;
      }

    } else {
      /* i don't belong to this Config */
      if (send_counter) {
	(void) pvm_recv(task_ids[Config.u_proc - 1], INTER_TASK_TAG);
	PVM_UPKCARDINAL(&elems_loaded, 1);
	return elems_loaded;
      } else {
	/* no data exchange, return 0 (has no special meaning) */
	return 0;
      }
    }
  }
}


/* the following 2 functions are for use in the construct that is written */
/* for the STORE-statement */

void PREPARE_UNPACKING
/* This routine is part of the STORE-statement. It makes sure that a message */
/* is received and at least one data-element is available to unpack. If the */
/* message is empty, a message from the next task is received. If no task */
/* is left in Config, data_available becomes FALSE. The counter num_unpacked */
/* has to be incremented after one data-element was unpacked. */
/* num_unpacked: (0)  number of unpacked data-items of current message */
/* num_data:     (0)  number of data-elements in current message */
/* t_ind:        (-1) index of task from which is currently processed */
/*                    message  */
/* values_available: (TRUE) becomes FALSE, if all tasks and all */
/*                          data-elements are processed */
/* num_total:      (0) when values_available is FALSE then here is the total */
/*                     number of data-items received */
/* type_size:        size of one data-element in bytes (result of sizeof(x)) */
/* Config:            Configuration of the data */
/* (the values in parenthes are the initialization-values before the first */
/* call of PREPARE_UNPACKING) */
#ifdef __STDC__
(CARDINAL *num_unpacked, CARDINAL *num_data, CARDINAL *t_ind, 
 BOOLEAN *values_available, unsigned int type_size, 
 CONFIGURATION Config)
#else
(num_unpacked, num_data, t_ind, values_available, type_size, Config)
CARDINAL *num_unpacked;
CARDINAL *num_data;
CARDINAL *t_ind;
BOOLEAN  *values_available;
unsigned int type_size;
CONFIGURATION Config;
#endif
{
  int bufid, bytes, dummy;

  while (*num_unpacked == *num_data && *values_available) {
    (*t_ind)++;
    if (*t_ind == Config.u_proc) {
      *values_available = FALSE;
    } else {
      *num_unpacked = 0;
      /* receive new message */
      bufid = pvm_recv(task_ids[*t_ind], INTER_TASK_TAG);
      /* get number of bytes of message */
      (void) pvm_bufinfo(bufid, &bytes, &dummy, &dummy);
      /* compute number of data-elements in message */
      *num_data = bytes / type_size;
    }
  }
}


void FREE_MSG_BUFFERS
/* frees all buffer of unreceived messages from PREPARE_UNPACKING */
#ifdef __STDC__
(CARDINAL t_ind, CONFIGURATION Config)
#else
(t_ind, Config) 
CARDINAL t_ind; CONFIGURATION Config;
#endif
{
  t_ind++;
  while (t_ind < Config.u_proc) {
    pvm_recv(task_ids[t_ind], INTER_TASK_TAG);
    t_ind++;
  }
}


/* ------------------------------------------------------------------------- */


/* ROUTINES for CONFIGURATIONs */


void INIT_CONFIGURATION
/* Initializes a Configuration with the value of Rank and the length of the  */
/* Configuration with 1. Also allocates memory for the Lower-, Upper- and    */
/* Len-array. After this procedure several calls of SET_BOUNDS (beginning    */
/* with dimension-index 0!) are needed to compute the value of Len_conf and  */
/* to initialize the Lower-, Upper- and Len-array.                           */
/* Finally a call of INIT_ID_RANGE is needed to distribute the virtual PEs   */
/* to the real Processors.                                                   */
#ifdef __STDC__
(char Name[], CONFIGURATION *Config, CARDINAL Rank)
#else
(Name, Config, Rank) char *Name; CONFIGURATION *Config; CARDINAL Rank;
#endif
{
  strncpy(Config->Name, Name, 256);
  Config->Rank = Rank;
  Config->Lower = (INTEGER *) malloc(sizeof(INTEGER) * Rank);
  Config->Upper = (INTEGER *) malloc(sizeof(INTEGER) * Rank);
  Config->Len   = (CARDINAL *) malloc(sizeof(CARDINAL) * Rank);
  Config->Len_conf = 1;  /* this number is modified by each call of SET_BOUNDS */
  Config->Generics = NULL;
  Config->GenericsParam = NULL;
}


void SET_BOUNDS
/* Initializes one field of the Lower-, Upper- and Len-array of Config. */
/* with the Index Dim. The value of Len_conf is updated.                */
/* Dim starts with 0 and  represents the most left dimension.           */
#ifdef __STDC__
(CONFIGURATION *Config, CARDINAL Dim, INTEGER Lwb, INTEGER Upb)
#else
(Config, Dim, Lwb, Upb)
CONFIGURATION *Config; CARDINAL Dim; INTEGER Lwb; INTEGER Upb;
#endif
{
  Config->Lower[Dim] = Lwb;
  Config->Upper[Dim] = Upb;
  Config->Len[Dim]   = Upb - Lwb + 1;
  Config->Len_conf  *= Config->Len[Dim];
}


void INIT_ID_RANGE
#ifdef __STDC__
(CONFIGURATION *Config, CARDINAL num_real, CARDINAL MaxDiff, CARDINAL MinPEs)
#else
(Config, num_real, MaxDiff, MinPEs) 
CONFIGURATION *Config;
CARDINAL      num_real;
CARDINAL      MaxDiff;
CARDINAL      MinPEs;
#endif
/* This procedure distributes the start_ID, end_ID and num_PEs to each      */
/* of the num_real tasks (except that there are less than num_real PEs      */
/* in this Configuration.                                                   */
/* The Configuration must already be initialized (uses Config->Len_conf). */
/* MaxDiff is the maximum difference in the number of PEs that are simulated */
/*         from two different Tasks. Is MUST be >= 1 !!! */
/* MinPEs  is the minmal number of PEs that are simulated by one Task. */
/*         MinPEs MUST be less than MaxDiff! If MinPEs is less than */
/*         MaxDiff/2, then it is granted that one task simulates at least */
/*         MinPEs, otherwise it may be less. */
{
  CARDINAL simulate;    /* number of Parts that are simulated by one Task  */
  CARDINAL one_more;    /* all Task with (me < one_more) simulate one Part */
			/* more than the other Tasks                       */
  CARDINAL j;           /* counter over tasks */
  CARDINAL act_ID;      /* start-ID of ID-range of actual Task */
  CARDINAL DivideDim;   /* dimension to divide the PEs */
  CARDINAL SizeOfPart;  /* Size of one Part when divided */
  CARDINAL mult;        /* factor to multiply SizeOfPart to get SizeOfPart */
			/* bigger than MinPEs */
  CARDINAL add;         /* 0 or 1 */
  CARDINAL NumParts;    /* number of Parts which is possible to divide */
			/* current Dimension in */
  CARDINAL Rest;        /* rest of PEs that are not enough for one Part */

  DivideDim = 0;             /* at the beginning,                           */
  NumParts = 1;              /* the whole Configuration is one part and the */
  SizeOfPart = LEN(*Config); /* size of one part is length of Configuration */
  while (SizeOfPart > MaxDiff) { /* with MaxDiff >= 1 termination is granted */
    /* compute NumParts und SizeOfPart */
    NumParts = NumParts * Config->Len[DivideDim];
    SizeOfPart = SizeOfPart / Config->Len[DivideDim];
    DivideDim++;
  }

  /* Now we have SizeOfPart, but with onedimensional Configurations it is    */
  /* always 1! Therefore we are now trying to increase it to make SizeOfPart */
  /* greater than MinPEs. */
  if (MinPEs > 0) {
    mult = MinPEs / SizeOfPart;
    add  = ((MinPEs % SizeOfPart) > 0 ? 1 : 0);
  } else {
    mult = 1;
    add  = 0;
  }
  if (SizeOfPart * (mult + add) <= MaxDiff) {
    /* it is possible to make SizeOfPart bigger than MinPEs without making */
    /* it bigger than MaxDiff */
    SizeOfPart = SizeOfPart * (mult + add);
  } else {
    SizeOfPart = SizeOfPart * mult;
  }

  /* now compute the new NumParts and the rest */
  NumParts = LEN(*Config) / SizeOfPart;
  Rest     = LEN(*Config) % SizeOfPart;

  simulate = NumParts / num_real;
  one_more = NumParts % num_real;

  if (simulate == 0)
    if (Rest > 0) 
      Config->u_proc = one_more + 1;
    else 
      Config->u_proc = one_more;
  else 
    Config->u_proc = num_real;
  
  /* allocate memory for the TASK_DATA-array */
  Config->ID_range = (TASK_DATA*) malloc(sizeof(TASK_DATA) * num_real);

  /* initialize loops */
  j = 0;
  act_ID = 1;

  /* loop 1: (simulate + 1) * SizeOfPart PEs per processor */
  while (j < one_more) {
    Config->ID_range[j].start_ID = act_ID;
    Config->ID_range[j].end_ID   = act_ID - 1 + (simulate + 1) * SizeOfPart;
    Config->ID_range[j].num_PEs  = (simulate + 1) * SizeOfPart;
    act_ID += (simulate + 1) * SizeOfPart;
    j++;
  }

  /* processor 'one_more' gets the Rest-PEs */
  Config->ID_range[j].start_ID = act_ID;
  Config->ID_range[j].end_ID   = act_ID -1 + simulate * SizeOfPart + Rest;
  Config->ID_range[j].num_PEs  = simulate * SizeOfPart + Rest;
  act_ID += simulate * SizeOfPart + Rest;
  j++;

  /* loop 2: only (simulate) * SizeOfPart PEs per processor */
  while (j < Config->u_proc) {
    Config->ID_range[j].start_ID = act_ID;
    Config->ID_range[j].end_ID   = act_ID - 1 + simulate * SizeOfPart;
    Config->ID_range[j].num_PEs  = simulate * SizeOfPart;
    act_ID += simulate * SizeOfPart;
    j++;
  }

  for (j = 0; j < Config->u_proc; j++)
    Config->ID_range[j].num_breaks = num_CR(*Config);

#ifdef TestIDdistribution
  { int i;
    WriteString("ID-Verteilung:");
    WriteLn();
    for (i = 0; i < Config->u_proc; i++) {
      WriteString("Start-ID von Task ");
      WriteInt( (INTEGER) i, 3);
      WriteString(": ");
      WriteCard(Config->ID_range[i].start_ID, 6);
      WriteString("     End-ID : ");
      WriteCard(Config->ID_range[i].end_ID, 6);
      WriteString("  Num-PEs:");
      WriteCard(Config->ID_range[i].num_PEs, 6);
      WriteLn();
    }
      WriteString("used processors: ");
      WriteCard(Config->u_proc, 3);
      WriteLn();
  }
#endif

} /* End of INIT_ID_RANGE */


void COPY_ID_RANGE 
/* is used instead of INIT_ID_RANGE when several Configuration share the */
/* same PEs                                                              */
#ifdef __STDC__
(CONFIGURATION *DestConfig, CONFIGURATION SourceConfig, CARDINAL num_real)
#else
(DestConfig, SourceConfig, num_real) 
CONFIGURATION *DestConfig; 
CONFIGURATION SourceConfig;
CARDINAL num_real;
#endif
{
  CARDINAL j;

  DestConfig->u_proc = SourceConfig.u_proc;
  /* allocate memory for the TASK_DATA-array */
  DestConfig->ID_range = (TASK_DATA*) malloc(sizeof(TASK_DATA) * num_real);
  /* copy ID_range data */
  for (j = 0; j < DestConfig->u_proc; j++) {
    DestConfig->ID_range[j].start_ID   = SourceConfig.ID_range[j].start_ID;
    DestConfig->ID_range[j].end_ID     = SourceConfig.ID_range[j].end_ID;
    DestConfig->ID_range[j].num_PEs    = SourceConfig.ID_range[j].num_PEs;
  }
  for (j=0; j < DestConfig->u_proc; j++)
    DestConfig->ID_range[j].num_breaks = num_CR(*DestConfig);
}


void INIT_STACKS
/* Inits all required Stacks of Config (active-stack and else-stack) */
#ifdef __STDC__
(CONFIGURATION *Config)
#else
(Config) CONFIGURATION *Config;
#endif
{
  INIT_ACTIVE_STACK(Config);
  INIT_ELSE_STACK(Config);
}


void COPY_STACKS
#ifdef __STDC__
(CONFIGURATION *DestConfig, CONFIGURATION SourceConfig)
#else
(DestConfig, SourceConfig)
CONFIGURATION *DestConfig; CONFIGURATION SourceConfig;
#endif
/* Copies the references to the original stacks of the configuration group */
/* which are hold from the first declared configuration in the group.      */
/* is used instead of INIT_STACKS when several Configurations share the */
/* same PEs                                                             */
{
  DestConfig->ActiveStack = SourceConfig.ActiveStack;
  DestConfig->ElseStack   = SourceConfig.ElseStack;
}


void FREE_CONFIGURATION
/* Frees the memory allocated for the Len-, Lower- and Upper-arrays */
#ifdef __STDC__
(CONFIGURATION *Config, BOOLEAN FreeStacks)
#else
(Config, FreeStacks) 
CONFIGURATION *Config; BOOLEAN FreeStacks;
#endif
{
  if (FreeStacks) {
    FREE_ACTIVE_STACK(Config);
    FREE_ELSE_STACK(Config);
  }
  free (Config->ID_range);
  free (Config->Lower);
  free (Config->Upper);
  free (Config->Len);
}


CARDINAL num_CR
/* returns the number of Carrige-Returns to print while printing data */
/* from the ID-range start_ID to end_ID of the task me */ 
#ifdef __STDC__
(CONFIGURATION Config)
#else
(Config) CONFIGURATION Config; 
#endif
{
  CARDINAL line_breaks, Dimension, Divisor;

  line_breaks = 0;
  FOR_ALL_LOCAL_PEs(Config) {
    for (Dimension = 0, Divisor = 1; Dimension < RANK(Config); Dimension++) {
      Divisor *= LEN1(Config,Dimension+1);
      if (((Config.ID_range[me].start_ID + PE) % Divisor) == 0) {
	/* new-line is necessary */
	line_breaks++;
      } /* end if */
    } /* end for Dimension */
  }

  return line_breaks;
}

/* Functions for connections ---------------------------------------------- */

void INIT_CONNECTION
#ifdef __STDC__
(CONNECTION *Connection, CONFIGURATION *SourceConfig,
  CONFIGURATION *DestinationConfig)
#else
(Connection, SourceConfig, DestinationConfig)
CONNECTION *Connection; CONFIGURATION *SourceConfig;
CONFIGURATION *DestinationConfig;
#endif
/* Initialises the connection with the source and destination configuration. */
{
  Connection->SourceConfig = SourceConfig;
  Connection->DestinationConfig = DestinationConfig;
  /*  Connection->Links = NULL;
   *  Do not do it! In case the connection declaration is surrounded by a
   *  FOR-loop, all up to now computet connections will be lost!
   *  The Pointer is set to NULL when the configuration is declared.
   */
}

void REGISTER_LINK
#ifdef __STDC__
(CONNECTION *Connection, INTEGER *SourceArray, INTEGER *DestinationArray)
#else
(Connection, SourceArray, DestinationArray)
CONNECTION *Connection; INTEGER *SourceArray; INTEGER *DestinationArray;
#endif
/* Stores a link of the connection. */
/* Attention: stores the real IDs of the connection! */
{
  LINK *Link;
  if ( IS_IN_CONFIG(*(Connection->SourceConfig), SourceArray) &&
       IS_IN_CONFIG(*(Connection->DestinationConfig), DestinationArray) &&
       (IS_IN_LOCAL_DIM_RANGE(*(Connection->SourceConfig), SourceArray) ||
	IS_IN_LOCAL_DIM_RANGE(*(Connection->DestinationConfig), DestinationArray)) )
  {
    Link = (LINK *) malloc(sizeof(LINK));
    Link->Next = Connection->Links;
    Link->SourceID = DIM2ID(*(Connection->SourceConfig), SourceArray);
    Link->DestinationID = DIM2ID(*(Connection->DestinationConfig),
      DestinationArray);
    Connection->Links = Link;
  }
}

LINK_ARRAY *GET_LINKS
#ifdef __STDC__
(CONNECTION Connection)
#else
(Connection)
CONNECTION Connection;
#endif
{
  return (LINK_ARRAY*) Connection.Links;
}


void FREE_CONNECTION
#ifdef __STDC__
(CONNECTION *Connection)
#else
(Connection)
CONNECTION *Connection;
#endif
{
  LINK_ARRAY *Link;
  CARDINAL i;

  Link = Connection->Links;

  for (i = 0; i < Connection->SourceConfig->u_proc; i++) {
    if (Link->FromTask[i].NumConnections != 0) {
      free(Link->FromTask[i].PEnumbers);
    }
  }

  for (i = 0; i < Connection->DestinationConfig->u_proc; i++) {
    if (Link->ToTask[i].NumConnections != 0) {
      free(Link->ToTask[i].PEnumbers);
    }
  }

  Connection->Links = NULL;
}


/* connections with index ----- */

void INIT_PARAMETER_CONNECTION
#ifdef __STDC__
(PARAMETER_CONNECTION *Connection, CONFIGURATION *SourceConfig,
  CONFIGURATION *DestinationConfig)
#else
(Connection, SourceConfig, DestinationConfig)
PARAMETER_CONNECTION *Connection;
CONFIGURATION *SourceConfig;
CONFIGURATION *DestinationConfig;
#endif
/* Initialises the connection with the source and destination configuration. */
{
  Connection->SourceConfig = SourceConfig;
  Connection->DestinationConfig = DestinationConfig;
  /*  Connection->Links = NULL;
   *  Do not do it! In case the connection declaration is surrounded by a
   *  FOR-loop, all up to now computet connections will be lost!
   *  The Pointer is set to NULL when the configuration is declared.
   */
}

void REGISTER_PARAMETER_LINK
#ifdef __STDC__
(PARAMETER_CONNECTION *Connection, INTEGER Index, INTEGER *SourceArray,
  INTEGER *DestinationArray)
#else
(Connection, Index, SourceArray, DestinationArray)
CONNECTION *Connection; INTEGER Index; INTEGER *SourceArray;
INTEGER *DestinationArray;
#endif
/* Stores a link of the connection. */
{
  PARAMETER_LINK *ParameterLink;
  LINK *Link;

  if ( IS_IN_CONFIG(*(Connection->SourceConfig), SourceArray) &&
       IS_IN_CONFIG(*(Connection->DestinationConfig), DestinationArray) &&
       (IS_IN_LOCAL_DIM_RANGE(*(Connection->SourceConfig), SourceArray) ||
	    IS_IN_LOCAL_DIM_RANGE(*(Connection->DestinationConfig), DestinationArray)) ) {
    /* search for the connection with the index 'Index' */
    ParameterLink = Connection->ParameterLinks;
    while (ParameterLink != NULL && ParameterLink->Parameter != Index) {
      ParameterLink = ParameterLink->Next;
    }
    /* if index not found, then creat a new connection with this index */
    if (ParameterLink == NULL) {
      ParameterLink = (PARAMETER_LINK *) malloc(sizeof(PARAMETER_LINK));
      ParameterLink->Next = Connection->ParameterLinks;
      ParameterLink->Parameter = Index;
      Connection->ParameterLinks = ParameterLink;
    }
    Link = (LINK *) malloc(sizeof(LINK));
    Link->Next = ParameterLink->Links;
    Link->SourceID = DIM2ID(*(Connection->SourceConfig), SourceArray);
    Link->DestinationID = DIM2ID(*(Connection->DestinationConfig),
      DestinationArray);
    ParameterLink->Links = Link;
  }
}

LINK_ARRAY *GET_PARAMETER_LINKS
#ifdef __STDC__
(PARAMETER_CONNECTION Connection, INTEGER Parameter)
#else
(Connection, Parameter)
PARAMETER_CONNECTION Connection; INTEGER Parameter;
#endif
{
  PARAMETER_LINK *ParameterLinks;

  ParameterLinks = Connection.ParameterLinks;
  while (ParameterLinks != NULL && ParameterLinks->Parameter != Parameter) {
    ParameterLinks = ParameterLinks->Next;
  }
  if (ParameterLinks != NULL) {
    return (LINK_ARRAY*) ParameterLinks->Links;
  } else {
    return NULL;
  }
}

void FREE_PARAMETER_CONNECTION
#ifdef __STDC__
(PARAMETER_CONNECTION *Connection)
#else
(Connection)
PARAMETER_CONNECTION *Connection;
#endif
{
  PARAMETER_LINK *ParaLink, *ParaLink2;
  LINK_ARRAY *Link;
  CARDINAL i;

  ParaLink = Connection->ParameterLinks;
  while (ParaLink != NULL) {  

    Link = ParaLink->Links;

    for (i = 0; i < Connection->SourceConfig->u_proc; i++) {
      if (Link->FromTask[i].NumConnections != 0) {
	free(Link->FromTask[i].PEnumbers);
      }
    }
    
    for (i = 0; i < Connection->DestinationConfig->u_proc; i++) {
      if (Link->ToTask[i].NumConnections != 0) {
	free(Link->ToTask[i].PEnumbers);
      }
    }

    ParaLink2 = ParaLink;
    ParaLink = ParaLink->Next;
    free(ParaLink2);
  }
  Connection->ParameterLinks = NULL;
}

/* generic connections without index ----- */

void INIT_GENERIC_CONNECTION
#ifdef __STDC__
(char Connection[], CONFIGURATION *Configuration)
#else
(Connection, Configuration)
char Connection[];
CONFIGURATION *Configuration;
#endif
{
  GENERIC_CONNECTIONS_PTR Connect;

  Connect = (GENERIC_CONNECTIONS_PTR) malloc(sizeof(GENERIC_CONNECTIONS));
  strncpy(Connect->ConnectionName, Connection, 256);
  Connect->Configuration = Configuration;
  Connect->Links = NULL;
  Connect->Next = ALL_GENERIC_CONNECTIONS;
  ALL_GENERIC_CONNECTIONS = Connect;
}

void REGISTER_GENERIC_LINK
#ifdef __STDC__
(char ConnectionName[], CONFIGURATION *Configuration, INTEGER *SourceArray,
  INTEGER *DestinationArray)
#else
(ConnectionName, Configuration, SourceArray, DestinationArray)
char ConnectionName[]; CONFIGURATION *Configuration;
INTEGER *SourceArray; INTEGER *DestinationArray;
#endif
/* Stores a link of the generic connection. */
{
  GENERIC_CONNECTIONS_PTR Connection;
  LINK *Link;

  /* search generic connection */
  Connection = ALL_GENERIC_CONNECTIONS;
  while (Connection != NULL && (
         strcmp(ConnectionName, Connection->ConnectionName) != 0 ||
         strcmp(Configuration->Name, Connection->Configuration->Name) != 0)) {
    Connection = Connection->Next;
  }
  /* check if link is in configuration and local dim-range */
  if ( IS_IN_CONFIG(*(Connection->Configuration), SourceArray) &&
       IS_IN_CONFIG(*(Connection->Configuration), DestinationArray) &&
       (IS_IN_LOCAL_DIM_RANGE(*(Connection->Configuration), SourceArray) ||
	IS_IN_LOCAL_DIM_RANGE(*(Connection->Configuration), DestinationArray)))
  {
    Link = (LINK *) malloc(sizeof(LINK));
    Link->Next = Connection->Links;
    Link->SourceID = DIM2ID(*(Connection->Configuration), SourceArray);
    Link->DestinationID = DIM2ID(*(Connection->Configuration),
      DestinationArray);
    Connection->Links = Link;
  }
}

LINK_ARRAY *GET_GENERIC_LINKS
#ifdef __STDC__
(char Connection[], CONFIGURATION *Configuration)
#else
(Connection, Configuration)
char Connection[];
CONFIGURATION *Configuration;
#endif
{
  GENERIC_CONNECTIONS_PTR All_Generic;

  All_Generic = ALL_GENERIC_CONNECTIONS;
  while (All_Generic != NULL && (
         strcmp(Connection, All_Generic->ConnectionName) != 0 ||
         strcmp(Configuration->Name, All_Generic->Configuration->Name) != 0)) {
    All_Generic = All_Generic->Next;
  }
  if (All_Generic != NULL) {
    return (LINK_ARRAY*) All_Generic->Links;
  } else {
    return NULL;
  }
}

void FREE_GENERIC_CONNECTION
#ifdef __STDC__
(char Connection[], CONFIGURATION *Configuration)
#else
(Connection, Configuration)
char Connection[];
CONFIGURATION *Configuration;
#endif
{
  GENERIC_CONNECTIONS_PTR Connect1, Connect2;
  LINK_ARRAY *Link;
  CARDINAL i;

  /* search generic connection */
  Connect1 = ALL_GENERIC_CONNECTIONS;
  Connect2 = NULL;
  while (Connect1 != NULL && (
         strcmp(Connection, Connect1->ConnectionName) != 0 ||
         strcmp(Configuration->Name, Connect1->Configuration->Name) != 0)) {
    Connect2 = Connect1; /* pointer to predecessor */
    Connect1 = Connect1->Next;
  }
  /* take connection out of the list */
  if (Connect2 == NULL) { /* connection is the first in the list */
    ALL_GENERIC_CONNECTIONS = Connect1->Next;
  } else {
    Connect2->Next = Connect1->Next;
  }
  /* free links */
  Link = Connect1->Links;

  for (i = 0; i < Configuration->u_proc; i++) {
    if (Link->FromTask[i].NumConnections != 0) {
      free(Link->FromTask[i].PEnumbers);
    }
  }

  for (i = 0; i < Configuration->u_proc; i++) {
    if (Link->ToTask[i].NumConnections != 0) {
      free(Link->ToTask[i].PEnumbers);
    }
  }

  /* free connection */
  free(Connect1);  
}

void FREE_ALL_GENERIC_CONNECTIONS()
{
  GENERIC_CONNECTIONS_PTR Connect1, Connect2;
  LINK_ARRAY *Link;
  CARDINAL i;

  Connect1 = ALL_GENERIC_CONNECTIONS;
  while (Connect1 != NULL) {
    /* free links */
    Link = Connect1->Links;

    for (i = 0; i < Connect1->Configuration->u_proc; i++) {
      if (Link->FromTask[i].NumConnections != 0) {
	free(Link->FromTask[i].PEnumbers);
      }
    }

    for (i = 0; i < Connect1->Configuration->u_proc; i++) {
      if (Link->ToTask[i].NumConnections != 0) {
	free(Link->ToTask[i].PEnumbers);
      }
    }

    Connect2 = Connect1;
    Connect1 = Connect1->Next;
    /* free connection */
    free(Connect2);
  }
  ALL_GENERIC_CONNECTIONS = NULL;
}

/* generic connections with index ----- */

void INIT_GENERIC_PARAMETER_CONNECTION
#ifdef __STDC__
(char ConnectionName[], CONFIGURATION *Configuration)
#else
(Connection, Configuration)
char Connection[];
CONFIGURATION *Configuration;
#endif
{
  GENERIC_PARAMETER_CONNECTIONS_PTR Connection;

  Connection = (GENERIC_PARAMETER_CONNECTIONS_PTR)
    malloc(sizeof(GENERIC_PARAMETER_CONNECTIONS));
  strncpy(Connection->ConnectionName, ConnectionName, 256);
  Connection->Configuration = Configuration;
  Connection->ParameterLinks = NULL;
  Connection->Next = ALL_GENERIC_PARAMETER_CONNECTIONS;
  ALL_GENERIC_PARAMETER_CONNECTIONS = Connection;
}

void REGISTER_GENERIC_PARAMETER_LINK
#ifdef __STDC__
(char ConnectionName[], CONFIGURATION *Configuration, INTEGER Parameter,
  INTEGER *SourceArray, INTEGER *DestinationArray)
#else
(ConnectionName, Configuration, Parameter, SourceArray, DestinationArray)
char ConnectionName[]; CONFIGURATION *Configuration; INTEGER Parameter;
INTEGER *SourceArray; INTEGER *DestinationArray;
#endif
/* Stores a link of the generic parameter connection. */
{
  GENERIC_PARAMETER_CONNECTIONS_PTR Connection;
  PARAMETER_LINK *ParameterLink;
  LINK *Link;

  /* search generic connection */
  Connection = ALL_GENERIC_PARAMETER_CONNECTIONS;
  while (Connection != NULL && (
         strcmp(ConnectionName, Connection->ConnectionName) != 0 ||
         strcmp(Configuration->Name, Connection->Configuration->Name) != 0)) {
    Connection = Connection->Next;
  }
  /* check if link is in configuration and local dim-range */
  if ( IS_IN_CONFIG(*(Connection->Configuration), SourceArray) &&
       IS_IN_CONFIG(*(Connection->Configuration), DestinationArray) &&
       (IS_IN_LOCAL_DIM_RANGE(*(Connection->Configuration), SourceArray) ||
	IS_IN_LOCAL_DIM_RANGE(*(Connection->Configuration), DestinationArray)))
  {
    /* search for the connection with the index 'Parameter' */
    ParameterLink = Connection->ParameterLinks;
    while (ParameterLink != NULL && ParameterLink->Parameter != Parameter) {
      ParameterLink = ParameterLink->Next;
    }
    /* if index not found, then creat a new connection with this index */
    if (ParameterLink == NULL) {
      ParameterLink = (PARAMETER_LINK *) malloc(sizeof(PARAMETER_LINK));
      ParameterLink->Next = Connection->ParameterLinks;
      ParameterLink->Parameter = Parameter;
      Connection->ParameterLinks = ParameterLink;
    }
    Link = (LINK *) malloc(sizeof(LINK));
    Link->Next = ParameterLink->Links;
    Link->SourceID = DIM2ID(*(Connection->Configuration), SourceArray);
    Link->DestinationID = DIM2ID(*(Connection->Configuration),
      DestinationArray);
    ParameterLink->Links = Link;
  }
}

LINK_ARRAY *GET_GENERIC_PARAMETER_LINKS
#ifdef __STDC__
(char Connection[], CONFIGURATION *Configuration, INTEGER Parameter)
#else
(Connection, Configuration, Parameter)
char Connection[];
CONFIGURATION *Configuration;
INTEGER Parameter;
#endif
{
  GENERIC_PARAMETER_CONNECTIONS_PTR All_Generic;
  PARAMETER_LINK *ParameterLinks;

  All_Generic = ALL_GENERIC_PARAMETER_CONNECTIONS;
  while (All_Generic != NULL && (
         strcmp(Connection, All_Generic->ConnectionName) != 0 ||
         strcmp(Configuration->Name, All_Generic->Configuration->Name) != 0)) {
    All_Generic = All_Generic->Next;
  }
  ParameterLinks = All_Generic->ParameterLinks;
  while (ParameterLinks != NULL && ParameterLinks->Parameter != Parameter) {
    ParameterLinks = ParameterLinks->Next;
  }
  if (ParameterLinks != NULL) {
    return (LINK_ARRAY*) ParameterLinks->Links;
  } else {
    return NULL;
  }
}

void FREE_GENERIC_PARAMETER_CONNECTION
#ifdef __STDC__
(char Connection[], CONFIGURATION *Configuration)
#else
(Connection, Configuration)
char Connection[];
CONFIGURATION *Configuration;
#endif
{
  GENERIC_PARAMETER_CONNECTIONS_PTR Connect1, Connect2;
  PARAMETER_LINK *ParaLink1, *ParaLink2;
  LINK_ARRAY *Link;
  CARDINAL i;

  /* search generic parameter connection */
  Connect1 = ALL_GENERIC_PARAMETER_CONNECTIONS;
  Connect2 = NULL;
  while (Connect1 != NULL && (
         strcmp(Connection, Connect1->ConnectionName) != 0 ||
         strcmp(Configuration->Name, Connect1->Configuration->Name) != 0)) {
    Connect2 = Connect1; /* pointer to predecessor */
    Connect1 = Connect1->Next;
  }
  /* found -> take connection out of the list */
  if (Connect1 != NULL) {
    if (Connect2 == NULL) { /* connection is the first in the global list */
      ALL_GENERIC_PARAMETER_CONNECTIONS = Connect1->Next;
    } else {
      Connect2->Next = Connect1->Next;
    }
    ParaLink1 = Connect1->ParameterLinks;
    while (ParaLink1 != NULL) {  
      /* free Links */
      Link = ParaLink1->Links;

      for (i = 0; i < Configuration->u_proc; i++) {
	if (Link->FromTask[i].NumConnections != 0) {
	  free(Link->FromTask[i].PEnumbers);
	}
      }

      for (i = 0; i < Configuration->u_proc; i++) {
	if (Link->ToTask[i].NumConnections != 0) {
	  free(Link->ToTask[i].PEnumbers);
	}
      }

      ParaLink2 = ParaLink1;
      ParaLink1 = ParaLink1->Next;
      free(ParaLink2);
    }
    free(Connect1); 
  }
}


void FREE_ALL_GENERIC_PARAMETER_CONNECTIONS()
{
  GENERIC_PARAMETER_CONNECTIONS_PTR Connect1, Connect2;
  PARAMETER_LINK *ParamLink1, *ParamLink2;
  LINK_ARRAY *Link;
  CARDINAL i;

  Connect1 = ALL_GENERIC_PARAMETER_CONNECTIONS;
  while (Connect1 != NULL) {
    /* free parameter-lists */
    ParamLink1 = Connect1->ParameterLinks;
    while (ParamLink1 != NULL) {
      /* free links */
      Link = ParamLink1->Links;

      for (i = 0; i < Connect1->Configuration->u_proc; i++) {
	if (Link->FromTask[i].NumConnections != 0) {
	  free(Link->FromTask[i].PEnumbers);
	}
      }

      for (i = 0; i < Connect1->Configuration->u_proc; i++) {
	if (Link->ToTask[i].NumConnections != 0) {
	  free(Link->ToTask[i].PEnumbers);
	}
      }

      ParamLink2 = ParamLink1;
      ParamLink1 = ParamLink1->Next;
      free(ParamLink2);
    }

    Connect2 = Connect1;
    Connect1 = Connect1->Next;
    /* free connection */
    free(Connect2);
  }

  ALL_GENERIC_PARAMETER_CONNECTIONS = NULL;
}


/* only needed in the following two procedures */
typedef struct {
  CARDINAL SourceID, DestID;
} SOURCE_DEST_PAIR;

/* global variables needed for the comparison function */
CONFIGURATION CompSourceConfig;
CONFIGURATION CompDestConfig;


int CompareSourceDestPair
#ifdef __STDC__
(const void* SourceDestPair1, const void* SourceDestPair2)
#else
(SourceDestPair1, SourceDestPair2)
const void* SourceDestPair1;
const void* SourceDestPair2;
#endif
{
  CARDINAL t_ind_SID1, t_ind_SID2; /* index of task holding source-id      */
  CARDINAL t_ind_DID1, t_ind_DID2; /* index of task holding destination-id */

  t_ind_SID1 = TASK_INDEX_HOLDING_ID(
		 ((SOURCE_DEST_PAIR*) SourceDestPair1)->SourceID,
		 CompSourceConfig);
  t_ind_SID2 = TASK_INDEX_HOLDING_ID(
		 ((SOURCE_DEST_PAIR*) SourceDestPair2)->SourceID,
		 CompSourceConfig);

  if (t_ind_SID1 < t_ind_SID2) {
    return -1;
  } else {
  if (t_ind_SID1 > t_ind_SID2) {
    return +1;
  } else {
  /* equal */
  t_ind_DID1 = TASK_INDEX_HOLDING_ID(
                 ((SOURCE_DEST_PAIR*) SourceDestPair1)->DestID,
		 CompDestConfig);
  t_ind_DID2 = TASK_INDEX_HOLDING_ID(
		 ((SOURCE_DEST_PAIR*) SourceDestPair2)->DestID,
		 CompDestConfig);
  if (t_ind_DID1 < t_ind_DID2) {
    return -1;
  } else {
  if (t_ind_DID1 > t_ind_DID2) {
    return +1;
  } else {
  /* equal! */
  if (((SOURCE_DEST_PAIR*) SourceDestPair1)->SourceID < ((SOURCE_DEST_PAIR*) SourceDestPair2)->SourceID) {
    return -1;
  } else {
  if (((SOURCE_DEST_PAIR*) SourceDestPair1)->SourceID > ((SOURCE_DEST_PAIR*) SourceDestPair2)->SourceID) {
    return +1;
  } else {
  /* equal!! */
  if (((SOURCE_DEST_PAIR*) SourceDestPair1)->DestID < ((SOURCE_DEST_PAIR*) SourceDestPair2)->DestID) {
    return -1;
  } else {
  if (((SOURCE_DEST_PAIR*) SourceDestPair1)->DestID > ((SOURCE_DEST_PAIR*) SourceDestPair2)->DestID) {
    return +1;
  } else {
  /* equal!!! */
    return 0;
  }
 }}}}}}} /* close all ifs */
}



/* function to organize links in a way that messages to each task can be */
/* easily produced */
LINK_ARRAY *REORGANIZE_LINKS
#ifdef __STDC__
(LINK *Links, CONFIGURATION SourceConfig, CONFIGURATION DestConfig)
#else
(Links, SourceConfig, DestConfig) 
LINK *Links;
CONFIGURATION SourceConfig;
CONFIGURATION DestConfig;
#endif
{
  LINK *LinkPtr, *OldLinkPtr;   /* to process LINK-list */
  LINK_ARRAY *NewLinks;
  CARDINAL TotalNumConnections; /* total number of connections between two PEs */
  SOURCE_DEST_PAIR *SortArray; /* to sort the LINKs */
  CARDINAL SortIndex;
  CARDINAL first, last, index;
  CARDINAL t_ind, t_ind2;


  /* count the number of elements in the LINK-list */
  LinkPtr = Links;
  TotalNumConnections = 0;
  while (LinkPtr != NULL) {
    TotalNumConnections++;
    LinkPtr = LinkPtr->Next;
  }

  NewLinks = (LINK_ARRAY*) malloc(sizeof(LINK_ARRAY));

  if (TotalNumConnections == 0) {
    /* no connections, set all values to 0 */
    CARDINAL i;

    NewLinks->ToTask   = (void*) malloc(sizeof(TASK_LINKS) * n_tasks);
    NewLinks->FromTask = (void*) malloc(sizeof(TASK_LINKS) * n_tasks);
    for (i = 0; i < n_tasks; i++) {
      NewLinks->ToTask[i].NumConnections = 0;
      NewLinks->ToTask[i].PEnumbers      = NULL;
    }
    for (i = 0; i < n_tasks; i++) {
      NewLinks->FromTask[i].NumConnections = 0;
      NewLinks->FromTask[i].PEnumbers      = NULL;
    }
    return NewLinks;
  }

  /* allocate memory for temporary sort-array */
  SortArray = (void*) malloc(sizeof(SOURCE_DEST_PAIR) *
			     TotalNumConnections);

  /* copy LINK-list to SortArray and free memory of LINK-list */
  LinkPtr = Links;
  SortIndex = 0;
  while (LinkPtr != NULL) {
    SortArray[SortIndex].SourceID = LinkPtr->SourceID;
    SortArray[SortIndex].DestID   = LinkPtr->DestinationID;
    SortIndex++;
    OldLinkPtr = LinkPtr;
    LinkPtr    = LinkPtr->Next;
    /* with big connections this gets slower and slower and ... */
    /* therefore we are a bit sloppy */
    /* free(OldLinkPtr); */
  }

  /* set global variables for comparison */
  CompSourceConfig = SourceConfig;
  CompDestConfig   = DestConfig;
  /* sort the array */
  qsort(SortArray,                /* void *base   */
	TotalNumConnections,      /* size_t nmemb */
	sizeof(SOURCE_DEST_PAIR), /* size_t size  */
	CompareSourceDestPair);   /* comparison function */

  /* construct new datastructure out of SortArray */
  NewLinks->ToTask   = (void*) malloc(sizeof(TASK_LINKS) * DestConfig.u_proc);
  NewLinks->FromTask = (void*) malloc(sizeof(TASK_LINKS) * SourceConfig.u_proc);

  first = 0;
  for (t_ind = 0; t_ind < SourceConfig.u_proc; t_ind++) {
    if (t_ind != me) {
      if (first < TotalNumConnections &&
	  t_ind == TASK_INDEX_HOLDING_ID(SortArray[first].SourceID, SourceConfig)) {

	last = first;
	/* count number of elements in this block */
	while (last < TotalNumConnections &&
	       SortArray[last].SourceID >= SourceConfig.ID_range[t_ind].start_ID &&
	       SortArray[last].SourceID <= SourceConfig.ID_range[t_ind].end_ID ) {
	  last++;
	}
	NewLinks->FromTask[t_ind].NumConnections = last - first;
	NewLinks->FromTask[t_ind].PEnumbers = 
	  (void*) malloc(sizeof(SOURCE_DEST_PAIR) * (last - first));
	/* copy DestinationIDs and transfer to PE-indexnumbers */
	for (index = first; index < last; index++) {
	  NewLinks->FromTask[t_ind].PEnumbers[index - first] =
	    SortArray[index].DestID - DestConfig.ID_range[me].start_ID;
	}
	first = last;

      } else { /* no connections from t_ind */
	NewLinks->FromTask[t_ind].NumConnections = 0;
	NewLinks->FromTask[t_ind].PEnumbers      = NULL;
      }

    } else { /* t_ind == me */
      for (t_ind2 = 0; t_ind2 < DestConfig.u_proc; t_ind2++) {
	if (first < TotalNumConnections &&
	    t_ind == TASK_INDEX_HOLDING_ID(SortArray[first].SourceID,
					    SourceConfig) 
	    && t_ind2 == TASK_INDEX_HOLDING_ID(SortArray[first].DestID,
						DestConfig))
	{

	  last = first;
	  /* count number of elements of subblock */
	  while (last < TotalNumConnections
	    && SortArray[last].SourceID >= SourceConfig.ID_range[me].start_ID
	    && SortArray[last].SourceID <= SourceConfig.ID_range[me].end_ID
	    && SortArray[last].DestID >= DestConfig.ID_range[t_ind2].start_ID
	    && SortArray[last].DestID <= DestConfig.ID_range[t_ind2].end_ID)
	  {
	    last++;
	  }

	  NewLinks->ToTask[t_ind2].NumConnections = last - first;
	  NewLinks->ToTask[t_ind2].PEnumbers =
	    (void*) malloc(sizeof(SOURCE_DEST_PAIR) * (last - first));
	  /* copy from SortArray to PENumbers */
	  for (index = first; index < last; index++) {
	    NewLinks->ToTask[t_ind2].PEnumbers[index - first] =
	      SortArray[index].SourceID - SourceConfig.ID_range[me].start_ID;
	  }

	  if (t_ind2 == me ) {
	    /* do the same for the FromTask-entry */
	    NewLinks->FromTask[t_ind2].NumConnections = last - first;
	    NewLinks->FromTask[t_ind2].PEnumbers =
	      (void*) malloc(sizeof(SOURCE_DEST_PAIR) * (last - first));

	    /* copy also in the FromTask-entry */
	    for (index = first; index < last; index++) {
	      NewLinks->FromTask[t_ind2].PEnumbers[index - first] =
		SortArray[index].DestID - DestConfig.ID_range[me].start_ID;  
	    }
	  }

	  first = last;

	} else { /* no connections to t_ind2 */
	  NewLinks->ToTask[t_ind2].NumConnections = 0;
	  NewLinks->ToTask[t_ind2].PEnumbers      = NULL;
	  if (t_ind2 == me ) {
	    /* do the same for the FromTask-entry */
	    NewLinks->FromTask[t_ind2].NumConnections = 0;
	    NewLinks->FromTask[t_ind2].PEnumbers      = NULL;
	  }
	}

      } /* end for t_ind2 */
    } /* end else */
  } /* for t_ind */

  return NewLinks;
}


void REORGANIZE_CONNECTION
#ifdef __STDC__
(CONNECTION *Connect)
#else
(Connect) CONNECTION *Connect;
#endif
{
  Connect->Links = REORGANIZE_LINKS(Connect->Links, *(Connect->SourceConfig),
				    *(Connect->DestinationConfig));
}


void REORGANIZE_PARAMETER_CONNECTION
#ifdef __STDC__
(PARAMETER_CONNECTION *Connect)
#else
(Connect) PARAMETER_CONNECTION *Connect;
#endif
{
  PARAMETER_LINK *ParaLink;

  ParaLink = Connect->ParameterLinks;
  while (ParaLink != NULL) {  
    ParaLink->Links = REORGANIZE_LINKS(ParaLink->Links, 
				       *(Connect->SourceConfig),
				       *(Connect->DestinationConfig));
    ParaLink = ParaLink->Next;
  }
}


void REORGANIZE_GENERIC_CONNECTION
#ifdef __STDC__
(char ConnectionName[], CONFIGURATION *Configuration)
#else
(ConnectionName, Configuration)
char Connection[];
CONFIGURATION *Configuration;
#endif
{
  GENERIC_CONNECTIONS_PTR Connection;

  /* search generic connection */
  Connection = ALL_GENERIC_CONNECTIONS;
  while (Connection != NULL && (
         strcmp(ConnectionName, Connection->ConnectionName) != 0 ||
         strcmp(Configuration->Name, Connection->Configuration->Name) != 0)) {
    Connection = Connection->Next;
  }
  if (Connection != NULL) {
    Connection->Links = REORGANIZE_LINKS(Connection->Links, 
					 *(Connection->Configuration),
					 *(Connection->Configuration));
  }
}


void REORGANIZE_GENERIC_PARAMETER_CONNECTION
#ifdef __STDC__
(char ConnectionName[], CONFIGURATION *Configuration)
#else
(Connection, Configuration)
char Connection[];
CONFIGURATION *Configuration;
#endif
{
  GENERIC_PARAMETER_CONNECTIONS_PTR Connection;
  PARAMETER_LINK *ParameterLink;

  /* search generic connection */
  Connection = ALL_GENERIC_PARAMETER_CONNECTIONS;
  while (Connection != NULL && (
         strcmp(ConnectionName, Connection->ConnectionName) != 0 ||
         strcmp(Configuration->Name, Connection->Configuration->Name) != 0)) {
    Connection = Connection->Next;
  }
  if (Connection != NULL) {
    ParameterLink = Connection->ParameterLinks;
    /* do it for all Parameters */
    while (ParameterLink != NULL) {  
      ParameterLink->Links = REORGANIZE_LINKS(ParameterLink->Links, 
					      *(Connection->Configuration),
					      *(Connection->Configuration));
      ParameterLink = ParameterLink->Next;
    }
  }
}
    



/* Functions for active stack --------------------------------------------- */

void INIT_ACTIVE_STACK
/* Allocates memory for the ActiveStack-Pointer of Config and initializes */
/* all PEs to active.                                                      */
#ifdef __STDC__
(CONFIGURATION *Config)
#else
(Config) CONFIGURATION *Config;
#endif
{
  if (me < Config->u_proc) {
    Config->ActiveStack = (ACTIVE_STACK_PTR) malloc(sizeof(ACTIVE_STACK));
    /* assuming eight bits per byte */
    Config->ActiveStack->ElemSize = (LOCAL_NUM_PEs(*Config) + 7) / 8;
    Config->ActiveStack->StackElem = NULL;
    PUSH_ACTIVE_ALL(*Config); /* create stackelement and set all to active */
  } else {
    Config->ActiveStack = (ACTIVE_STACK_PTR) malloc(sizeof(ACTIVE_STACK));
    Config->ActiveStack->ElemSize = 0;
    Config->ActiveStack->StackElem = NULL;
  }
}


void FREE_ACTIVE_STACK
/* frees the memory of the active stack of Config */
#ifdef __STDC__
(CONFIGURATION *Config)
#else
(Config) CONFIGURATION *Config;
#endif
{
  if (me < Config->u_proc) {
    POP_ACTIVE(*Config);
    free(Config->ActiveStack);
  }
}


void PUSH_ACTIVE_ALL
/* pushes the actual state of each PE on the active-stack and creates a new */
/* element on top with all PEs set active                                   */
#ifdef __STDC__
(CONFIGURATION Config)
#else
(Config) CONFIGURATION Config;
#endif
{
  BOOLEAN_ELEM_PTR StackElem;

  if (me < Config.u_proc) {
    StackElem = (BOOLEAN_ELEM_PTR) malloc(sizeof(BOOLEAN_ELEM));
    StackElem->Elem = (CHAR*) malloc(Config.ActiveStack->ElemSize);
    /* assuming eight bits per char, all bits are filled with '1' */
    (void) memset(StackElem->Elem, 255, Config.ActiveStack->ElemSize);
    StackElem->Previous = Config.ActiveStack->StackElem;
    Config.ActiveStack->StackElem = StackElem;
  }
}


void PUSH_ACTIVE
/* pushes the actual active-state of each PE on the active-stack and creates */
/* a new element on top of the stack which is a copy of the old element */
#ifdef __STDC__
(CONFIGURATION Config)
#else
(Config) CONFIGURATION Config;
#endif
{
  BOOLEAN_ELEM_PTR StackElem;

  if (me < Config.u_proc) {
    StackElem = (BOOLEAN_ELEM_PTR) malloc(sizeof(BOOLEAN_ELEM));
    StackElem->Elem = (CHAR*) malloc(Config.ActiveStack->ElemSize);
    /* assuming eight bits per char, bits are copied from old stack */
    (void) memcpy(StackElem->Elem, Config.ActiveStack->StackElem->Elem,
		  Config.ActiveStack->ElemSize);
    StackElem->Previous = Config.ActiveStack->StackElem;
    Config.ActiveStack->StackElem = StackElem;
  }
}


void POP_ACTIVE
/* pops the top element of the active-stack and frees memory */
#ifdef __STDC__
(CONFIGURATION Config)
#else
(Config) CONFIGURATION Config;
#endif
{
  BOOLEAN_ELEM_PTR StackElem;

  if (me < Config.u_proc) {
    StackElem = Config.ActiveStack->StackElem;
    Config.ActiveStack->StackElem = StackElem->Previous;
    free(StackElem->Elem);
    free(StackElem);
  }
}


void TOP_ACTIVE
#ifdef __STDC__
(CONFIGURATION Config)
#else
(Config) CONFIGURATION Config;
#endif
/* Is same as POP_ACTIVE(); PUSH_ACTIVE();. Just copies the active status */
/* from the previous stack element into the actual stack element. */
{
  (void) memcpy(Config.ActiveStack->StackElem->Elem,
    Config.ActiveStack->StackElem->Previous->Elem,
    Config.ActiveStack->ElemSize);
}


BOOLEAN_ELEM_PTR READ_ACTIVE
#ifdef __STDC__
(CONFIGURATION Config)
#else
(Config) CONFIGURATION Config;
#endif
/* Reads the pointer to the actual active stack element */
{
  return Config.ActiveStack->StackElem;
}


void RESET_ACTIVE
#ifdef __STDC__
(CONFIGURATION Config, BOOLEAN_ELEM_PTR Pointer)
#else
(Config, Pointer) CONFIGURATION Config; BOOLEAN_ELEM_PTR Pointer;
#endif
/* Resets the active stack to the element with the pointer */
{
  while (Config.ActiveStack->StackElem != Pointer) {
    POP_ACTIVE(Config);
  }
}


void EXIT_ALL_ACTIVES
#ifdef __STDC__
(CONFIGURATION Config, BOOLEAN_ELEM_PTR ActivePtr, BOOLEAN_ELEM_PTR ElsePtr)
#else
(Config, Pointer) CONFIGURATION Config; BOOLEAN_ELEM_PTR Pointer; 
#endif
/* Sets all active processors of the configuration up to the (active */
/* stack) pointer to inactive. Also for all active processors their  */
/* else-stack is deactivated. This is used for the parallel LOOP to  */
/* disactivate PEs up to the end of the LOOP.                        */
{
  BOOLEAN_ELEM_PTR TempPtr;

  FOR_ALL_LOCAL_ACTIVES(Config) {
    TempPtr = Config.ActiveStack->StackElem;
    while (TempPtr != ActivePtr) {
      TempPtr->Elem[PE / 8] &= (~(1 << (PE % 8)));
      TempPtr = TempPtr->Previous;
    };
    TempPtr = Config.ElseStack->StackElem;
    while (TempPtr != ElsePtr) {
      TempPtr->Elem[PE / 8] &= (~(1 << (PE % 8)));
      TempPtr = TempPtr->Previous;
    };
  }
}

/* not used yet */
void INVERT_ACTIVE
/* inverts the active status of these PEs which are now active or */
/* inactive and were active before */
#ifdef __STDC__
(CONFIGURATION Config)
#else
(Config) CONFIGURATION Config;
#endif
{
  CHAR *old_elem, *top_elem;
  int j;

  if (me < Config.u_proc) {
    top_elem = Config.ActiveStack->StackElem->Elem;
    old_elem = Config.ActiveStack->StackElem->Previous->Elem;
    for (j=0; j<Config.ActiveStack->ElemSize; j++) {
      top_elem[j] = top_elem[j] ^ old_elem[j];
    }
  }
}

void SET_INACTIVE /* local */
/* sets the status of one Processor of Config to inactive */
/* the Parameter local_Processor has the range            */
/* from 0 to Config.ID_range[me].num_PEs - 1 (!)          */
#ifdef __STDC__
(CONFIGURATION Config, CARDINAL local_Processor)
#else
(Config, local_Processor) CONFIGURATION Config; CARDINAL local_Processor;
#endif
{
  if (me < Config.u_proc) {
    Config.ActiveStack->StackElem->Elem[local_Processor / 8] &=
      (~(1 << (local_Processor % 8)));
  }
}


BOOLEAN LOCAL_IS_ACTIVE
/* returns the active-status of one Processor of Config */
/* the Parameter local_Processor has the range           */
/* from 0 to Config.ID_range[me].num_PEs - 1 (!)         */
#ifdef __STDC__
(CONFIGURATION Config, CARDINAL local_Processor)
#else
(Config, local_Processor) CONFIGURATION Config; CARDINAL local_Processor; 
#endif
{
  if (me < Config.u_proc) {
    return (Config.ActiveStack->StackElem->Elem[local_Processor / 8] &
	    (1 << (local_Processor % 8))) != 0;
  }
  else {
    return FALSE;
  }
}


BOOLEAN LOCAL_IS_ANY_ACTIVE
/* returns true if almost one PE has the status "active" */
#ifdef __STDC__
(CONFIGURATION Config)
#else
(Config) CONFIGURATION Config;
#endif
{
  CARDINAL Index;
  CHAR Select_Bits[7] = {1,3,7,15,31,63,127 };

  /* it's not necessary to check if this processor belongs to Config, */
  /* because it is checked before the call of this function */

  /* up to the last but one element it is possible to check eight processors
   * simultaneously, because if at least one is active, the value is not null
   */
  for (Index = 0; Index < Config.ActiveStack->ElemSize - (LOCAL_NUM_PEs(Config) % 8 != 0); Index++) {
    if (Config.ActiveStack->StackElem->Elem[Index] != (CHAR) 0) {
      return TRUE;
    }
  }

  if (Index == Config.ActiveStack->ElemSize - 1)
    return ((Config.ActiveStack->StackElem->Elem[Index]
	     & Select_Bits[LOCAL_NUM_PEs(Config) % 8 - 1]) != 0);

  return FALSE;
}


/* Functions for else stack ----------------------------------------------- */


void INIT_ELSE_STACK
#ifdef __STDC__
(CONFIGURATION *Config)
#else
(Config) CONFIGURATION *Config;
#endif
/* Initialises the else stack */
{
  if (me < Config->u_proc) {
    Config->ElseStack = (ELSE_STACK_PTR) malloc(sizeof(ELSE_STACK));
    /* assuming eight bits per byte */
    Config->ElseStack->ElemSize = (Config->ID_range[me].num_PEs + 7) / 8;
    Config->ElseStack->StackElem = NULL;
  }
}

void FREE_ELSE_STACK
#ifdef __STDC__
(CONFIGURATION *Config)
#else
(Config) CONFIGURATION *Config;
#endif
/* Frees the else stack. */
{
  if (me < Config->u_proc) {
    while (Config->ElseStack->StackElem != NULL) {
      POP_ELSE(*Config);
    }
    free(Config->ElseStack);
  }
}

void PUSH_ELSE
#ifdef __STDC__
(CONFIGURATION Config)
#else
(Config) CONFIGURATION Config;
#endif
/* Push the actual else status and copies the actual active status. */
{
  BOOLEAN_ELEM_PTR StackElem;

  if (me < Config.u_proc) {
    StackElem = (BOOLEAN_ELEM_PTR) malloc(sizeof(BOOLEAN_ELEM));
    StackElem->Elem = malloc(Config.ElseStack->ElemSize);
    /* assuming eight bits per char, bits are copied from the active stack */
    (void) memcpy(StackElem->Elem, Config.ActiveStack->StackElem->Elem,
		  Config.ElseStack->ElemSize);
    StackElem->Previous = Config.ElseStack->StackElem;
    Config.ElseStack->StackElem = StackElem;
  }
}

void POP_ELSE
#ifdef __STDC__
(CONFIGURATION Config)
#else
(Config) CONFIGURATION Config;
#endif
/* Pop the previous else status. */
{
  BOOLEAN_ELEM_PTR StackElem;

  if (me < Config.u_proc) {
    StackElem = Config.ElseStack->StackElem;
    Config.ElseStack->StackElem = StackElem->Previous;
    free(StackElem->Elem);
    free(StackElem);
  }
}

BOOLEAN_ELEM_PTR READ_ELSE
#ifdef __STDC__
(CONFIGURATION Config)
#else
(Config) CONFIGURATION Config;
#endif
/* Reads the pointer to the actual else stack element */
{
  if (me < Config.u_proc)
    return Config.ElseStack->StackElem;
  else
    return NIL;
}

void RESET_ELSE
#ifdef __STDC__
(CONFIGURATION Config, BOOLEAN_ELEM_PTR Pointer)
#else
(Config, Pointer) CONFIGURATION Config; BOOLEAN_ELEM_PTR Pointer;
#endif
/* Resets the else stack to the element with the pointer */
{
  if (me < Config.u_proc) {
    while (Config.ElseStack->StackElem != Pointer) {
      POP_ACTIVE(Config);
    }
  }    
}

void COPY_ELSE_TO_ACTIVE
#ifdef __STDC__
(CONFIGURATION Config)
#else
(Config) CONFIGURATION Config;
#endif
/* Copy the status in the else stack into the active stack. */
{
  if (me < Config.u_proc) {
    /* assuming eight bits per char, bits are copied from the else stack */
    (void) memcpy(Config.ActiveStack->StackElem->Elem,
		  Config.ElseStack->StackElem->Elem, Config.ElseStack->ElemSize);
  }
}

void UNSET_ELSE
#ifdef __STDC__
(CONFIGURATION Config, CARDINAL Processor)
#else
(Config, Processor) CONFIGURATION Config; CARDINAL Processor; 
#endif
/* Sets a processor in the range from 0 to LOCAL_num_PEs(Config) - 1 for the */
/* ELSE-part of a CASE-statement to inactive. */
{
  if (me < Config.u_proc) {
    Config.ElseStack->StackElem->Elem[Processor / 8] &=
      (~(1 << (Processor % 8)));
  }
}


/* Functions for checks --------------------------------------------------- */


BOOLEAN IS_IN_CONFIG
/* checks if an specific PE given by a array of dimension-indices is in the */
/* range of the configuration */
#ifdef __STDC__
(CONFIGURATION Config, INTEGER *IndexArray)
#else
(Config, IndexArray) CONFIGURATION Config; INTEGER *IndexArray;
#endif
/* Checks, whether an index is in the configuration bounds or not. */
{
  BOOLEAN IsInConfig;
  CARDINAL Dimension;

  IsInConfig = TRUE;
  for (Dimension = 1; Dimension <= RANK(Config); Dimension++) {
    IsInConfig = IsInConfig &&
      LOWER(Config, Dimension) <= IndexArray[Dimension - 1] &&
      IndexArray[Dimension - 1] <= UPPER(Config, Dimension);
  }
  return IsInConfig;
}


BOOLEAN IS_IN_LOCAL_ID_RANGE
/* checks if an given id is in the local range of id-numbers that this */
/* (real) processor holds */
#ifdef __STDC__
(CONFIGURATION Config, CARDINAL id)
#else
(Config, id) CONFIGURATION Config; CARDINAL id; 
#endif
{
  if (me < Config.u_proc) {
    return (id >= Config.ID_range[me].start_ID) &&
           (id <= Config.ID_range[me].end_ID);
  } else {
    return FALSE;
  }
}


CARDINAL TASK_INDEX_HOLDING_ID
#ifdef __STDC__
(CARDINAL id, CONFIGURATION Config)
#else
(id, Config) CARDINAL id; CONFIGURATION Config;
#endif
{
  CARDINAL i, index;

  index = Config.u_proc;
  i = 0;
  do {
    if (Config.ID_range[i].start_ID <= id &&
	Config.ID_range[i].end_ID   >= id) {
      index = i;
    }
    i++;
  } while (i < Config.u_proc);
  if (index == Config.u_proc) {
    Id_Range_Error(id, Config);
  }
  return index;
}


/* Parallaxis specific functions ------------------------------------------ */

CARDINAL DIM2ID
#ifdef __STDC__
(CONFIGURATION Config, INTEGER *Dim)
#else
(Config, Dim) CONFIGURATION Config; INTEGER *Dim;
#endif
/* Expects an index and returns the corresponding ID number in the range */
/* from 1 to LEN(Config). 'Dim' is in the range from 1 to RANK(Config).  */
{
  CARDINAL Id, Dimension;

  Id = 0;
  if (IS_IN_CONFIG(Config, Dim)) {
    for (Dimension = RANK(Config); Dimension >= 1; Dimension--) {
      Id = Id * LEN1(Config, Dimension) +
        (Dim[Dimension - 1] - LOWER(Config, Dimension));
    }
    Id += 1;
  }
  return Id;
}


INTEGER *DIM
#ifdef __STDC__
(CONFIGURATION Config, CARDINAL Dim)
#else
(Config, Dim) CONFIGURATION Config; CARDINAL Dim;
#endif
/* Returns a vector with the indices of the dimension 'Dim'. 'Dim' is in */
/* the range from 1 to RANK(Config).                                     */
{
  INTEGER *Result;
  CARDINAL Dimension, Divisor;

  /* !!! Aenderungen: "LOCAL_NUM_PEs" */
  Result = (INTEGER *) malloc(sizeof(INTEGER) * LOCAL_NUM_PEs(Config));
  Divisor = 1;
  for (Dimension = 1; Dimension < Dim; Dimension++) {
    Divisor *= LEN1(Config, Dimension);
  }
  if (1 <= Dim && Dim <= RANK(Config)) {
    /* !!! Aenderungen: "LOCAL_NUM_PEs" und "(PE + Config.ID_Sc...- 1)" */
    for (PE = 0; PE < LOCAL_NUM_PEs(Config); PE++) {
      Result[PE] = ((PE + Config.ID_range[me].start_ID - 1) / Divisor) % LEN1(Config, Dim) + LOWER(Config, Dim);
    }
  }
  return Result;
}


CARDINAL *ID
#ifdef __STDC__
(CONFIGURATION Config)
#else
(Config) CONFIGURATION Config;
#endif
/* Returns a vector with the ID */
{
  CARDINAL *Result;

  /* !!! Aenderungen: zweimal LOCAL... und einmal ...[me].start_ID (+1 weg) */
  if (me < Config.u_proc) {
    Result = (CARDINAL *) malloc(sizeof(CARDINAL) * LOCAL_NUM_PEs(Config));
    for (PE = 0; PE < LOCAL_NUM_PEs(Config); PE++) {
      Result[PE] = Config.ID_range[me].start_ID + PE;
    }
  } else {
    Result = NULL;
  }
  return Result;
}



/*********** Implementation of Communication-routines ***********/


BOOLEAN IS_ANY_ACTIVE
/* reduces the active-status of Config locally on each real Processor */
/* and then between all tasks */
#ifdef __STDC__
(CONFIGURATION Config)
#else
(Config) CONFIGURATION Config;
#endif
{
  BOOLEAN  active_status, my_active_status, global_active_status;
  CARDINAL t_ind;

  if (me == 0) { /* i am process 0 */
    /* receive from all other processes in Config and reduce it */
    global_active_status = LOCAL_IS_ANY_ACTIVE(Config);
    for (t_ind = 1; t_ind < Config.u_proc; t_ind++) {
      (void) pvm_recv(task_ids[t_ind], INTER_TASK_TAG);
      PVM_UPKBOOLEAN(&active_status, 1);
      /* reduction: */ 
      global_active_status = global_active_status || active_status;
    }
 
    /* send global_active_status to all tasks except the first */
    (void) pvm_initsend(PvmDataDefault);
    PVM_PKBOOLEAN(&global_active_status, 1);
    (void) pvm_mcast(&task_ids[1], n_tasks - 1, INTER_TASK_TAG); /* index 1!!! */
  }

  if (me >= 1 && me < Config.u_proc) {
    /* i am a process which belongs to this configuration */
    /* send active-status to process 0 */
    my_active_status = LOCAL_IS_ANY_ACTIVE(Config);
    (void) pvm_initsend(PvmDataDefault);
    PVM_PKBOOLEAN(&my_active_status, 1);
    (void) pvm_send(parent_tid, INTER_TASK_TAG);

    /* receive active-status from parent */
    (void) pvm_recv(parent_tid, INTER_TASK_TAG);
    PVM_UPKBOOLEAN(&global_active_status, 1);
  }

  if (me >= Config.u_proc) {
    /* i am a process which doesn't belong to this configuration */
    /* receive active-status from parent */
    (void) pvm_recv(parent_tid, INTER_TASK_TAG);
    PVM_UPKBOOLEAN(&global_active_status, 1);
  }

  return global_active_status;

} /* end IS_ANY_ACTIVE */


/* ------------------------------------------------------------------------- */
/* ------------------- Module IO ------------------------------------------- */
/* ------------------------------------------------------------------------- */

/* message-signals for I/O are defined in "shared_macros.h" */

/*
 *      Implementation of own error-funcions
 */

void Id_Range_Error
#ifdef __STDC__
(CARDINAL id, CONFIGURATION Config)
#else
(id, Config) CARDINAL id; CONFIGURATION Config;
#endif
{
  (void) pvm_initsend(PvmDataDefault);
  PVM_PKSIG(IdRangeErrorSig);
  PVM_PKCARDINAL(&id, 1);
  PVM_PKSTRING(Config.Name);

  printf("id-range-error: no id %lu in Configuration %s\n", id, Config.Name);
  (void) pvm_send(io_tid, IO_TAG);

  EndParallaxisError(-1);
}


/*
 *	Implementation of two compiler error-functions
 */

void CaseError
#ifdef __STDC__
(char file[], int line)
#else
(file, line) char file[]; int line;
#endif
{
  if (me == 0) {
    (void) pvm_initsend(PvmDataDefault);
    PVM_PKSIG(CaseErrorSig);
    (void) pvm_pkstr(file);
    (void) pvm_pkint(&line, 1, 1);
    (void) pvm_send(io_tid, IO_TAG);
  }

  EndParallaxis(-2, StartTime, u_sec, u_msec, s_sec, s_msec);
}


void ReturnError
#ifdef __STDC__
(char file[], int line)
#else
(file, line) char file[]; int line;
#endif
{
  if (me == 0) {
    (void) pvm_initsend(PvmDataDefault);
    PVM_PKSIG(ReturnErrorSig);
    (void) pvm_pkstr(file);
    (void) pvm_pkint(&line, 1, 1);
    (void) pvm_send(io_tid, IO_TAG);
  }

  EndParallaxis(-3, StartTime, u_sec, u_msec, s_sec, s_msec);
}


INTEGER CHECK_PE_NUMBER
#ifdef __STDC__
(CONFIGURATION Config, INTEGER PeNumber, char file[], int line)
#else
(Config, PeNumber, file, line)
CONFIGURATION Config; INTEGER PeNumber;  char file[]; int line;
#endif
/* Checks, if a PE-number is in the range of the configuration. */
/* On success, the PE-number is returned, otherwise the programm */
/* terminates. */
{
  if (0 <= PeNumber && PeNumber < LEN(Config)) {
    return PeNumber;
  } else {
    if (me == 0) {
      (void) pvm_initsend(PvmDataDefault);
      PVM_PKSIG(PeRangeErrorSig);
      (void) pvm_pkstr(file);
      (void) pvm_pkint(&line, 1, 1);
      PVM_PKCARDINAL(&PeNumber, 1);
      (void) pvm_send(io_tid, IO_TAG);

      EndParallaxisError(-3);
    }
  }

  return (INTEGER) 0; /* only to avoid compiler-warning */
}


/* I/O-functions */


void OpenInput
#ifdef __STDC__
(CHAR *s)
#else
(s) CHAR *s;
#endif
{
  if (me == 0) {
    /* i am process 0 */
    /* send Message to Open Input from s */
    (void) pvm_initsend(PvmDataDefault);
    PVM_PKSIG(OpenInputSig);
    PVM_PKSTRING(s);
    (void) pvm_send(io_tid, IO_TAG);
  }

  /* all: receive Done-statusvariable */
  (void) pvm_recv(io_tid, IO_TAG);
  PVM_UPKBOOLEAN(&Done, 1);
}


void OpenOutput
#ifdef __STDC__
(CHAR *s)
#else
(s) CHAR *s;
#endif
{
  if (me == 0) {
    /* i am process 0 */
    /* send Message to Open Input from s */
    (void) pvm_initsend(PvmDataDefault);
    PVM_PKSIG(OpenOutputSig);
    PVM_PKSTRING(s);
    (void) pvm_send(io_tid, IO_TAG);
  }

  /* all: receive Done-statusvariable */
  (void) pvm_recv(io_tid, IO_TAG);
  PVM_UPKBOOLEAN(&Done, 1);
}


void CloseInput()
{
  if (me == 0) {
    /* i am process 0 */
    /* send Message to close Input */
    (void) pvm_initsend(PvmDataDefault);
    PVM_PKSIG(CloseInputSig);
    (void) pvm_send(io_tid, IO_TAG);
  }

  /* all: receive Done-statusvariable */
  (void) pvm_recv(io_tid, IO_TAG);
  PVM_UPKBOOLEAN(&Done, 1);
}


void CloseOutput()
{
  if (me == 0) {
    /* i am process 0 */
    /* send Message to Open Input from s */
    (void) pvm_initsend(PvmDataDefault);
    PVM_PKSIG(CloseOutputSig);
    (void) pvm_send(io_tid, IO_TAG);
  }

  /* all: receive Done-statusvariable */
  (void) pvm_recv(io_tid, IO_TAG);
  PVM_UPKBOOLEAN(&Done, 1);
}


/* ------- Write-Routines ---------------- */


void WriteLn
#ifdef __STDC__
(void)
#else
()
#endif
{
  if (me == 0) {
    (void) pvm_initsend(PvmDataDefault);
    PVM_PKSIG(WriteLnSig);
    (void) pvm_send(io_tid, IO_TAG);
  }

  (void) pvm_recv(io_tid, IO_TAG);
  PVM_UPKBOOLEAN(&Done, 1);
}


void WriteCard
/* prints a scalar CARDINAL value within 'length' print spaces to */
/* stdout; the variable Done is set when successful */
#ifdef __STDC__
(CARDINAL card, CARDINAL length)
#else
(card, length) CARDINAL card; CARDINAL length;
#endif
{
  if (me == 0) {
    (void) pvm_initsend(PvmDataDefault);
    PVM_PKSIG(WriteCardSig);
    PVM_PKCARDINAL(&card, 1);
    PVM_PKCARDINAL(&length, 1);
    (void) pvm_send(io_tid, IO_TAG);
  }

  (void) pvm_recv(io_tid, IO_TAG);
  PVM_UPKBOOLEAN(&Done, 1);
}


void WriteInt
/* prints a scalar INTEGER value within 'length' print spaces to */
/* stdout; the variable Done is set when successful */
#ifdef __STDC__
(INTEGER int_value, CARDINAL length)
#else
(int_value, length) INTEGER int_value; CARDINAL length;
#endif
{
  if (me == 0) {
    (void) pvm_initsend(PvmDataDefault);
    PVM_PKSIG(WriteIntSig);
    PVM_PKINTEGER(&int_value, 1);
    PVM_PKCARDINAL(&length, 1);
    (void) pvm_send(io_tid, IO_TAG);
  }

  (void) pvm_recv(io_tid, IO_TAG);
  PVM_UPKBOOLEAN(&Done, 1);
}


void WriteReal
#ifdef __STDC__
(REAL real_value, CARDINAL length)
#else
(real_value, length) REAL real_value; CARDINAL length;
#endif
{
  if (me == 0) {
    (void) pvm_initsend(PvmDataDefault);
    PVM_PKSIG(WriteRealSig);
    PVM_PKREAL(&real_value, 1);
    PVM_PKCARDINAL(&length, 1);
    (void) pvm_send(io_tid, IO_TAG);
  }

  (void) pvm_recv(io_tid, IO_TAG);
  PVM_UPKBOOLEAN(&Done, 1);
}


void WriteFixPt
#ifdef __STDC__
(REAL real_value, CARDINAL length, CARDINAL decimals)
#else
(real_value, length, decimals) REAL real_value; CARDINAL length, decimals;
#endif
{
  if (me == 0) {
    (void) pvm_initsend(PvmDataDefault);
    PVM_PKSIG(WriteFixPtSig);
    PVM_PKREAL(&real_value, 1);
    PVM_PKCARDINAL(&length, 1);
    PVM_PKCARDINAL(&decimals, 1);
    (void) pvm_send(io_tid, IO_TAG);
  }

  (void) pvm_recv(io_tid, IO_TAG);
  PVM_UPKBOOLEAN(&Done, 1);
}


void WriteBool
#ifdef __STDC__
(BOOLEAN bool_value, CARDINAL length)
#else
(bool_value, length) BOOLEAN bool_value; CARDINAL length;
#endif
{
  if (me == 0) {
    (void) pvm_initsend(PvmDataDefault);
    PVM_PKSIG(WriteBoolSig);
    PVM_PKBOOLEAN(&bool_value, 1);
    PVM_PKCARDINAL(&length, 1);
    (void) pvm_send(io_tid, IO_TAG);
  }

  (void) pvm_recv(io_tid, IO_TAG);
  PVM_UPKBOOLEAN(&Done, 1);
}


void Write
#ifdef __STDC__
(CHAR char_value)
#else
(char_value) CHAR char_value;
#endif
{
  if (me == 0) {
    (void) pvm_initsend(PvmDataDefault);
    PVM_PKSIG(WriteCharSig);
    PVM_PKCHAR(&char_value, 1);
    (void) pvm_send(io_tid, IO_TAG);
  }

  (void) pvm_recv(io_tid, IO_TAG);
  PVM_UPKBOOLEAN(&Done, 1);
}


void WriteString
#ifdef __STDC__
(CHAR *string_value)
#else
(string_value) CHAR *string_value;
#endif
{
  if (me == 0) {
    (void) pvm_initsend(PvmDataDefault);
    PVM_PKSIG(WriteStringSig);
    PVM_PKSTRING(string_value);
    (void) pvm_send(io_tid, IO_TAG);
  }

  (void) pvm_recv(io_tid, IO_TAG);
  PVM_UPKBOOLEAN(&Done, 1);
}


/* ------------ Vector-Write-Routines -------------- */


void collect_format_messages
#ifdef __STDC__
(CONFIGURATION Config, CARDINAL *item_count)
#else
(Config, item_count)
CONFIGURATION Config;
CARDINAL *item_count;
#endif
{
  CARDINAL count, t_ind, n, Dimension, Divisor, format_info, i;

  /* first collect all format-messages and send them to the i/o-process */
  /* pack own format-info in buffer */
  count = 0;
  *item_count = 0;
  PVM_PKCARDINAL(&Config.ID_range[me].num_breaks, 1);
  FOR_ALL_LOCAL_PEs(Config) {
    if (LOCAL_IS_ACTIVE(Config, PE)) {
      (*item_count)++; /* increment counter of items */
      count++;       /* increment counter of data after last newline */
    }

    /* all PEs even the inactive ones */
    /* check if at this position is a new-line necessary */
    for (Dimension = 0, Divisor = 1; Dimension < RANK(Config); Dimension++) {
      Divisor *= LEN1(Config,Dimension+1);
      if (((Config.ID_range[me].start_ID + PE) % Divisor) == 0) {
	/* new-line is necessary */
	PVM_PKCARDINAL(&count, 1); /* store format-data */
	count = 0;                 /* reset counter */
      } /* end if */
    } /* end for Dimension */

  } /* end FOR_ALL_LOCAL_PEs */

  count++; /* terminate format_field correctly */
  PVM_PKCARDINAL(&count, 1);

  /* now receive from the other tasks and pack it in global message */
  for (t_ind = 1; t_ind < Config.u_proc; t_ind++) {
    (void) pvm_recv(task_ids[t_ind], INTER_TASK_TAG);
    PVM_UPKCARDINAL(&n, 1);
    PVM_PKCARDINAL(&n, 1);
    for (i=0; i<=n; i++) {            /* unpack n+1 format-infos */
      PVM_UPKCARDINAL(&format_info, 1);
      PVM_PKCARDINAL(&format_info, 1);
    }
  } /* end for t_ind  */
}


#define COLLECT_DATA_MESSAGES(CONFIG,VALUE,ITEM_COUNT,PVM_PKROUTINE,PVM_UPKROUTINE,TYPE) \
{                                                                       \
  CARDINAL t_ind, n, i;                                                 \
  TYPE data;                                                            \
                                                                        \
  /* collect all data-messages and send them to the i/o-process */      \
                                                                        \
  /* pack own data in buffer */                                         \
  PVM_PKCARDINAL(&ITEM_COUNT, 1);                                       \
  FOR_ALL_LOCAL_ACTIVES(CONFIG) {                                       \
    PVM_PKROUTINE(&VALUE[PE], 1); /* pack value in send-buffer */       \
  }                                                                     \
                                                                        \
  /* now receive from the other tasks and pack it in global message */	\
  for (t_ind = 1; t_ind < CONFIG.u_proc; t_ind++) {			\
    (void) pvm_recv(task_ids[t_ind], INTER_TASK_TAG);		        \
    PVM_UPKCARDINAL(&n, 1);						\
    PVM_PKCARDINAL(&n, 1);                                              \
    for (i=0; i<n; i++) {                                               \
      PVM_UPKROUTINE(&data, 1); 					\
      PVM_PKROUTINE(&data, 1);                                          \
    }									\
  } /* end for t_ind  */                                                \
}


void send_format_message
#ifdef __STDC__
(CONFIGURATION Config, CARDINAL *item_count)
#else
(Config, item_count) 
CONFIGURATION Config; CARDINAL *item_count;
#endif
{
  CARDINAL count;
  CARDINAL Dimension, Divisor;

  count = 0;
  *item_count = 0;
  (void) pvm_initsend(PvmDataDefault);
  PVM_PKCARDINAL(&Config.ID_range[me].num_breaks, 1);

  FOR_ALL_LOCAL_PEs(Config) {
    if (LOCAL_IS_ACTIVE(Config, PE)) {
      (*item_count)++; /* increment counter of packed items */
      count++;      /* increment counter of packed data after last newline */
    }

    /* all PEs even the inactive ones */
    /* check if at this position is a new-line necessary */
    for (Dimension = 0, Divisor = 1; Dimension < RANK(Config); Dimension++) {
      Divisor *= LEN1(Config,Dimension+1);
      if (((Config.ID_range[me].start_ID + PE) % Divisor) == 0) {
	/* new-line is necessary */
	PVM_PKCARDINAL(&count, 1); /* store format-data */
	count = 0;                 /* reset counter */
      } /* end if */
    } /* end for Dimension */

  } /* end FOR_ALL_LOCAL_PEs */

  count++; /* terminate format_field correctly */
  PVM_PKCARDINAL(&count, 1);
  (void) pvm_send(parent_tid, INTER_TASK_TAG); /* send format-message */
}


void WriteCard1
#ifdef __STDC__
(CONFIGURATION Config, CARDINAL *card, CARDINAL length)
#else
(Config, card, length) CONFIGURATION Config; CARDINAL *card; CARDINAL length;
#endif
{
  CARDINAL item_count;      /* counter packed data-items */

  if (me == 0) { /* i am process 0 */
    (void) pvm_initsend(PvmDataDefault);
    PVM_PKSIG(WriteCardVectSig);          /* pack WriteCardVect-Signal */

    PVM_PKCARDINAL(&length, 1);
    PVM_PKCARDINAL(&Config.u_proc, 1);
    collect_format_messages(Config, &item_count);
    /* the same with all data-messages */
    COLLECT_DATA_MESSAGES(Config, card, item_count, PVM_PKCARDINAL, PVM_UPKCARDINAL, CARDINAL);
    /* collecting is complete, now send message to i/o-process */
    pvm_send(io_tid, IO_TAG);
  } /* end if (me == 0) */

  if (me >= 1 && me < Config.u_proc) {
    /* i am a process which belongs to this configuration */
    /* send the format-message to parent */
    send_format_message(Config, &item_count);
    /* now the quite same for the data */
    (void) pvm_initsend(PvmDataDefault);
    PVM_PKCARDINAL(&item_count, 1);  /* number of data-items */
    FOR_ALL_LOCAL_ACTIVES(Config) {
      /* all actives pack their value into the message */
      PVM_PKCARDINAL(&card[PE], 1);  /* pack value in send-buffer */
    }
    (void) pvm_send(parent_tid, INTER_TASK_TAG); /* send format-buffer */
  }

  /* all processors receive Done-Status from i/o-process */
  (void) pvm_recv(io_tid, IO_TAG);
  PVM_UPKBOOLEAN(&Done, 1);
} /* end WriteCard1 */


void WriteInt1
#ifdef __STDC__
(CONFIGURATION Config, INTEGER *int_value, CARDINAL length)
#else
(Config, int_value, length) 
CONFIGURATION Config; INTEGER *int_value; CARDINAL length;
#endif
{
  CARDINAL item_count;      /* counter packed data-items */

  if (me == 0) { /* i am process 0 */
    /* send WriteIntVect-Signal */
    (void) pvm_initsend(PvmDataDefault);
    PVM_PKSIG(WriteIntVectSig);

    PVM_PKCARDINAL(&length, 1);
    PVM_PKCARDINAL(&Config.u_proc, 1);
    collect_format_messages(Config, &item_count);
    /* the same with all data-messages */
    COLLECT_DATA_MESSAGES(Config, int_value, item_count, PVM_PKINTEGER, PVM_UPKINTEGER, INTEGER);
    /* collecting is complete, now send message to i/o-process */
    pvm_send(io_tid, IO_TAG);
  } /* end if (me == 0) */

  if (me >= 1 && me < Config.u_proc) {
    /* i am a process which belongs to this configuration */
    /* send the format-message to parent */
    send_format_message(Config, &item_count);
    /* now the quite same for the data */
    (void) pvm_initsend(PvmDataDefault);
    PVM_PKCARDINAL(&item_count, 1);  /* number of data-items */
    FOR_ALL_LOCAL_ACTIVES(Config) {
      /* all actives pack their value into the message */
      PVM_PKINTEGER(&int_value[PE], 1);  /* pack value in send-buffer */
    }
    (void) pvm_send(parent_tid, INTER_TASK_TAG); /* send format-buffer */
  }

  /* all processors receive Done-Status from i/o-process */
  (void) pvm_recv(io_tid, IO_TAG);
  PVM_UPKBOOLEAN(&Done, 1);
} /* end WriteInt1 */


void WriteReal1
#ifdef __STDC__
(CONFIGURATION Config, REAL *real_value, CARDINAL length)
#else
(Config, real_value, length) 
CONFIGURATION Config; REAL *real_value; CARDINAL length;
#endif
{
  CARDINAL item_count;      /* counter packed data-items */

  if (me == 0) { /* i am process 0 */
    /* send WriteRealVect-Signal */
    (void) pvm_initsend(PvmDataDefault);
    PVM_PKSIG(WriteRealVectSig);

    PVM_PKCARDINAL(&length, 1);
    PVM_PKCARDINAL(&Config.u_proc, 1);
    collect_format_messages(Config, &item_count);
    /* the same with all data-messages */
    COLLECT_DATA_MESSAGES(Config, real_value, item_count, PVM_PKREAL, PVM_UPKREAL, REAL);
    /* collecting is complete, now send message to i/o-process */
    pvm_send(io_tid, IO_TAG);
  } /* end if (me == 0) */

  if (me >= 1 && me < Config.u_proc) {
    /* i am a process which belongs to this configuration */
    /* send the format-message to parent */
    send_format_message(Config, &item_count);
    /* now the quite same for the data */
    (void) pvm_initsend(PvmDataDefault);
    PVM_PKCARDINAL(&item_count, 1);  /* number of data-items */
    FOR_ALL_LOCAL_ACTIVES(Config) {
      /* all actives pack their value into the message */
      PVM_PKREAL(&real_value[PE], 1);  /* pack value in send-buffer */
    }
    (void) pvm_send(parent_tid, INTER_TASK_TAG); /* send format-buffer */
  }

  /* all processors receive Done-Status from i/o-process */
  (void) pvm_recv(io_tid, IO_TAG);
  PVM_UPKBOOLEAN(&Done, 1);
} /* end WriteReal1 */


void WriteFixPt1
#ifdef __STDC__
(CONFIGURATION Config, REAL *real_value, CARDINAL length, CARDINAL decimals)
#else
(Config, real_value, length, decimals) 
CONFIGURATION Config; REAL *real_value; CARDINAL length; CARDINAL decimals;
#endif
{
  CARDINAL item_count;      /* counter packed data-items */

  if (me == 0) { /* i am process 0 */
    /* send WriteFixPtVect-Signal */
    (void) pvm_initsend(PvmDataDefault);
    PVM_PKSIG(WriteFixPtVectSig);

    PVM_PKCARDINAL(&length, 1);
    PVM_PKCARDINAL(&decimals, 1);
    PVM_PKCARDINAL(&Config.u_proc, 1);
    collect_format_messages(Config, &item_count);
    /* the same with all data-messages */
    COLLECT_DATA_MESSAGES(Config, real_value, item_count, PVM_PKREAL, PVM_UPKREAL, REAL);
    /* collecting is complete, now send message to i/o-process */
    pvm_send(io_tid, IO_TAG);
  } /* end if (me == 0) */

  if (me >= 1 && me < Config.u_proc) {
    /* i am a process which belongs to this configuration */
    /* send the format-message to parent */
    send_format_message(Config, &item_count);
    /* now the quite same for the data */
    (void) pvm_initsend(PvmDataDefault);
    PVM_PKCARDINAL(&item_count, 1);  /* number of data-items */
    FOR_ALL_LOCAL_ACTIVES(Config) {
      /* all actives pack their value into the message */
      PVM_PKREAL(&real_value[PE], 1);  /* pack value in send-buffer */
    }
    (void) pvm_send(parent_tid, INTER_TASK_TAG); /* send format-buffer */
  }

  /* all processors receive Done-Status from i/o-process */
  (void) pvm_recv(io_tid, IO_TAG);
  PVM_UPKBOOLEAN(&Done, 1);
} /* end WriteFixPt1 */


void WriteBool1
#ifdef __STDC__
(CONFIGURATION Config, BOOLEAN *bool_value, CARDINAL length)
#else
(Config, bool_value, length) 
CONFIGURATION Config; BOOLEAN *bool_value; CARDINAL length;
#endif
{
  CARDINAL item_count;      /* counter packed data-items */

  if (me == 0) { /* i am process 0 */
    /* send WriteBoolVect-Signal */
    (void) pvm_initsend(PvmDataDefault);
    PVM_PKSIG(WriteBoolVectSig);

    PVM_PKCARDINAL(&length, 1);
    PVM_PKCARDINAL(&Config.u_proc, 1);
    collect_format_messages(Config, &item_count);
    /* the same with all data-messages */
    COLLECT_DATA_MESSAGES(Config, bool_value, item_count, PVM_PKBOOLEAN, PVM_UPKBOOLEAN, BOOLEAN);
    /* collecting is complete, now send message to i/o-process */
    pvm_send(io_tid, IO_TAG);
  } /* end if (me == 0) */

  if (me >= 1 && me < Config.u_proc) {
    /* i am a process which belongs to this configuration */
    /* send the format-message to parent */
    send_format_message(Config, &item_count);
    /* now the quite same for the data */
    (void) pvm_initsend(PvmDataDefault);
    PVM_PKCARDINAL(&item_count, 1);  /* number of data-items */
    FOR_ALL_LOCAL_ACTIVES(Config) {
      /* all actives pack their value into the message */
      PVM_PKBOOLEAN(&bool_value[PE], 1);  /* pack value in send-buffer */
    }
    (void) pvm_send(parent_tid, INTER_TASK_TAG); /* send format-buffer */
  }

  /* all processors receive Done-Status from i/o-process */
  (void) pvm_recv(io_tid, IO_TAG);
  PVM_UPKBOOLEAN(&Done, 1);
} /* end WriteBool1 */


void Write1
#ifdef __STDC__
(CONFIGURATION Config, CHAR *char_value)
#else
(Config, char_value) 
CONFIGURATION Config; CHAR *char_value;
#endif
{
  CARDINAL item_count;      /* counter packed data-items */

  if (me == 0) { /* i am process 0 */
    /* send WriteCharVect-Signal */
    (void) pvm_initsend(PvmDataDefault);
    PVM_PKSIG(WriteCharVectSig);

    PVM_PKCARDINAL(&Config.u_proc, 1);
    collect_format_messages(Config, &item_count);
    /* the same with all data-messages */
    COLLECT_DATA_MESSAGES(Config, char_value, item_count, PVM_PKCHAR, PVM_UPKCHAR, CHAR);
    /* collecting is complete, now send message to i/o-process */
    pvm_send(io_tid, IO_TAG);
  } /* end if (me == 0) */

  if (me >= 1 && me < Config.u_proc) {
    /* i am a process which belongs to this configuration */
    /* send the format-message to parent */
    send_format_message(Config, &item_count);
    /* now the quite same for the data */
    (void) pvm_initsend(PvmDataDefault);
    PVM_PKCARDINAL(&item_count, 1);  /* number of data-items */
    FOR_ALL_LOCAL_ACTIVES(Config) {
      /* all actives pack their value into the message */
      PVM_PKCHAR(&char_value[PE], 1);  /* pack value in send-buffer */
    }
    (void) pvm_send(parent_tid, INTER_TASK_TAG); /* send format-buffer */
  }

  /* all processors receive Done-Status from i/o-process */
  (void) pvm_recv(io_tid, IO_TAG);
  PVM_UPKBOOLEAN(&Done, 1);
} /* end Write1 */


void WriteString1
#ifdef __STDC__
(CONFIGURATION Config, CHAR **string_value)
#else
(Config, string_value) 
CONFIGURATION Config; CHAR **string_value;
#endif
{
  CARDINAL t_ind, n, i;
  CHAR string[1024];
  CARDINAL item_count;      /* counter packed data-items */

  if (me == 0) { /* i am process 0 */
    /* send WriteStringVect-Signal */
    (void) pvm_initsend(PvmDataDefault);
    PVM_PKSIG(WriteStringVectSig);

    PVM_PKCARDINAL(&Config.u_proc, 1);
    collect_format_messages(Config, &item_count);
    /* the same with all data-messages */
    /* pack own data in buffer */                                         
    PVM_PKCARDINAL(&item_count, 1);
    FOR_ALL_LOCAL_ACTIVES(Config) {
      PVM_PKSTRING(string_value[PE]); /* pack value in send-buffer */
    }

    /* now receive from the other tasks and pack it in global message*/ 
    for (t_ind = 1; t_ind < Config.u_proc; t_ind++) {                    
      (void) pvm_recv(task_ids[t_ind], INTER_TASK_TAG);                       
      PVM_UPKCARDINAL(&n, 1);
      PVM_PKCARDINAL(&n, 1);
      for (i = 0; i<n; i++) {
	PVM_UPKSTRING(string);
	PVM_PKSTRING(string);                                             
      } /* end for i */
    } /* end for t_ind  */                                               
                                                                       
    /* collecting is complete, now send message to i/o-process */        
    pvm_send(io_tid, IO_TAG);
  } /* end if (me == 0) */

  if (me >= 1 && me < Config.u_proc) {
    /* i am a process which belongs to this configuration */
    /* send the format-message to parent */
    send_format_message(Config, &item_count);
    /* now the quite same for the data */
    (void) pvm_initsend(PvmDataDefault);
    PVM_PKCARDINAL(&item_count, 1);  /* number of data-items */
    FOR_ALL_LOCAL_ACTIVES(Config) {
      /* all actives pack their value into the message */
      PVM_PKSTRING(string_value[PE]);  /* pack value in send-buffer */
    }
    (void) pvm_send(parent_tid, INTER_TASK_TAG); /* send format-buffer */
  }

  /* all processors receive Done-Status from i/o-process */
  (void) pvm_recv(io_tid, IO_TAG);
  PVM_UPKBOOLEAN(&Done, 1);
} /* end WriteString1 */


/* ----------- End of Vector-Write-Routines --------- */


/* ------- Read-Routines ---------------- */


void Read
#ifdef __STDC__
(CHAR *c)
#else
(c) CHAR *c;
#endif
{
  if (me == 0) {
    /* i am task 0 */
    (void) pvm_initsend(PvmDataDefault);
    PVM_PKSIG(ReadCharSig);
    (void) pvm_send(io_tid, IO_TAG);
  }

  (void) pvm_recv(io_tid, IO_TAG);
  PVM_UPKCHAR(c, 1);
  PVM_UPKBOOLEAN(&Done, 1);
}


void ReadCard
#ifdef __STDC__
(CARDINAL *c)
#else
(c) CARDINAL *c;
#endif
{
  if (me == 0) {
    /* i am task 0 */
    (void) pvm_initsend(PvmDataDefault);
    PVM_PKSIG(ReadCardSig);
    (void) pvm_send(io_tid, IO_TAG);
  }

  (void) pvm_recv(io_tid, IO_TAG);
  PVM_UPKCARDINAL(c, 1);
  PVM_UPKCHAR(&termCH, 1);
  PVM_UPKBOOLEAN(&Done, 1);
}


void ReadInt
#ifdef __STDC__
(INTEGER *i)
#else
(i) INTEGER *i;
#endif
{
  if (me == 0) {
    /* i am task 0 */
    (void) pvm_initsend(PvmDataDefault);
    PVM_PKSIG(ReadIntSig);
    (void) pvm_send(io_tid, IO_TAG);
  }

  (void) pvm_recv(io_tid, IO_TAG);
  PVM_UPKINTEGER(i, 1);
  PVM_UPKCHAR(&termCH, 1);
  PVM_UPKBOOLEAN(&Done, 1);
}


void ReadReal
#ifdef __STDC__
(REAL *r)
#else
(r) REAL *r;
#endif
{
  if (me == 0) {
    /* i am task 0 */
    (void) pvm_initsend(PvmDataDefault);
    PVM_PKSIG(ReadRealSig);
    (void) pvm_send(io_tid, IO_TAG);
  }

  (void) pvm_recv(io_tid, IO_TAG);
  PVM_UPKREAL(r, 1);
  PVM_UPKCHAR(&termCH, 1);
  PVM_UPKBOOLEAN(&Done, 1);
}


void ReadBool
#ifdef __STDC__
(BOOLEAN *b)
#else
(b) BOOLEAN *b;
#endif
{
  if (me == 0) {
    /* i am task 0 */
    (void) pvm_initsend(PvmDataDefault);
    PVM_PKSIG(ReadBoolSig);
    (void) pvm_send(io_tid, IO_TAG);
  }

  (void) pvm_recv(io_tid, IO_TAG);
  PVM_UPKBOOLEAN(b, 1);
  PVM_UPKCHAR(&termCH, 1);
  PVM_UPKBOOLEAN(&Done, 1);
}


void ReadString
#ifdef __STDC__
(CHAR *s)
#else
(s) CHAR *s;
#endif
{
  if (me == 0) {
    /* i am task 0 */
    (void) pvm_initsend(PvmDataDefault);
    PVM_PKSIG(ReadStringSig);
    (void) pvm_send(io_tid, IO_TAG);
  }

  (void) pvm_recv(io_tid, IO_TAG);
  PVM_UPKSTRING(s);
  PVM_UPKCHAR(&termCH, 1);
  PVM_UPKBOOLEAN(&Done, 1);
}


void collect_number_of_items
/* collects messages which contain number of items to read for each task */
/* and send it to the I/O-process. This procedure is called only by task 0 */
#ifdef __STDC__
(CONFIGURATION Config, int readsignal)
#else
(Config, readsignal) 
CONFIGURATION Config; int readsignal;
#endif
{
  CARDINAL t_ind, count, item_count;

  (void) pvm_initsend(PvmDataDefault);
  PVM_PKSIG(readsignal);          /* pack signal which kind of data to read */
  PVM_PKCARDINAL(&Config.u_proc, 1);

  /* compute own number of items to read */
  item_count = 0;
  FOR_ALL_LOCAL_ACTIVES(Config) {
    item_count++;
  }
  PVM_PKCARDINAL(&item_count, 1);

  /* receive number from other tasks and pack it all in one message */
  for (t_ind = 1; t_ind < Config.u_proc; t_ind++) {
    (void) pvm_recv(task_ids[t_ind], INTER_TASK_TAG);
    PVM_UPKCARDINAL(&count, 1);
    PVM_PKCARDINAL(&count, 1);
  }
  /* send numbers to the I/O-process */
  (void) pvm_send(io_tid, IO_TAG);
}


void send_number_of_items
/* compute number of items to read and send it to task 0 */
/* this procedure is called by all tasks except number 0 */
#ifdef __STDC__
(CONFIGURATION Config)
#else
(Config) CONFIGURATION Config;
#endif
{
  CARDINAL item_count;

  (void) pvm_initsend(PvmDataDefault);
  item_count = 0;
  FOR_ALL_LOCAL_ACTIVES(Config) {
    item_count++;
  }
  PVM_PKCARDINAL(&item_count, 1);
  (void) pvm_send(parent_tid, INTER_TASK_TAG);
}


#define RECEIVE_DATA(CONFIG, VALUE, TYPE, PKROUTINE, UPKROUTINE) \
{                                                                \
  TYPE data;							 \
  CARDINAL d_length, t_ind, i;					 \
								 \
  if (me == 0) { /* i am process 0 */				 \
    (void) pvm_recv(io_tid, IO_TAG);				 \
								 \
    PVM_UPKCARDINAL(&d_length, 1);				 \
    FOR_ALL_LOCAL_ACTIVES(CONFIG) {				 \
      UPKROUTINE(&VALUE[PE], 1);				 \
    }								 \
								 \
    for (t_ind = 1; t_ind < (CONFIG).u_proc; t_ind++) {		 \
      (void) pvm_initsend(PvmDataDefault);			 \
      PVM_UPKCARDINAL(&d_length, 1);				 \
      for (i = 0; i < d_length; i++) {				 \
	UPKROUTINE(&data, 1);					 \
	PKROUTINE(&data, 1);					 \
      }								 \
      (void) pvm_send(task_ids[t_ind], INTER_TASK_TAG);		 \
    }								 \
  }								 \
								 \
  if (me >= 1 && me < (CONFIG).u_proc) {			 \
    /* i am a process which belongs to this configuration */	 \
    (void) pvm_recv(parent_tid, INTER_TASK_TAG);		 \
    FOR_ALL_LOCAL_ACTIVES(CONFIG) {				 \
      UPKROUTINE(&VALUE[PE], 1);				 \
    }								 \
  }								 \
								 \
  (void) pvm_recv(io_tid, IO_TAG);				 \
  PVM_UPKCHAR(&termCH, 1);                                       \
  PVM_UPKBOOLEAN(&Done, 1);					 \
}
      

void ReadCard1
#ifdef __STDC__
(CONFIGURATION Config, CARDINAL *card_value)
#else
(Config, card_value) CONFIGURATION Config, CARDINAL *card_value;
#endif
{
  if (me == 0) { /* i am process 0 */
    collect_number_of_items(Config, ReadCardVectSig);
  } /* end if (me == 0) */

  if (me >= 1 && me < Config.u_proc) {
    /* i am a process which belongs to this configuration */
    /* send the format-message to parent */
    send_number_of_items(Config);
  }

  /* receive data-message from parent-process */
  RECEIVE_DATA(Config, card_value, CARDINAL,
	       PVM_PKCARDINAL, PVM_UPKCARDINAL);
} /* end ReadCard1 */



void ReadInt1
#ifdef __STDC__
(CONFIGURATION Config, INTEGER *int_value)
#else
(Config, int_value) CONFIGURATION Config, INTEGER *int_value;
#endif
{
  if (me == 0) { /* i am process 0 */
    collect_number_of_items(Config, ReadIntVectSig);
  } /* end if (me == 0) */

  if (me >= 1 && me < Config.u_proc) {
    /* i am a process which belongs to this configuration */
    /* send the format-message to parent */
    send_number_of_items(Config);
  }

  /* receive data-message from parent-process */
  RECEIVE_DATA(Config, int_value, INTEGER,
	       PVM_PKINTEGER, PVM_UPKINTEGER);
} /* end ReadInt1 */


void ReadReal1
#ifdef __STDC__
(CONFIGURATION Config, REAL *real_value)
#else
(Config, real_value) CONFIGURATION Config, REAL *real_value;
#endif
{
  if (me == 0) { /* i am process 0 */
    collect_number_of_items(Config, ReadRealVectSig);
  } /* end if (me == 0) */

  if (me >= 1 && me < Config.u_proc) {
    /* i am a process which belongs to this configuration */
    /* send the format-message to parent */
    send_number_of_items(Config);
  }

  /* receive data-message from parent-process */
  RECEIVE_DATA(Config, real_value, REAL,
	       PVM_PKREAL, PVM_UPKREAL);
} /* end ReadReal1 */


void ReadBool1
#ifdef __STDC__
(CONFIGURATION Config, BOOLEAN *bool_value)
#else
(Config, bool_value) CONFIGURATION Config, BOOLEAN *bool_value;
#endif
{
  if (me == 0) { /* i am process 0 */
    collect_number_of_items(Config, ReadBoolVectSig);
  } /* end if (me == 0) */

  if (me >= 1 && me < Config.u_proc) {
    /* i am a process which belongs to this configuration */
    /* send the format-message to parent */
    send_number_of_items(Config);
  }

  /* receive data-message from parent-process */
  RECEIVE_DATA(Config, bool_value, BOOLEAN,
	       PVM_PKBOOLEAN, PVM_UPKBOOLEAN);
} /* end ReadBool1 */


void collect_noit_format
/* collects messages which contain number of items to read for each task */
/* and format-information */
/* and sends it to the I/O-process. This procedure is called only by task 0 */
#ifdef __STDC__
(CONFIGURATION Config)
#else
(Config, item_count) 
CONFIGURATION Config;
#endif
{
  CARDINAL t_ind, count, item_count, num_breaks, i, format_info;
  CARDINAL Dimension, Divisor;

  (void) pvm_initsend(PvmDataDefault);
  PVM_PKSIG(ReadCharVectSig);     /* pack signal to read CHARs */
  PVM_PKCARDINAL(&Config.u_proc, 1);

  PVM_PKCARDINAL(&Config.ID_range[me].num_breaks, 1);
  FOR_ALL_LOCAL_PEs(Config) {
    if (LOCAL_IS_ACTIVE(Config, PE)) {
      item_count++; /* increment counter of packed items */
      count++;      /* increment counter of packed data after last newline */
    }

    /* all PEs even the inactive ones */
    /* check if at this position is a new-line necessary */
    for (Dimension = 0, Divisor = 1; Dimension < RANK(Config); Dimension++) {
      Divisor *= LEN1(Config,Dimension+1);
      if (((Config.ID_range[me].start_ID + PE) % Divisor) == 0) {
	/* new-line is necessary */
	PVM_PKCARDINAL(&count, 1); /* pack format-data */
	count = 0;                 /* reset counter */
      } /* end if */
    } /* end for Dimension */

  } /* end FOR_ALL_LOCAL_PEs */

  count++; /* terminate format-message correctly */
  PVM_PKCARDINAL(&count, 1);
  PVM_PKCARDINAL(&item_count, 1);

  /* receive messages from other tasks and pack it all in one message */
  for (t_ind = 1; t_ind < Config.u_proc; t_ind++) {
    (void) pvm_recv(task_ids[t_ind], INTER_TASK_TAG);
    PVM_UPKCARDINAL(&num_breaks, 1);
    PVM_PKCARDINAL(&num_breaks, 1);
    for (i = 0; i < num_breaks + 2; i++) { /* + 2 because of item_count */
      PVM_UPKCARDINAL(&format_info, 1);
      PVM_PKCARDINAL(&format_info, 1);
    }
  }
  /* send format-info to the I/O-process */
  (void) pvm_send(io_tid, IO_TAG);
}


void send_noit_format
/* sends message which contains number of items to read and */
/* format-information to the parent-process. */
/* This procedure is called by all tasks except task 0 */
#ifdef __STDC__
(CONFIGURATION Config)
#else
(Config, item_count) 
CONFIGURATION Config;
#endif
{
  CARDINAL count, item_count;
  CARDINAL Dimension, Divisor;

  (void) pvm_initsend(PvmDataDefault);

  PVM_PKCARDINAL(&(Config.ID_range[me].num_breaks), 1);
  FOR_ALL_LOCAL_PEs(Config) {
    if (LOCAL_IS_ACTIVE(Config, PE)) {
      item_count++; /* increment counter of packed items */
      count++;      /* increment counter of packed data after last newline */
    }

    /* all PEs even the inactive ones */
    /* check if at this position is a new-line necessary */
    for (Dimension = 0, Divisor = 1; Dimension < RANK(Config); Dimension++) {
      Divisor *= LEN1(Config,Dimension+1);
      if (((Config.ID_range[me].start_ID + PE) % Divisor) == 0) {
	/* new-line is necessary */
	PVM_PKCARDINAL(&count, 1); /* pack format-data */
	count = 0;                 /* reset counter */
      } /* end if */
    } /* end for Dimension */

  } /* end FOR_ALL_LOCAL_PEs */

  count++; /* terminate format-message correctly */
  PVM_PKCARDINAL(&count, 1);
  PVM_PKCARDINAL(&item_count, 1);

  /* send message to the parent-process */
  (void) pvm_send(parent_tid, INTER_TASK_TAG);
}


void Read1
#ifdef __STDC__
(CONFIGURATION Config, CHAR *char_value)
#else
(Config, char_value) CONFIGURATION Config, CHAR *char_value;
#endif
{
  if (me == 0) { /* i am process 0 */
    collect_noit_format(Config);
  } /* end if (me == 0) */

  if (me >= 1 && me < Config.u_proc) {
    /* i am a process which belongs to this configuration */
    /* send the format-message to parent */
    send_noit_format(Config);
  }

  /* receive data-message from parent-process */
  RECEIVE_DATA(Config, char_value, CHAR,
	       PVM_PKCHAR, PVM_UPKCHAR);
} /* end Read1 */


void ReadString1
#ifdef __STDC__
(CONFIGURATION Config, CHAR **string_value)
#else
(Config, char_value) CONFIGURATION Config, CHAR **string_value;
#endif
{
  CHAR data[1024];
  CARDINAL d_length, t_ind, i;

  if (me == 0) { /* i am process 0 */
    collect_number_of_items(Config, ReadCardVectSig);

    /* receive data-message from parent-process */
    (void) pvm_recv(io_tid, IO_TAG);

    PVM_UPKCARDINAL(&d_length, 1);
    FOR_ALL_LOCAL_ACTIVES(Config) {
      PVM_UPKSTRING(string_value[PE]);
    }

    /* compose data-messages to other tasks */
    for (t_ind = 1; t_ind < (Config).u_proc; t_ind++) {
      (void) pvm_initsend(PvmDataDefault);
      PVM_UPKCARDINAL(&d_length, 1);
      for (i = 0; i < d_length; i++) {
	PVM_UPKSTRING(data);
	PVM_PKSTRING(data);
      }
      (void) pvm_send(task_ids[t_ind], INTER_TASK_TAG);
    }
  } /* end if (me == 0) */


  if (me >= 1 && me < Config.u_proc) {
    /* i am a process which belongs to this configuration */
    /* send the format-message to parent */
    send_number_of_items(Config);

    /* receive data from parent */
    (void) pvm_recv(parent_tid, INTER_TASK_TAG);
    FOR_ALL_LOCAL_ACTIVES(Config) {
      PVM_UPKSTRING(string_value[PE]);
    }
  }

  /* all tasks receive termCH and Done */
  (void) pvm_recv(io_tid, IO_TAG);
  PVM_UPKCHAR(&termCH, 1);
  PVM_UPKBOOLEAN(&Done, 1);
}



/* ---------------------------------------------------------------------- */
/* functions for time-measuring */
/* ---------------------------------------------------------------------- */

#define InitTime(x) getrusage(RUSAGE_SELF, &x);
/* call this procedure to start time-measuring and keep x to use it with StopTime */

void StopTime
#ifdef __STDC__
(struct rusage Zeit1, long *u_sec, long *u_msec, long *s_sec, long* s_msec)
#else
(Zeit1, u_sec, u_msec, s_sec, s_msec)
struct rusage Zeit1; 
long *u_sec;
long *u_msec;
long *s_sec;
long* s_msec;
#endif
/* stops time measuring */
/* Zeit1 is the starting-time and has to be the parameter x of InitTime */
/* returns the elapsed time in user- (u_xxx) and system-mode (s_xxx)    */
/* in seconds (x_sec) and microseconds (x_msec)                         */
{
  struct rusage Zeit2;

  getrusage(RUSAGE_SELF, &Zeit2);

  if (Zeit2.ru_utime.tv_usec < Zeit1.ru_utime.tv_usec) {
    *u_msec = Zeit1.ru_utime.tv_usec - Zeit2.ru_utime.tv_usec;
    *u_sec  = Zeit2.ru_utime.tv_sec  - Zeit1.ru_utime.tv_sec - 1;
  } else {
    *u_msec = Zeit2.ru_utime.tv_usec - Zeit1.ru_utime.tv_usec;
    *u_sec  = Zeit2.ru_utime.tv_sec  - Zeit1.ru_utime.tv_sec;
  }
  if (Zeit2.ru_stime.tv_usec < Zeit1.ru_stime.tv_usec) {
    *s_msec = Zeit1.ru_stime.tv_usec - Zeit2.ru_stime.tv_usec;
    *s_sec  = Zeit2.ru_stime.tv_sec  - Zeit1.ru_stime.tv_sec - 1;
  } else {
    *s_msec = Zeit2.ru_stime.tv_usec - Zeit1.ru_stime.tv_usec;
    *s_sec  = Zeit2.ru_stime.tv_sec  - Zeit1.ru_stime.tv_sec;
  }
}


void EndParallaxisError
#ifdef __STDC__
(int ExitCode)
#else
(ExitCode) int ExitCode;
#endif
{
  (void) pvm_initsend(PvmDataDefault);
  PVM_PKSIG(TermErrorSig);
  pvm_pkint(&ExitCode, 1, 1);
  (void) pvm_send(parent_tid, IO_TAG);

  pvm_exit();
  exit(ExitCode);
}


void EndParallaxis
#ifdef __STDC__
(int ExitCode, struct rusage Zeit1, long u_sec, long u_msec, 
 long s_sec, long s_msec)
#else
(ExitCode, Zeit1, u_sec, u_msec, s_sec, s_msec) 
int ExitCode;
struct rusage Zeit1;
long u_sec;
long u_msec;
long s_sec;
long s_msec;
#endif
{
  if (me == 0) {
    CARDINAL i;
    long time_val;

    (void) pvm_initsend(PvmDataDefault);
    PVM_PKSIG(TermSig);
    (void) pvm_pkint(&ExitCode, 1, 1);
    (void) pvm_pklong(&u_sec, 1, 1);
    (void) pvm_pklong(&u_msec, 1, 1);
    (void) pvm_pklong(&s_sec, 1, 1);
    (void) pvm_pklong(&s_msec, 1, 1);

    /* receive from all other tasks time values */
    for (i = 1; i < n_tasks; i++) {
      int j;
      (void) pvm_recv(task_ids[i], TIME_TAG);
      for (j = 0; j < 4; j++) {
	(void) pvm_upklong(&time_val, 1, 1);
	(void) pvm_pklong(&time_val, 1, 1);
      }
    }

    (void) pvm_send(io_tid, IO_TAG);

  } else {
    /* send time values to task 0 */
    (void) pvm_initsend(PvmDataDefault);
    (void) pvm_pklong(&u_sec, 1, 1);
    (void) pvm_pklong(&u_msec, 1, 1);
    (void) pvm_pklong(&s_sec, 1, 1);
    (void) pvm_pklong(&s_msec, 1, 1);
    (void) pvm_send(parent_tid, TIME_TAG);
  }  

  pvm_exit();
  exit(ExitCode);
}


/***********************************************************************/
/*************************** Main program ******************************/
/***********************************************************************/

/* The main()-function is generated in the program module itself. Directly */
/* after the start of main() there's called the function INIT_MODULE() to */
/* initialize some variables, e.g. the argument pointers. Then the code of */
/* the parallaxis program is executed, and before the end of the main() */
/* function, the function EXIT_MODULE() is called to free some memory. */


/*********** intern Parallaxis variables *****************/
int my_tid;       /* own tid                             */
int parent_tid;   /* tid of parent-process               */
int io_tid;       /* tid of io-process                   */
CARDINAL n_tasks; /* number of real processors           */
int *task_ids;    /* array containing tids of all tasks  */
int me;           /* index of own tid in task_ids        */
CARDINAL PE;      /* counter over all local PEs          */
int message_tag;  /* tag used for data-exchange-messages */
INTEGER *SOURCE_ARRAY;      /* for computation of connections */
INTEGER *DESTINATION_ARRAY; /* dito                           */
CONFIGURATION CONFIG_1;     /* for REDUCE-functions           */
GENERIC_CONNECTIONS_PTR ALL_GENERIC_CONNECTIONS;
GENERIC_PARAMETER_CONNECTIONS_PTR ALL_GENERIC_PARAMETER_CONNECTIONS;
CARDINAL MaxDiff;
CARDINAL MinPEs;
long u_sec, u_msec, s_sec, s_msec;      /* for elapsed time */
struct rusage StartTime;

unsigned int SYSTEM_argc;
char **SYSTEM_argv;
char **SYSTEM_envp;

BOOLEAN Done;
char termCH;


void INIT_MODULE
#ifdef __STDC__
(int argc, char *argv[], char *envp[])
#else
(argc, argv, envp)
int argc;
char *argv[], *envp[];
#endif
{
  INIT_PVM(&my_tid, &parent_tid, &io_tid, &n_tasks, &task_ids, &me,
	   &MaxDiff, &MinPEs);               /* also sets SYSTEM_argc/argv */

  INIT_CONFIGURATION("CONFIG_1", &CONFIG_1, 1);
  SET_BOUNDS(&CONFIG_1, 0, 0, 0);
  INIT_STACKS(&CONFIG_1);

  ALL_GENERIC_CONNECTIONS = NULL;
  ALL_GENERIC_PARAMETER_CONNECTIONS = NULL;

  srand48(time(NULL)); /* init random number generator */

  InitTime(StartTime);
}


void EXIT_MODULE
#ifdef __STDC__
(void)
#else
()
#endif
{
  StopTime(StartTime, &u_sec, &u_msec, &s_sec, &s_msec);

  FREE_ALL_GENERIC_CONNECTIONS();
  FREE_ALL_GENERIC_PARAMETER_CONNECTIONS();

  FREE_CONFIGURATION(&CONFIG_1, TRUE);

  EndParallaxis(0, StartTime, u_sec, u_msec, s_sec, s_msec);
}
