/*************************************************************************
 *
 * Parallaxis Compiler Projects
 * ----------------------------
 *
 * Name  :  Source.c
 * Author:  lampke, keller
 * Date  :  Thu Sep  8 22:53:17 1994
 *
 * Description:
 * ------------
 * Source code is read in blocks when ever scanner
 *          need a new junk.
 *          Actual filename is stored in Scanner.Attribute.Position
 *
 ************************************************************************/

#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include "System.h"			  /* Reuse-Library */
#include "Idents.h"			  /* Reuse-Library */
#include "Source.h"
#include "FileIO.h"
#include "Scanner.h"
#include "parallaxis.h"

static BOOLEAN isInitialised = FALSE;


/*******************************************************************
 *  Open the file with the given name for input, remember the 
 *  name and return a filehandle.
 *******************************************************************/

int BeginSource( char *FileName )
{
  Attribute.Position.File = MakeIdent( FileName, strlen( FileName ) );
  return FileIO_OpenInputHandle( FileName );
}


/*******************************************************************
 *  Read a block of Size from File in the Buffer. If the last 
 *  character read was a newline, add an additional space. 
 *  Return the total number of read characters.
 *******************************************************************/

int GetLine( int File, char * Buffer, int Size )
{
  register int n = read( File, Buffer, Size );

#if 1	   /* was "#ifdef Dialog". Is it really needed? */
# define IgnoreChar ' '
/* Add dummy after newline character in order to supply a lookahead for rex. */
/* This way newline tokens are recognized without typing an extra line.      */
  if ((n > 0) && (Buffer [n - 1] == '\n'))
    Buffer [n ++] = IgnoreChar;
#endif

  return n;
}

/*******************************************************************
 *  Close the file.
 *******************************************************************/

void CloseSource( int File )
{
  FileIO_CloseHandle (File);
}


/*******************************************************************
 *  Initialize this module by initializing module FileIO. The 
 *  "current" file is set to stdin, so remember this name.
 *******************************************************************/

void Source_Init( void )
{
  if ( !isInitialised ) {
    isInitialised = TRUE;
    Attribute.Position.File = MakeIdent( "stdin", 5 );
    FileIO_Init();
  } /* endif */
}

