/*************************************************************************
 *
 * Parallaxis Compiler Projects
 * ----------------------------
 *
 * name  :  p3c.c
 * desc  :  main program for Parallaxis-3-Compiler
 * author:  lampke, keller
 *
 * Return-Values:
 *   -3: compilation aborted with semantic errors
 *   -2: compilation aborted with syntax errors
 *   -1: general error (missing files, invalid usage, ...)
 *    0: successfull compilation (probably warnings)
 *    1: only parser was run (using option -Z1)
 *    2: only parser and semantic analysis was run (using option -Z2)
 *
 ************************************************************************/
#define  DATE "Date: Mon Nov 20 17:49:12 1995"
#define  VERSION "$Revision: 1.16 $"

#include "parallaxis.h"
#include "Code.h"
#include "Eval.h"
#include "Parser.h"
#include "Scanner.h"
#include "Tree.h"
#include "Symbols.h"
#include "GetDefMods.h"
#include "Options.h"
#include "Source.h"
#include "Statist.h"
#include "language.h"
/*
 * local variables
 */
static int ReturnValue;

/*
 * global function implementation
 */
int main( int argc, char *argv[] )
{
  char *sourceFile;
  BOOLEAN t, m, inputfiles;

  Source_Init();
  GetDefMods_Init();
  Options_Init();
  Statist_Init();

  BeginTree();
  BeginSymbols();
  inputfiles = Options_CheckArguments( argc, argv );
  error_messages_language=english;
  if ( Options_IsOptionSet( Options_edeutsch))error_messages_language=deutsch;
  if ( Options_IsOptionSet( Options_efrancais))error_messages_language=francais;
  if ( Options_IsOptionSet( Options_eenglish))error_messages_language=english;
  if ( Options_IsOptionSet( Options_eVerbose ))
  {
    fprintf(stderr, "\nParallaxis-III Compiler Beta Release, %s\n", VERSION);
    fprintf(stderr, "University of Stuttgart, %s\n\n", DATE);
  }

  if (!inputfiles) exit(-1);		  /* No input files specified */

  t = Options_IsOptionSet( Options_eTimeStats );
  m = Options_IsOptionSet( Options_eMemStats );
  if ( (sourceFile = Options_GetSourceFile()) != NULL ) {
    BeginFile( sourceFile );
  } /* endif */
  if ( m ) { Statist_WriteMemory(); }
  if ( t || m ) { Statist_WritePhase( "Parser" ); }
  (void) Parser();
  if ( t ) { Statist_WriteStepTime(); }
  if ( m ) { Statist_WriteMemory(); }
  if ( t || m ) { Statist_WritePhase( "Get Definition Modules" ); }
  TreeRoot = GetDefMods_GetDefModules( TreeRoot );
  CloseParser();
  if ( t ) { Statist_WriteStepTime(); }
  if ( m ) { Statist_WriteMemory(); }

  if ( ErrorGetCount() < 1 ) {

#ifdef DEVELOP
if (Options_IsOptionSet(Options_eCompileOnly)) {
  fprintf(stderr, "Stopped after compiling !\n");
  ReturnValue = 1;
} else {
#endif

    if ( t || m ) { Statist_WritePhase( "Semantic Analysis" ); }
    BeginEval();
    Eval( TreeRoot );
    CloseEval();
    if ( t ) { Statist_WriteStepTime(); }
    if ( m ) { Statist_WriteMemory(); }

    if ( ErrorGetCount() < 1 ) {

#ifdef DEVELOP
if (Options_IsOptionSet(Options_eEvalOnly)) {
  fprintf(stderr, "Stopped after semantic check !\n");
  ReturnValue = 2;
} else {
#endif

      if ( t || m ) { Statist_WritePhase( "Coding" ); }
      BeginCode();
      CodeCompUnits( TreeRoot );
      CloseCode();
      if ( t ) { Statist_WriteStepTime(); }
      if ( m ) { Statist_WriteMemory();}
      if ( ErrorGetCount() < 1 ) {
        if ( t ) { Statist_WriteCompTime(); }
	if ( Options_IsOptionSet( Options_eVerbose ))
          fprintf( stderr, "\nTranslation successful completed !\n" );
        ReturnValue = 0;
      } else {
        WriteMessages( stderr , Options_IsOptionSet( Options_eWarnings ) );
        fprintf( stderr, "\nTranslation completed with %d warning%s !\n",
          ErrorGetCount(), ErrorGetCount() == 1 ? "" : "s" );
        ReturnValue = 0;
      } /* endif */

#ifdef DEVELOP
} /* endif EvalOnly*/
#endif

    } else {
      WriteMessages( stderr , Options_IsOptionSet( Options_eWarnings ) );
      switch (error_messages_language) {
      case deutsch:
        fprintf( stderr, "\n%d semantische%s Fehler trat%s auf!\n",
                 ErrorGetCount(),ErrorGetCount() == 1 ? "r" : "",  
                                 ErrorGetCount() == 1 ? "" : "en");
        break;
      case english:
      default:
        fprintf( stderr, "\n%d semantic error%s occured !\n", ErrorGetCount(),
                 ErrorGetCount() == 1 ? "" : "s" );
        break;
      }
      ReturnValue = -3;
    } /* endif ErrorGetCount*/

#ifdef DEVELOP
} /* endif CompileOnly*/
#endif

  } else {
    WriteMessages( stderr , Options_IsOptionSet( Options_eWarnings ) );
    switch (error_messages_language) {
    case deutsch:
      fprintf( stderr, "\n%d Syntaxfehler trat%s auf ! Keine semantische Analyse !\n",
	      ErrorGetCount(), ErrorGetCount() == 1 ? "" : "en" );
      break;
    case english:
    default:
      fprintf( stderr, "\n%d syntax error%s occured ! No semantic analysis !\n",
	      ErrorGetCount(), ErrorGetCount() == 1 ? "" : "s" );
      break;
    }
    ReturnValue = -2;
  }


#ifdef DEVELOP
  if (Options_IsOptionSet(Options_eWriteTree)) WriteTree(stdout, TreeRoot);
  if (Options_IsOptionSet(Options_eCheckTree)) CheckTree(TreeRoot);
  if (Options_IsOptionSet(Options_eQueryTree)) QueryTree(TreeRoot);
#endif

  CloseTree();
  CloseSymbols();
  return ReturnValue;
}

