/*************************************************************************
 *
 * Parallaxis Compiler Projects
 * ----------------------------
 *
 * name  :  LoopStack
 * desc  :  Stack for the administration of the environments generated for
 *	    the translation of LOOP and EXIT statements.
 * author:  weiss
 *
 ************************************************************************/

#ifndef _LoopStack_H
#define _LoopStack_H

#include "Idents.h"                       /* Reuse-Library */

typedef struct E_L {
    tIdent config;
    struct E_L *next;
    struct E_L *prev;
} tConfig_List;

typedef struct {
    tIdent parloop_config;
    tIdent label;
    ULONG config_num;
    tConfig_List *first;
    tConfig_List *last;
} tLoopStack;

/*************************************************************************
 * name: LoopStack_Push
 * in  : Ident, LoopStack with this label ident is put on stack
 * out :
 ************************************************************************/

extern void LoopStack_Push(tIdent Ident, tIdent parloop_config);

/*************************************************************************
 * name: LoopStack_OpenEnv
 * in  : Ident, remember opening of environment with this configname
 * out :
 ************************************************************************/

extern void LoopStack_OpenEnv(tIdent Ident);

/*************************************************************************
 * name: LoopStack_CloseEnv
 * in  :
 * out :
 ************************************************************************/

extern void LoopStack_CloseEnv();

/*************************************************************************
 * name: LoopStack_GetTopLabel
 * in  :
 * out :
 ************************************************************************/

extern tIdent LoopStack_GetTopLabel();

/*************************************************************************
 * name: LoopStack_GetTopConfigNum
 * in  :
 * out :
 ************************************************************************/

extern ULONG LoopStack_GetTopConfigNum();

/*************************************************************************
 * name: LoopStack_GetTopFirstConfig
 * in  :
 * out :
 ************************************************************************/

extern tConfig_List *LoopStack_GetTopFirstConfig();

/*************************************************************************
 * name: LoopStack_GetTopActConfig
 * in  : list, the config name of list is returned and set then set
 *	 to next list element
 * out :
 ************************************************************************/

extern tIdent LoopStack_GetTopActConfig(tConfig_List **list);

/*************************************************************************
 * name: LoopStack_IsTopLastConfig
 * in  : list, checks if the current config name exists again in the
 *	 further list, if not list is set to next list element
 * out :
 ************************************************************************/

extern BOOLEAN LoopStack_IsTopLastConfig(tConfig_List **list);

/*************************************************************************
 * name: LoopStack_GetTopLoopConfig
 * in  :
 * out :
 ************************************************************************/

extern tIdent LoopStack_GetTopLoopConfig();

/*************************************************************************
 * name: LoopStack_GetTopLastConfig
 * in  : list, searches backwards in the list for the
 *	 config name that matches the name in LOOP
 * out : list is set to the found element
 ************************************************************************/

extern ULONG LoopStack_GetTopLastConfig(tConfig_List **list);

/*************************************************************************
 * name: LoopStack_GetTopNextLoopConfig
 * in  :
 * out :
 ************************************************************************/

extern ULONG LoopStack_GetTopNextLoopConfig(tConfig_List **list, ULONG num);

/*************************************************************************
 * name: LoopStack_Pop
 * in  :
 * out :
 ************************************************************************/

extern void LoopStack_Pop();

/*************************************************************************
 * name: LoopStack_Top
 * in  :
 * out :
 ************************************************************************/

extern tLoopStack *LoopStack_Top();

/*************************************************************************
 * name: LoopStack_Free
 * in  :
 * out :
 ************************************************************************/

extern void LoopStack_Free();

/*************************************************************************
 * name: LoopStack_Init
 * in  :
 * out :
 ************************************************************************/

extern void LoopStack_Init();

#endif
