/*************************************************************************
 *
 * Parallaxis Compiler Projects
 * ----------------------------
 *
 * name  :  AssocTab.c
 * desc  :  associate ideentifiers with objects of arbitrary type
 * author:  lampke
 *
 ************************************************************************/

#include "DynArray.h"			  /* Reuse-Library */
#include "Idents.h"			  /* Reuse-Library */
#include "AssocTab.h"

static long ActualSize;
static unsigned int ActMaxAssoc;
static void **AssocTab;

/*
 * global function implementation
 */
/*------- Begin ------------------------------------------------------------*/

void AssocTab_Begin( void )
{
  unsigned int Index;
  long ActMaxIdent;

  ActualSize = 1;
  ActMaxIdent = MaxIdent();
  while (ActualSize <= ActMaxIdent ) {
    ActualSize = ActualSize * 2;
  }
  MakeArray( (char **) &AssocTab, &ActualSize, sizeof( void * ) );
  ActMaxAssoc = ActualSize - 1;
  for ( Index = 0; Index <= ActMaxAssoc; Index ++ ) {
    AssocTab[Index] = NULL;
  }
}

/*------- PutAssoc ---------------------------------------------------------*/

void AssocTab_PutAssoc( tIdent Ident, void *Object )
{
  unsigned long Index, OldMaxAssoc;
  long ActMaxIdent;

  if ( Ident > ActMaxAssoc ) {
    ActMaxIdent = MaxIdent();
    while ( ActualSize <= ActMaxIdent ) {
      ExtendArray( (char **) &AssocTab, & ActualSize, sizeof( void * ) );
    }
    OldMaxAssoc = ActMaxAssoc;
    ActMaxAssoc = ActualSize - 1;
    for ( Index = OldMaxAssoc + 1; Index <= ActMaxAssoc; Index++ ) {
      AssocTab[Index] = NULL;
    }
  }
  AssocTab[Ident] = Object;
} 

/*------- GetAssoc ---------------------------------------------------------*/

void AssocTab_GetAssoc( tIdent Ident, void **Object )
{
  if ( Ident >= ActMaxAssoc ) {
    *Object = NULL;
  } else {
    *Object = AssocTab[Ident];
  } /* endif */
}

/*------- Close ------------------------------------------------------------*/

void AssocTab_Close( void )
{
  ReleaseArray( (char **) &AssocTab, &ActualSize, sizeof( void * ) );
}
