/*************************************************************************
 *
 * Parallaxis Compiler Projects
 * ----------------------------
 *
 * name  :  Errors.h
 * desc  :  handels all occuring errors by collecting them
 *          and giving a sorted output
 * author:  lampke
 *
 ************************************************************************/
#ifndef Errors_h
#define Errors_h

#include <stdio.h>
#include "parallaxis.h"
#include "Positions.h"

#define xxNoText               0
#define xxSyntaxError          1       /* error codes          */
#define xxExpectedTokens       2
#define xxRestartPoint         3
#define xxTokenInserted        4
#define xxTokensIgnored        5
#define xxModNotFound          6
#define xxCyclicDefMods        7
#define xxFileNameMismatch     8
#define xxTooManyErrors        9
#define xxSemanticError       10
#define xxAssignIncomp        11
#define xxOperandIncomp       12
#define xxConfigIncomp        13
#define xxResultTypeIncomp    14
#define xxTypeExpected        15
#define xxTypeFound           16
#define xxTypeLeft            17
#define xxTypeRight           18
#define xxIllegalOperator     19
#define xxNotDeclared         20
#define xxNotInRecord         21
#define xxExprNotConst        22
#define xxTypeOverFlow        23
#define xxSetOverFlow         24
#define xxValueRange          25
#define xxDivisionByZero      26
#define xxIllegalArgCount     27
#define xxArgumentTypeIncomp  28
#define xxVectorExpected      29
#define xxFunction            30
#define xxArgNumber           31
#define xxArgsExpected        32
#define xxArgsFound           33
#define xxNoReduceFunction    34
#define xxIllegalProcExpr     35
#define xxStructTypeReq       36
#define xxOutOfLongRange      37
#define xxFunctionAsProcedure 38
#define xxNoOpenVector        39
#define xxWrongConfigParams   40
#define xxConnectionExpected  41
#define xxIllegalReduceNumber 42
#define xxTooManyReduces      43
#define xxVectorIndex         44
#define xxScalarExpected      45
#define xxArrayExpected       46
#define xxConfigExpected      47
#define xxReturnValueExpected 48
#define xxReturnValueWrong    49
#define xxConnectIndex        50
#define xxGenericConfig       51
#define xxRankMismatch        52
#define xxIdentRedeclared     53
#define xxNoGenericConfig     54
#define xxWrongGenericParams  55
#define xxVectorInInit        56
#define xxNoTypeCast          57
#define xxGenericConnectMixed 58
#define xxInvalidFilename     59
#define xxConfigCycle         60
#define xxInvalidConfigLen    61
#define xxNoTypeName          62
#define xxVariableUsedAsFunc  63
#define xxVarParNotLval       64
#define xxLvalExpected        65
#define xxBooleanExpected     66

#define xxFatal                1       /* error classes        */
#define xxRestriction          2
#define xxError                3
#define xxWarning              4
#define xxRepair               5
#define xxNote                 6
#define xxInformation          7

#define xxNone                 0
#define xxInteger              1       /* info classes         */
#define xxShort                2
#define xxLong                 3
#define xxReal                 4
#define xxBoolean              5
#define xxCharacter            6
#define xxString               7
#define xxSet                  8
#define xxBITSET               9
#define xxIdent               10

extern void (*Errors_Exit)( void );
                        /* Refers to a procedure that specifies         */
                        /* what to do if 'ErrorClass' = Fatal.          */
                        /* Default: terminate program execution.        */

extern void StoreMessages( BOOLEAN Store );
                        /* Messages are stored if 'Store' = TRUE        */
                        /* for printing with the routine 'WriteMessages'*/
                        /* otherwise they are printed immediately.      */
                        /* If 'Store'=TRUE the message store is cleared.*/

extern void ErrorMessage( int ErrorCode, int ErrorClass, tPosition Position );
                        /* Report a message represented by an integer   */
                        /* 'ErrorCode' and classified by 'ErrorClass'.  */

extern void ErrorMessageI( int ErrorCode, int ErrorClass, tPosition Position,
                           int InfoClass, void *Info );
                        /* Like the previous routine with additional    */
                        /* information of type 'InfoClass' at the       */
                        /* address 'Info'.                              */

extern void Message( char * ErrorText, int ErrorClass, tPosition Position );
                        /* Report a message represented by a string     */
                        /* 'ErrorText' and classified by 'ErrorClass'.  */

extern void MessageI( char * ErrorText, int ErrorClass, tPosition Position,
                      int InfoClass, void *Info );
                        /* Like the previous routine with additional    */
                        /* information of type 'InfoClass' at the       */
                        /* address 'Info'.                              */

extern void WriteMessages( FILE * File , BOOLEAN NoWarnings );
                        /* The stored messages are sorted by their      */
                        /* source position and printed on 'File'.       */

extern USHORT ErrorGetCount( void );
                        /* Get number of error registered               */

extern void Errors_Init( void );

#endif
