/*************************************************************************
 *
 * Parallaxis Compiler Projects
 * ----------------------------
 *
 * name  :  Values.h
 * desc  :  allow for computing and manageing of constants
 * author:  lampke
 *
 ************************************************************************/

#ifndef Values_h
#define Values_h

#include "StringMem.h"			  /* Reuse-Library */
#include "parallaxis.h"
#include "Positions.h"
#include "BigSet.h"
/*
#include "Tree.h"
#include "Symbols.h"
 */

typedef enum {
   eNoValue, eCardinal, eInteger, eReal, eBoolean,
   eChar, eStringChar, eString, eBitset, eEnum,
   eNilType, eStdProc, eType
} tValueKind;

typedef struct {
   int         code;
   tPosition   pos;
   union {
   tIdent      ident;
   ADDRESS     type;
   ULONG       integer;
   }           info;
} tErrorDesc;

typedef struct {
   tValueKind  Kind;
   union {
      tErrorDesc  error;
      ULONG       cardinalValue;
      LONG        integerValue;
      REAL        realValue;
      BOOLEAN     booleanValue;
      UCHAR       charValue;
      tStringRef  stringValue;
      BITSET      bitsetValue;
      void       *enumValue;              /* tObject  */
      ULONG       stdProcValue;           /* tStdProc */
      void       *typeValue;              /* tType    */
   } value;
} tValue;

extern tValue  Values_Error,
               Values_Nil,
               Values_True,
               Values_False,
               Values_MinChar,
               Values_MaxChar,
               Values_Zero,
               Values_MaxCardinal,
               Values_MinInteger,
               Values_MaxInteger,
               Values_MinReal,
               Values_MaxReal,
               Values_Pi,
               Values_EOL;

/*************************************************************************
 * name :  ComputeConstant
 * in   :  tree of expression
 *         environment of expression
 *         pointer to value, which will be computed
 * out  :  TRUE, if expression is constant and value 'value'
 *         FALSE, if expression is not constant
 ************************************************************************/
extern BOOLEAN Values_ComputeConstant( ADDRESS /* tTree */ tree,
                                       ADDRESS /* tEnv */ env,
                                       tValue *value );

/*************************************************************************
 * name :  TypeOfValue
 * in   :  pointer to value
 * out  :  type representation
 ************************************************************************/
extern ADDRESS /* tType */ Values_TypeOfValue( tValue *value );

/*************************************************************************
 * name :  ValueToInt
 * in   :  pointer to value
 * out  :  integer aquivalent to value
 ************************************************************************/
extern LONG Values_ValueToInt( tValue *value );

/*************************************************************************
 * name :  ValueAsString
 * in   :  pointer to value, buffer
 * out  :  buffer
 ************************************************************************/
extern char *Values_ValueAsString( tValue *value, char *buffer );

/*************************************************************************
 * name :  WriteValue
 * in   :  pointer to value, which is to be written on file 'f'
 * out  :
 ************************************************************************/
extern void Values_WriteValue( FILE *f, tValue *value );

extern void Values_Init( void );

#endif
