.TH P3 1 "15 August 1994"
.SH NAME
p3 - Parallaxis-III User Interface
.SH SYNOPSIS

.B p3
{ 
.I options
} 
[
.I file
]
\&.\|.\|.


.SH DESCRIPTION
.B p3
is the user interface of the Parallaxis-III compiler system.
The purpose of 
.B p3 
is to provide a useful and familiar interface between the Parallaxis
compiler p3c(1) itself and the user. The usage of 
.B p3
is very similar to the usage of other known compilers such as cc(1).

The Parallaxis compiler p3c converts Parallaxis-III programs to an
intermediate language, namely C and/or MPL (Maspar Parallel Application 
Language). To get an executable program, there also has to be called a
backend compiler. p3 does this process automatically, so
usually you just have to invoke
.B p3 
without knowing which compiler in fact is called. Nevertheless you can
control the behaviour of 
.B p3
by giving additional options. Every argument beginning with `-' is
treated as an option, probably with additional parameters. Every other
argument is treated as a filename.

First the given filenames are checked for their extension. If a
filename has no extension,
.B p3
tries to append `.pd' and  `.pm'
to it and checks, if files with the resulting filenames exist. If yes,
these files are added to the list of filenames.

Then every file with a name ending in `.pd' or `.pm' is compiled
with p3c, where every `.pd'-file is converted to a header file with
extension `.h' and every `.pm'-file is converted to a C source file
with extension `.c' and/or to an MPL source file with extension
`.m'.  These resulting files and the remaining unknown files are
finally passed to the backend compiler.

.SH OPTIONS
.TP
.B -c
Don't call the linker. This option is passed to the backend compiler. No
library paths or libraries are put on the command line of the backend
compiler. Compare with option "-p".
.TP
.B -C
Use p3c as Parallaxis compiler. This should generate C-code for
simulation on a workstation.
.TP
.B -casts
Let p3c generate additional casts to make the resulting C-programs lint
compliant.
.TP
.BI -cc\  name
Use backend compiler
.I name
instead of the default one. This can also be achieved by setting
the environment variable P3CC.
.TP
.B -g
Let p3c generate debugging information; that's additional directives
in the code and a crossreference file with filename extension
`.px'. This file is used by the Parallaxis debugger.

This option is also passed as "-g" to the backend compiler.
.TP
.BR -h ,\  -H ,\  -help
Print usage of
.B p3.
.TP
.B -headers
Generate a `.h'-header-file for each imported definition module. Usually
the `.pd'-files are read and included in the compilation process, but no
`.h'-file is generated unless you compile a `.pd'-file for itself.
.TP
.BI -I\  path
Add 
.I path
to the list of import paths. This path list is used
when searching for `.pd'-files. Usually the Parallaxis standard include
paths are automatically appended to this list. If you don't want this,
use also option -nop3inc or -nodefaults.

The resulting path list is also passed to the backend
compiler for finding the header files. If you don't like this
behaviour, use option -P instead of -I.

.TP
.BI -indent\  i
Set code indentation of p3c to
.I i
blanks.
.TP
.BI -k\  option
Pass
.I option
directly to the backend compiler. In this way you can pass options to
the backend compiler which would normally be misinterpreted by 
.BR p3 .
Example:
.RS 8
p3 -cc gcc -k -I-
.I file
.RE
.RS
If you append the option to pass directly after -k, you must omit the
`-' from it, e.g.
.RE
.RS 8
p3 -cc gcc -kI-
.I file
.RE
.TP
.BI -L\  path
Add 
.I path
to the list of library paths. This list is only used by the backend
compiler. Usually the Parallaxis standard library paths are
automatically appended to this list. If you don't want this behaviour,
use option -nop3lib or -nodefaults in addition.
.TP
.BR -m ,\  -mem
Let p3c print statistics about its memory usage during compilation.
.TP
.BR -mpl ,\  -MPL
Use p3c_MPL as Parallaxis compiler. This should generate C- and/or
MPL-code for the MasPar computer.
.TP
.BR -n ,\  -nocompile
Don't call compilers, just print the commandlines that would be used
to do so.
.TP
.B -nop3inc
Don't append Parallaxis standard import paths to the path list
built by options -I and -P. Usually these paths are automatically
appended.
.TP
.B -nop3lib
Don't append Parallaxis standard library paths to the path list built
by option -L. Usually these paths are automatically appended.
.TP
.B -nodefaults
Synonym for `-nop3inc -nop3lib'.
.TP
.BI -o\  name
Create executable with filename
.I name
instead of `a.out'. This option is simply passed to the backend
compiler.
.TP
.B -p
Don't call backend compiler, do only Parallaxis compilation. Compare
with option "-c".
.TP
.BI -P\  path
Add
.I path
to the list of import paths. This path list is used
when searching for .pd files. Usually the Parallaxis standard include
paths are automatically appended to this list. If you don't want this
behaviour, use also option -nop3inc or -nodefaults.

In difference to option -I, the resulting path list is *not* passed to
the backend compiler.
.TP
.BR -pvm ,\  -PVM
Use p3c_PVM as Parallaxis compiler. This should generate C-code for
simulation on several workstations or the Intel Paragon using PVM.
.TP
.BR -r ,\  -rchecks
Don't let p3c generate code for runtime range checks. By default p3c
generates code to check for ranges at assignments to subrange
type variables, array index references, processor element references,
etc. 
.TP
.BR -t ,\  -time
Let p3c print statistics about the time consumption during the
Parallaxis compilation process.
.TP
.BR -tt ,\  -total
Print statistics about the time consumption during the whole
compilation process, including backend compilation.
.TP
.B -v
Print version of
.B p3
and the commandlines of the resulting compiler calls.
.TP
.B -vv
Like -v, but also passes "-v" to the backend compiler.
.TP
.B -w
Don't generate warnings during Parallaxis compilation process. The
backend compiler may still give warnings.
.TP
.B -ww
Like -w, but also passes "-w" to the backend compiler.

.SH OPTIONS (available if MPL output possible)
.TP
.BR -s ,\  -small
Generate a small MPL-only model. The executable must be smaller than
128KB. Usually there's created an MPL- and a C-part.

.RE
Every other option is passed to the backend compiler. 

Parameters usually can be appended directly to the option.
.RS 3
p3 -Ipath
.RE
is the same as
.RS 3
p3 -I path
.RE

As p3 can not decide if an unknown option has parameters or not, it
assumes no parameters. So you probably have to quote such options
together with their parameters. E.g.
.RS 3
p3 "-xyz parameters"
.RE

.SH FILES
.TP 20
*.pd
Definition modules. These files are converted to `.h'-header-files,
but not propagated to the backend compiler.
.TP 20
*.pm
Implementation or program modules. These files are converted to
`.c'-files (and/or `.m'-files respectively). They are passed to the
backend compiler for further compilation.
.TP 20
+++LIB+++/+++SYSLIB+++
Library holding the Parallaxis related stuff for C-version.
.TP 20
+++LIB+++/+++SYSLIBMPL+++
Library holding the Parallaxis related stuff for MPL-version.
.TP 20
+++LIB+++/+++SYSLIBPVM+++
Library holding the Parallaxis related stuff for PVM-version.
.TP 20
+++INC+++/*.h
Header files for precompiled Parallaxis modules.

.SH ENVIRONMENT

Usually there are meaningful defaults compiled into
.B p3
when installed at a site. But you can override some of them by setting
some environment variables.

.TP 10
P3OPT
Default options to be set on every invocation of 
.B p3.
These options are prepended to the options given on the commandline,
thus giving the possibility to override default options of P3OPT on
the commandline. Example: P3OPT set to "-PVM" and p3 called with
option "-C". This results in an implicit call of p3 with options "-PVM
-C" where "-C" overrides "-PVM". The compilation will take place for
C.

.TP 10
P3INC
Parallaxis standard import paths. A list of paths where
to find `.pd'-files. These are automatically appended to the import
path list built by options -P and -I unless one of -nop3inc or
-nodefaults is given.

.RS
.PD 0
Defaults:
.LP 
C-version:   +++P3INC+++
.LP
MPL-version: +++P3INCMPL+++
.LP
PVM-version: +++P3INCPVM+++
.RE

.TP 10
P3LIB
Parallaxis standard import paths. A list of paths where
to find libraries to include. These are automatically appended to the
library path list built by option -L unless one of -nop3lib or
-nodefaults is given. 

.RS
.PD 0
Defaults:
.LP 
C-version:   +++P3LIB+++
.LP
MPL-version: +++P3LIBMPL+++
.LP
PVM-version: +++P3LIBPVM+++
.RE

.TP 10
P3CC
Parallaxis standard backend compiler. This compiler is used if no
option -cc is given. 

.RS
.PD 0
Defaults:
.LP 
C-version:   +++P3CC+++
.LP
MPL-version: +++P3CCMPL+++
.LP
PVM-version: +++P3CCPVM+++
.RE


When setting P3INC or P3LIB you can also give lists of paths by
seperating the paths with commas, blanks, semicolons or colons. In
addition, you may include other environment variables in your
pathlists. Note: the environment variables replace the default paths.
If you like to have the default paths in addition, you have to give a
pseudo path "@" somewhere in your environment variable's pathlist. The
"@" will be replaced by the default paths compiled into 
.B p3.
Example: 

.RS 3
P3LIB = $HOME/lib,@,$PVM_ROOT/lib/$PVM_ARCH
.RE

.SH EXAMPLES

Suppose you have a module system with modules a and b, each consisting
of a definition and implementation module (a.pd, a.pm, b.pd, b.pm) and a
program module p (p.pm). p imports from a and b, a imports from b and
b imports from a globally supplied module g which resides in a library
(libg.a) found in the Parallaxis standard library paths.

Then you can compile the whole system by typing

.RS 3
p3 a b p -lg
.RE

which creates an executable a.out. This is equivalent to typing

.RS 3
p3 a.pd a.pm b.pd b.pm p.pm -lg
.RE

Now you change something in a.pd and in b.pm. You know you have
to change also something in p.pm but you like to compile a and b first
to see if there are any errors. But the system is not consistent at
the moment, so no executable should be created. Do this by 

.RS 3
p3 a.pd b.pm -c
.RE

Here "-c" is an option passed to the backend C-compiler and prevents
it from linking. Though you don't need to mention the other modules.
Let's assume everything went alright and you made the changes to p.pm.
Now you want again an executable and call it `p'. As a and b may be large
modules you don't want to compile them again, so you just type

.RS 3
p3 a.o b.o p -o p -lg
.RE

Now you've discovered that the supplied library g is not exactly what
you need but builds a useful basis to start an own module g. As your
site provides the source code g.pd and g.pm to module g, you copy it
to a new directory $HOME/newg and start your own development. After lots
of work you have finished and want to compile the whole system again,
now using your own module g.

.RS 3
p3 -I $HOME/newg a.o b.o p.o $HOME/newg/g -o p
.RE

note that you no longer need -lg as you don't use the global library
anymore. Instead you compile your own library. But you need to give
the -I for the C-compiler to find the appropriate `.h'-file generated
in the same p3 call.

Now you've transferred your module g to a library libg.a in the known
directory $HOME/newg. (This is usually done by "ar cr libg.a g.o; ranlib
libg.a".) Now you want to compile your system all over again, from
start to end, using this library:

.RS 3
p3 -I $HOME/newg a b p -o p -L $HOME/newg -lg
.RE

This does not use the global library because the standard paths are
checked at the end, so your own library is found first. To be sure you
could also add a "-nop3lib" to the p3-call above, but this may cause
other problems because the Parallaxis standard library libp3 wouldn't
be found anymore.

Now as you've finished your own library you always like to include it
instead of the global one. So you set the environment variable P3LIB
and P3INC to the newg-path. Don't forget to include at least that
standard path where the Parallaxis system library resides or else the
linker won't find it. Then the compiler call from above reduces to

.RS 3
p3 a b p -o p -lg
.RE

which is exactly the same as when using the global library.

For bigger projects the use of make and makefiles is recommended.


.SH SEE ALSO
.BR p3c (1).

.SH BUGS

p3 does not delete old `.px'-Files when compiling. Thus an
old `.px'-File may exist even when compiling without option -g.

p3 is still under development. Especially the MPL part is not
implemented satisfactory, as there may be two backend compilers
involved. Though the part about passing arguments to the backend
compiler (option -k) as well as the naming of backend compilers
(option -cc, environment variable P3CC) may be
subject to modification.


