/*************************************************************************
 *
 * Parallaxis Compiler Projects
 * ----------------------------
 *
 * Name  :  Parameter
 * Author:  lampke, keller
 * Date  :  Fri Sep  2 21:54:16 1994
 *
 * Description:
 * ------------
 * Generates Parameter identifiers.

 ************************************************************************/

#include "Idents.h"			  /* Reuse-Library */
#include "parallaxis.h"
#include "GenIdents.h"
#include "Symbols.h"
#include "Parameter.h"

static BOOLEAN isInitialised = FALSE;

/*
 * global variables
 */

tCObjects Parameter_ValueOpens = NoCObjects;
tCObjects Parameter_VAROpens   = NoCObjects;

/*---------------------------------------------------------------------------*/

/*
 * global function implementation
 */

/**************************************************************************
 *  Generate the identifiers (names) for the additional parameters 
 *  containing the lengths of open arrays in a procedure parameter list.
 **************************************************************************/

void Parameter_GenParams( tCObjects OpenParams )
{
  while ( OpenParams != NoCObjects ) {
    OpenParams->SymCObjects.CObject = GenIdents_Param();
    OpenParams = OpenParams->SymCObjects.Next;
  } /* endwhile */
}


/**************************************************************************
 *  If the argument is an additional parameter of an open array declared
 *  VAR, search for it in the VAROpens-list, else (additional parameter of
 *  an open array not declared VAR) in the ValueOpens.
 **************************************************************************/

tIdent Parameter_GetParam( tObject M2Object )
{
  if (Symbols_IsVAR( M2Object ) ) {
    return( Symbols_LookUp( M2Object, Parameter_VAROpens ));
  } else {
    return( Symbols_LookUp( M2Object, Parameter_ValueOpens ));
 } /* endif */
}


/*---------------------------------------------------------------------------*/

/****************************/
/*  Initialize this module  */
/****************************/

void Parameter_Init( void )
{
  if ( !isInitialised ) {
    isInitialised = TRUE;

    GenIdents_Init();

    Parameter_ValueOpens = NoCObjects;
    Parameter_VAROpens = NoCObjects;
  } /* endif */
}
