/*************************************************************************
 *
 * Parallaxis Compiler Projects
 * ----------------------------
 *
 * name  :  Types.h
 * desc  :  general handling of types and type compatibility
 * author:  lampke
 *
 ************************************************************************/

#ifndef Types_h
#define Types_h

#include "parallaxis.h"
#include "BigSet.h"
#include "Values.h"
#include "Tree.h"
#include "Symbols.h"

#define Types_MinInteger   MINLONG
#define Types_MaxInteger   MAXLONG
#define Types_MaxCardinal  MAXULONG
#define Types_MinChar      MINUCHAR
#define Types_MaxChar      MAXUCHAR
#define Types_MaxBitset    (BITSPERBITSET-1)
#define Types_MinReal      MINREAL
#define Types_MaxReal      MAXREAL

/*************************************************************************
 * name :  IsAssignCompatible
 * in   :  lvalue type and expression type
 * out  :  TRUE, if assignment is correct
 *         FALSE otherwise
 ************************************************************************/
extern BOOLEAN Types_IsAssignCompatible(tType lvalue, tType expr);

/*************************************************************************
 * name :  IsParameterCompatible
 * in   :  lvalue type and expression type
 * out  :  TRUE, if pass as parameter is correct
 *         FALSE otherwise
 ************************************************************************/
extern BOOLEAN Types_IsParameterCompatible(tType lvalue, tType expr);

/*************************************************************************
 * name :  IsReturnCompatible
 * in   :  lvalue type and expression type
 * out  :  TRUE, if return is correct
 *         FALSE otherwise
 ************************************************************************/
extern BOOLEAN Types_IsReturnCompatible(tType lvalue, tType expr);

/*************************************************************************
 * name :  VectorsCompatible    
 * in   :  two vector types
 * out  :  TRUE, if configurations are compible (identical, same config
             group, both open)
 *         FALSE otherwise
 ************************************************************************/
extern BOOLEAN Types_VectorsCompatible(tType vt1, tType vt2);

/*************************************************************************
 * name :  ResultType
 * in   :  operator and operand types
 * out  :  type of operator result
 ************************************************************************/
extern tType Types_ResultType(INDEX operator, tType operand1, tType operand2);

/*************************************************************************
 * name :  StdProcResultType
 * in   :  standard procedure and parameter types
 * out  :  type of procedure result
 ************************************************************************/
extern tType Types_StdProcResultType(tType proc, tTypes params);

/*************************************************************************
 * name :  WriteType
 * in   :  buffer to take string representation, type to be printed
 * out  :  pointer to buffer
 ************************************************************************/
extern CHAR *Types_WriteType(tType type, CHAR *buffer);

/*************************************************************************
 * name :  TypeSize
 * in   :  type
 * out  :  number of bytes needed to store type 'type'
 ************************************************************************/
extern ULONG Types_TypeSize(tType type);

/*************************************************************************
 * name :  GetLwb
 * in   :  type and pointer to value
 * out  :  TRUE if type is ordinal and 'value' contains the lower bound
 *         FALSE otherwise
 ************************************************************************/
extern BOOLEAN Types_GetLwb(tType type, tValue *value);

/*************************************************************************
 * name :  GetUpb
 * in   :  type and pointer to value
 * out  :  TRUE if type is ordinal and 'value' contains the upper bound
 *         FALSE otherwise
 ************************************************************************/
extern BOOLEAN Types_GetUpb(tType type, tValue *value);

/*************************************************************************
 * name :  Cast
 * in   :  two types and operator
 * out  :  TRUE if type cast must be done, when 'operator' is applied to
 *         an operant of type 't2' and the result should be of type 't1'
 *         FALSE otherwise
 ************************************************************************/
extern BOOLEAN Types_Cast(INDEX Operator, tType t1, tType t2);

extern void Types_Init(void);

#endif
