/* Muss natuerlich durch Optionen verwaltet werden nur zum Test*/
/*************************************************************************
 *
 * Parallaxis Compiler Projects
 * ----------------------------
 *
 * name  :  Errors.c
 * desc  :  handels all occuring errors by collecting them
 *          and giving a sorted output
 * author:  lampke
 *
 ************************************************************************/

#include <stdlib.h>
#include "Memory.h"			  /* Reuse-Library */
#include "Sets.h"			  /* Reuse-Library */
#include "Idents.h"			  /* Reuse-Library */
#include "BigSet.h"
#include "Errors.h"
#include "language.h"
static BOOLEAN isInitialised = FALSE;

#define  MaxError       100

/*
 * local function prototypes
 */
static void yyExit      (void);

static void WriteHead   (tPosition Position, int ErrorClass);
static void WriteCode   (int ErrorCode);
static void WriteInfo   (int InfoClass, void *Info);
static void WriteMessage(BOOLEAN IsErrorCode, int ErrorCode, int ErrorClass,
                          tPosition Position, int InfoClass, void *Info);
static void StoreMessage(BOOLEAN IsErrorCode, int ErrorCode, int ErrorClass,
                          tPosition Position, int InfoClass, void *Info);
static int  IsLess      (const void *i, const void *j);

/*
 * local variables
 */
typedef struct {
  tPosition    Position;
  BOOLEAN      IsErrorCode;
  SHORT        ErrorNumber;
  SHORT        SubErrorNumber;
  SHORT        ErrorCode;
  SHORT        ErrorClass;
  SHORT        InfoClass;
  union {
    LONG      vInteger;
    SHORT     vShort;
    LONG      vLong;
    float     vReal;
    BOOLEAN   vBoolean;
    UCHAR     vCharacter;
    tStringRef vString;
    tSet     *vSet;
    tIdent    vIdent;
  } Info;
} tError;

static tError   ErrorTable [MaxError + 1];
static USHORT   ErrorCount = 0;
static USHORT   MessageCount = 0;
static BOOLEAN  IsStore = TRUE;
static void   (*HandleMessage)(BOOLEAN IsErrorCode, int ErrorCode, int ErrorClass, tPosition Position, int InfoClass, void *Info) = StoreMessage;
static FILE    *Out;
static BOOLEAN  PrintWarnings = TRUE;


/*
 * local function implementation
 */
static void yyExit(void)
{
   exit(1);
}

/*------- WriteHead --------------------------------------------------------*/
static void WriteHead(tPosition Position, int ErrorClass)
{
  WritePosition (Out, Position);
  (void)fputs(": ", Out);
  switch (ErrorClass) {
  case xxFatal         : (void)fputs("Fatal       ", Out); break;
  case xxRestriction   : (void)fputs("Einschraenkung ", Out); break;
  case xxError         : (void)fputs("Error       ", Out); break;
  case xxWarning       : (void)fputs("Warnung     ", Out); break;
  case xxRepair        : (void)fputs("Reparatur      ", Out); break;
  case xxNote          : (void)fputs("Bemerkung        ", Out); break;
  case xxInformation   : (void)fputs("Information ", Out); break;
  default              : (void)fprintf(Out, "Error class: %d ", ErrorClass);
  }
}

/*------- WriteCode --------------------------------------------------------*/
/*The macro errlang allows simple implementation of error-messages in different languages
  Just expand the macro for your desired language and add the errormessages to
  the appropriate cases of ErrorCode
*/
#define errlang(engmessag, deutschmeldung) \
switch (error_messages_language) { \
				   case english: \
				     (void)fputs(engmessag, Out); break; \
				   case deutsch: \
				     (void)fputs(deutschmeldung, Out); break; \
				 }; break;

static void WriteCode(int ErrorCode)
{
  switch (ErrorCode) {
  case xxNoText: break;
  case xxSyntaxError:
     errlang("syntax error",
             "Syntaktischer Fehler" )
  case xxExpectedTokens:
     errlang("expected tokens",
             "Erwartete Token" )
  case xxRestartPoint:
     errlang("restart point",
             "Restart Point" )
  case xxTokenInserted:
     errlang("token inserted",
             "Token eingefuegt" )
  case xxTokensIgnored:
     errlang("tokens ignored",
             "Token ignoriert" )
  case xxModNotFound:
     errlang("definition module not found",
             "Definition module nicht gefunden" )
  case xxCyclicDefMods:
     errlang("cyclic dependency",
             "Zyklische Abhaengigkeit" )
  case xxFileNameMismatch:
     errlang("filename - modname mismatch ",
             "Filename und Modulname passen nicht zueinander" )
  case xxTooManyErrors:
     errlang("too many errors",
             "Zu viele Fehler" )
  case xxSemanticError:
     errlang("semantic error",
             "Semantischer Fehler" )
  case xxAssignIncomp:
     errlang("assignment of incompatible types",
             "Zuweisung inkompatibler Typen" )
  case xxOperandIncomp:
     errlang("operands of incompatible types",
             "Operanden sind zueiander nicht Typkompatibel" )
  case xxConfigIncomp:
     errlang("operands have to be of same configuration",
             "Die Operanden muessen von derselben Konfiguration sein" )
  case xxResultTypeIncomp:
     errlang("incompatible result type",
             "Inkompatibler Ergebnistyp" )
  case xxTypeExpected:
     errlang("Expected type",
             "Erwarteter Typ" )
  case xxTypeFound:
     errlang("Found type",
             "Gefundener Typ" )
  case xxTypeLeft:
     errlang("Left type",
             "Linker Typ" )
  case xxTypeRight:
     errlang("Right type",
             "Rechter Typ" )
  case xxIllegalOperator:
     errlang("operator illegal for these types",
             "Der Operator ist illegal fuer diesen Typ" )
  case xxNotDeclared:
     errlang("Not declared",
             "Nicht vereinbart" )
  case xxNotInRecord:
     errlang("Not in record",
             "Nicht im Record" )
  case xxExprNotConst:
     errlang("Expression not constant",
             "Ausdruck muss konstant sein" )
  case xxTypeOverFlow:
     errlang("Exceeded maximum type range",
             "Maximaler Wertebereich dieses Typs ueberschritten" )
  case xxSetOverFlow:
     errlang("Exceeded maximum number of set elements",
             "Die maximal moegliche Maechtigkeit der Menge wurde ueberschritten" )
  case xxValueRange:
     errlang("legal range",
             "erlaubter Bereich" )
  case xxDivisionByZero:
     errlang("Division by zero",
             "Division durch Null" )
  case xxIllegalArgCount:
     errlang("wrong number of arguments",
             "Falsche Anzahl der Argumente" )
  case xxArgumentTypeIncomp:
     errlang("incompatible argument type at parameter ",
             "Inkompatibler Argumenttyp beim Parameter" )
  case xxVectorExpected:
     errlang("vector type expected",
             "Vektortyp erwartet" )
  case xxFunction:
     errlang("function",
             "Funktion" )
  case xxArgNumber:
     errlang("parameter",
             "Parameter" )
  case xxArgsExpected:
     errlang("parameters expected",
             "erwartete Parameter" )
  case xxArgsFound:
     errlang("parameters found",
             "gefundene Parameter" )
  case xxNoReduceFunction:
     errlang("not a reduce function",
             "Keine Reduce-Funktion" )
  case xxIllegalProcExpr:
     errlang("ID must be of integer type",
             "ID muss vom Typ INTEGER sein" )
  case xxStructTypeReq:
     errlang("cannot translate forward reference to non structured type",
             "Ich kann keinen Vorwaertsbezug eines nicht strukturierten Typs uebersetzen" )
  case xxOutOfLongRange:
     errlang("constant out of integer range",
             "Die Konstante ist ausserhalb des Wertebereichs von INTEGER" )
  case xxFunctionAsProcedure:
     errlang("function used as procedure",
             "Die Funktion wurde als Prozedur verwendet" )
  case xxNoOpenVector:
     errlang("no open vector in parameters specified",
             "Kein Open Vector als Parameter spezifiziert" )
  case xxWrongConfigParams:
     errlang("incompatible configurations in open vector parameters",
             "inkompatible Konfiguration in offenen Vektorparametern" )
  case xxConnectionExpected:
     errlang("connection expected",
             "Connection erwartet")
  case xxIllegalReduceNumber:
     errlang("number of reduces must be null or one",
             "Die Anzahl der Reduktionen muss Null oder eins sein")
  case xxTooManyReduces:
     errlang("too many reduces specified",
             "Zu viele Reduktionen spezifiziert" )
  case xxVectorIndex:
     errlang("vector index to scalar array",
             "Vektor Index bei einem scalaren Feld" )
  case xxScalarExpected:
     errlang("scalar type expected",
             "Skalarer Typ erwartet" )
  case xxArrayExpected:
     errlang("array type expected",
             "Array-Typ erwartet" )
  case xxConfigExpected:
     errlang("configuration name expected",
             "Name der Konfiguration erwartet" )
  case xxReturnValueExpected:
     errlang("no return value specified",
             "Kein Rueckgabewert spezifiziert" )
  case xxReturnValueWrong:
     errlang("return value specified",
             "Rueckgabewert spezifiziert" )
  case xxConnectIndex:
     errlang("connection and connection index do not match",
             "CONNECTION und CONNECTION-Index passen nicht zueinander" )
  case xxGenericConfig:
     errlang("generic configuration expected",
             "Generische Configuration erwartet" )
  case xxRankMismatch:
     errlang("rank of respecified and generic configuration do not match",
             "Rang der neudefinierten und der generischen Konfiguration stimmen nicht ueberein" )
  case xxIdentRedeclared:
     errlang("redeclared identifier in same scope",
             "Doppelt deklarierter Bezeichner in selben Bereich" )
  case xxNoGenericConfig:
     errlang("no generic configuration in parameters specified",
             "Keine generische Konfiguration als Parameter spezifiziert" )
  case xxWrongGenericParams:
     errlang("incompatible configurations in generic configuration parameters",
             "Inkompatible Configuration bei den generischen Configurationsparametern" )
  case xxVectorInInit:
     errlang("vector data in record resp. array initialisation",
             "Vektor-Daten im Record bzw. in der Array-Initialisierung" )
  case xxNoTypeCast:
     errlang("neither type cast nor record or array initialisation",
             "Weder Typumwandlung noch Record- odr Arrayinitialisierung" )
  case xxGenericConnectMixed:
     errlang("generic connection with different configurations",
             "Generische Connection mit verschiedenen Konfigurationen" )
  case xxInvalidFilename:
     errlang("file name and module name do not match",
             "Name der Datei und Name des Moduls stimmen nicht ueberein" )
  case xxConfigCycle:
     errlang("cycle in configuration group",
             "Zyklus in der Konfigurationsgruppe" )
  case xxInvalidConfigLen:
     errlang("configuration group members must have same size",
             "Die Mitglieder der Konfigurationsgruppe mussen gleich gross sein" )
       case xxNoTypeName:
     errlang("no type name",
             "kein Typname" )
  case xxVariableUsedAsFunc:
     errlang("Variable used as function-call",
             "Sie haben eine Variable zum Funktionsaufruf verwendet" )
  case xxVarParNotLval:
     errlang("Expression not suitable for call-by-reference",
	     "Expression nicht geeignet fuer Call-by-reference")
  case xxLvalExpected:
     errlang("Lvalue expected",
	     "erwarte Lvalue")
  case xxBooleanExpected:
     errlang("Expression must be of type boolean",
	     "Expression muss den Typ Boolean haben")
  default:
    (void)fprintf(Out, " error code: %d", ErrorCode); break;
  }
}

/*------- WriteInfo --------------------------------------------------------*/
static void WriteInfo(int InfoClass, void *Info)
{
   SHORT i;

   if (InfoClass == xxNone)
      return;
   (void)fputs(": ", Out);
   switch (InfoClass) {
   case xxInteger       : (void)fprintf(Out, "%d", *(int *)Info); break;
   case xxShort         : i = *(SHORT *)Info; (void)fprintf(Out, "%hd", i); break;
   case xxLong          : (void)fprintf(Out, "%ld", *(LONG *)Info); break;
   case xxReal          : (void)fprintf(Out, "%e", *(float *)Info); break;
   case xxBoolean       : (void)fprintf(Out, "%c", *(BOOLEAN *)Info ? 'T' : 'F'); break;
   case xxCharacter     : (void)fprintf(Out, "%c", *(char *)Info); break;
   case xxString        : (void)fputs((char *)Info, Out); break;
   case xxSet           : WriteSet(Out, (tSet *)Info); break;
   case xxBITSET        : BITSET_WriteSet(Out, *(LONG *)Info); break;
   case xxIdent         : WriteIdent(Out, *(tIdent *)Info); break;
   default              : (void)fprintf(Out, "info class: %d", InfoClass);
   }
}

/*------- WriteMessage -----------------------------------------------------*/
static void WriteMessage(BOOLEAN IsErrorCode, int ErrorCode, int ErrorClass,
                          tPosition Position, int InfoClass, void *Info)
{
  if (ErrorClass != xxWarning || PrintWarnings) {
    WriteHead(Position, ErrorClass);
    if (IsErrorCode) {
       WriteCode(ErrorCode);
    } else {
       WriteIdent(Out, ErrorCode);
    }
    WriteInfo(InfoClass, Info);
    (void)fputc('\n', Out);
    if (ErrorClass == xxFatal && !IsStore) {
      (*Errors_Exit)();
    }
  }
}

/*------- StoreMessage -----------------------------------------------------*/
static void StoreMessage(BOOLEAN IsErrorCode, int ErrorCode, int ErrorClass,
                          tPosition Position, int InfoClass, void *Info)
{
   static USHORT lastErrorNumber = 0;
   static USHORT lastSubErrorNumber = 0;
   tError *With;
   ULONG   i;
   BOOLEAN found = FALSE;

   for (i=1; i <= MessageCount && !found; i++) {
      With = &ErrorTable[i];
      if (!Compare(With->Position, Position) &&
           With->ErrorCode  == ErrorCode        &&
           With->ErrorClass == ErrorClass       &&
           With->InfoClass  == InfoClass) {
         found = TRUE;
      } /* endif */
   } /* endfor */

   if (found) {
      if (ErrorClass == xxFatal || ErrorClass == xxError ||
          ErrorClass == xxRestriction || ErrorClass == xxWarning) {
         ErrorCount--;
      } /* endif */
   } else {
      if (MessageCount < MaxError) {
         MessageCount++;
         if (ErrorClass != xxInformation) {
            lastErrorNumber    = MessageCount;
            lastSubErrorNumber = 0;
         } else {
            lastSubErrorNumber++;
         } /* endif */
         With = &ErrorTable[MessageCount];
         With->Position       = Position;
         With->IsErrorCode    = IsErrorCode;
         With->ErrorNumber    = lastErrorNumber;
         With->SubErrorNumber = lastSubErrorNumber;
         With->ErrorCode      = ErrorCode;
         With->ErrorClass     = ErrorClass;
         With->InfoClass      = InfoClass;
         switch (With->InfoClass) {
         case xxInteger    : With->Info.vInteger   = *(int *) Info; break;
         case xxShort      : With->Info.vShort     = *(SHORT *) Info; break;
         case xxLong       : With->Info.vLong      = *(LONG *) Info; break;
         case xxReal       : With->Info.vReal      = *(float *) Info; break;
         case xxBoolean    : With->Info.vBoolean   = *(BOOLEAN *) Info; break;
         case xxCharacter  : With->Info.vCharacter = *(UCHAR *) Info; break;
         case xxString     : With->Info.vString    = PutString(Info, strlen(Info)); break;
         case xxSet        : With->Info.vSet = (tSet *)Alloc((unsigned long) sizeof(tSet));
                             MakeSet(With->Info.vSet, Size((tSet *)Info));
                             Assign(With->Info.vSet, (tSet *)Info); break;
         case xxBITSET     : With->Info.vLong      = *(LONG *) Info; break;
         case xxIdent      : With->Info.vIdent     = *(tIdent *)Info; break;
         }
      } else {
         With = &ErrorTable[MessageCount];

         With->Position    = NoPosition;
         With->IsErrorCode = TRUE;
         With->ErrorCode   = xxTooManyErrors;
         With->ErrorClass  = xxRestriction;
         With->InfoClass   = xxNone;
      }
   } /* endif */
   if (ErrorClass == xxFatal) {
      WriteMessages(stderr, FALSE);
      (*Errors_Exit)();
   }
}

/*------- IsLess -----------------------------------------------------------*/
static int IsLess(const void *i, const void *j)
{
   int r;

   if (((tError *)i)->ErrorNumber == ((tError *)j)->ErrorNumber) {
      return ((tError *)i)->SubErrorNumber - ((tError *)j)->SubErrorNumber;
   } else {
      r = Compare(((tError *)i)->Position, ((tError *)j)->Position);
      return r != 0 ? r : ((tError *)i)->ErrorNumber - ((tError *)j)->ErrorNumber;
   } /* endif */
}

/*
 * Global variables
 */
void (*Errors_Exit)(void) = yyExit;

/*
 * global function implementation
 */
/*------- StoreMessages ----------------------------------------------------*/
void StoreMessages(BOOLEAN Store)
{
   if (Store) {
      HandleMessage = StoreMessage;
      MessageCount  = 0;
   } else {
      HandleMessage = WriteMessage;
   }
   IsStore = Store;
}

/*------- ErrorMessage -----------------------------------------------------*/
void ErrorMessage(int ErrorCode, int ErrorClass, tPosition Position)
{
   if (ErrorClass == xxFatal || ErrorClass == xxError ||
       ErrorClass == xxRestriction || ErrorClass == xxWarning) {
      ErrorCount++;
   } /* endif */
   (* HandleMessage)(TRUE, ErrorCode, ErrorClass, Position, xxNone, NULL);
}

/*------- ErrorMessageI ----------------------------------------------------*/
void ErrorMessageI(int ErrorCode, int ErrorClass, tPosition Position,
                    int InfoClass, void *Info)
{
   if (ErrorClass == xxFatal || ErrorClass == xxError ||
       ErrorClass == xxRestriction || ErrorClass == xxWarning) {
      ErrorCount++;
   } /* endif */
   (* HandleMessage)(TRUE, ErrorCode, ErrorClass, Position, InfoClass, Info);
}

/*------- Message ----------------------------------------------------------*/
void Message(char *ErrorText, int ErrorClass, tPosition Position)
{
   if (ErrorClass == xxFatal || ErrorClass == xxError ||
       ErrorClass == xxRestriction || ErrorClass == xxWarning) {
      ErrorCount++;
   } /* endif */
   (* HandleMessage)(FALSE, MakeIdent(ErrorText, strlen(ErrorText)),
                      ErrorClass, Position, xxNone, NULL);
}

/*------- MessageI ---------------------------------------------------------*/
void MessageI(char *ErrorText, int ErrorClass, tPosition Position,
               int InfoClass, void *Info)
{
   if (ErrorClass == xxFatal || ErrorClass == xxError ||
       ErrorClass == xxRestriction || ErrorClass == xxWarning) {
      ErrorCount++;
   } /* endif */
   (* HandleMessage)(FALSE, MakeIdent(ErrorText, strlen(ErrorText)),
                      ErrorClass, Position, InfoClass, Info);
}

/*------- WriteMessages ----------------------------------------------------*/
void WriteMessages(FILE * File, BOOLEAN Warnings)
{
   USHORT i;
   void *Info;
   char  s[256];

   PrintWarnings = Warnings;
   qsort((void *)&ErrorTable[1],
          (ErrorTable[MessageCount].ErrorCode != xxTooManyErrors) ?
          MessageCount : MessageCount-1,
          sizeof(tError), IsLess);
   Out = File;
   for (i = 1; i <= MessageCount; i ++) {
      tError *With = &ErrorTable[i];

      switch (With->InfoClass) {
      case xxInteger    : Info = (void *)&With->Info.vInteger ; break;
      case xxShort      : Info = (void *)&With->Info.vShort   ; break;
      case xxLong       : Info = (void *)&With->Info.vLong    ; break;
      case xxReal       : Info = (void *)&With->Info.vReal    ; break;
      case xxBoolean    : Info = (void *)&With->Info.vBoolean ; break;
      case xxCharacter  : Info = (void *)&With->Info.vCharacter; break;
      case xxString     : StGetString(With->Info.vString, s); Info = s; break;
      case xxSet        : Info = (void *)With->Info.vSet      ; break;
      case xxBITSET     : Info = (void *)&With->Info.vLong    ; break;
      case xxIdent      : Info = (void *)&With->Info.vIdent   ; break;
      }
      WriteMessage(With->IsErrorCode, With->ErrorCode, With->ErrorClass,
                    With->Position, With->InfoClass, Info);
   }
   Out = stderr;
}

/*------- ErrorGetCount ----------------------------------------------------*/
USHORT ErrorGetCount(void)
{
   return ErrorCount;
}

void Errors_Init(void)
{
   if (!isInitialised) {
      isInitialised = TRUE;
      Positions_Init();
      ErrorCount = 0;
      Out = stderr;
   } /* endif */
}

