/*************************************************************************
 *
 * Parallaxis Compiler Projects
 * ----------------------------
 *
 * name  :  Options.h
 * desc  :  Handling of command line options and arguments
 * author:  lampke
 *
 ************************************************************************/

#ifndef Options_h
#define Options_h

#include "parallaxis.h"

#if 0		      /* Option "/" for MSDOS not supported, use "-" too. */
#ifdef MSDOS
#define OPTIONSIGN   '/'
#else
#define OPTIONSIGN   '-'
#endif
#endif

typedef enum {
    Options_eCasts,
    Options_eDebug,
    Options_eInclude,
    Options_eIndent,
    Options_eMemStats,
    Options_eChecks,
    Options_eTimeStats,
    Options_eWarnings,
    Options_eVerbose,
    Options_edeutsch,
    Options_eenglish,
    Options_efrancais,
#ifdef DEVELOP
    Options_eWriteTree,
    Options_eWriteSymbolsTree,
    Options_eQueryTree,
    Options_eCheckTree,
    Options_eCompileOnly,
    Options_eEvalOnly,
    Options_eDebugPuma,
#endif
    Options_Options
} tOptions;

/*************************************************************************
 * name :  CheckArguments
 * in   :
 * out  :
 ************************************************************************/
extern BOOLEAN Options_CheckArguments( int argc, char *argv[] );

/*************************************************************************
 * name :  GetSourceFile
 * in   :
 * out  :
 ************************************************************************/
extern char *Options_GetSourceFile( void );

/*************************************************************************
 * name :  GetSourceIdent
 * in   :
 * out  :
 ************************************************************************/
extern ULONG Options_GetSourceIndent( void );

/*************************************************************************
 * name :  IsOptionSet
 * in   :
 * out  :
 * Remark: the appropriate option bit is set if the feature is turned on,
 *         not if the command line switch is given. E.g. Options_eChecks is
 *         set, if range checks should be generated and unset, if the switch
 *         -r ("don't generate range checks") is given. 
 ************************************************************************/
extern BOOLEAN Options_IsOptionSet( tOptions Option );

extern void Options_Init( void );

#endif
