(***********************************************************************
 *   Program:     test.m  (MPL-source)
 *   Version:     1.0
 *   Created:     4.9.95
 *   Last change: Tue Sep 12 00:28:13 1995
 *   Author:      Claudia Stumpp
 ***********************************************************************)
MODULE claudi;


(* ****************          CONSTANTS            ************** *)

CONST h = 3;	(* height of torus *)
      w = 4;	(* width of torus  *)


(* ****************          TYPES                ************** *)

(* TYPE vector_of_char = VECTOR OF CHAR; *)
(* not allowed, but compiler accepts *)

TYPE day = (mo,di,mi,do,fr,sa,so);
     workday = [mo..fr];

(* **************** CONFIGURATIONS AND CONNECTIONS ************* *)

(* grid *)
CONFIGURATION grid [1..3],[1..4];
CONNECTION north: grid[i,j] -> grid[i-1,j];
      	   south: grid[i,j] -> grid[i+1,j];
           east : grid[i,j] -> grid[i,j+1];
           west : grid[i,j] -> grid[i,j-1];

(* torus *)
CONFIGURATION torus [0..h-1],[0..w-1];
CONNECTION north: torus[i,j] -> torus[(i-1) MOD h,j];
           south: torus[i,j] -> torus[(i+1) MOD h,j];
           east : torus[i,j] -> torus[i, (j+1) MOD w];
           west : torus[i,j] -> torus[i, (j-1) MOD w];

(* tree *)
CONFIGURATION tree [1..7];
CONNECTION lchild: tree[i] <-> tree[2*i]   : parent;
           rchild: tree[i] <-> tree[2*i+1] : parent;

(* hypercube *)
CONFIGURATION hyper [0..1],[0..1],[0..1],[0..1];
CONNECTION go[1]: hyper[i,j,k,l] -> hyper[(i+1) MOD 2,j,k,l];
           go[2]: hyper[i,j,k,l] -> hyper[i,(j+1) MOD 2,k,l];
           go[3]: hyper[i,j,k,l] -> hyper[i,j,(k+1) MOD 2,l];
           go[4]: hyper[i,j,k,l] -> hyper[i,j,k,(l+1) MOD 2];

(* list *)
CONFIGURATION list [1..8];
CONNECTION left: list[i] <-> list[i-1] : right;

(* grid mix tree *)
CONFIGURATION mix_grid [0..2],[0..3];
              mix_tree [1..12];
CONNECTION mix: mix_grid[i,j] -> mix_tree[i*4 + j];

(* char_list *)
CONFIGURATION char_list [mo..so];
CONNECTION left: char_list[i] <-> char_list[i-1] : right;



(* ****************          VARIABLES            ************** *)

(* gr_ means grid, to_ torus, tr_ tree, hy_ hyper, li_ list, cl_ char_list *)
VAR gr_real    : grid OF REAL;
    gr_integer : grid OF INTEGER;
    gr_cardinal: grid OF CARDINAL;
    gr_char    : grid OF CHAR;
    gr_array   : ARRAY [1..3] OF grid OF INTEGER;
    gr_boolean : grid OF BOOLEAN;
    
    to_real    : torus OF REAL;
    to_integer : torus OF INTEGER;
    to_cardinal: torus OF CARDINAL;
    to_char    : torus OF CHAR;
    to_array   : ARRAY [1..3] OF torus OF INTEGER;
    to_boolean : torus OF BOOLEAN;
    
    tr_real    : tree OF REAL;
    tr_integer : tree OF INTEGER;
    tr_cardinal: tree OF CARDINAL;
    tr_char    : tree OF CHAR;
    tr_array   : ARRAY [1..3] OF tree OF INTEGER;
    tr_boolean : tree OF BOOLEAN;

    hy_real    : hyper OF REAL;
    hy_integer : hyper OF INTEGER;
    hy_cardinal: hyper OF CARDINAL;
    hy_char    : hyper OF CHAR;
    hy_array   : ARRAY [1..3] OF hyper OF INTEGER;
    hy_boolean : hyper OF BOOLEAN;

    li_real    : list OF REAL;
    li_integer : list OF INTEGER;
    li_cardinal: list OF CARDINAL;
    li_char    : list OF CHAR;
    li_array   : ARRAY [1..3] OF list OF INTEGER;
    li_boolean : list OF BOOLEAN;


    li_day     : list OF day;
    li_workday : list OF workday;

    cl_real    : list OF REAL;
    cl_integer : list OF INTEGER;
    cl_cardinal: list OF CARDINAL;
    cl_char    : list OF CHAR;
    cl_array   : ARRAY [1..3] OF list OF INTEGER;
    cl_boolean : list OF BOOLEAN;

(***********************************************************************
 *   Procedure:    PROCEDURE init
 *                                                                     
 *   Parameters:   -
 *
 *   Result:       -
 *
 *   Description:                                                      
 *   ------------                                                      
 *   Initializes vector all variables, mainly with their grid-id.
 ***********************************************************************)
PROCEDURE init;       (* initializes all variables *)
BEGIN
  gr_real    := FLOAT(ID(grid));
  gr_integer := ID(grid);
  gr_cardinal:= ID(grid);
  gr_char    := CHR(ID(grid));
  gr_array[1]:= ID(grid);
  gr_array[2]:= ID(grid) + 12;
  gr_array[3]:= ID(grid) + 24;
  gr_boolean := ODD(ID(grid));
  
  to_real    := FLOAT(ID(torus));
  to_integer := ID(torus);
  to_cardinal:= ID(torus);
  to_char    := CHR(ID(torus));
  to_array[1]:= ID(torus);
  to_array[2]:= ID(torus) + 12;
  to_array[3]:= ID(torus) + 24;
  to_boolean := ODD(ID(torus));

  tr_real    := FLOAT(ID(tree));
  tr_integer := ID(tree);
  tr_cardinal:= ID(tree);
  tr_char    := CHR(ID(tree));
  tr_array[1]:= ID(tree);
  tr_array[2]:= ID(tree) + 12;
  tr_array[3]:= ID(tree) + 24;
  tr_boolean := ODD(ID(tree));

  hy_real    := FLOAT(ID(hyper));
  hy_integer := ID(hyper);
  hy_cardinal:= ID(hyper);
  hy_char    := CHR(ID(hyper));
  hy_array[1]:= ID(hyper);
  hy_array[2]:= ID(hyper) + 12;
  hy_array[3]:= ID(hyper) + 24;
  hy_boolean := ODD(ID(hyper));

  li_real    := FLOAT(ID(list));
  li_integer := ID(list);
  li_cardinal:= ID(list);
  li_char    := CHR(ID(list));
  li_array[1]:= ID(list);
  li_array[2]:= ID(list) + 12;
  li_array[3]:= ID(list) + 24;
  li_boolean := ODD(ID(list));

  cl_real    := FLOAT(ID(list));
  cl_integer := ID(list);
  cl_cardinal:= ID(list);
  cl_char    := CHR(ID(list));
  cl_array[1]:= ID(list);
  cl_array[2]:= ID(list) + 12;
  cl_array[3]:= ID(list) + 24;
  cl_boolean := ODD(ID(list));
	        
END init;


(***********************************************************************
 *   Procedure:    PROCEDURE general_functions
 *                                                                     
 *   Parameters:   a    : any vector of integer 
 *                 b    : any vector of real    
 *                 c    : any vector of char    
 *                 d    : any vector of cardinal
 *                 e    : any vector of boolean
 *                 arr  : open array of integer (scalar)
 *                 print: boolean (if solutions should be printed)
 *
 *   Result:       -
 *
 *   Description:                                                      
 *   ------------                                                      
 *   Tests all general function with scalar AND vector variables. Following
 *   functions are tested: float,trunc,abs,chr,val,ord,max,min,size,high,
 *   odd,even,cap. In this moment, only the syntax is tested but no 
 *   semantics!
 ***********************************************************************)
PROCEDURE general_functions(a    : VECTOR OF INTEGER;
                            b    : VECTOR OF REAL;
			    c    : VECTOR OF CHAR;
			    d    : VECTOR OF CARDINAL;
			    e    : VECTOR OF BOOLEAN;
			    arr  : ARRAY OF INTEGER;
			    print: BOOLEAN);
                           
(* note: open array of VECTOR not implemented: arr: ARRAY VECTOR OF INTEGER *)

TYPE vector_of_char = grid OF CHAR;

VAR x,hx: INTEGER;
    y,hy: REAL;
    z,hz: CHAR;
    w,hw: CARDINAL;
    sbol: BOOLEAN;
    ha  : VECTOR OF INTEGER;
    hb  : VECTOR OF REAL;
    hc  : VECTOR OF CHAR;
    hd  : VECTOR OF CARDINAL;
    vbol: VECTOR OF BOOLEAN;
    hgrid: grid OF CHAR;
    hgridint: grid OF INTEGER;

    hday    : day;
    hworkday: workday;

BEGIN

  x := 3;
  y := 3.5;
  z := 's';
  sbol := TRUE;

  hy := FLOAT(x); (* scalar *)
  hb := FLOAT(a); (* vector *)
  hx := TRUNC(y); (* scalar *)
  ha := TRUNC(b); (* vector *)
  IF print THEN
    WriteString("float from scalar:");WriteFixPt(hy,2,1);WriteLn;
    WriteString("float from vector:");WriteLn;WriteFixPt(hb,2,1);WriteLn;
    WriteString("trunc from scalar:");WriteInt(hx,2);WriteLn;
    WriteString("trunc from vector:");WriteLn;WriteInt(ha,2);WriteLn;
  END; (*IF*)

  hy := ABS(y); (* scalar *)
  hb := ABS(b); (* vector *)
  hx := ABS(x); (* scalar *)
  ha := ABS(a); (* vector *)
  hz := CHR(x+65); (* scalar *)
  hc := CHR(a+65); (* vector *)
  IF print THEN
    WriteString("abs from scalar-real:");WriteFixPt(hy,2,1);WriteLn;
    WriteString("abs from vector-real:");WriteLn;WriteFixPt(hb,2,1);WriteLn;
    WriteString("abs from scalar-integer:");WriteInt(hx,2);WriteLn;
    WriteString("abs from vector-integer:");WriteLn;WriteInt(ha,2);WriteLn;
    WriteString("chr from scalar:");Write(hz);WriteLn;
    WriteString("chr from vector:");WriteLn;Write(hc);WriteLn;
  END; (*IF*)

  hz := VAL(CHAR,x+65); (* scalar *)
(*
  hgrid := VAL(vector_of_char,x); (* vector *)  not correct in test.c!!!
  (* not implemented: VAL(vector_of_char,a) *) 
  (* not possible: VAL(VECTOR OF CHAR,a or x) *)

*)			          
  hx := ORD(z); (* scalar *)
  ha := ORD(c); (* vector *)
  IF print THEN
    WriteString("val from scalar:");Write(hz);WriteLn;
    WriteString("ord from scalar:");WriteInt(hx,2);WriteLn;
    WriteString("ord from vector:");WriteInt(ha,2);WriteLn;
  END; (*IF*)
      
  hday := fr;
  hx := ORD(hday); (* scalar *)
  ha := ORD(hday); (* vector *)
  IF print THEN
    WriteString("ord from scalar:");WriteInt(hx,2);WriteLn;
    WriteString("ord from vector:");WriteLn;WriteInt(ha,2);WriteLn;
  END; (*IF*)
    
  hx := ORD(sbol); (* scalar *)
  ha := ORD(d);    (* vector *)
  IF print THEN
    WriteString("ord from scalar:");WriteInt(hx,2);WriteLn;
    WriteString("ord from vector:");WriteLn;WriteInt(ha,2);WriteLn;
  END; (*IF*)
  
  hz := MAX(CHAR); (* scalar *) 
  hc := MAX(CHAR); (* vector *)
  hx := MAX(INTEGER); (* scalar *)
  ha := MAX(INTEGER); (* vector *)
  hw := MAX(CARDINAL); (* scalar *)
  hd := MAX(CARDINAL); (* vector *)
  sbol := MAX(BOOLEAN); (* scalar *)
  vbol := MAX(BOOLEAN); (* vector *)
  hday := MAX(day); (* scalar *)
  IF print THEN
    WriteString("max from scalar:");Write(hz);WriteLn;
    WriteString("max from vector:");WriteLn;Write(hc);WriteLn;
    WriteString("max from scalar:");WriteInt(hx,2);WriteLn;
    WriteString("max from vector:");WriteLn;WriteInt(ha,2);WriteLn;
    WriteString("max from scalar:");WriteCard(hw,2);WriteLn;
    WriteString("max from vector:");WriteLn;WriteCard(hd,2);WriteLn;
    WriteString("max from scalar:");WriteBool(sbol,2);WriteLn;
    WriteString("max from vector:");WriteLn;WriteBool(vbol,2);WriteLn;
  END; (*IF*)
    
  hday := MIN(workday); (* scalar *)
  (* not possible: hworkday:= MIN(workday); *) 
  
  hz := MIN(CHAR); (* scalar *) 
  hc := MIN(CHAR); (* vector *)
  hx := MIN(INTEGER); (* scalar *)
  ha := MIN(INTEGER); (* vector *)
  hw := MIN(CARDINAL); (* scalar *)
  hd := MIN(CARDINAL); (* vector *)
  sbol := MIN(BOOLEAN); (* scalar *)
  vbol := MIN(BOOLEAN); (* vector *)
  IF print THEN
    WriteString("min from scalar:");Write(hz);WriteLn;
    WriteString("min from vector:");WriteLn;Write(hc);WriteLn;
    WriteString("min from scalar:");WriteInt(hx,2);WriteLn;
    WriteString("min from vector:");WriteLn;WriteInt(ha,2);WriteLn;
    WriteString("min from scalar:");WriteCard(hw,2);WriteLn;
    WriteString("min from vector:");WriteLn;WriteCard(hd,2);WriteLn;
    WriteString("min from scalar:");WriteBool(sbol,2);WriteLn;
    WriteString("min from vector:");WriteLn;WriteBool(vbol,2);WriteLn;
  END; (*IF*)

  
  hx := SIZE(CHAR); (* scalar *)
(*
  hgridint := SIZE(vector_of_char); (* vector *) not correct in test.c!!!
  (* not possible: SIZE(VECTOR OF CHAR) *)
*)
  IF print THEN
    WriteString("size from scalar:");WriteInt(hx,2);WriteLn;
  END; (*IF*)

  hx := HIGH(arr); (* scalar *)
  ha := HIGH(arr); (* vector *)
  sbol := ODD(x);  (* scalar *)
  vbol := ODD(a);  (* vector *)
  IF print THEN
    WriteString("high from scalar:");WriteInt(hx,2);WriteLn;
    WriteString("high from vector:");WriteLn;WriteInt(ha,2);WriteLn;
    WriteString("odd from scalar:");WriteBool(sbol,2);WriteLn;
    WriteString("odd from vector:");WriteLn;WriteBool(vbol,2);WriteLn;
  END; (*IF*)
 
  sbol := EVEN(z); (* scalar *)
  vbol := EVEN(c); (* vector *)
  hz := CAP(z); (* scalar *)
  hc := CAP(c); (* vector *)
  IF print THEN
    WriteString("even from scalar:");WriteBool(sbol,2);WriteLn;
    WriteString("even from vector:");WriteLn;WriteBool(vbol,2);WriteLn;
    WriteString("cap from scalar:");Write(hz);WriteLn;
    WriteString("cap from vector:");WriteLn;Write(hc);WriteLn;
  END; (*IF*)

END general_functions;



(***********************************************************************
 *   Procedure:    PROCEDURE test_general_functions
 *                                                                     
 *   Parameters:   -
 *
 *   Result:       -
 *
 *   Description:                                                      
 *   ------------                                                      
 *   Calls test of the general functions with different types of vector.
 ***********************************************************************)
PROCEDURE test_general_functions(print: BOOLEAN);
VAR i: INTEGER;
    test_array : ARRAY [1..100] OF INTEGER;
BEGIN
  FOR i:=1 TO 100 DO   (* init array *)
   test_array[i] := i;
  END; (*FOR*)

  IF print THEN 
    WriteString("-------------------------------------------------");
    WriteLn;
    WriteString("Test of general functions with configuration grid");
    WriteLn;
    WriteString("-------------------------------------------------");
    WriteLn;
  END; (*IF*)
  general_functions(gr_integer, gr_real, gr_char, gr_cardinal, 
                    gr_boolean, test_array, print);

  IF print THEN 
    WriteString("--------------------------------------------------");
    WriteLn;
    WriteString("Test of general functions with configuration torus");
    WriteLn;
    WriteString("--------------------------------------------------");
    WriteLn;
  END; (*IF*)
  general_functions(to_integer, to_real, to_char, to_cardinal,  
                    to_boolean, test_array, print);

  IF print THEN 
    WriteString("-------------------------------------------------");
    WriteLn;
    WriteString("Test of general functions with configuration tree");
    WriteLn;
    WriteString("-------------------------------------------------");
    WriteLn;
  END; (*IF*)
  general_functions(tr_integer, tr_real, tr_char, tr_cardinal, 
                    tr_boolean, test_array, print);

  IF print THEN 
    WriteString("--------------------------------------------------");
    WriteLn;
    WriteString("Test of general functions with configuration hyper");
    WriteLn;
    WriteString("--------------------------------------------------");
    WriteLn;
  END; (*IF*)
  general_functions(hy_integer, hy_real, hy_char, hy_cardinal, 
                    hy_boolean, test_array, print);
		    
  IF print THEN 
    WriteString("-------------------------------------------------");
    WriteLn;
    WriteString("Test of general functions with configuration list");
    WriteLn;
    WriteString("-------------------------------------------------");
    WriteLn;
  END; (*IF*)
  general_functions(li_integer, li_real, li_char, li_cardinal, 
                    li_boolean, test_array, print);

  IF print THEN 
    WriteString("------------------------------------------------------");
    WriteLn;
    WriteString("Test of general functions with configuration char_list");
    WriteLn;
    WriteString("------------------------------------------------------");
    WriteLn;
  END; (*IF*)
  general_functions(cl_integer, cl_real, cl_char, cl_cardinal, 
                    cl_boolean, test_array, print);
END test_general_functions;



(***********************************************************************
 *   Procedure:    PROCEDURE mathematical_functions
 *                                                                     
 *   Parameters:   a    : any vector of integer 
 *                 b    : any vector of real    
 *                 c    : any vector of char    
 *                 d    : any vector of cardinal
 *                 e    : any vector of boolean
 *                 arr  : open array of integer (scalar)
 *                 print: boolean (if solutions should be printed)
 *
 *   Result:       -
 *
 *   Description:                                                      
 *   ------------                                                      
 *   Tests all general function with scalar AND vector variables.
 ***********************************************************************)
PROCEDURE mathematical_functions(vec: VECTOR OF REAL; print: BOOLEAN);
VAR hvec1 : VECTOR OF REAL;
    hvec2 : VECTOR OF REAL;
    scal  : REAL;
    hscal1: REAL;
    hscal2: REAL;
BEGIN
  scal := ABS(RandomReal()*FLOAT(RandomInt()));

  hscal1:= sqrt(scal); (* scalar *)
  hvec1 := sqrt(vec);  (* vector *)
  hscal2:= exp(RandomReal());  (* scalar *)
  hvec2 := exp(vec);   (* vector *)
  IF print THEN
    WriteString("sqrt from scalar:");WriteFixPt(hscal1,2,2);WriteLn;
    WriteString("sqrt from vector:");WriteLn;WriteFixPt(hvec1,2,2);WriteLn;
    WriteString("exp from scalar:");WriteFixPt(hscal2,2,2);WriteLn;
    WriteString("exp from vector:");WriteLn;WriteFixPt(hvec2,2,2);WriteLn;
  END; (*IF*)

  hscal1:= ln(scal); (* scalar *)
  hvec1 := ln(vec);  (* vector *)
  hscal2:= sin(scal);(* scalar *)
  hvec2 := sin(vec); (* vector *)
  IF print THEN
    WriteString("ln from scalar:");WriteFixPt(hscal1,2,2);WriteLn;
    WriteString("ln from vector:");WriteLn;WriteFixPt(hvec1,2,2);WriteLn;
    WriteString("sin from scalar:");WriteFixPt(hscal2,2,2);WriteLn;
    WriteString("sin from vector:");WriteLn;WriteFixPt(hvec2,2,2);WriteLn;
  END; (*IF*)
  
  hscal1:= arcsin(RandomReal()); (* scalar *)
  hvec1 := arcsin(RandomReal(vec));  (* vector *)
  hscal2:= cos(scal);(* scalar *)
  hvec2 := cos(vec); (* vector *)
  IF print THEN
    WriteString("arcsin from scalar:");WriteFixPt(hscal1,2,2);WriteLn;
    WriteString("arcsin from vector:");WriteLn;WriteFixPt(hvec1,2,2);WriteLn;
    WriteString("cos from scalar:");WriteFixPt(hscal2,2,2);WriteLn;
    WriteString("cos from vector:");WriteLn;WriteFixPt(hvec2,2,2);WriteLn;
  END; (*IF*)

  hscal1:= arccos(RandomReal()); (* scalar *)
  hvec1 := arccos(RandomReal(vec));  (* vector *)
  hscal2:= tan(scal);(* scalar *)
  hvec2 := tan(vec); (* vector *)
  IF print THEN
    WriteString("arccos from scalar:");WriteFixPt(hscal1,2,2);WriteLn;
    WriteString("arccos from vector:");WriteLn;WriteFixPt(hvec1,2,2);WriteLn;
    WriteString("tan from scalar:");WriteFixPt(hscal2,2,2);WriteLn;
    WriteString("tan from vector:");WriteLn;WriteFixPt(hvec2,2,2);WriteLn;
  END; (*IF*)

  hscal1:= arctan(scal); (* scalar *)
  hvec1 := arctan(vec);  (* vector *)
  hscal2:= arctan2(scal,RandomReal());(* scalar *)
  hvec2 := arctan2(vec,RandomReal()); (* vector *)
  IF print THEN
    WriteString("arctan from scalar:");WriteFixPt(hscal1,2,2);WriteLn;
    WriteString("arctan from vector:");WriteLn;WriteFixPt(hvec1,2,2);WriteLn;
    WriteString("arctan2 from scalar:");WriteFixPt(hscal2,2,2);WriteLn;
    WriteString("arctan2 from vector:");WriteLn;WriteFixPt(hvec2,2,2);WriteLn;
  END; (*IF*)




END mathematical_functions;


(***********************************************************************
 *   Procedure:    PROCEDURE test_mathematical_functions
 *                                                                     
 *   Parameters:   print_mathematical_functions: true-all results are printed
 *
 *   Result:       -
 *
 *   Description:                                                      
 *   ------------                                                      
 *   Calls test of the mathematical functions with different types of vector.
 ***********************************************************************)
PROCEDURE test_mathematical_functions(print: BOOLEAN);
BEGIN

  IF print THEN 
    WriteString("-------------------------------------------------");
    WriteLn;
    WriteString("Test of mathematical functions with configuration grid");
    WriteLn;
    WriteString("-------------------------------------------------");
    WriteLn;
  END; (*IF*)
  mathematical_functions(gr_real,print);

  IF print THEN 
    WriteString("--------------------------------------------------");
    WriteLn;
    WriteString("Test of mathematical functions with configuration torus");
    WriteLn;
    WriteString("--------------------------------------------------");
    WriteLn;
  END; (*IF*)
  mathematical_functions(to_real,print);

  IF print THEN 
    WriteString("-------------------------------------------------");
    WriteLn;
    WriteString("Test of mathematical functions with configuration tree");
    WriteLn;
    WriteString("-------------------------------------------------");
    WriteLn;
  END; (*IF*)
  mathematical_functions(tr_real,print);

  IF print THEN 
    WriteString("--------------------------------------------------");
    WriteLn;
    WriteString("Test of mathematical functions with configuration hyper");
    WriteLn;
    WriteString("--------------------------------------------------");
    WriteLn;
  END; (*IF*)
  mathematical_functions(hy_real,print);
		    
  IF print THEN 
    WriteString("-------------------------------------------------");
    WriteLn;
    WriteString("Test of mathematical functions with configuration list");
    WriteLn;
    WriteString("-------------------------------------------------");
    WriteLn;
  END; (*IF*)
  mathematical_functions(li_real,print);

  IF print THEN 
    WriteString("------------------------------------------------------");
    WriteLn;
    WriteString("Test of mathematical functions with configuration char_list");
    WriteLn;
    WriteString("------------------------------------------------------");
    WriteLn;
  END; (*IF*)
  mathematical_functions(cl_real,print);

END test_mathematical_functions;
(***********************************************************************
 *   Procedure:    main program
 *                                                                     
 *   Parameters:   -
 *
 *   Result:       -
 *
 *   Description:                                                      
 *   ------------                                                      
 *   calls all testing procedures
 ***********************************************************************)
VAR print_general_functions:      BOOLEAN;
    print_mathematical_functions: BOOLEAN;
BEGIN
  
  print_general_functions := FALSE;
  print_mathematical_functions := TRUE;
  
  init;
  test_general_functions(print_general_functions);
  WriteString("Hallo"); WriteLn;
  test_mathematical_functions(print_mathematical_functions);
  
END claudi.
