/*************************************************************************
 *
 * Parallaxis Compiler Projects
 * ----------------------------
 *
 * Name  :  GenIdents.h
 * Author:  lampke, keller
 * Date  :  Tue Sep 27 14:26:08 1994
 *
 * Description:
 * ------------
 * Generate (new) identifiers. These are modified versions of existing
 * identifiers (e.g. in the case of name conflicts) or automatically
 * generated new identifiers like labels or help variables.
 *
 ************************************************************************/

#ifndef GenIdents_h
#define GenIdents_h

#include "Idents.h"			  /* Reuse-Library */
#include "parallaxis.h"

/*************************************************************************
 * name: GenIdents_MakeNewIdent
 * in  : Name of the identififier
 * out : A definitely new id (realized by calling GenIdents_Rename() if the
 *       name already existed)
 ************************************************************************/
extern tIdent GenIdents_MakeNewIdent( char *ReqIdName );

/*************************************************************************
 * name: GenIdents_MakeQualified
 * in  :
 * out : Module_Ident
 ************************************************************************/
extern tIdent GenIdents_MakeQualified( tIdent Module, tIdent Ident );

/*************************************************************************
 * name: GenIdents_RenameField
 * in  :
 * out : C_0_Ident
 ************************************************************************/
extern tIdent GenIdents_RenameField( tIdent Ident );

/*************************************************************************
 * name: GenIdents_Rename
 * in  :
 * out : C_nnn_Ident
 ************************************************************************/
extern tIdent GenIdents_Rename( tIdent tIdent );

/*************************************************************************
 * name: GenIdents_Label
 * in  :
 * out : EXIT_nnn
 ************************************************************************/
extern tIdent GenIdents_Label( void );

/*************************************************************************
 * name: GenIdents_Selector
 * in  :
 * out : StructurOrUnion
 ************************************************************************/
extern tIdent GenIdents_Selector( CHAR StructOrUnion, ULONG Nr );

/*************************************************************************
 * name: GenIdents_Param
 * in  :
 * out : O_nnn
 ************************************************************************/
extern tIdent GenIdents_Param( void );

/*************************************************************************
 * name: GenIdent_With
 * in  :
 * out : W_nnn
 ************************************************************************/
extern tIdent GenIdents_With( void );

/*************************************************************************
 * name: GenIdents_Struct1
 * in  :
 * out : Module__nnn, if Module = single char; Module_nnn else
 ************************************************************************/
extern tIdent GenIdents_Struct1( tIdent module, ULONG nr );

/*************************************************************************
 * name: GenIdents_Struct2
 * in  :
 * out : S_nnn
 ************************************************************************/
extern tIdent GenIdents_Struct2( void );

/*************************************************************************
 * name: GenIdents_GlobalPtr
 * in  :
 * out : G_nnn
 ************************************************************************/
extern tIdent GenIdents_GlobalPtr( tIdent Ident );

/*************************************************************************
 * name: GenIdents_LocalPtr
 * in  :
 * out : L_nnn
 ************************************************************************/
extern tIdent GenIdents_LocalPtr( void );

/*************************************************************************
 * name: GenIdents_Bound
 * in  :
 * out : B_nnn
 ************************************************************************/
extern tIdent GenIdents_Bound( void );

/*************************************************************************
 * name: GenIdents_Return
 * in  :
 * out : R_nnn
 ************************************************************************/
extern tIdent GenIdents_Return( void );

/*************************************************************************
 * name: GenIdents_String
 * in  :
 * out : X_nnn
 ************************************************************************/
extern tIdent GenIdents_String( void );

/*************************************************************************
 * name: GenIdents_Opaque
 * in  :
 * out : Module_Ident -> Ident
 ************************************************************************/
extern tIdent GenIdents_Opaque( tIdent typeName );

/*************************************************************************
 * name: GenIdents_HelpVar
 * in  :
 * out : H_nnn
 ************************************************************************/
extern tIdent GenIdents_HelpVar( void );

extern void GenIdents_Init( void );

#endif
