/*************************************************************************
 *
 * Parallaxis Compiler Projects
 * ----------------------------
 *
 * name  :  LoopStack.c
 * desc  :  stack for parallel statements nested in LOOP
 * author:  weiss
 *
 ************************************************************************/

#include <stdlib.h>
#include "DynArray.h"			  /* Reuse-Library */
#include "Idents.h"			  /* Reuse-Library */
#include "parallaxis.h"
#include "LoopStack.h"

static BOOLEAN isInitialised = FALSE;

/*
 * local
 */

static BOOLEAN isInLOOP = FALSE;

#define InitialConfigStackSize 32

static long ConfigStackPtr, ConfigStackSize;

static tConfigStack *ConfigStack;

#define InitialActivesetsSize 6

#define InitialLoopStackSize 16

static long LoopStackPtr, LoopStackSize;

static tLoopStack *LoopStack;

/*
 * global function implementation
 */

/*---------------------------------------------------------------------------*/

void ConfigStack_Push(tIdent config_ident)
{
  if (ConfigStackPtr == ConfigStackSize) {
    ExtendArray((char **) &ConfigStack, &ConfigStackSize, sizeof(tConfigStack));
  }
  ConfigStack[ConfigStackPtr].config_ident = config_ident;
  if (ConfigStackPtr == 0) {
    ConfigStack[ConfigStackPtr].activeset_num = 0;
  } else {
    ConfigStack[ConfigStackPtr].activeset_num =
	ConfigStack[ConfigStackPtr-1].activeset_num+1;
  }

  ConfigStackPtr++;
}

/*---------------------------------------------------------------------------*/

ULONG ConfigStack_GetNextActivesetNum(long *config_stack_index)
{
  long index = *config_stack_index+1;

  while(index < ConfigStackPtr) {
    if (ConfigStack[index].config_ident == ConfigStack_GetReturnId()) {
      *config_stack_index = index;
      return ConfigStack[index].activeset_num;
    } else {
      index++;
    }
  }

  return ConfigStack[*config_stack_index].activeset_num;
}

/*---------------------------------------------------------------------------*/

ULONG ConfigStack_GetTopActivesetNum()
{
  return ConfigStack[ConfigStackPtr-1].activeset_num;
}

/*---------------------------------------------------------------------------*/

tIdent ConfigStack_GetReturnId()
{
  return ConfigStack[0].config_ident;
}

/*---------------------------------------------------------------------------*/

void ConfigStack_Pop()
{
  if (ConfigStackPtr != 0) {
    ConfigStackPtr--;
  }
}

/*---------------------------------------------------------------------------*/

void LoopStack_Push(tIdent label_ident, tIdent parloop_config)
{
  if (LoopStackPtr == LoopStackSize) {
    ExtendArray((char **) &LoopStack, &LoopStackSize, sizeof(tLoopStack));
  }
  LoopStack[LoopStackPtr].parloop_config = parloop_config;
  LoopStack[LoopStackPtr].label_ident = label_ident;
  LoopStack[LoopStackPtr].activesets_size = InitialActivesetsSize;
  LoopStack[LoopStackPtr].activesets_ptr = 0;
  MakeArray((char **) &(LoopStack[LoopStackPtr].activesets),
	    &(LoopStack[LoopStackPtr].activesets_size),
	    sizeof(long));
  LoopStackPtr++;
  isInLOOP = TRUE;
}

/*---------------------------------------------------------------------------*/

void LoopStack_Pop()
{
  if (LoopStackPtr != 0) {
    if (LoopStackPtr == 1) {
      isInLOOP = FALSE;
    }
    LoopStackPtr--;
  }
}

/*---------------------------------------------------------------------------*/

void LoopStack_OpenEnv(tIdent config_ident)
{
  tLoopStack *top = LoopStack_Top();

  ConfigStack_Push(config_ident);
  if (isInLOOP) {
    if (top->activesets_ptr == top->activesets_size) {
      ExtendArray((char **) &(top->activesets), &(top->activesets_size), sizeof(long));
    }
    top->activesets[top->activesets_ptr] = ConfigStackPtr-1;
    top->activesets_ptr++;
  }
}

/*---------------------------------------------------------------------------*/

void LoopStack_CloseEnv()
{
  tLoopStack *top = LoopStack_Top();

  ConfigStack_Pop();
  if (isInLOOP) {
    if (top->activesets_ptr != 0) {
      top->activesets_ptr--;
    }
  }
}

/*---------------------------------------------------------------------------*/

ULONG LoopStack_GetTopLastActivesetIndex()
{
  return LoopStack_Top()->activesets_ptr-1;
}

/*---------------------------------------------------------------------------*/

BOOLEAN LoopStack_IsTopLastConfig(long index, long to_index)
{
  tLoopStack *top = LoopStack_Top();
  tIdent act_config_ident = ConfigStack[top->activesets[index]].config_ident;
  long i;

  for (i = index-1; i >= to_index; i--) {
    if (ConfigStack[top->activesets[i]].config_ident == act_config_ident) {
      return FALSE;
    }
  }

  return TRUE;
}

/*---------------------------------------------------------------------------*/

tIdent LoopStack_GetTopConfig(long index)
{
  return ConfigStack[LoopStack_Top()->activesets[index]].config_ident;
}

/*---------------------------------------------------------------------------*/

ULONG LoopStack_GetTopActivesetNum(long index)
{
  return ConfigStack[LoopStack_Top()->activesets[index]].activeset_num;
}

/*---------------------------------------------------------------------------*/

tIdent LoopStack_GetTopLabel()
{
  return LoopStack_Top()->label_ident;
}

/*---------------------------------------------------------------------------*/

tIdent LoopStack_GetTopLoopConfig()
{
  return LoopStack_Top()->parloop_config;
}

/*---------------------------------------------------------------------------*/

ULONG LoopStack_GetTopNextActivesetNum(long *act_index)
{
  tLoopStack *top = LoopStack_Top();
  long index = *act_index+1;

  while (index <= LoopStack_GetTopLastActivesetIndex()) {
    if (ConfigStack[top->activesets[index]].config_ident == top->parloop_config) {
      *act_index = index;
      return ConfigStack[top->activesets[index]].activeset_num;
    } else {
      index++;
    }
  }
  return ConfigStack[top->activesets[*act_index]].activeset_num;
}  
  
/*---------------------------------------------------------------------------*/

tLoopStack *LoopStack_Top()
{
  if (LoopStackPtr > 0) {
    return &LoopStack[LoopStackPtr-1];
  } else {
    return NULL;
  }
}

/*---------------------------------------------------------------------------*/

void ConfigStack_Free()
{
  ReleaseArray((char **) &ConfigStack, &ConfigStackSize, sizeof(tConfigStack));
}

/*---------------------------------------------------------------------------*/

void ConfigStack_Init()
{
  if (!isInitialised) {
    ConfigStackSize = InitialConfigStackSize;
    ConfigStackPtr = 0;
    MakeArray((char **) &ConfigStack, &ConfigStackSize, sizeof(tConfigStack));
  }
}

/*---------------------------------------------------------------------------*/

void LoopStack_Free()
{
  ConfigStack_Free();
  ReleaseArray((char **) &LoopStack, &LoopStackSize, sizeof(tLoopStack));
}

/*---------------------------------------------------------------------------*/

void LoopStack_Init()
{
  if (!isInitialised) {

    ConfigStack_Init();

    LoopStackSize = InitialLoopStackSize;
    LoopStackPtr = 0;
    MakeArray((char **) &LoopStack, &LoopStackSize, sizeof(tLoopStack));

    isInitialised = TRUE;
  }
}
