/*************************************************************************
 *
 * Parallaxis Compiler Projects
 * ----------------------------
 *
 * name  :  LoopStack
 * desc  :  Stack for the administration of the environments generated for
 *	    the translation of LOOP and EXIT statements.
 * author:  weiss
 *
 ************************************************************************/

#ifndef _LoopStack_H
#define _LoopStack_H

#include "Idents.h"                       /* Reuse-Library */

typedef struct {
    tIdent config_ident; /* name of vector config */
    ULONG activeset_num; /* number of activeset in nesting level */
} tConfigStack;

typedef struct {
    tIdent parloop_config;  /* config of loop, NoIdent if no parallel LOOP */
    tIdent label_ident;	    /* label to jump to on EXIT */
    long activesets_size;
    long activesets_ptr;
    long *activesets;	    /* activesets belonging to this LOOP */
} tLoopStack;

/*************************************************************************
 * name: ConfigStack_Push
 * in  : config_ident, element with this config ident is put on stack
 * out :
 ************************************************************************/

extern void ConfigStack_Push(tIdent config_ident);

/*************************************************************************
 * name: ConfigStack_GetNextActivesetNum
 * in  : search from config_stack_index
 * out : if a new corresponding config is found config_stack_index is
 *	 set to the index of the config found
 *	 number of activeset found corresponding to config ReturnId,
 *	 if no corresponding activeset is found the activeset number
 *       of config_stack_index is returned
 ************************************************************************/

extern ULONG ConfigStack_GetNextActivesetNum(long *config_stack_index);

/*************************************************************************
 * name: ConfigStack_GetTopActivesetNum
 * in  : activeset number of top element of ConfigStack
 * out :
 ************************************************************************/

extern ULONG ConfigStack_GetTopActivesetNum();

/*************************************************************************
 * name: ConfigStack_GetReturnId
 * in  : configuration name of the return type of a parallel function
 * out :
 ************************************************************************/

extern tIdent ConfigStack_GetReturnId();

/*************************************************************************
 * name: ConfigStack_Pop
 * in  : 
 * out :
 ************************************************************************/

extern void ConfigStack_Pop();

/*************************************************************************
 * name: LoopStack_Push
 * in  : label_ident, parloop_config, LoopStack with this label ident and
 *	 config ident is put on stack
 * out :
 ************************************************************************/

extern void LoopStack_Push(tIdent label_ident, tIdent parloop_config);

/*************************************************************************
 * name: LoopStack_Pop
 * in  :
 * out :
 ************************************************************************/

extern void LoopStack_Pop();

/*************************************************************************
 * name: LoopStack_OpenEnv
 * in  : Ident, remember opening of environment with this configname
 * out :
 ************************************************************************/

extern void LoopStack_OpenEnv(tIdent Ident);

/*************************************************************************
 * name: LoopStack_CloseEnv
 * in  :
 * out :
 ************************************************************************/

extern void LoopStack_CloseEnv();

/*************************************************************************
 * name: LoopStack_GetTopLastActivesetIndex
 * in  :
 * out :
 ************************************************************************/

extern ULONG LoopStack_GetTopLastActivesetIndex();

/*************************************************************************
 * name: LoopStack_IsTopLastConfig
 * in  : index, to_index, we search backwards from index to to_index
 *	 for another config ident
 * out :
 ************************************************************************/

extern BOOLEAN LoopStack_IsTopLastConfig();

/*************************************************************************
 * name: LoopStack_GetTopConfig
 * in  : index, return config ident of activeset in current loopstack element
 * out :
 ************************************************************************/

extern tIdent LoopStack_GetTopConfig(long index);

/*************************************************************************
 * name: LoopStack_GetTopActivesetNum
 * in  : index, return num of activeset in current loopstack element
 * out :
 ************************************************************************/

extern ULONG LoopStack_GetTopActivesetNum(long index);

/*************************************************************************
 * name: LoopStack_GetTopLabel
 * in  :
 * out :
 ************************************************************************/

extern tIdent LoopStack_GetTopLabel();

/*************************************************************************
 * name: LoopStack_GetTopLoopConfig
 * in  : 
 * out :
 ************************************************************************/

extern tIdent LoopStack_GetTopLoopConfig();

/*************************************************************************
 * name: LoopStack_GetTopNextActivesetNum
 * in  : act_index, search for other activesets with current config ident
 *	 from act_index to last index
 * out :
 ************************************************************************/

extern ULONG LoopStack_GetTopNextActivesetNum(long *act_index);


/*************************************************************************
 * name: LoopStack_Top
 * in  :
 * out :
 ************************************************************************/

extern tLoopStack *LoopStack_Top();

/*************************************************************************
 * name: LoopStack_Free
 * in  :
 * out :
 ************************************************************************/

extern void LoopStack_Free();

/*************************************************************************
 * name: LoopStack_Init
 * in  :
 * out :
 ************************************************************************/

extern void LoopStack_Init();

#endif
