/*************************************************************************
 *
 * Parallaxis Compiler Projects
 * ----------------------------
 *
 * name   :  StdProcs.c
 * desc   :  handling of buildin functions and procedures
 * Authors:  lampke
 *
 ************************************************************************/

#include <stdlib.h>
#include "DynArray.h"			  /* Reuse-Library */
#include "Idents.h"			  /* Reuse-Library */
#include "StdProcs.h"

static BOOLEAN isInitialised = FALSE;

typedef struct {
   tStdProc proc;
   BITSET   params;
} tParamCount;

/*
 * local variables
 */

/*
 * number of allowed parameters
 * 01 = no parameters, 02 = 1 parameter, 04 = 2 parameters, 08 = 3 parameters
 * if a standard procedure has a variable number of parameters, add the numbers
 * e.g. INC(x) and INC(x,y) : 02 + 04 = 06
 */
static tParamCount paramCounts[] = {
   { eSystemAbs,         0x02 },
   { eSystemCap,         0x02 },
   { eSystemChr,         0x02 },
   { eSystemDim,         0x04 },
   { eSystemDim2Id,      0x04 },
   { eSystemEven,        0x02 },
   { eSystemFloat,       0x02 },
   { eSystemHigh,        0x02 },
   { eSystemId,          0x02 },
   { eSystemId2Dim,      0x08 },
   { eSystemLen,         0x06 },
   { eSystemLower,       0x04 },
   { eSystemMax,         0x02 },
   { eSystemMin,         0x02 },
   { eSystemOdd,         0x02 },
   { eSystemOrd,         0x02 },
   { eSystemRank,        0x02 },
   { eSystemSize,        0x02 },
   { eSystemTrunc,       0x02 },
   { eSystemUpper,       0x04 },
   { eSystemVal,         0x04 },
   { eSystemDec,         0x06 },
   { eSystemExcl,        0x04 },
   { eSystemHalt,        0x01 },
   { eSystemInc,         0x06 },
   { eSystemIncl,        0x04 },
   { eSystemNew,         0x02 },
   { eSystemDispose,     0x02 },
   { eSystemAdr,         0x02 },
   { eSystemTSize,       0x02 },
/*
 * Reduce functions
 */
   { eSystemSum,         0x02 },
   { eSystemProduct,     0x02 },
   { eSystemAnd,         0x02 },
   { eSystemOr,          0x02 },
   { eSystemFirst,       0x02 },
   { eSystemLast,        0x02 },
/*
 * MathLib
 */
   { eSystemSqrt,        0x02 },
   { eSystemExponent,    0x02 },
   { eSystemLn,          0x02 },
   { eSystemSin,         0x02 },
   { eSystemArcSin,      0x02 },
   { eSystemCos,         0x02 },
   { eSystemArcCos,      0x02 },
   { eSystemTan,         0x02 },
   { eSystemArcTan,      0x02 },
   { eSystemArcTan2,     0x04 },
   { eSystemCeiling,     0x02 },
   { eSystemFloor,       0x02 },
   { eSystemRound,       0x02 },
   { eSystemCardRandom,  0x03 },
   { eSystemIntRandom,   0x03 },
   { eSystemRealRandom,  0x03 },
   { eSystemBoolRandom,  0x03 },
   { eSystemCharRandom , 0x03 },
/*
 * SystemIO
 */
   { eSystemOpenOutput,  0x02 },
   { eSystemCloseOutput, 0x01 },
   { eSystemOpenInput,   0x02 },
   { eSystemCloseInput,  0x01 },    
   { eSystemWrite,       0x02 },
   { eSystemWriteInt,    0x04 },
   { eSystemWriteCard,   0x04 },
   { eSystemWriteReal,   0x04 },
   { eSystemWriteFixPt,  0x08 },
   { eSystemWriteBoolean,0x04 },
   { eSystemWriteString, 0x02 },
   { eSystemWriteLn,     0x01 },
   { eSystemRead,        0x02 },
   { eSystemReadInt,     0x02 },
   { eSystemReadCard,    0x02 },
   { eSystemReadReal,    0x02 },
   { eSystemReadBoolean, 0x02 },
   { eSystemReadString,  0x02 },
/*
 * Commandline Arguments
 */
   { eSystemArgc,        0x01 },
   { eSystemArgv,        0x04 }
};


/*
 * global function implementation
 */
/*------- ParamCount -------------------------------------------------------*/

BOOLEAN StdProcs_ParamCount(tStdProc proc, unsigned long param, BITSET *params)
{
  unsigned long i;

  for (i = 0; i < ARRAYSIZE(paramCounts); i++) {
    if (paramCounts[i].proc == proc) {
      *params = paramCounts[i].params;
      return BITSET_IsElement(*params, param);
    }
  }
  *params = EMPTYBITSET;
  return FALSE;
}

void StdProcs_Init(void)
{
  if (!isInitialised) {
    isInitialised = TRUE;
    BigSet_Init();
  }
}

